// Copyright (c) 2012-2015 The Bitcoin Core developers
// Copyright (c) 2015-2023 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#ifndef NEXA_BLOOM_H
#define NEXA_BLOOM_H

#include "consensus/consensus.h"
#include "serialize.h"

#include <vector>

class COutPoint;
class uint256;
class CTransaction;

typedef std::shared_ptr<const CTransaction> CTransactionRef;

//! 20,000 items with fp rate < 0.1% or 10,000 items and <0.0001%
static const unsigned int MAX_HASH_FUNCS = 50;

//! Don't insert any script data in the bloom filter whose size is less than this
static const unsigned int MIN_BLOOM_DATA_SIZE = 16;

/**
 * First two bits of nFlags control how much IsRelevantAndUpdate actually updates
 * The remaining bits are reserved
 */
enum bloomflags
{
    BLOOM_UPDATE_NONE = 0,
    BLOOM_UPDATE_ALL = 1,
    // Only adds outpoints to the filter if the output is a pay-to-pubkey/pay-to-multisig script
    BLOOM_UPDATE_P2PUBKEY_ONLY = 2,
    BLOOM_UPDATE_MASK = 3,
};

/**
 * BloomFilter is a probabilistic filter which SPV clients provide
 * so that we can filter the transactions we send them.
 *
 * This allows for significantly more efficient transaction and block downloads.
 *
 * Because bloom filters are probabilistic, a SPV node can increase the false-
 * positive rate, making us send it transactions which aren't actually its,
 * allowing clients to trade more bandwidth for more privacy by obfuscating which
 * keys are controlled by them.
 */
class CBloomFilter
{
protected:
    std::vector<unsigned char> vData;
    bool isFull;
    bool isEmpty;
    unsigned int nHashFuncs;
    unsigned int nTweak;
    unsigned char nFlags;

    unsigned int Hash(unsigned int nHashNum, const std::vector<unsigned char> &vDataToHash) const;

    // Private constructor for CRollingBloomFilter, no restrictions on size
    CBloomFilter(unsigned int nElements, double nFPRate, unsigned int nTweak);
    friend class CRollingBloomFilter;

    /** Helper function to set up bloom filter from desired number of elements and false positive rate.
        Used by the constructors. Checks that the variables in sane ranges. */
    void setup(unsigned int nElements,
        double nFPRate,
        unsigned int nTweakIn,
        unsigned char nFlagsIn,
        bool size_constrained,
        uint32_t nMaxFilterSize);
    /** Alternative setup function that guarantees the false positive rate does not exceed nFPRate. */
    void setupGuaranteeFPR(unsigned int nElements,
        double nFPRate,
        unsigned int nTweakIn,
        unsigned char nFlagsIn,
        uint32_t nMaxFilterSize);

    //! Checks for empty and full filters to avoid wasting cpu
    void UpdateEmptyFull();

public:
    bool IsEmpty() const { return isEmpty; }
    //! for testing only
    bool IsFull() const { return isFull; }
    //! for testing only
    unsigned int vDataSize() const { return vData.size(); }
    /**
     * Creates a new bloom filter which will provide the given fp rate when filled with the given number of elements
     * Note that if the given parameters will result in a filter outside the bounds of the protocol limits,
     * the filter created will be as close to the given parameters as possible within the protocol limits.
     * This will apply if nFPRate is very low or nElements is unreasonably high.
     * nTweak is a constant which is added to the seed value passed to the hash function
     * It should generally always be a random value (and is largely only exposed for unit testing)
     * nFlags should be one of the BLOOM_UPDATE_* enums (not _MASK)
     */
    CBloomFilter(unsigned int nElements,
        double nFPRate,
        unsigned int nTweak,
        unsigned char nFlagsIn,
        uint32_t nMaxFilterSize = SMALLEST_MAX_BLOOM_FILTER_SIZE);
    /**
     * When fGuaranteeFPR == true, it is guaranteed that the false positive rate will not exceed nFPRate.
     */
    CBloomFilter(unsigned int nElements,
        double nFPRate,
        unsigned int nTweak,
        unsigned char nFlagsIn,
        bool fGuaranteeFPR,
        uint32_t nMaxFilterSize);
    CBloomFilter() : isFull(true), isEmpty(true), nHashFuncs(0), nTweak(0), nFlags(0) {}
    ADD_SERIALIZE_METHODS;

    template <typename Stream, typename Operation>
    inline void SerializationOp(Stream &s, Operation ser_action)
    {
        READWRITE(vData);
        READWRITE(nHashFuncs);
        READWRITE(nTweak);
        READWRITE(nFlags);
        if (ser_action.ForRead())
        {
            UpdateEmptyFull();
        }
    }

    void insert(const std::vector<unsigned char> &vKey);
    void insert(const COutPoint &outpoint);
    void insert(const uint256 &hash);

    bool contains(const std::vector<unsigned char> &vKey) const;
    bool contains(const COutPoint &outpoint) const;
    bool contains(const uint256 &hash) const;

    void clear();
    void reset(unsigned int nNewTweak);

    //! True if the size is <= SMALLEST_MAX_BLOOM_FILTER_SIZE and the number of hash functions is <= MAX_HASH_FUNCS
    //! (catch a filter which was just deserialized which was too big)
    bool IsWithinSizeConstraints() const;

#ifndef LIGHT // limit dependencies -- we do not need these functions in light clients because they support the server
    // side of the filter protocol.

    //! Scans output scripts for matches and adds those outpoints to the filter
    //! for spend detection. Returns true if any output matched, or the txid
    //! matches.
    bool MatchAndInsertOutpoints(const CTransactionRef &tx);

    //! Scan inputs to see if the spent outpoints are a match, or the input
    //! scripts contain matching elements.
    bool MatchInputs(const CTransactionRef &tx);

    //! Check if the transaction is relevant for any reason.
    //! Also adds any outputs which match the filter to the filter (to match
    //! their spending txes)
    bool IsRelevantAndUpdate(const CTransactionRef &tx) { return MatchAndInsertOutpoints(tx) || MatchInputs(tx); }
#endif
};

/**
 * RollingBloomFilter is a probabilistic "keep track of most recently inserted" set.
 * Construct it with the number of items to keep track of, and a false-positive
 * rate. Unlike CBloomFilter, by default nTweak is set to a cryptographically
 * secure random value for you. Similarly rather than clear() the method
 * reset() is provided, which also changes nTweak to decrease the impact of
 * false-positives.
 *
 * contains(item) will always return true if item was one of the last N to 1.5*N
 * insert()'ed ... but may also return true for items that were not inserted.
 *
 * It needs around 1.8 bytes per element per factor 0.1 of false positive rate.
 * (More accurately: 3/(log(256)*log(2)) * log(1/fpRate) * nElements bytes)
 */
class CRollingBloomFilter
{
public:
    // A random bloom filter calls GetRand() at creation time.
    // Don't create global CRollingBloomFilter objects, as they may be
    // constructed before the randomizer is properly initialized.
    CRollingBloomFilter(unsigned int nElements, double nFPRate);

    void insert(const std::vector<unsigned char> &vKey);
    void insert(const uint256 &hash);
    void insert(const COutPoint &outpoint);
    bool contains(const std::vector<unsigned char> &vKey) const;
    bool contains(const uint256 &hash) const;
    bool contains(const COutPoint &outpoint) const;

    void reset();

    //! for testing only
    unsigned vDataSize() const { return data.size(); }

private:
    int nEntriesPerGeneration;
    int nEntriesThisGeneration;
    int nGeneration;
    std::vector<uint64_t> data;
    unsigned int nTweak;
    int nHashFuncs;
};

#endif // NEXA_BLOOM_H
