// Copyright (c) 2019-2022 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.
#include "electrum/electrumserver.h"
#include "electrum/rostrum.h"
#include "init.h"
#include "util.h"
#include "utilprocess.h"

#include <chrono>
#include <string>

#if BOOST_OS_LINUX
#include <signal.h>
#endif

//! give the program a second to complain about startup issues, such as invalid
//! parameters.
static bool startup_check(const SubProcess &p)
{
    using namespace std::chrono_literals;
    std::this_thread::sleep_for(1s);
    if (p.IsRunning())
    {
        // process hasn't exited, good.
        return true;
    }
    LOGA("Electrum: startup check failed, server exited within 1 second");
    return false;
}

static void log_args(const std::string &path, const std::vector<std::string> &args)
{
    if (!Logging::LogAcceptCategory(ELECTRUM))
    {
        return;
    }

    std::stringstream ss;
    ss << path;
    for (auto &a : args)
    {
        ss << " " << a;
    }
    LOGA("Electrum: spawning %s", ss.str());
}

namespace electrum
{
ElectrumServer::ElectrumServer() : started(false), stop_requested(false) {}
ElectrumServer::~ElectrumServer()
{
    if (started)
        Stop();

    if (process_thread.joinable())
    {
        process_thread.join();
    }
}

//! called when rostrum produces a line in stdout/stderr
static void callb_logger(const std::string &line) { LOGA("Electrum: %s", line); }
bool ElectrumServer::Start(int rpcport, int p2pport, const std::string &network)
{
    if (mapArgs.count("-electrum")) // arg is explicitly specified
    {
        if (!GetBoolArg("-electrum", true))
        {
            LOGA("Electrum: Disabled. Not starting server.");
            return true;
        }
        return Start(rostrum_path(), rostrum_args(rpcport, p2pport, network));
    }
    else // arg is not specified, so run rostrum only if it exists
    {
        try
        {
            // always run rostrum by default on testnet, mainet but not on regtest
            if (!GetBoolArg("-regtest", false))
            {
                return Start(rostrum_path(), rostrum_args(rpcport, p2pport, network));
            }
            else
            {
                LOGA("Not starting electrum by default on regtest.");
                return true;
            }
        }
        catch (std::runtime_error &e) // no rostrum found.  This is ok because the user did not explicitly ask to start
        {
            LOGA("Electrum NOT STARTED: Error %s.  On platforms unsupported by Rostrum this may be benign.", e.what());
            return true;
        }
    }
}

bool ElectrumServer::Start(const std::string &path, const std::vector<std::string> &args)
{
    stop_requested = false;
    DbgAssert(!started, return false);
    log_args(path, args);
    std::unique_lock<std::mutex> lock(process_cs);
    process.reset(new SubProcess(path, args, callb_logger, callb_logger));

    process_thread = std::thread(
        [this]()
        {
            LOGA("Electrum: Starting server");
            try
            {
                this->process->Run();
            }
            catch (const subprocess_error &e)
            {
                LOGA("Electrum: Server not running: %s, exit status %d, termination signal %d", e.what(), e.exit_status,
                    e.termination_signal);
            }
            catch (...)
            {
                LOGA("Electrum: Unknown error running server");
            }
            this->started = false;

            if (!stop_requested && GetBoolArg("-electrum.shutdownonerror", false))
            {
                // The electrum server exit was not initiated by us, so it
                // must have stopped due to some error.

                LOGA("Electrum: Nexad is configured to exit when "
                     "electrum exits on error. Initiating shutdown.");
                StartShutdown();
            }
        });
    started = startup_check(*process);
    return started;
}

static void stop_server(SubProcess &p, int timeoutInSeconds)
{
    if (!p.IsRunning())
    {
        return;
    }
    LOGA("Electrum: Stopping server");

    try
    {
        p.Interrupt();
    }
    catch (const subprocess_error &e)
    {
        LOGA("Electrum: %s", e.what());
        p.Terminate();
        return;
    }

    using namespace std::chrono_literals;
    using namespace std::chrono;

    auto timeout = seconds(timeoutInSeconds);
    auto start = system_clock::now();
    while (p.IsRunning())
    {
        if ((system_clock::now() - start) < timeout)
        {
            std::this_thread::sleep_for(1s);
            continue;
        }
        LOGA("Electrum: Timed out waiting for clean shutdown (%s seconds)", timeout.count());
        p.Terminate();
        return;
    }
}

void ElectrumServer::Stop(int timeoutInSeconds)
{
    stop_requested = true;
    if (!started)
    {
        return;
    }
    try
    {
        std::unique_lock<std::mutex> lock(process_cs);
        stop_server(*process, timeoutInSeconds);
    }
    catch (const std::exception &e)
    {
        LOGA("Electrum: Error stopping server %s", e.what());
    }
    process_thread.join();
    started = false;
}

bool ElectrumServer::IsRunning() const
{
    std::unique_lock<std::mutex> lock(process_cs);
    if (!bool(process))
    {
        return false;
    }
    return process->IsRunning();
}

void ElectrumServer::NotifyNewBlock()
{
    std::unique_lock<std::mutex> lock(process_cs);
    if (!bool(process))
    {
        return;
    }
#if BOOST_OS_LINUX
    process->SendSignal(SIGUSR1);
#endif
}

ElectrumServer &ElectrumServer::Instance()
{
    static ElectrumServer instance;
    return instance;
}

} // namespace electrum
