// Copyright (c) 2009-2010 Satoshi Nakamoto
// Copyright (c) 2009-2015 The Bitcoin Core developers
// Copyright (c) 2018-2023 The Bitcoin Unlimited developers
// Copyright (C) 2019-2020 Tom Zander <tomz@freedommail.ch>
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include "net_processing.h"

#include "addrman.h"
#include "blockrelay/blockrelay_common.h"
#include "blockrelay/compactblock.h"
#include "blockrelay/graphene.h"
#include "blockrelay/mempool_sync.h"
#include "blockrelay/thinblock.h"
#include "blockstorage/blockstorage.h"
#include "capd/capd.h"
#include "chain.h"
#include "connmgr.h"
#include "dosman.h"
#include "electrum/rostrum.h"
#include "expedited.h"
#include "extversionkeys.h"
#include "main.h"
#include "merkleblock.h"
#include "nodestate.h"
#include "requestManager.h"
#include "respend/dsproof.h"
#include "respend/dsproofstorage.h"
#include "timedata.h"
#include "txadmission.h"
#include "validation/validation.h"
#include "validationinterface.h"
#include "version.h"
#include "wallet/grouptokencache.h"


extern std::atomic<int64_t> nTimeBestReceived;
extern std::atomic<int> nPreferredDownload;
extern std::atomic<int> nSyncStarted;
extern uint64_t nMaxReceiveBufferSize;
extern uint64_t nMaxSendBufferSize;
extern CTweak<unsigned int> maxBlocksInTransitPerPeer;
extern CTweak<uint64_t> grapheneMinVersionSupported;
extern CTweak<uint64_t> grapheneMaxVersionSupported;
extern CTweak<uint64_t> grapheneFastFilterCompatibility;
extern CTweak<uint64_t> mempoolSyncMinVersionSupported;
extern CTweak<uint64_t> mempoolSyncMaxVersionSupported;
extern CTweak<uint64_t> syncMempoolWithPeers;
extern CTweak<uint32_t> randomlyDontInv;
extern CTweak<uint32_t> doubleSpendProofs;
extern CTweak<bool> extVersionEnabled;
extern CTweak<bool> allowp2pTxVal;
extern bool nDropMessages;

bool HandleHeaderPathMessage(CDataStream &vRecv, CNode *pfrom, uint32_t msgCookie);

/** How many inbound connections will we track before pruning entries */
const uint32_t MAX_INBOUND_CONNECTIONS_TRACKED = 10000;
/** maximum size (in bytes) of a batched set of transactions */
static const uint32_t MAX_TXN_BATCH_SIZE = 10000;

extern std::string DebuggerToString(CValidationDebugger &debugger);

// Used for serializing calls to ReqestNextBlocksToDownload() so that only
// one thread at a time can execute this function.
CCriticalSection cs_requestNextBlocks;

bool CanDirectFetch(const Consensus::Params &consensusParams)
{
    return chainActive.Tip()->GetBlockTime() > GetAdjustedTime() - consensusParams.nPowTargetSpacing * 20;
}

void UpdatePreferredDownload(CNode *node)
{
    CNodeStateAccessor state(nodestate, node->GetId());
    DbgAssert(state != nullptr, return);
    nPreferredDownload.fetch_sub(state->fPreferredDownload);

    // Whether this node should be marked as a preferred download node.
    state->fPreferredDownload = !node->fOneShot && !node->fClient;
    // BU allow downloads from inbound nodes; this may have been limited to stop attackers from connecting
    // and offering a bad chain.  However, we are connecting to multiple nodes and so can choose the most work
    // chain on that basis.
    // state->fPreferredDownload = (!node->fInbound || node->fWhitelisted) && !node->fOneShot && !node->fClient;
    // LOG(NET, "node %s preferred DL: %d because (%d || %d) && %d && %d\n", node->GetLogName(),
    //   state->fPreferredDownload, !node->fInbound, node->fWhitelisted, !node->fOneShot, !node->fClient);

    nPreferredDownload.fetch_add(state->fPreferredDownload);
}

bool PeerHasHeader(const CNodeState *state, CBlockIndex *pindex)
{
    if (pindex == nullptr)
        return false;
    if (state->pindexBestKnownBlock && pindex == state->pindexBestKnownBlock->GetAncestor(pindex->height()))
        return true;
    if (state->pindexBestHeaderSent && pindex == state->pindexBestHeaderSent->GetAncestor(pindex->height()))
        return true;
    return false;
}

template <class T>
void static ProcessGetData(CNode *pfrom,
    const Consensus::Params &consensusParams,
    std::deque<std::pair<T, uint32_t> > &vInv)
{
    std::vector<T> vNotFound;
    CDataStream ss(SER_NETWORK, PROTOCOL_VERSION);

    auto it = vInv.begin();
    uint32_t replyCookieCount = 0;
    int cur = 0;
    int sz = vInv.size();
    uint32_t msgCookie = 0;
    LOG(NET, "Processing %d INVs\n", sz);
    while (it != vInv.end())
    {
        // If the cookie has changed, restart the reply count.  This might happen in weird cases where the invs
        // were deferred.
        uint32_t nextMsgCookie = it->second;
        if (nextMsgCookie != msgCookie)
            replyCookieCount = 0;
        msgCookie = nextMsgCookie;
        cur++;
        bool lastInv = (sz == cur);
        uint32_t oneReplyPerInvCookie = (lastInv ? (msgCookie | 0xFFFF) : (msgCookie + replyCookieCount));

        // Don't bother if send buffer is too full to respond anyway except we always consume at least one inv.
        // Otherwise we could get into an infinite loop where we never
        // process a message that's too large for our send buffer size.
        if (cur != 0 && pfrom->nSendSize >= nMaxSendBufferSize + ss.size())
        {
            break;
        }
        if (shutdown_threads.load() == true)
        {
            // I am in the middle of communicating with this node, so I want to make sure it gets an explicit close
            // rather than timing out since I am breaking the request/reply protocol by quitting.
            pfrom->CloseSocketDisconnect();
            return;
        }

        // start processing inventory here
        const T &inv = it->first;

        if (inv.type == MSG_BLOCK || inv.type == MSG_FILTERED_BLOCK || inv.type == MSG_CMPCT_BLOCK)
        {
            auto *mi = LookupBlockIndex(inv.hash);
            if (mi)
            {
                bool fSend = false;
                unsigned char noSendReason = 0;
                std::string noSendStr = std::string("");

                static const int nOneMonth = 30 * 24 * 60 * 60;
                if (chainActive.Contains(mi))
                {
                    fSend = true;
                }
                else
                {
                    // To prevent fingerprinting attacks, only send blocks outside of the active
                    // chain if they are valid, and no more than a month older (both in time, and in
                    // best equivalent proof of work) than the best header chain we know about.
                    {
                        READLOCK(cs_mapBlockIndex);
                        fSend = mi->IsValid(BLOCK_VALID_SCRIPTS) && (pindexBestHeader != NULL) &&
                                (pindexBestHeader.load()->GetBlockTime() - mi->GetBlockTime() < nOneMonth) &&
                                (GetBlockProofEquivalentTime(
                                     *pindexBestHeader, *mi, *pindexBestHeader, consensusParams) < nOneMonth);
                    }
                    if (!fSend)
                    {
                        LOG(NET, "%s: ignoring request from peer=%s for old block that isn't in the main chain\n",
                            __func__, pfrom->GetLogName());
                        noSendStr = "Not on main chain";
                        noSendReason = REJECT_FORK;
                    }

                    // TODO: in the future we can throttle old block requests by setting send=false if we are out
                    // of bandwidth
                }
                // disconnect node in case we have reached the outbound limit for serving historical blocks
                // never disconnect whitelisted nodes
                if (fSend && CNode::OutboundTargetReached(true) &&
                    (((pindexBestHeader != nullptr) &&
                         (pindexBestHeader.load()->GetBlockTime() - mi->GetBlockTime() > nOneMonth)) ||
                        inv.type == MSG_FILTERED_BLOCK) &&
                    !pfrom->fWhitelisted)
                {
                    LOG(NET, "historical block serving limit reached, disconnect peer %s\n", pfrom->GetLogName());
                    // disconnect node because it wants info that we are not going to give it
                    pfrom->fDisconnect = true;
                    fSend = false;
                    noSendStr = "Historical block limit reached";
                    noSendReason = REJECT_RETRY;
                }
                // Avoid leaking prune-height by never sending blocks below the
                // NODE_NETWORK_LIMITED threshold.
                // Add two blocks buffer extension for possible races
                if (fSend && !pfrom->fWhitelisted &&
                    ((((nLocalServices & NODE_NETWORK_LIMITED) == NODE_NETWORK_LIMITED) &&
                        ((nLocalServices & NODE_NETWORK) != NODE_NETWORK) &&
                        (chainActive.Tip()->height() - mi->height() > (int)NODE_NETWORK_LIMITED_MIN_BLOCKS + 2))))
                {
                    LOG(NET, "Ignore block request below NODE_NETWORK_LIMITED threshold from peer=%d\n",
                        pfrom->GetId());
                    // disconnect node and prevent it from stalling (would
                    // otherwise wait for the missing block)
                    pfrom->fDisconnect = true;
                    fSend = false;
                    noSendStr = "Block too old";
                    noSendReason = REJECT_LIMITED;
                }
                if (!(mi->nStatus & BLOCK_HAVE_DATA))
                {
                    fSend = false;
                    noSendStr = "Block pruned";
                    noSendReason = REJECT_OBSOLETE;
                }
                // Pruned nodes may have deleted the block, so check whether
                // it's available before trying to send.
                if (fSend)
                {
                    // Send block from disk
                    const ConstCBlockRef pblock = ReadBlockFromDisk(mi, consensusParams);
                    if (!pblock)
                    {
                        // its possible that I know about it but haven't stored it yet
                        LOG(THIN, "unable to load block %s from disk\n",
                            mi->phashBlock ? mi->phashBlock->ToString() : "");
                        fSend = false;
                        noSendStr = "Block is processing";
                        noSendReason = REJECT_WAITING;
                    }
                    else
                    {
                        if (inv.type == MSG_BLOCK)
                        {
                            LOG(NET, "Sending block via getdata message to %s\n", pfrom->GetLogName());
                            pfrom->blocksSent += 1;
                            pfrom->PushMessageWithCookie(NetMsgType::BLOCK, oneReplyPerInvCookie, *pblock);
                            replyCookieCount++;
                        }
                        else if (inv.type == MSG_CMPCT_BLOCK)
                        {
                            LOG(CMPCT, "Sending compactblock via getdata message to %s\n", pfrom->GetLogName());
                            SendCompactBlock(pblock, pfrom, msgCookie, inv.type);
                            replyCookieCount++;
                        }
                        else // MSG_FILTERED_BLOCK)
                        {
                            LOG(CMPCT, "Sending filtered block via getdata message to %s\n", pfrom->GetLogName());
                            LOCK(pfrom->cs_filter);
                            if (pfrom->pfilter)
                            {
                                CMerkleBlock merkleBlock(*pblock, *pfrom->pfilter);
                                pfrom->PushMessageWithCookie(
                                    NetMsgType::MERKLEBLOCK, msgCookie + replyCookieCount, merkleBlock);
                                replyCookieCount++;
                                pfrom->blocksSent += 1;
                                LOG(NET, "Bloom filtered block matched: %d tx\n", merkleBlock.vMatchedTxn.size());
                                // CMerkleBlock just contains hashes, so also push any transactions in the block the
                                // client did not see. This avoids hurting performance by pointlessly requiring a
                                // round-trip.
                                //
                                // Note that there is currently no way for a node to request any single transactions
                                // we didn't send here - they must either disconnect and retry or request the full
                                // block. Thus, the protocol spec specified allows for us to provide duplicate txn
                                // here, however we MUST always provide at least what the remote peer needs
                                typedef std::pair<unsigned int, uint256> PairType;
                                int txSz = merkleBlock.vMatchedTxn.size();
                                int txCount = 0;
                                for (PairType &pair : merkleBlock.vMatchedTxn)
                                {
                                    txCount++;
                                    bool lastTx = (txCount == txSz);
                                    pfrom->txsSent += 1;
                                    uint32_t replyCookie = msgCookie + replyCookieCount;
                                    if (lastInv && lastTx)
                                        replyCookie = msgCookie | 0xFFFF;
                                    pfrom->PushMessageWithCookie(NetMsgType::TX, replyCookie, pblock->vtx[pair.first]);
                                    replyCookieCount++;
                                }
                            }
                            else
                            {
                                fSend = false;
                                noSendReason = REJECT_MALFORMED;
                                noSendStr = std::string("Merkle block requested but no filter installed");
                            }
                        }
                    }
                }
                if (!fSend)
                {
                    std::string strCommand = NetMsgType::BLOCK;
                    uint32_t replyCookie = msgCookie + replyCookieCount;
                    if (lastInv)
                        replyCookie = msgCookie | 0xFFFF;
                    pfrom->PushMessageWithCookie(
                        NetMsgType::REJECT, replyCookie, strCommand, noSendReason, noSendStr, inv.hash);
                }
            }
            else
            {
                // Trigger the peer node to send a getblocks request for the next batch of inventory
                if (inv.hash == pfrom->hashContinue)
                {
                    // Bypass PushInventory, this must send even if redundant,
                    // and we want it right after the last block so they don't
                    // wait for other stuff first.
                    std::vector<T> oneInv;
                    oneInv.push_back(T(MSG_BLOCK, chainActive.Tip()->GetBlockHash()));
                    pfrom->PushMessageWithCookie(NetMsgType::INV, oneReplyPerInvCookie + replyCookieCount, oneInv);
                    replyCookieCount++;
                    pfrom->hashContinue.SetNull();
                }
                else
                {
                    vNotFound.push_back(inv);
                }
            }
        }
        else if (inv.type == MSG_DOUBLESPENDPROOF && doubleSpendProofs.Value() == true)
        {
            DoubleSpendProof dsp = mempool.doubleSpendProofStorage()->lookup(inv.hash);
            if (!dsp.isEmpty())
            {
                CDataStream ssDSP(SER_NETWORK, PROTOCOL_VERSION);
                ssDSP.reserve(600);
                ssDSP << dsp;
                pfrom->PushMessageWithCookie(NetMsgType::DSPROOF, oneReplyPerInvCookie, ssDSP);
            }
            else
            {
                pfrom->PushMessageWithCookie(NetMsgType::REJECT, oneReplyPerInvCookie, std::string(NetMsgType::DSPROOF),
                    REJECT_INVALID, std::string("dsproof requested was not found"));
            }
            replyCookieCount++;
        }
        else if (inv.IsKnownType())
        {
            std::shared_ptr<CDataStream> pStream = nullptr;
            std::vector<std::shared_ptr<CDataStream> > vStream;
            CTransactionRef ptx = nullptr;

            // Check for stream from relay memory first since this is the most efficient data to send.
            vStream = maprelay.Find(inv.hash);
            if (vStream.empty())
            {
                ptx = CommitQGet(inv.hash);
                if (!ptx)
                {
                    ptx = mempool.get(inv.hash);
                }
            }

            if (vStream.empty() && !ptx)
            {
                vNotFound.push_back(inv);
            }
            else
            {
                // If we found a txn or a stream then push it
                while (!vStream.empty() || ptx)
                {
                    if (!vStream.empty())
                    {
                        pStream = vStream.back();
                        vStream.pop_back();
                    }

                    if (pfrom->txConcat)
                    {
                        if (pStream)
                            ss << *pStream;
                        else
                            ss << *ptx;

                        // Send the concatenated txns if we're over the limit. We don't want to batch
                        // too many and end up delaying the send.
                        if (ss.size() > MAX_TXN_BATCH_SIZE)
                        {
                            // TODO, what if this is the last one
                            pfrom->PushMessageWithCookie(NetMsgType::TX, msgCookie + replyCookieCount, ss);
                            replyCookieCount++;
                            ss.clear();
                        }
                    }
                    else
                    {
                        // Or if this is not a peer that supports
                        // concatenation then send the transaction right away.
                        if (pStream)
                            pfrom->PushMessageWithCookie(NetMsgType::TX, msgCookie + replyCookieCount, *pStream);
                        else
                            pfrom->PushMessageWithCookie(NetMsgType::TX, msgCookie + replyCookieCount, ptx);
                        replyCookieCount++;
                    }
                    pfrom->txsSent += 1;
                    ptx = nullptr;
                }
            }
        }

        // Track requests for our stuff.
        GetMainSignals().Inventory(inv.hash);

        // Increment at the end in case we break anywhere in this loop because if we increment before
        // processing an inv then when we erase further downstream we may end up erasing an inv
        // we never actually processed.
        it++;
    }
    // Send the batched transactions if any to send.
    if (!ss.empty())
    {
        uint32_t lastMsgCookie = msgCookie | 0xFFFF;
        pfrom->PushMessageWithCookie(NetMsgType::TX, lastMsgCookie, ss);
        replyCookieCount++;
    }

    // Erase all messages inv's we processed
    vInv.erase(vInv.begin(), it);

    if (!vNotFound.empty())
    {
        // Let the peer know that we didn't find what it asked for, so it doesn't
        // have to wait around forever. Currently only SPV clients actually care
        // about this message: it's needed when they are recursively walking the
        // dependencies of relevant unconfirmed transactions. SPV clients want to
        // do that because they want to know about (and store and rebroadcast and
        // risk analyze) the dependencies of transactions relevant to them, without
        // having to download the entire memory pool.
        pfrom->PushMessageWithCookie(NetMsgType::NOTFOUND, msgCookie | 0xFFFF, vNotFound);
    }
}

template <class T>
void static ProcessExtGetData(CNode *pfrom,
    const Consensus::Params &consensusParams,
    std::deque<std::pair<T, uint32_t> > &vInv)
{
    std::vector<T> vNotFound;
    CDataStream ss(SER_NETWORK, PROTOCOL_VERSION);

    auto it = vInv.begin();
    uint32_t replyCookieCount = 0;
    int cur = 0;
    int sz = vInv.size();
    uint32_t msgCookie = 0;
    LOG(NET, "Processing %d EXTINVs\n", sz);
    while (it != vInv.end())
    {
        // If the cookie has changed, restart the reply count.  This might happen in weird cases where the invs
        // were deferred.
        uint32_t nextMsgCookie = it->second;
        if (nextMsgCookie != msgCookie)
            replyCookieCount = 0;
        msgCookie = nextMsgCookie;
        cur++;

        // NOTE: commented out for now, but may be used in the future.
        // bool lastInv = (sz == cur);
        // uint32_t oneReplyPerInvCookie = (lastInv ? (msgCookie | 0xFFFF) : (msgCookie + replyCookieCount));

        // Don't bother if send buffer is too full to respond anyway except we always consume at least one inv.
        // Otherwise we could get into an infinite loop where we never
        // process a message that's too large for our send buffer size.
        if (cur != 0 && pfrom->nSendSize >= nMaxSendBufferSize + ss.size())
        {
            break;
        }
        if (shutdown_threads.load() == true)
        {
            // I am in the middle of communicating with this node, so I want to make sure it gets an explicit close
            // rather than timing out since I am breaking the request/reply protocol by quitting.
            pfrom->CloseSocketDisconnect();
            return;
        }

        // start processing inventory here
        const T &inv = it->first;

        if (inv.type == MSG_TOKENINFO)
        {
            // Get the token info associated with the group id and return
            // a tokeninfo object.
            CGroupTokenID groupID(inv.hash);
            const std::vector<std::string> vInfo = tokencache.GetTokenDesc(groupID);
            if (vInfo.size() >= DESC_ENTRY_SIZE)
            {
                CTokenInfo ti;
                ti.groupId = groupID.bytes();
                ti.name = std::move(vInfo[1]);
                ti.ticker = std::move(vInfo[0]);
                ti.url = std::move(vInfo[2]);
                ti.hash = uint256S(vInfo[3]);
                try
                {
                    ti.decimals = std::stoi(vInfo[4]);
                }
                catch (...)
                {
                    pfrom->PushMessageWithCookie(NetMsgType::REJECT, replyCookieCount,
                        std::string(NetMsgType::TOKENINFO), REJECT_INVALID,
                        strprintf(
                            "decimals information requested was not decoded for group id: %s", inv.ToString().c_str()),
                        inv.hash);
                    continue;
                }
                ti.genesisAddress = tokenmint.GetTokenGenesis(groupID);
                pfrom->PushMessage(NetMsgType::TOKENINFO, ti);
            }
            else
            {
                pfrom->PushMessageWithCookie(NetMsgType::REJECT, replyCookieCount, std::string(NetMsgType::TOKENINFO),
                    REJECT_INVALID,
                    strprintf("token information requested was not found for id: %s", inv.ToString().c_str()),
                    inv.hash);
            }
            replyCookieCount++;
        }
        else if (inv.type == MSG_EXT_TX)
        {
            std::shared_ptr<CDataStream> pStream = nullptr;
            CTransactionRef ptx = nullptr;
            std::vector<CTransactionRef> vTx;
            std::vector<std::shared_ptr<CDataStream> > vStream;

            // Check for stream from relay memory first since this is the most efficient data to send.
            CDataStream ss1(0, 0);
            for (auto c : inv.hash)
                ss1 << c;
            uint64_t cheaphash = ser_readdata64(ss1);
            vStream = maprelay.Find(cheaphash);
#ifdef DEBUG
            // This -debug test code allows us to fall through to check the different types of searches
            // for transactions. If we didn't do this we very rarely, if ever, check the commit queue
            // or the txpool for transactions since we almost always get them from the relay map.
            static std::atomic<uint32_t> count{0};
            count++;
            if (count >= 10)
            {
                vStream.clear();
            }
#endif
            if (vStream.empty())
            {
                vTx = CommitQGet(cheaphash);
#ifdef DEBUG
                if (count >= 15)
                {
                    count = 0;
                    vTx.clear();
                }
#endif
                if (vTx.empty())
                {
                    vTx = mempool.get(cheaphash);
                }
            }

            if (vStream.empty() && vTx.empty())
            {
                vNotFound.push_back(inv);
            }
            else
            {
                // If we found a txn or a stream then push it
                while (!vStream.empty() || !vTx.empty())
                {
                    if (!vStream.empty())
                    {
                        pStream = vStream.back();
                        vStream.pop_back();
                    }
                    else if (!vTx.empty())
                    {
                        ptx = vTx.back();
                        vTx.pop_back();
                    }

                    if (pfrom->txConcat)
                    {
                        if (pStream)
                        {
                            ss << *pStream;
                        }
                        else
                        {
                            ss << *ptx;
                        }
                        // Send the concatenated txns if we're over the limit. We don't want to batch
                        // too many and end up delaying the send.
                        if (ss.size() > MAX_TXN_BATCH_SIZE)
                        {
                            // TODO, what if this is the last one
                            pfrom->PushMessageWithCookie(NetMsgType::TX, msgCookie + replyCookieCount, ss);
                            replyCookieCount++;
                            ss.clear();
                        }
                    }
                    else
                    {
                        // Or if this is not a peer that supports
                        // concatenation then send the transaction right away.
                        if (pStream)
                            pfrom->PushMessageWithCookie(NetMsgType::TX, msgCookie + replyCookieCount, *pStream);
                        else
                            pfrom->PushMessageWithCookie(NetMsgType::TX, msgCookie + replyCookieCount, ptx);
                        replyCookieCount++;
                    }

                    pfrom->txsSent += 1;
                    pStream = nullptr;
                }
            }
        }
        else
        {
            vNotFound.push_back(inv);
        }

        // TODO: Track requests for our extended getdata.  Currently we don't need to do
        // this because we don't share EXTINV's with other peers yet.
        // GetMainSignals().Inventory(inv.hash);

        // Increment at the end in case we break anywhere in this loop because if we increment before
        // processing an inv then when we erase further downstream we may end up erasing an inv
        // we never actually processed.
        it++;
    }

    // Send the batched transactions if any to send.
    if (!ss.empty())
    {
        uint32_t lastMsgCookie = msgCookie | 0xFFFF;
        pfrom->PushMessageWithCookie(NetMsgType::TX, lastMsgCookie, ss);
        replyCookieCount++;
    }

    // Erase all messages extinv's we processed
    vInv.erase(vInv.begin(), it);

    if (!vNotFound.empty())
    {
        // Let the peer know that we didn't find what it asked for, so it doesn't
        // have to wait around forever. Currently only SPV clients actually care
        // about this message: it's needed when they are recursively walking the
        // dependencies of relevant unconfirmed transactions. SPV clients want to
        // do that because they want to know about (and store and rebroadcast and
        // risk analyze) the dependencies of transactions relevant to them, without
        // having to download the entire memory pool.
        pfrom->PushMessageWithCookie(NetMsgType::NOTFOUND, msgCookie | 0xFFFF, vNotFound);
    }
}

template <class T>
static bool processInvMsgs(CNode *pfrom, CDataStream &vRecv, std::vector<T> &vInv)
{
    vRecv >> vInv;
    size_t vInvSize = vInv.size();
    LOG(NET, "Received INV list of size %d\n", vInvSize);

    // Message Consistency Checking
    //   Check size == 0 to be intolerant of an empty and useless request.
    //   Validate that INVs are a valid type and not null.
    if (vInvSize > MAX_INV_SZ || vInv.empty())
    {
        dosMan.Misbehaving(pfrom, 20);
        return error("message inv size() = %u", vInvSize);
    }

    // Allow whitelisted peers to send data other than blocks in blocks only mode if whitelistrelay is true
    if (pfrom->fWhitelisted && fWhiteListRelay)
        fBlocksOnly = false;

    for (unsigned int nInv = 0; nInv < vInvSize; nInv++)
    {
        if (shutdown_threads.load() == true)
            return false;

        const T &inv = vInv[nInv];
        if (!((inv.type == MSG_TX) || (inv.type == MSG_BLOCK) || inv.type == MSG_DOUBLESPENDPROOF))
        {
            LOG(NET, "message inv invalid type = %u hash %s", inv.type, inv.hash.ToString());
            return false;
        }
        else if (inv.hash.IsNull())
        {
            LOG(NET, "message inv has null hash %s", inv.type, inv.hash.ToString());
            return false;
        }

        if (inv.type == MSG_BLOCK)
        {
            bool fAlreadyHaveBlock = AlreadyHaveBlock(inv);
            LOG(NET, "got BLOCK inv: %s  %s peer=%d\n", inv.ToString(), fAlreadyHaveBlock ? "have" : "new", pfrom->id);

            requester.UpdateBlockAvailability(pfrom->GetId(), inv.hash);
            // RE !IsInitialBlockDownload(): We do not want to get the block if the system is executing the initial
            // block download because
            // blocks are stored in block files in the order of arrival.  So grabbing blocks "early" will cause new
            // blocks to be sprinkled
            // throughout older block files.  This will stop those files from being pruned.
            // !IsInitialBlockDownload() can be removed if
            // a better block storage system is devised.
            if ((!fAlreadyHaveBlock && !IsInitialBlockDownload()) ||
                (!fAlreadyHaveBlock && Params().NetworkIDString() == "regtest"))
            {
                // Since we now only rely on headers for block requests, if we get an INV from an older node or
                // if there was a very large re-org which resulted in a revert to block announcements via INV,
                // we will instead request the header rather than the block.  This is safer and prevents an
                // attacker from sending us fake INV's for blocks that do not exist or try to get us to request
                // and download fake blocks.
                pfrom->PushMessage(NetMsgType::GETHEADERS, chainActive.GetLocator(pindexBestHeader), inv.hash);
            }
            else
            {
                LOG(NET,
                    "skipping request of block %s.  already have: %d  importing: %d  reindex: %d  "
                    "isChainNearlySyncd: %d\n",
                    inv.hash.ToString(), fAlreadyHaveBlock, fImporting, fReindex, IsChainNearlySyncd());
            }
        }
        else if (inv.type == MSG_TX)
        {
            bool fMayAlreadyHaveTx = TxMayAlreadyHave(inv.type, inv.hash);
            // LOG(NET, "got inv: %s  %d peer=%s\n", inv.ToString(), fMayAlreadyHaveTx ? "have" : "new",
            // pfrom->GetLogName());
            LOG(NET, "got inv: %s  have: %d peer=%s\n", inv.ToString(), fMayAlreadyHaveTx, pfrom->GetLogName());

            pfrom->AddInventoryKnown(inv);
            if (fBlocksOnly)
            {
                LOG(NET, "transaction (%s) inv sent in violation of protocol peer=%d\n", inv.hash.ToString(),
                    pfrom->id);
            }
            // RE !IsInitialBlockDownload(): during IBD, its a waste of bandwidth to grab transactions, they will
            // likely be included in blocks that we IBD download anyway.  This is especially important as
            // transaction volumes increase.
            else if (!fMayAlreadyHaveTx && !IsInitialBlockDownload())
            {
                requester.AskFor(CInv(inv.type, inv.hash), pfrom);
            }
        }
        else if (inv.type == MSG_DOUBLESPENDPROOF && doubleSpendProofs.Value() == true)
        {
            std::vector<T> vGetData;
            vGetData.push_back(inv);
            pfrom->PushMessage(NetMsgType::GETDATA, vGetData);
        }

        // Track requests for our stuff.
        GetMainSignals().Inventory(inv.hash);
    }

    return true;
}

static void handleAddressAfterInit(CNode *pfrom)
{
    if (!pfrom->fInbound)
    {
        // Advertise our address
        if (fListen && !IsInitialBlockDownload())
        {
            CAddress addr = GetLocalAddress(&pfrom->addr);
            FastRandomContext insecure_rand;
            if (addr.IsRoutable())
            {
                LOG(NET, "ProcessMessages: advertising address %s\n", addr.ToString());
                pfrom->PushAddress(addr, insecure_rand);
            }
            else if (IsPeerAddrLocalGood(pfrom))
            {
                addr.SetIP(pfrom->addrLocal);
                LOG(NET, "ProcessMessages: advertising address %s\n", addr.ToString());
                pfrom->PushAddress(addr, insecure_rand);
            }
        }
        // Get recent addresses
        pfrom->fGetAddr = true;
        pfrom->PushMessage(NetMsgType::GETADDR);
        addrman.Good(pfrom->addr);
    }
    else
    {
        if (((CNetAddr)pfrom->addr) == (CNetAddr)pfrom->addrFrom_advertised)
        {
            addrman.Add(pfrom->addrFrom_advertised, pfrom->addrFrom_advertised);
            addrman.Good(pfrom->addrFrom_advertised);
        }
    }
}

static void enableSendHeaders(CNode *pfrom)
{
    // Tell our peer we prefer to receive headers rather than inv's
    // We send this to non-NODE NETWORK peers as well, because even
    // non-NODE NETWORK peers can announce blocks (such as pruning
    // nodes)
    pfrom->PushMessage(NetMsgType::SENDHEADERS);
}

static void enableCompactBlocks(CNode *pfrom)
{
    // Tell our peer that we support compact blocks
    if (IsCompactBlocksEnabled())
    {
        bool fHighBandwidth = false;
        uint64_t nVersion = 1;
        pfrom->PushMessage(NetMsgType::SENDCMPCT, fHighBandwidth, nVersion);
    }
}

bool ProcessMessage(CNode *pfrom,
    std::string strCommand,
    uint32_t msgCookie,
    CDataStream &vRecv,
    int64_t nStopwatchTimeReceived)
{
    int64_t receiptTime = GetTime();
    const CChainParams &chainparams = Params();
    unsigned int msgSize = vRecv.size(); // BU for statistics
    UpdateRecvStats(pfrom, strCommand, msgSize, nStopwatchTimeReceived);
    LOG(NET, "received: %s (%u bytes) peer=%s\n", SanitizeString(strCommand), msgSize, pfrom->GetLogName());

    // if -dropmessagetest is enabled then randomly drop the processing of a message
    if (nDropMessages && GetRand(nDropMessages) == 0)
    {
        LOGA("dropmessagestest DROPPING RECV MESSAGE\n");
        return true;
    }

    if (!(nLocalServices & NODE_BLOOM) &&
        (strCommand == NetMsgType::FILTERLOAD || strCommand == NetMsgType::FILTERADD ||
            strCommand == NetMsgType::FILTERCLEAR))
    {
        LOG(NET, "Inconsistent bloom filter settings peer %s\n", pfrom->GetLogName());
        pfrom->fDisconnect = true;
        return false;
    }

    bool grapheneVersionCompatible = true;
    try
    {
        NegotiateGrapheneVersion(pfrom);
        NegotiateFastFilterSupport(pfrom);
    }
    catch (const std::runtime_error &e)
    {
        grapheneVersionCompatible = false;
    }

    // ------------------------- BEGIN INITIAL COMMAND SET PROCESSING
    if (strCommand == NetMsgType::VERSION)
    {
        int64_t nTime;
        CAddress addrMe;
        uint64_t nNonce = 1;
        vRecv >> pfrom->nVersion >> pfrom->nServices >> nTime >> addrMe;

        // Update thin type peer counters. This should be at the top here before we have any
        // potential disconnects, because on disconnect the counters will then get decremented.
        thinrelay.AddPeers(pfrom);

        if (pfrom->nVersion < MIN_PEER_PROTO_VERSION)
        {
            // ban peers older than this proto version
            pfrom->PushMessageWithCookie(NetMsgType::REJECT, msgCookie | 0xFFFF, strCommand, REJECT_OBSOLETE,
                strprintf("Protocol Version must be %d or greater", MIN_PEER_PROTO_VERSION));
            dosMan.Misbehaving(pfrom, 100);
            return error("Using obsolete protocol version %i - banning peer=%s version=%s", pfrom->nVersion,
                pfrom->GetLogName(), pfrom->cleanSubVer);
        }

        if (!vRecv.empty())
            vRecv >> pfrom->addrFrom_advertised >> nNonce;
        if (!vRecv.empty())
        {
            vRecv >> LIMITED_STRING(pfrom->strSubVer, MAX_SUBVERSION_LENGTH);
            pfrom->cleanSubVer = SanitizeString(pfrom->strSubVer);

            // Track the user agent string
            {
                LOCK(cs_mapInboundConnectionTracker);

                // Remove a random entry if we've gotten too big.
                if (mapInboundConnectionTracker.size() >= MAX_INBOUND_CONNECTIONS_TRACKED)
                {
                    size_t nIndex = GetRandInt(mapInboundConnectionTracker.size() - 1);
                    auto rand_iter = std::next(mapInboundConnectionTracker.begin(), nIndex);
                    mapInboundConnectionTracker.erase(rand_iter);
                }

                // Add the subver string.
                mapInboundConnectionTracker[(CNetAddr)pfrom->addr].userAgent = pfrom->cleanSubVer;
            }
        }
        if (!vRecv.empty())
            vRecv >> pfrom->nStartingHeight;
        if (!vRecv.empty())
            vRecv >> pfrom->fRelayTxes; // set to true after we get the first filter* message
        else
            pfrom->fRelayTxes = true;

        // Disconnect if we connected to ourself
        if (nNonce == nLocalHostNonce && nNonce > 1)
        {
            LOGA("connected to self at %s, disconnecting\n", pfrom->addr.ToString());
            pfrom->fDisconnect = true;
            return true;
        }

        pfrom->addrLocal = addrMe;
        if (pfrom->fInbound && addrMe.IsRoutable())
        {
            SeenLocal(addrMe);
        }

        // Be shy and don't send version until we hear
        if (pfrom->fInbound)
            pfrom->PushVersion();


        // set nodes not relaying blocks and tx and not serving (parts) of the historical blockchain as "clients"
        pfrom->fClient = (!(pfrom->nServices & NODE_NETWORK) && !(pfrom->nServices & NODE_NETWORK_LIMITED));

        // set nodes not capable of serving the complete blockchain history as "limited nodes"
        pfrom->m_limited_node = (!(pfrom->nServices & NODE_NETWORK) && (pfrom->nServices & NODE_NETWORK_LIMITED));

        // Potentially mark this peer as a preferred download peer.
        UpdatePreferredDownload(pfrom);

        // only send extversion message if both peers are using the protocol
        if ((nLocalServices & NODE_EXTVERSION) && (pfrom->nServices & NODE_EXTVERSION) &&
            extVersionEnabled.Value() == true)
        {
            // BU expedited procecessing requires the exchange of the listening port id
            // The former BUVERSION message has now been integrated into the xmap field in CExtversionMessage.

            // prepare extversion message. This must be sent before we send a verack message in the new spec
            CExtversionMessage xver;
            xver.set_u64c(XVer::EXTVERSION_VERSION_KEY, EXTVERSION_VERSION_VALUE);
            xver.set_u64c(XVer::BU_LISTEN_PORT, GetListenPort());
            xver.set_u64c(XVer::BU_MSG_IGNORE_CHECKSUM, 1); // we will ignore 0 value msg checksums
            xver.set_u64c(XVer::BU_GRAPHENE_MAX_VERSION_SUPPORTED, grapheneMaxVersionSupported.Value());
            xver.set_u64c(XVer::BU_GRAPHENE_MIN_VERSION_SUPPORTED, grapheneMinVersionSupported.Value());
            xver.set_u64c(XVer::BU_GRAPHENE_FAST_FILTER_PREF, grapheneFastFilterCompatibility.Value());
            xver.set_u64c(XVer::BU_MEMPOOL_SYNC, syncMempoolWithPeers.Value());
            xver.set_u64c(XVer::BU_MEMPOOL_SYNC_MAX_VERSION_SUPPORTED, mempoolSyncMaxVersionSupported.Value());
            xver.set_u64c(XVer::BU_MEMPOOL_SYNC_MIN_VERSION_SUPPORTED, mempoolSyncMinVersionSupported.Value());
            xver.set_u64c(XVer::BU_XTHIN_VERSION, 2); // xthin version
            xver.set_u64c(XVer::BU_TXN_CONCATENATION, 1);
            if (capdPoolSize.Value() != 0)
                xver.set_u64c(XVer::BU_CAPD_VERSION, 1); // capd version
            xver.set_u64c(XVer::BU_ADDRV2_SUPPORT, 1); // signal support for ADDRV2 messages
            xver.set_u64c(XVer::BU_INV2_SUPPORT, 1); // signal support for INV2 messages

            electrum::set_extversion_flags(xver, chainparams.NetworkIDString());

            pfrom->extversionExpected = true;
            pfrom->PushMessageWithCookie(NetMsgType::EXTVERSION, msgCookie | 0xFFFF, xver);
        }
        else
        {
            // Send VERACK handshake message
            pfrom->PushMessageWithCookie(NetMsgType::VERACK, msgCookie | 0xFFFF);
        }

        LOG(NET, "receive version message: %s: version %d, blocks=%d, us=%s, peer=%s\n", pfrom->cleanSubVer,
            pfrom->nVersion, pfrom->nStartingHeight, addrMe.ToString(), pfrom->GetLogName());

        int64_t nTimeOffset = nTime - GetTime();
        pfrom->nTimeOffset = nTimeOffset;
        AddTimeData(pfrom->addr, nTimeOffset);

        // Feeler connections exist only to verify if address is online.
        if (pfrom->fFeeler)
        {
            // Should never occur but if it does correct the value.
            // We can't have an inbound "feeler" connection, so the value must be improperly set.
            DbgAssert(pfrom->fInbound == false, pfrom->fFeeler = false);
            if (pfrom->fInbound == false)
            {
                LOG(NET, "Disconnecting feeler to peer %s\n", pfrom->GetLogName());
                pfrom->fDisconnect = true;
            }
        }
    }

    else if ((pfrom->nVersion == 0 || pfrom->tVersionSent < 0) && !pfrom->fWhitelisted)
    {
        // Must have a version message before anything else
        dosMan.Misbehaving(pfrom, 1);
        pfrom->fDisconnect = true;
        return error("%s receieved before VERSION message - disconnecting peer=%s", strCommand, pfrom->GetLogName());
    }

    else if (strCommand == NetMsgType::EXTVERSION && extVersionEnabled.Value() == true)
    {
        // set expected to false, we got the message
        pfrom->extversionExpected = false;
        if (pfrom->fSuccessfullyConnected == true)
        {
            dosMan.Misbehaving(pfrom, 1);
            pfrom->fDisconnect = true;
            return error("odd peer behavior: received verack message before extversion, disconnecting \n");
        }

        LOCK(pfrom->cs_extversion);
        vRecv >> pfrom->extversion;

        if (pfrom->addrFromPort != 0)
        {
            LOG(NET,
                "Encountered odd node that sent BUVERSION before XVERSION. Ignoring duplicate addrFromPort "
                "setting. peer=%s version=%s\n",
                pfrom->GetLogName(), pfrom->cleanSubVer);
        }

        pfrom->ReadConfigFromExtversion();

        // Send VERACK handshake message
        pfrom->PushMessageWithCookie(NetMsgType::VERACK, msgCookie | 0xFFFF);
    }

    else if (!pfrom->fSuccessfullyConnected && GetTime() - pfrom->tVersionSent > VERACK_TIMEOUT &&
             pfrom->tVersionSent >= 0)
    {
        // If verack is not received within timeout then disconnect.
        // The peer may be slow so disconnect them only, to give them another chance if they try to re-connect.
        // If they are a bad peer and keep trying to reconnect and still do not VERACK, they will eventually
        // get banned by the connection slot algorithm which tracks disconnects and reconnects.
        pfrom->fDisconnect = true;
        LOG(NET, "ERROR: disconnecting - VERACK not received within %d seconds for peer=%s version=%s\n",
            VERACK_TIMEOUT, pfrom->GetLogName(), pfrom->cleanSubVer);

        // update connection tracker which is used by the connection slot algorithm.
        LOCK(cs_mapInboundConnectionTracker);
        CNetAddr ipAddress = (CNetAddr)pfrom->addr;
        mapInboundConnectionTracker[ipAddress].nEvictions += 1;
        mapInboundConnectionTracker[ipAddress].nLastEvictionTime = GetTime();
        mapInboundConnectionTracker[ipAddress].userAgent = pfrom->cleanSubVer;

        return true; // return true so we don't get any process message failures in the log.
    }

    else if (strCommand == NetMsgType::VERACK)
    {
        if (pfrom->fSuccessfullyConnected == true)
        {
            dosMan.Misbehaving(pfrom, 1);
            pfrom->fDisconnect = true;
            return error("duplicate verack messages");
        }
        pfrom->SetRecvVersion(std::min(pfrom->nVersion, PROTOCOL_VERSION));

        if (pfrom->extversionExpected.load() == true)
        {
            // if we expected extversion but got a verack it is possible there is a service bit
            // mismatch so we should send a verack response because the peer might not
            // support extversion

            // Send VERACK handshake message
            pfrom->PushMessageWithCookie(NetMsgType::VERACK, msgCookie | 0xFFFF);
        }

        handleAddressAfterInit(pfrom);
        enableSendHeaders(pfrom);
        enableCompactBlocks(pfrom);

        // Tell the peer what maximum xthin bloom filter size we will consider acceptable.
        // FIXME: integrate into extversion as well
        if (pfrom->ThinBlockCapable() && IsThinBlocksEnabled())
        {
            pfrom->PushMessage(NetMsgType::FILTERSIZEXTHIN, nXthinBloomFilterSize);
        }

        // This step done after final handshake
        connmgr->CheckAndRequestExpeditedBlocks(pfrom);

        pfrom->fSuccessfullyConnected = true;
    }

    else if (strCommand == NetMsgType::XUPDATE)
    {
        CExtversionMessage xUpdate;
        vRecv >> xUpdate;
        // check for peer trying to change non-changeable key
        for (auto entry : xUpdate.xmap)
        {
            if (XVer::IsChangableKey(entry.first))
            {
                LOCK(pfrom->cs_extversion);
                pfrom->extversion.xmap[entry.first] = xUpdate.xmap[entry.first];
            }
        }
    }

    else if (strCommand == NetMsgType::SENDHEADERS)
    {
        CNodeStateAccessor(nodestate, pfrom->GetId())->fPreferHeaders = true;
    }

    else if (strCommand == NetMsgType::FILTERSIZEXTHIN)
    {
        if (pfrom->ThinBlockCapable())
        {
            uint32_t nSize = 0;
            vRecv >> nSize;
            pfrom->nXthinBloomfilterSize.store(nSize);

            // As a safeguard don't allow a smaller max bloom filter size than the default max size.
            if (!pfrom->nXthinBloomfilterSize || (pfrom->nXthinBloomfilterSize < SMALLEST_MAX_BLOOM_FILTER_SIZE))
            {
                pfrom->PushMessageWithCookie(NetMsgType::REJECT, msgCookie | 0xFFFF, strCommand, REJECT_INVALID,
                    std::string("filter size was too small"));
                LOG(NET, "Disconnecting %s: bloom filter size too small\n", pfrom->GetLogName());
                pfrom->fDisconnect = true;
                return false;
            }
        }
        else
        {
            pfrom->fDisconnect = true;
            return false;
        }
    }

    else if (strCommand == NetMsgType::PING)
    {
        uint64_t nonce = 0;
        vRecv >> nonce;
        // although PONG was enabled in BIP31, all clients should handle it at this point
        // and unknown messages are silently dropped.  So for simplicity, always respond with PONG
        // Echo the message back with the nonce. This allows for two useful features:
        //
        // 1) A remote node can quickly check if the connection is operational
        // 2) Remote nodes can measure the latency of the network thread. If this node
        //    is overloaded it won't respond to pings quickly and the remote node can
        //    avoid sending us more work, like chain download requests.
        //
        // The nonce stops the remote getting confused between different pings: without
        // it, if the remote node sends a ping once per second and this node takes 5
        // seconds to respond to each, the 5th ping the remote sends would appear to
        // return very quickly.
        pfrom->PushMessageWithCookie(NetMsgType::PONG, msgCookie | 0xFFFF, nonce);
    }

    // ------------------------- END INITIAL COMMAND SET PROCESSING
    else if (!pfrom->fSuccessfullyConnected)
    {
        LOG(NET, "Ignoring command %s that comes in before initial handshake is finished. peer=%s version=%s\n",
            strCommand, pfrom->GetLogName(), pfrom->cleanSubVer);

        // Ignore any other commands early in the handshake
        return false;
    }
    else if (strCommand == NetMsgType::ADDR || strCommand == NetMsgType::ADDRV2)
    {
        std::vector<CAddress> vAddr;
        vRecv >> vAddr;

        if (vAddr.size() > 1000)
        {
            dosMan.Misbehaving(pfrom, 20);
            return error("%s message size() = %u", strCommand, vAddr.size());
        }

        // To avoid malicious flooding of our address table, only allow unsolicited ADDR messages to
        // insert the connecting IP.  We need to allow this IP to be inserted, or there is no way for that node
        // to tell the network about itself if its behind a NAT.

        // Digression about how things work behind a NAT:
        //     Node A periodically ADDRs node B with the address that B reported to A as A's own
        //     address (in the VERSION message).
        //
        // The purpose of using exchange here is to atomically set to false and also get whether I asked for an addr
        if (!pfrom->fGetAddr.exchange(0) && pfrom->fInbound)
        {
            bool reportedOwnAddr = false;
            CAddress ownAddr;
            for (CAddress &addr : vAddr)
            {
                // server listen port will be different.  We want to compare IPs and then use provided port
                if ((CNetAddr)addr == (CNetAddr)pfrom->addr)
                {
                    ownAddr = addr;
                    reportedOwnAddr = true;
                    break;
                }
            }
            if (reportedOwnAddr)
            {
                vAddr.resize(1); // Get rid of every address the remote node tried to inject except itself.
                vAddr[0] = ownAddr;
            }
            else
            {
                // Today unsolicited ADDRs are not illegal, but we should consider misbehaving on this (if we add logic
                // to unmisbehaving over time), because a few unsolicited ADDRs are ok from a DOS perspective but lots
                // are not.
                // dosMan.Misbehaving(pfrom, 1);
                return true; // We don't want to process any other addresses, but giving them is not an error
            }
        }

        // Store the new addresses
        std::vector<CAddress> vAddrOk;
        int64_t nNow = GetAdjustedTime();
        int64_t nSince = nNow - 10 * 60;
        FastRandomContext insecure_rand;
        for (CAddress &addr : vAddr)
        {
            if (shutdown_threads.load() == true)
            {
                return false;
            }

            if (addr.nTime <= 100000000 || addr.nTime > nNow + 10 * 60)
                addr.nTime = nNow - 5 * 24 * 60 * 60;
            pfrom->AddAddressKnown(addr);
            bool fReachable = IsReachable(addr);
            if (addr.nTime > nSince && vAddr.size() <= 10 && addr.IsRoutable())
            {
                // Relay to a limited number of other nodes
                {
                    LOCK(cs_vNodes);
                    // Use deterministic randomness to send to the same nodes for 24 hours
                    // at a time so the addrKnowns of the chosen nodes prevent repeats
                    static uint256 hashSalt;
                    if (hashSalt.IsNull())
                        hashSalt = GetRandHash();
                    uint64_t hashAddr = addr.GetHash();
                    uint256 hashRand = ArithToUint256(
                        UintToArith256(hashSalt) ^ (hashAddr << 32) ^ ((GetTime() + hashAddr) / (24 * 60 * 60)));
                    hashRand = Hash(BEGIN(hashRand), END(hashRand));
                    std::multimap<uint256, CNode *> mapMix;
                    for (CNode *pnode : vNodes)
                    {
                        unsigned int nPointer;
                        memcpy(&nPointer, &pnode, sizeof(nPointer));
                        uint256 hashKey = ArithToUint256(UintToArith256(hashRand) ^ nPointer);
                        hashKey = Hash(BEGIN(hashKey), END(hashKey));
                        mapMix.insert(std::make_pair(hashKey, pnode));
                    }
                    int nRelayNodes = fReachable ? 2 : 1; // limited relaying of addresses outside our network(s)
                    for (std::multimap<uint256, CNode *>::iterator mi = mapMix.begin();
                         mi != mapMix.end() && nRelayNodes-- > 0; ++mi)
                        ((*mi).second)->PushAddress(addr, insecure_rand);
                }
            }
            // Do not store addresses outside our network
            if (fReachable)
                vAddrOk.push_back(addr);
        }
        addrman.Add(vAddrOk, pfrom->addr, 2 * 60 * 60);
        if (pfrom->fOneShot)
        {
            LOG(NET, "Disconnecting %s: one shot\n", pfrom->GetLogName());
            pfrom->fDisconnect = true;
        }
    }

    // Ignore this message if sent from a node advertising a version earlier than the first CB release (70014)
    else if (strCommand == NetMsgType::SENDCMPCT)
    {
        bool fHighBandwidth = false;
        uint64_t nVersion = 0;
        vRecv >> fHighBandwidth >> nVersion;

        // The network currently only supports version 1
        // May need to be updated in the future if other clients deploy a new version
        pfrom->fSupportsCompactBlocks = nVersion == 1;

        // Increment compact block peer counter.
        thinrelay.AddCompactBlockPeer(pfrom);
    }

    else if (strCommand == NetMsgType::INV)
    {
        // This message is not part of a request/response so even if processing this results in a request,
        // we will not "reply" with the passed cookie.

        if (fImporting || fReindex)
            return true;

        if (pfrom->fPeerWantsINV2)
        {
            std::vector<CInv2> vInv;
            return processInvMsgs(pfrom, vRecv, vInv);
        }
        else
        {
            std::vector<CInv> vInv;
            return processInvMsgs(pfrom, vRecv, vInv);
        }
    }

    // Peer wants old style inventory (CInv) messages
    else if (strCommand == NetMsgType::GETDATA && !pfrom->fPeerWantsINV2)
    {
        if (fImporting || fReindex)
        {
            LOG(NET, "received getdata from %s but importing\n", pfrom->GetLogName());
            return true;
        }

        std::vector<CInv> vInv;
        vRecv >> vInv;
        // check size == 0 to be intolerant of an empty and useless request
        if ((vInv.size() > MAX_INV_SZ) || (vInv.size() == 0))
        {
            dosMan.Misbehaving(pfrom, 20);
            return error("message getdata size() = %u", vInv.size());
        }

        // Validate that INVs are a valid type
        std::deque<std::pair<CInv, uint32_t> > invDeque;
        for (unsigned int nInv = 0; nInv < vInv.size(); nInv++)
        {
            const CInv &inv = vInv[nInv];
            if (!((inv.type == MSG_TX) || (inv.type == MSG_BLOCK) || (inv.type == MSG_FILTERED_BLOCK) ||
                    (inv.type == MSG_CMPCT_BLOCK) || inv.type == MSG_DOUBLESPENDPROOF))
            {
                dosMan.Misbehaving(pfrom, 20, BanReasonInvalidInventory);
                return error("message inv invalid type = %u", inv.type);
            }

            // Make basic checks
            if (inv.type == MSG_CMPCT_BLOCK)
            {
                if (!requester.CheckForRequestDOS(pfrom, chainparams))
                    return false;
            }

            invDeque.push_back(std::pair<CInv, uint32_t>(inv, msgCookie));
            // These will be deferred, so we need to provide a different cookie for each one.
            if ((inv.type == MSG_BLOCK) || (inv.type == MSG_CMPCT_BLOCK))
                msgCookie++;
        }

        if (fDebug || (invDeque.size() != 1))
            LOG(NET, "received getdata (%u invsz) peer=%s\n", invDeque.size(), pfrom->GetLogName());

        if ((fDebug && invDeque.size() > 0) || (invDeque.size() == 1))
            LOG(NET, "received getdata for: %s peer=%s\n", invDeque[0].first.ToString(), pfrom->GetLogName());

        // Run process getdata and process as much of the getdata's as we can before taking the lock
        // and appending the remainder to the vRecvGetData queue.
        ProcessGetData(pfrom, chainparams.GetConsensus(), invDeque);
        if (!invDeque.empty())
        {
            LOCK(pfrom->csRecvGetData);
            pfrom->vRecvGetData.insert(pfrom->vRecvGetData.end(), invDeque.begin(), invDeque.end());
        }
    }

    // Peer wants new style inventory (CInv2) messages
    else if (strCommand == NetMsgType::GETDATA && pfrom->fPeerWantsINV2)
    {
        if (fImporting || fReindex)
        {
            LOG(NET, "received getdata from %s but importing\n", pfrom->GetLogName());
            return true;
        }

        std::vector<CInv2> vInv;
        vRecv >> vInv;
        // check size == 0 to be intolerant of an empty and useless request
        if ((vInv.size() > MAX_INV_SZ) || (vInv.size() == 0))
        {
            dosMan.Misbehaving(pfrom, 20);
            return error("message getdata size() = %u", vInv.size());
        }

        // Validate that INVs are a valid type
        std::deque<std::pair<CInv2, uint32_t> > invDeque;
        for (unsigned int nInv = 0; nInv < vInv.size(); nInv++)
        {
            const CInv2 &inv = vInv[nInv];
            if (!((inv.type == MSG_TX) || (inv.type == MSG_BLOCK) || (inv.type == MSG_FILTERED_BLOCK) ||
                    (inv.type == MSG_CMPCT_BLOCK) || inv.type == MSG_DOUBLESPENDPROOF))
            {
                dosMan.Misbehaving(pfrom, 20, BanReasonInvalidInventory);
                return error("message inv invalid type = %u", inv.type);
            }

            // Make basic checks
            if (inv.type == MSG_CMPCT_BLOCK)
            {
                if (!requester.CheckForRequestDOS(pfrom, chainparams))
                    return false;
            }

            invDeque.push_back(std::pair<CInv2, uint32_t>(inv, msgCookie));
            // These will be deferred, so we need to provide a different cookie for each one.
            if ((inv.type == MSG_BLOCK) || (inv.type == MSG_CMPCT_BLOCK))
                msgCookie++;
        }

        if (fDebug || (invDeque.size() != 1))
            LOG(NET, "received getdata (%u invsz) peer=%s\n", invDeque.size(), pfrom->GetLogName());

        if ((fDebug && invDeque.size() > 0) || (invDeque.size() == 1))
            LOG(NET, "received getdata for: %s peer=%s\n", invDeque[0].first.ToString(), pfrom->GetLogName());

        // Run process getdata and process as much of the getdata's as we can before taking the lock
        // and appending the remainder to the vRecvGetData queue.
        ProcessGetData(pfrom, chainparams.GetConsensus(), invDeque);
        if (!invDeque.empty())
        {
            LOCK(pfrom->csRecvGetData);
            pfrom->vRecvGetData2.insert(pfrom->vRecvGetData2.end(), invDeque.begin(), invDeque.end());
        }
    }

    // Peer wants new style inventory (CExtInv) messages
    else if (strCommand == NetMsgType::EXTGETDATA)
    {
        if (fImporting || fReindex)
        {
            LOG(NET, "received getdata from %s but importing\n", pfrom->GetLogName());
            return true;
        }

        std::vector<CExtInv> vInv;
        vRecv >> vInv;
        // check size == 0 to be intolerant of an empty and useless request
        if ((vInv.size() > MAX_INV_SZ) || (vInv.size() == 0))
        {
            dosMan.Misbehaving(pfrom, 20);
            return error("message extgetdata size() = %u", vInv.size());
        }

        // Validate that EXTINVs are a valid type
        std::deque<std::pair<CExtInv, uint32_t> > invDeque;
        for (unsigned int nInv = 0; nInv < vInv.size(); nInv++)
        {
            const CExtInv &inv = vInv[nInv];
            if (!inv.IsKnownType())
            {
                dosMan.Misbehaving(pfrom, 20, BanReasonInvalidInventory);
                return error("message extinv invalid type = %u", inv.type);
            }

            invDeque.push_back(std::pair<CExtInv, uint32_t>(inv, msgCookie));
        }

        if (fDebug || (invDeque.size() != 1))
            LOG(NET, "received extgetdata (%u invsz) peer=%s\n", invDeque.size(), pfrom->GetLogName());

        if ((fDebug && invDeque.size() > 0) || (invDeque.size() == 1))
            LOG(NET, "received extgetdata for: %s peer=%s\n", invDeque[0].first.ToString(), pfrom->GetLogName());

        // Run process extgetdata and process as much of the extgetdata's as we can before taking the lock
        // and appending the remainder to the vRecvExtGetData queue.
        ProcessExtGetData(pfrom, chainparams.GetConsensus(), invDeque);
        if (!invDeque.empty())
        {
            LOCK(pfrom->csRecvGetData);
            pfrom->vRecvExtGetData.insert(pfrom->vRecvExtGetData.end(), invDeque.begin(), invDeque.end());
        }
    }

    else if (strCommand == NetMsgType::GETBLOCKS)
    {
        if (fImporting || fReindex)
            return true;

        CBlockLocator locator;
        uint256 hashStop;
        vRecv >> locator >> hashStop;

        // Find the last block the caller has in the main chain
        CBlockIndex *pindex = FindForkInGlobalIndex(chainActive, locator);

        // Send the rest of the chain
        if (pindex)
            pindex = chainActive.Next(pindex);
        int nLimit = 500;
        LOG(NET, "getblocks %d to %s limit %d from peer=%d\n", (pindex ? pindex->height() : -1),
            hashStop.IsNull() ? "end" : hashStop.ToString(), nLimit, pfrom->id);
        for (; pindex; pindex = chainActive.Next(pindex))
        {
            if (pindex->GetBlockHash() == hashStop)
            {
                LOG(NET, "  getblocks stopping at %d %s\n", pindex->height(), pindex->GetBlockHash().ToString());
                break;
            }
            // If pruning, don't inv blocks unless we have on disk and are likely to still have
            // for some reasonable time window (1 hour) that block relay might require.
            const int nPrunedBlocksLikelyToHave =
                MIN_BLOCKS_TO_KEEP - 3600 / chainparams.GetConsensus().nPowTargetSpacing;
            {
                READLOCK(cs_mapBlockIndex); // for nStatus
                if (fPruneMode && (!(pindex->nStatus & BLOCK_HAVE_DATA) ||
                                      pindex->height() <= chainActive.Tip()->height() - nPrunedBlocksLikelyToHave))
                {
                    LOG(NET, " getblocks stopping, pruned or too old block at %d %s\n", pindex->height(),
                        pindex->GetBlockHash().ToString());
                    break;
                }
            }
            if (pfrom->fPeerWantsINV2)
                pfrom->PushInventory(CInv2(MSG_BLOCK, pindex->GetBlockHash()));
            else
                pfrom->PushInventory(CInv(MSG_BLOCK, pindex->GetBlockHash()));

            if (--nLimit <= 0)
            {
                // When this block is requested, we'll send an inv that'll
                // trigger the peer to getblocks the next batch of inventory.
                LOG(NET, "  getblocks stopping at limit %d %s\n", pindex->height(), pindex->GetBlockHash().ToString());
                pfrom->hashContinue = pindex->GetBlockHash();
                break;
            }
        }
    }


    else if (strCommand == NetMsgType::GETHEADERS)
    {
        CBlockLocator locator;
        uint256 hashStop;
        vRecv >> locator >> hashStop;

        CBlockIndex *pindex = nullptr;
        if (locator.IsNull())
        {
            pindex = LookupBlockIndex(hashStop);
            if (!pindex)
                return true;
        }

        std::vector<CBlockHeader> vHeaders;
        {
            if (!locator.IsNull())
            {
                // Find the last block the caller has in the main chain
                pindex = FindForkInGlobalIndex(chainActive, locator);
                if (pindex)
                    pindex = chainActive.Next(pindex);
            }

            int nLimit = MAX_HEADERS_RESULTS;
            LOG(NET, "getheaders height %d for block %s from peer %s\n", (pindex ? pindex->height() : -1),
                hashStop.ToString(), pfrom->GetLogName());
            READLOCK(chainActive.cs_chainLock);
            for (; pindex; pindex = chainActive._Next(pindex))
            {
                vHeaders.push_back(pindex->GetBlockHeader());
                if (--nLimit <= 0 || pindex->GetBlockHash() == hashStop)
                    break;
            }
        }
        // pindex can be nullptr either if we sent chainActive.Tip() OR
        // if our peer has chainActive.Tip() (and thus we are sending an empty
        // headers message). In both cases it's safe to update
        // pindexBestHeaderSent to be our tip.
        {
            CNodeStateAccessor state(nodestate, pfrom->GetId());
            state->pindexBestHeaderSent = pindex ? pindex : chainActive.Tip();
        }
        pfrom->PushMessageWithCookie(NetMsgType::HEADERS, msgCookie | 0xFFFF, vHeaders);
    }


    else if (strCommand == NetMsgType::TX)
    {
        // Stop processing the transaction early if
        // We are in blocks only mode and peer is either not whitelisted or whitelistrelay is off
        if (fBlocksOnly && (!pfrom->fWhitelisted || !fWhiteListRelay))
        {
            LOG(NET, "transaction sent in violation of protocol peer=%d\n", pfrom->id);
            return true;
        }

        // Process as many concatenated txns as there may be in this message
        while (!vRecv.empty())
        {
            // Put the tx on the tx admission queue for processing
            CTxInputData txd;
            vRecv >> txd.tx;

            // Indicate that the tx was received and is about to be processed. Setting the processing flag
            // prevents us from re-requesting the txn during the time of processing and before mempool acceptance.
            CInv inv(MSG_TX, txd.tx->GetId());
            requester.Received(inv, pfrom);

            // Processing begins here where we enqueue the transaction.
            txd.nodeId = pfrom->id;
            txd.nodeName = pfrom->GetLogName();
            txd.whitelisted = pfrom->fWhitelisted;
            txd.msgCookie = msgCookie;
            EnqueueTxForAdmission(txd);

            pfrom->AddInventoryKnown(inv);
            requester.UpdateTxnResponseTime(inv, pfrom);
        }
    }


    else if (strCommand == NetMsgType::HEADERS) // Ignore headers received while importing
    {
        if (fImporting)
        {
            LOG(NET, "skipping processing of HEADERS because importing\n");
            return true;
        }
        if (fReindex)
        {
            LOG(NET, "skipping processing of HEADERS because reindexing\n");
            return true;
        }
        std::vector<CBlockHeader> headers;

        // Bypass the normal CBlock deserialization, as we don't want to risk deserializing 2000 full blocks.
        unsigned int nCount = ReadCompactSize(vRecv);
        if (nCount > MAX_HEADERS_RESULTS)
        {
            dosMan.Misbehaving(pfrom, 20);
            return error("headers message size = %u", nCount);
        }
        headers.resize(nCount);
        for (unsigned int n = 0; n < nCount; n++)
        {
            vRecv >> headers[n];
        }

        // Nothing interesting. Stop asking this peers for more headers.
        if (nCount == 0)
        {
            LOG(NET, "No more headers from peer %s\n", pfrom->GetLogName());
            return true;
        }

        CBlockIndex *pindexLast = nullptr;
        {
            // We need to handle appending the header and analyzing the unconnected ones sequentially, or
            // 2 simultaneously processed header messages may cause an out of order header to not be reconnected
            // when its parent arrives.

            // We are reusing csUnConnected headers to both force this code to be sequential and to protect
            // the unconnected headers data structure.
            LOCK(csUnconnectedHeaders);

            // Check all headers to make sure they are continuous before attempting to accept them.
            // This prevents and attacker from keeping us from doing direct fetch by giving us out
            // of order headers.
            bool fNewUnconnectedHeaders = false;
            uint256 hashLastBlock;
            hashLastBlock.SetNull();
            for (const CBlockHeader &header : headers)
            {
                const uint256 &hash = header.GetHash();
                if (mapUnConnectedHeaders.count(hash))
                    continue;

                // LOG(NET, "Received header %s from %s\n", header.GetId().ToString(), pfrom->GetLogName());
                // check that the first header has a previous block in the blockindex.
                if (hashLastBlock.IsNull())
                {
                    if (LookupBlockIndex(header.hashPrevBlock))
                        hashLastBlock = header.hashPrevBlock;
                }

                // Add this header to the map if it doesn't connect to a previous header
                if (header.hashPrevBlock != hashLastBlock)
                {
                    // If we still haven't finished downloading the initial headers during node sync and we get
                    // an out of order header then we must disconnect the node so that we can finish downloading
                    // initial headers from a diffeent peer. An out of order header at this point is likely an attack
                    // to prevent the node from syncing.
                    if (header.GetBlockTime() < GetAdjustedTime() - 24 * 60 * 60)
                    {
                        pfrom->fDisconnect = true;
                        return error("non-continuous-headers sequence during node sync - disconnecting peer=%s",
                            pfrom->GetLogName());
                    }
                    fNewUnconnectedHeaders = true;
                }

                // if we have an unconnected header then add every following header to the unconnected headers cache.
                if (fNewUnconnectedHeaders)
                {
                    // LOG(NET, "Header %s from %s is unconnected\n", hash.ToString(), pfrom->GetLogName());
                    if (mapUnConnectedHeaders.size() < MAX_UNCONNECTED_HEADERS)
                        mapUnConnectedHeaders[hash] = std::make_pair(header, GetTime());
                    else
                        LOG(NET, "Ignoring header %s -- too many unconnected headers.\n", hash.ToString());

                    // update hashLastUnknownBlock so that we'll be able to download the block from this peer even
                    // if we receive the headers, which will connect this one, from a different peer.
                    requester.UpdateBlockAvailability(pfrom->GetId(), hash);
                }

                hashLastBlock = hash;
            }
            // return without error if we have an unconnected header.  This way we can try to connect it when the next
            // header arrives.
            if (fNewUnconnectedHeaders)
                return true;

            {
                // If possible add any previously unconnected headers to the headers vector and remove any expired
                // entries.
                std::map<uint256, std::pair<CBlockHeader, int64_t> >::iterator mi = mapUnConnectedHeaders.begin();
                while (mi != mapUnConnectedHeaders.end())
                {
                    std::map<uint256, std::pair<CBlockHeader, int64_t> >::iterator toErase = mi;

                    // Add the header if it connects to the previous header
                    if (headers.back().GetHash() == (*mi).second.first.hashPrevBlock)
                    {
                        headers.push_back((*mi).second.first);
                        mapUnConnectedHeaders.erase(toErase);

                        // if you found one to connect then search from the beginning again in case there is another
                        // that will connect to this new header that was added.
                        mi = mapUnConnectedHeaders.begin();
                        continue;
                    }

                    // Remove any entries that have been in the cache too long.  Unconnected headers should only exist
                    // for a very short while, typically just a second or two.
                    int64_t nTimeHeaderArrived = (*mi).second.second;
                    const uint256 &headerHash = (*mi).first;
                    mi++;
                    if (GetTime() - nTimeHeaderArrived >= UNCONNECTED_HEADERS_TIMEOUT)
                    {
                        mapUnConnectedHeaders.erase(toErase);
                    }
                    // At this point we know the headers in the list received are known to be in order, therefore,
                    // check if the header is equal to some other header in the list. If so then remove it from the
                    // cache.
                    else
                    {
                        for (const CBlockHeader &header : headers)
                        {
                            if (header.GetHash() == headerHash)
                            {
                                mapUnConnectedHeaders.erase(toErase);
                                break;
                            }
                        }
                    }
                }
            }

            // Check and accept each header in dependency order (oldest block to most recent)
            int i = 0;
            for (const CBlockHeader &header : headers)
            {
                CValidationState state;
                if (!AcceptBlockHeader(header, state, chainparams, &pindexLast))
                {
                    // Disconnect any peers that give us a bad checkpointed header.
                    // This prevents us from getting stuck in IBD where we download the intial headers.
                    // If we don't disconnect then we could end up attempting to download headers
                    // only from peers that are not on our own fork.
                    if (state.GetRejectCode() == REJECT_CHECKPOINT)
                    {
                        pfrom->fDisconnect = true;
                    }

                    int nDos;
                    if (state.IsInvalid(nDos))
                    {
                        if (nDos > 0)
                        {
                            dosMan.Misbehaving(pfrom, nDos);
                        }
                    }

                    // all headers from this one forward reference a fork that we don't follow, so erase them
                    headers.erase(headers.begin() + i, headers.end());
                    nCount = headers.size();

                    break;
                }
                else
                    PV->UpdateMostWorkOurFork(header);

                i++;
            }
        }
        if (pindexLast)
            requester.UpdateBlockAvailability(pfrom->GetId(), pindexLast->GetBlockHash());

        if (nCount == MAX_HEADERS_RESULTS && pindexLast)
        {
            // Headers message had its maximum size; the peer may have more headers.
            // TODO: optimize: if pindexLast is an ancestor of chainActive.Tip or pindexBestHeader, continue
            // from there instead.
            LOG(NET, "more getheaders (%d) to end to peer=%s (startheight:%d)\n", pindexLast->height(),
                pfrom->GetLogName(), pfrom->nStartingHeight);
            pfrom->PushMessage(NetMsgType::GETHEADERS, chainActive.GetLocator(pindexLast), uint256());

            {
                int64_t now = GetTime();
                CNodeStateAccessor state(nodestate, pfrom->GetId());
                DbgAssert(state != nullptr, );
                if (state != nullptr)
                    state->nSyncStartTime = now; // reset the time because more headers needed
            }

            // During the process of IBD we need to update block availability for every connected peer. To do that we
            // request, from each NODE_NETWORK peer, a header that matches the last blockhash found in this recent set
            // of headers. Once the requested header is received then the block availability for this peer will get
            // updated.
            if (IsInitialBlockDownload())
            {
                std::vector<CNode *> vNodesCopy;
                {
                    LOCK(cs_vNodes);
                    vNodesCopy = vNodes;
                    for (CNode *pnode : vNodes)
                    {
                        pnode->AddRef();
                    }
                }

                for (CNode *pnode : vNodesCopy)
                {
                    if (!pnode->fClient && pnode != pfrom)
                    {
                        bool ask = false;
                        {
                            CNodeStateAccessor state(nodestate, pfrom->GetId());
                            DbgAssert(state != nullptr, ); // do not return, we need to release refs later.
                            if (state == nullptr)
                                continue;

                            ask = (state->pindexBestKnownBlock == nullptr ||
                                   pindexLast->chainWork() > state->pindexBestKnownBlock->chainWork());
                        } // let go of the CNodeState lock before we PushMessage since that is trapping op.

                        if (ask)
                        {
                            // We only want one single header so we pass a null for CBlockLocator.
                            pnode->PushMessage(NetMsgType::GETHEADERS, CBlockLocator(), pindexLast->GetBlockHash());
                            LOG(NET | BLK, "Requesting header for blockavailability, peer=%s block=%s height=%d\n",
                                pnode->GetLogName(), pindexLast->GetBlockHash().ToString().c_str(),
                                pindexBestHeader.load()->height());
                        }
                    }
                }

                // release refs
                for (CNode *pnode : vNodesCopy)
                    pnode->Release();
            }
        }

        bool fCanDirectFetch = CanDirectFetch(chainparams.GetConsensus());

        {
            CNodeStateAccessor state(nodestate, pfrom->GetId());
            DbgAssert(state != nullptr, return false);

            // During the initial peer handshake we must receive the initial headers which should be greater
            // than or equal to our block height at the time of requesting GETHEADERS. This is because the peer has
            // advertised a height >= to our own. Furthermore, because the headers max returned is as much as 2000 this
            // could not be a mainnet re-org.
            if (!state->fFirstHeadersReceived)
            {
                // We want to make sure that the peer doesn't just send us any old valid header. The block height of the
                // last header they send us should be equal to our block height at the time we made the GETHEADERS
                // request.
                if (pindexLast && state->nFirstHeadersExpectedHeight <= pindexLast->height())
                {
                    state->fFirstHeadersReceived = true;
                    LOG(NET, "Initial headers received for peer=%s\n", pfrom->GetLogName());
                }

                // Allow for very large reorgs (> 2000 blocks) on the nol test chain or other test net.
                if (Params().NetworkIDString() != "main" && Params().NetworkIDString() != "regtest")
                    state->fFirstHeadersReceived = true;
            }
        }

        // update the syncd status.  This should come before we make calls to requester.AskFor().
        IsChainNearlySyncdInit();
        IsInitialBlockDownloadInit();

        // If this set of headers is valid and ends in a block with at least as
        // much work as our tip, download as much as possible.
        if (fCanDirectFetch && pindexLast && pindexLast->IsValid(BLOCK_VALID_TREE) &&
            chainActive.Tip()->chainWork() <= pindexLast->chainWork())
        {
            // Set tweak value.  Mostly used in testing direct fetch.
            if (maxBlocksInTransitPerPeer.Value() != 0)
                pfrom->nMaxBlocksInTransit.store(maxBlocksInTransitPerPeer.Value());

            std::vector<CBlockIndex *> vToFetch;
            CBlockIndex *pindexWalk = pindexLast;
            // Calculate all the blocks we'd need to switch to pindexLast.
            while (pindexWalk && !chainActive.Contains(pindexWalk))
            {
                vToFetch.push_back(pindexWalk);
                pindexWalk = pindexWalk->pprev;
            }

            // Download as much as possible, from earliest to latest.
            unsigned int nAskFor = 0;
            for (auto pindex_iter = vToFetch.rbegin(); pindex_iter != vToFetch.rend(); pindex_iter++)
            {
                CBlockIndex *pindex = *pindex_iter;
                // pindex must be nonnull because we populated vToFetch a few lines above
                CInv inv(MSG_BLOCK, pindex->GetBlockHash());
                if (!AlreadyHaveBlock(inv))
                {
                    requester.AskFor(inv, pfrom);
                    LOG(REQ, "AskFor block via headers direct fetch %s (%d) peer=%d\n",
                        pindex->GetBlockHash().ToString(), pindex->height(), pfrom->id);
                    nAskFor++;
                }
                // We don't care about how many blocks are in flight.  We just need to make sure we don't
                // ask for more than the maximum allowed per peer because the request manager will take care
                // of any duplicate requests.
                if (nAskFor >= pfrom->nMaxBlocksInTransit.load())
                {
                    LOG(NET, "Large reorg, could only direct fetch %d blocks\n", nAskFor);
                    break;
                }
            }
            if (nAskFor > 1)
            {
                LOG(NET, "Downloading blocks toward %s (%d) via headers direct fetch\n",
                    pindexLast->GetBlockHash().ToString(), pindexLast->height());
            }
        }

        CheckBlockIndex(chainparams.GetConsensus());
    }

    // Handle Xthinblocks and Thinblocks
    else if (strCommand == NetMsgType::GET_XTHIN && !fImporting && !fReindex && IsThinBlocksEnabled())
    {
        if (!requester.CheckForRequestDOS(pfrom, chainparams))
            return false;

        CBloomFilter filterMemPool;
        uint256 hash;
        int invType = 0;

        if (pfrom->fPeerWantsINV2)
        {
            CInv2 inv2;
            vRecv >> inv2 >> filterMemPool;
            invType = inv2.type;
            hash = inv2.hash;
        }
        else
        {
            CInv inv;
            vRecv >> inv >> filterMemPool;
            invType = inv.type;
            hash = inv.hash;
        }

        // Message consistency checking
        if (hash.IsNull())
        {
            dosMan.Misbehaving(pfrom, 100);
            return error("invalid get_xthin type=%u hash=%s", invType, hash.ToString());
        }


        // Validates that the filter is reasonably sized.
        LoadFilter(pfrom, &filterMemPool);
        {
            auto *invIndex = LookupBlockIndex(hash);
            if (!invIndex)
            {
                dosMan.Misbehaving(pfrom, 100);
                return error("Peer %srequested nonexistent block %s", pfrom->GetLogName(), hash.ToString());
            }

            const Consensus::Params &consensusParams = Params().GetConsensus();
            const ConstCBlockRef pblock = ReadBlockFromDisk(invIndex, consensusParams);
            if (!pblock)
            {
                // We don't have the block yet, although we know about it.
                return error("Peer %s requested block %s that cannot be read", pfrom->GetLogName(), hash.ToString());
            }
            else
            {
                SendXThinBlock(pblock, msgCookie, pfrom, invType);
            }
        }
    }
    else if (strCommand == NetMsgType::GET_THIN && !fImporting && !fReindex && IsThinBlocksEnabled())
    {
        if (!requester.CheckForRequestDOS(pfrom, chainparams))
            return false;

        uint256 hash;
        int invType = 0;

        if (pfrom->fPeerWantsINV2)
        {
            CInv2 inv2;
            vRecv >> inv2;
            invType = inv2.type;
            hash = inv2.hash;
        }
        else
        {
            CInv inv;
            vRecv >> inv;
            invType = inv.type;
            hash = inv.hash;
        }


        // Message consistency checking
        if (hash.IsNull())
        {
            dosMan.Misbehaving(pfrom, 100);
            return error("invalid get_thin type=%u hash=%s", invType, hash.ToString());
        }

        auto *invIndex = LookupBlockIndex(hash);
        if (!invIndex)
        {
            dosMan.Misbehaving(pfrom, 100);
            return error("Peer %srequested nonexistent block %s", pfrom->GetLogName(), hash.ToString());
        }

        const Consensus::Params &consensusParams = Params().GetConsensus();
        const ConstCBlockRef pblock = ReadBlockFromDisk(invIndex, consensusParams);
        if (!pblock)
        {
            // We don't have the block yet, although we know about it.
            return error("Peer %s requested block %s that cannot be read", pfrom->GetLogName(), hash.ToString());
        }
        else
        {
            SendXThinBlock(pblock, msgCookie, pfrom, invType);
        }
    }
    else if (strCommand == NetMsgType::XPEDITEDREQUEST)
    {
        return HandleExpeditedRequest(vRecv, pfrom);
    }
    else if (strCommand == NetMsgType::XPEDITEDBLK)
    {
        // ignore the expedited message unless we are at the chain tip...
        if (!fImporting && !fReindex && !IsInitialBlockDownload())
        {
            LOCK(pfrom->cs_thintype);
            if (!HandleExpeditedBlock(vRecv, pfrom))
                return false;
        }
    }

    else if (strCommand == NetMsgType::XTHINBLOCK && !fImporting && !fReindex && !IsInitialBlockDownload() &&
             IsThinBlocksEnabled())
    {
        LOCK(pfrom->cs_thintype);
        return CXThinBlock::HandleMessage(vRecv, pfrom, strCommand, 0);
    }


    else if (strCommand == NetMsgType::THINBLOCK && !fImporting && !fReindex && !IsInitialBlockDownload() &&
             IsThinBlocksEnabled())
    {
        LOCK(pfrom->cs_thintype);
        return CThinBlock::HandleMessage(vRecv, pfrom);
    }


    else if (strCommand == NetMsgType::GET_XBLOCKTX && !fImporting && !fReindex && !IsInitialBlockDownload() &&
             IsThinBlocksEnabled())
    {
        if (!requester.CheckForRequestDOS(pfrom, chainparams))
            return false;

        LOCK(pfrom->cs_thintype);
        return CXRequestThinBlockTx::HandleMessage(vRecv, pfrom);
    }


    else if (strCommand == NetMsgType::XBLOCKTX && !fImporting && !fReindex && !IsInitialBlockDownload() &&
             IsThinBlocksEnabled())
    {
        LOCK(pfrom->cs_thintype);
        return CXThinBlockTx::HandleMessage(vRecv, pfrom);
    }

    // Handle Graphene blocks
    else if (strCommand == NetMsgType::GET_GRAPHENE && !fImporting && !fReindex && IsGrapheneBlockEnabled() &&
             grapheneVersionCompatible)
    {
        if (!requester.CheckForRequestDOS(pfrom, chainparams))
            return false;

        LOCK(pfrom->cs_thintype);
        return HandleGrapheneBlockRequest(vRecv, pfrom, msgCookie, chainparams);
    }

    else if (strCommand == NetMsgType::GRAPHENEBLOCK && !fImporting && !fReindex && !IsInitialBlockDownload() &&
             IsGrapheneBlockEnabled() && grapheneVersionCompatible)
    {
        LOCK(pfrom->cs_thintype);
        return CGrapheneBlock::HandleMessage(vRecv, pfrom, msgCookie, strCommand, 0);
    }


    else if (strCommand == NetMsgType::GET_GRAPHENETX && !fImporting && !fReindex && !IsInitialBlockDownload() &&
             IsGrapheneBlockEnabled() && grapheneVersionCompatible)
    {
        if (!requester.CheckForRequestDOS(pfrom, chainparams))
            return false;

        LOCK(pfrom->cs_thintype);
        return CRequestGrapheneBlockTx::HandleMessage(vRecv, pfrom);
    }


    else if (strCommand == NetMsgType::GRAPHENETX && !fImporting && !fReindex && !IsInitialBlockDownload() &&
             IsGrapheneBlockEnabled() && grapheneVersionCompatible)
    {
        LOCK(pfrom->cs_thintype);
        return CGrapheneBlockTx::HandleMessage(vRecv, pfrom);
    }

    else if (strCommand == NetMsgType::GET_GRAPHENE_RECOVERY && IsGrapheneBlockEnabled() && grapheneVersionCompatible)
    {
        if (!requester.CheckForRequestDOS(pfrom, chainparams))
            return false;

        LOCK(pfrom->cs_thintype);
        return HandleGrapheneBlockRecoveryRequest(vRecv, pfrom, msgCookie, chainparams);
    }

    else if (strCommand == NetMsgType::GRAPHENE_RECOVERY && IsGrapheneBlockEnabled() && grapheneVersionCompatible)
    {
        if (!requester.CheckForRequestDOS(pfrom, chainparams))
            return false;

        LOCK(pfrom->cs_thintype);
        return HandleGrapheneBlockRecoveryResponse(vRecv, pfrom, chainparams);
    }

    // Handle Compact Blocks
    else if (strCommand == NetMsgType::CMPCTBLOCK && !fImporting && !fReindex && !IsInitialBlockDownload() &&
             IsCompactBlocksEnabled())
    {
        LOCK(pfrom->cs_thintype);
        return CompactBlock::HandleMessage(vRecv, msgCookie, pfrom);
    }
    else if (strCommand == NetMsgType::GETBLOCKTXN && !fImporting && !fReindex && !IsInitialBlockDownload() &&
             IsCompactBlocksEnabled())
    {
        if (!requester.CheckForRequestDOS(pfrom, chainparams))
            return false;

        LOCK(pfrom->cs_thintype);
        return CompactReRequest::HandleMessage(vRecv, msgCookie, pfrom);
    }
    else if (strCommand == NetMsgType::BLOCKTXN && !fImporting && !fReindex && !IsInitialBlockDownload() &&
             IsCompactBlocksEnabled())
    {
        LOCK(pfrom->cs_thintype);
        return CompactReReqResponse::HandleMessage(vRecv, msgCookie, pfrom);
    }
    else if (strCommand == NetMsgType::GETHEADERPATH)
    {
        return HandleHeaderPathMessage(vRecv, pfrom, msgCookie);
    }

    // Mempool synchronization request
    else if (strCommand == NetMsgType::GET_MEMPOOLSYNC)
    {
        return HandleMempoolSyncRequest(vRecv, pfrom, msgCookie);
    }

    else if (strCommand == NetMsgType::MEMPOOLSYNC)
    {
        return CMempoolSync::ReceiveMempoolSync(vRecv, pfrom, strCommand);
    }

    // Mempool synchronization transaction request
    else if (strCommand == NetMsgType::GET_MEMPOOLSYNCTX)
    {
        return CRequestMempoolSyncTx::HandleMessage(vRecv, msgCookie, pfrom);
    }

    else if (strCommand == NetMsgType::MEMPOOLSYNCTX)
    {
        return CMempoolSyncTx::HandleMessage(vRecv, msgCookie, pfrom);
    }


    // Handle full blocks
    else if (strCommand == NetMsgType::BLOCK && !fImporting && !fReindex) // Ignore blocks received while importing
    {
        CBlockRef pblock = MakeBlockRef();
        {
            uint64_t nCheckBlockSize = vRecv.size();
            vRecv >> *pblock;

            // Sanity check. The serialized block size should match the size that is in our receive queue.  If not
            // this could be an attack block of some kind.
            DbgAssert(pblock->GetBlockSize() <= nCheckBlockSize &&
                          nCheckBlockSize <= pblock->GetBlockSize() + CBlockHeader::MAX_NONCE_SIZE + 1,
                return true);
        }

        const uint256 hash = pblock->GetHash();
        LOG(BLK, "received block %s peer=%d\n", hash.ToString(), pfrom->id);
        UnlimitedLogBlock(*pblock, hash.ToString(), receiptTime);

        if (IsChainNearlySyncd()) // BU send the received block out expedited channels quickly
        {
            CValidationState state;
            if (CheckBlockHeader(chainparams.GetConsensus(), *pblock, state, true)) // block header is fine
                SendExpeditedBlock(*pblock, pfrom);
        }

        { // reset the getheaders time because block can consume all bandwidth
            int64_t now = GetTime();
            CNodeStateAccessor state(nodestate, pfrom->GetId());
            DbgAssert(state != nullptr, );
            if (state != nullptr)
                state->nSyncStartTime = now; // reset the time because more headers needed
        }
        pfrom->nPingUsecStart = GetStopwatchMicros(); // Reset ping time because block can consume all bandwidth

        // Message consistency checking
        // NOTE: consistency checking is handled by checkblock() which is called during
        //       ProcessNewBlock() during HandleBlockMessage.
        PV->HandleBlockMessage(pfrom, strCommand, pblock, hash);
    }


    else if (strCommand == NetMsgType::GETADDR)
    {
        // This asymmetric behavior for inbound and outbound connections was introduced
        // to prevent a fingerprinting attack: an attacker can send specific fake addresses
        // to users' AddrMan and later request them by sending getaddr messages.
        // Making nodes which are behind NAT and can only make outgoing connections ignore
        // the getaddr message mitigates the attack.
        if (!pfrom->fInbound)
        {
            LOG(NET, "Ignoring \"getaddr\" from outbound connection. peer=%d\n", pfrom->id);
            return true;
        }

        // Only send one GetAddr response per connection to reduce resource waste
        //  and discourage addr stamping of INV announcements.
        if (pfrom->fSentAddr)
        {
            LOG(NET, "Ignoring repeated \"getaddr\". peer=%d\n", pfrom->id);
            return true;
        }
        pfrom->fSentAddr = true;
        LOCK(pfrom->cs_vSend);
        pfrom->vAddrToSend.clear();
        std::vector<CAddress> vAddr = addrman.GetAddr();
        FastRandomContext insecure_rand;
        for (const CAddress &addr : vAddr)
            pfrom->PushAddress(addr, insecure_rand);
    }


    else if (strCommand == NetMsgType::MEMPOOL && !pfrom->fPeerWantsINV2)
    {
        if (CNode::OutboundTargetReached(false) && !pfrom->fWhitelisted)
        {
            LOG(NET, "mempool request with bandwidth limit reached, disconnect peer %s\n", pfrom->GetLogName());
            pfrom->fDisconnect = true;
            return true;
        }
        std::vector<uint256> vtxid;
        mempool.queryIds(vtxid);
        std::vector<CInv> vInv;

        // Because we have to take cs_filter after mempool.cs, in order to maintain locking order, we
        // need find out if a filter is present first before later doing the mempool.get().
        bool fHaveFilter = false;
        {
            LOCK(pfrom->cs_filter);
            fHaveFilter = pfrom->pfilter ? true : false;
        }

        uint32_t replyCount = 0;
        for (uint256 &hash : vtxid)
        {
            CInv inv(MSG_TX, hash);
            if (fHaveFilter)
            {
                CTransactionRef ptx = nullptr;
                ptx = mempool.get(inv.hash);
                if (ptx == nullptr)
                    continue; // another thread removed since queryHashes, maybe...

                LOCK(pfrom->cs_filter);
                if (!pfrom->pfilter->IsRelevantAndUpdate(ptx))
                    continue;
            }
            // By sending before we push the next one on the list, we guarantee that the pushmessage outside of this
            // loop will be called rather than this one at the end of the array.
            // That must happen so that the end-of-message-group cookie can be sent.
            if (vInv.size() == MAX_INV_SZ)
            {
                pfrom->PushMessageWithCookie(NetMsgType::INV, msgCookie + replyCount, vInv);
                replyCount++;
                vInv.clear();
            }
            vInv.push_back(inv);
        }
        if (vInv.size() > 0)
            pfrom->PushMessageWithCookie(NetMsgType::INV, msgCookie | 0xFFFF, vInv);
    }
    else if (strCommand == NetMsgType::MEMPOOL && pfrom->fPeerWantsINV2)
    {
        if (CNode::OutboundTargetReached(false) && pfrom->fWhitelisted)
        {
            LOG(NET, "mempool request with bandwidth limit reached, disconnect peer %s\n", pfrom->GetLogName());
            pfrom->fDisconnect = true;
            return true;
        }
        std::vector<uint256> vtxid;
        mempool.queryIds(vtxid);
        std::vector<CInv2> vInv;

        // Because we have to take cs_filter after mempool.cs, in order to maintain locking order, we
        // need find out if a filter is present first before later doing the mempool.get().
        bool fHaveFilter = false;
        {
            LOCK(pfrom->cs_filter);
            fHaveFilter = pfrom->pfilter ? true : false;
        }

        uint32_t replyCount = 0;
        for (uint256 &hash : vtxid)
        {
            CInv2 inv(MSG_TX, hash);
            if (fHaveFilter)
            {
                CTransactionRef ptx = nullptr;
                ptx = mempool.get(inv.hash);
                if (ptx == nullptr)
                    continue; // another thread removed since queryHashes, maybe...

                LOCK(pfrom->cs_filter);
                if (!pfrom->pfilter->IsRelevantAndUpdate(ptx))
                    continue;
            }
            // By sending before we push the next one on the list, we guarantee that the pushmessage outside of this
            // loop will be called rather than this one at the end of the array.
            // That must happen so that the end-of-message-group cookie can be sent.
            if (vInv.size() == MAX_INV_SZ)
            {
                pfrom->PushMessageWithCookie(NetMsgType::INV, msgCookie + replyCount, vInv);
                replyCount++;
                vInv.clear();
            }
            vInv.push_back(inv);
        }
        if (vInv.size() > 0)
            pfrom->PushMessageWithCookie(NetMsgType::INV, msgCookie | 0xFFFF, vInv);
    }

    else if (strCommand == NetMsgType::PONG)
    {
        int64_t pingUsecEnd = nStopwatchTimeReceived;
        uint64_t nonce = 0;
        size_t nAvail = vRecv.in_avail();
        bool bPingFinished = false;
        std::string sProblem;

        if (nAvail >= sizeof(nonce))
        {
            vRecv >> nonce;

            // Only process pong message if there is an outstanding ping (old ping without nonce should never pong)
            if (pfrom->nPingNonceSent != 0)
            {
                if (nonce == pfrom->nPingNonceSent)
                {
                    // Matching pong received, this ping is no longer outstanding
                    bPingFinished = true;
                    int64_t pingUsecTime = pingUsecEnd - pfrom->nPingUsecStart;
                    if (pingUsecTime > 0)
                    {
                        // Successful ping time measurement, replace previous
                        pfrom->nPingUsecTime = pingUsecTime;
                        pfrom->nMinPingUsecTime = std::min(pfrom->nMinPingUsecTime, pingUsecTime);
                    }
                    else
                    {
                        // This should never happen
                        sProblem = "Timing mishap";
                    }
                }
                else
                {
                    // Nonce mismatches are normal when pings are overlapping
                    sProblem = "Nonce mismatch";
                    if (nonce == 0)
                    {
                        // This is most likely a bug in another implementation somewhere; cancel this ping
                        bPingFinished = true;
                        sProblem = "Nonce zero";
                    }
                }
            }
            else
            {
                sProblem = "Unsolicited pong without ping";
            }
        }
        else
        {
            // This is most likely a bug in another implementation somewhere; cancel this ping
            bPingFinished = true;
            sProblem = "Short payload";
        }

        if (!(sProblem.empty()))
        {
            LOG(NET, "pong peer=%d: %s, %x expected, %x received, %u bytes\n", pfrom->id, sProblem,
                pfrom->nPingNonceSent, nonce, nAvail);
        }
        if (bPingFinished)
        {
            pfrom->nPingNonceSent = 0;
        }
    }


    else if (strCommand == NetMsgType::FILTERLOAD)
    {
        CBloomFilter filter;
        vRecv >> filter;
        LOG(NET, "Bloom Filter Installed: %d bytes\n", filter.vDataSize());

        if (!filter.IsWithinSizeConstraints())
        {
            // There is no excuse for sending a too-large filter
            dosMan.Misbehaving(pfrom, 100);
            return false;
        }

        LOCK(pfrom->cs_filter);
        delete pfrom->pfilter;
        pfrom->pfilter = new CBloomFilter(filter);
        if (!pfrom->pfilter->IsEmpty())
            pfrom->fRelayTxes = true;
        else
            pfrom->fRelayTxes = false;
    }


    else if (strCommand == NetMsgType::FILTERADD)
    {
        std::vector<unsigned char> vData;
        vRecv >> vData;

        // Nodes must NEVER send a data item > the max size for a script data object,
        // and thus, the maximum size any matched object can have) in a filteradd message

        if (!withinStackWidth(vData.size(), GetBlockScriptFlags(chainActive.Tip(), chainparams.GetConsensus())))
        {
            dosMan.Misbehaving(pfrom, 100);
        }
        else
        {
            LOCK(pfrom->cs_filter);
            if (pfrom->pfilter)
                pfrom->pfilter->insert(vData);
            else
                dosMan.Misbehaving(pfrom, 100);
        }
    }


    else if (strCommand == NetMsgType::FILTERCLEAR)
    {
        LOCK(pfrom->cs_filter);
        delete pfrom->pfilter;
        pfrom->pfilter = new CBloomFilter();
        pfrom->fRelayTxes = true;
    }

    else if (strCommand == NetMsgType::DSPROOF)
    {
        if (doubleSpendProofs.Value() == true)
        {
            LOG(DSPROOF, "Received a double spend proof from peer:%d\n", pfrom->GetId());
            uint256 dspHash;
            try
            {
                DoubleSpendProof dsp;
                vRecv >> dsp;
                if (dsp.isEmpty())
                    throw std::runtime_error("Double spend proof is empty");

                dspHash = dsp.GetHash();
                DoubleSpendProof::Validity validity;
                {
                    READLOCK(mempool.cs_txmempool);
                    validity = dsp.validate(mempool);
                }
                switch (validity)
                {
                case DoubleSpendProof::Valid:
                {
                    LOG(DSPROOF, "Double spend proof is valid from peer:%d\n", pfrom->GetId());
                    const auto ptx = mempool.addDoubleSpendProof(dsp);
                    if (ptx.get())
                    {
                        // find any descendants of this double spent transaction. If there are any
                        // then we must also forward this double spend proof to any SPV peers that
                        // want to know about this tx or its descendants.
                        CTxMemPool::setEntries setDescendants;
                        {
                            READLOCK(mempool.cs_txmempool);
                            CTxMemPool::indexed_transaction_set::const_iterator iter = mempool.mapTx.find(ptx->GetId());
                            if (iter == mempool.mapTx.end())
                            {
                                break;
                            }
                            else
                            {
                                mempool._CalculateDescendants(iter, setDescendants);
                            }
                        }

                        // added to mempool correctly, then forward to nodes.
                        broadcastDspInv(ptx, dspHash, &setDescendants);
                    }
                    break;
                }
                case DoubleSpendProof::MissingUTXO:
                case DoubleSpendProof::MissingTransaction:
                    LOG(DSPROOF, "Double spend proof is orphan: postponed\n");
                    mempool.doubleSpendProofStorage()->addOrphan(dsp, pfrom->GetId());
                    break;
                case DoubleSpendProof::Invalid:
                    throw std::runtime_error(strprintf("Double spend proof didn't validate (%s)", dspHash.ToString()));
                default:
                    return false;
                }
            }
            catch (const std::exception &e)
            {
                LOG(DSPROOF, "Failure handling double spend proof. Peer: %d Reason: %s\n", pfrom->GetId(), e.what());
                if (!dspHash.IsNull())
                    mempool.doubleSpendProofStorage()->markProofRejected(dspHash);
                return false;
            }
        }
    }

    else if (strCommand == NetMsgType::REQTXVAL)
    {
        if (allowp2pTxVal.Value() == false)
        {
            return true;
        }
        uint64_t nonce;
        CTransaction tx;
        vRecv >> nonce;
        vRecv >> tx;
        CTransactionRef ptx(MakeTransactionRef(std::move(tx)));
        const uint256 &hashTx = ptx->GetId();

        bool fOverrideFees = false;
        TransactionClass txClass = TransactionClass::DEFAULT;

        CCoinsViewCache &view = *pcoinsTip;
        bool fHaveChain = false;
        {
            for (size_t i = 0; !fHaveChain && i < tx.vout.size(); i++)
            {
                CoinAccessor existingCoin(view, COutPoint(hashTx, i));
                fHaveChain = !existingCoin->IsSpent();
            }
        }
        bool fHaveMempool = mempool.exists(hashTx);
        CValidationDebugger debugger;
        if (fHaveMempool)
        {
            std::string err = "transaction already in mempool";
            pfrom->PushMessageWithCookie(NetMsgType::RESTXVAL, msgCookie | 0xFFFF, nonce, err);
            return true;
        }
        else if (fHaveChain)
        {
            std::string err = "transaction already in block chain";
            pfrom->PushMessageWithCookie(NetMsgType::RESTXVAL, msgCookie | 0xFFFF, nonce, err);
            return true;
        }
        CValidationState state;
        bool fMissingInputs = false;
        std::vector<COutPoint> vCoinsToUncache;
        bool isRespend = false;
        ParallelAcceptToMemoryPool(mempool, state, std::move(ptx), AreFreeTxnsAllowed(), &fMissingInputs, fOverrideFees,
            txClass, vCoinsToUncache, &isRespend, &debugger);

        std::string result = DebuggerToString(debugger);
        pfrom->PushMessageWithCookie(NetMsgType::RESTXVAL, msgCookie | 0xFFFF, nonce, result);
    }
    else if ((capdPoolSize.Value() > 0) && pfrom->IsCapdEnabled() &&
             (strCommand.substr(0, 4) == NetMsgType::CAPDPREFIX))
    {
        capdProtocol.HandleCapdMessage(pfrom, strCommand, msgCookie, vRecv, nStopwatchTimeReceived);
    }

    else if (strCommand == NetMsgType::REJECT)
    {
        // BU: Request manager: this was restructured to not just be active in fDebug mode so that the request manager
        // can be notified of request rejections.
        try
        {
            std::string strMsg;
            unsigned char ccode;
            std::string strReason;
            uint256 hash;

            vRecv >> LIMITED_STRING(strMsg, CMessageHeader::COMMAND_SIZE) >> ccode >>
                LIMITED_STRING(strReason, MAX_REJECT_MESSAGE_LENGTH);
            std::ostringstream ss;
            ss << strMsg << " code " << itostr(ccode) << ": " << strReason;

            // BU: Check request manager reject codes
            if (strMsg == NetMsgType::BLOCK || strMsg == NetMsgType::TX)
            {
                vRecv >> hash;
                ss << ": hash " << hash.ToString();

                // We need to see this reject message in either "req" or "net" debug mode
                LOG(REQ | NET, "Reject %s\n", SanitizeString(ss.str()));

                if (strMsg == NetMsgType::BLOCK)
                {
                    requester.Rejected(CInv(MSG_BLOCK, hash), pfrom, ccode);
                }
                else if (strMsg == NetMsgType::TX)
                {
                    requester.Rejected(CInv(MSG_TX, hash), pfrom, ccode);
                }
            }
            // if (fDebug) {
            // ostringstream ss;
            // ss << strMsg << " code " << itostr(ccode) << ": " << strReason;

            // if (strMsg == NetMsgType::BLOCK || strMsg == NetMsgType::TX)
            //  {
            //    ss << ": hash " << hash.ToString();
            //  }
            // LOG(NET, "Reject %s\n", SanitizeString(ss.str()));
            // }
        }
        catch (const std::ios_base::failure &)
        {
            // Avoid feedback loops by preventing reject messages from triggering a new reject message.
            LOG(NET, "Unparseable reject message received\n");
            LOG(REQ, "Unparseable reject message received\n");
        }
    }

    else
    {
        // Ignore unknown commands for extensibility
        LOG(NET, "Unknown command \"%s\" from peer=%d\n", SanitizeString(strCommand), pfrom->id);
    }

    return true;
}

bool ProcessMessages(CNode *pfrom)
{
    const CChainParams &chainparams = Params();
    // if (fDebug)
    //    LOGA("%s(%u messages)\n", __func__, pfrom->vRecvMsg.size());

    //
    // Message format
    //  (4) message start
    //  (12) command
    //  (4) size
    //  (4) checksum
    //  (x) data
    //
    bool fOk = true;

    // Check getdata requests first if there are no priority messages waiting.
    if (!fPriorityRecvMsg.load())
    {
        if (!pfrom->vRecvGetData.empty())
        {
            std::deque<std::pair<CInv, uint32_t> > invDeque;
            {
                // Quickly lock and swap
                LOCK(pfrom->csRecvGetData);
                if (!pfrom->vRecvGetData.empty())
                {
                    pfrom->vRecvGetData.swap(invDeque);
                }
            }
            ProcessGetData(pfrom, chainparams.GetConsensus(), invDeque);
            if (!invDeque.empty())
            {
                LOCK(pfrom->csRecvGetData);
                pfrom->vRecvGetData.insert(pfrom->vRecvGetData.end(), invDeque.begin(), invDeque.end());
            }
        }
        if (!pfrom->vRecvGetData2.empty())
        {
            std::deque<std::pair<CInv2, uint32_t> > invDeque;
            {
                // Quickly lock and swap
                LOCK(pfrom->csRecvGetData);
                if (!pfrom->vRecvGetData2.empty())
                {
                    pfrom->vRecvGetData2.swap(invDeque);
                }
            }
            ProcessGetData(pfrom, chainparams.GetConsensus(), invDeque);
            if (!invDeque.empty())
            {
                LOCK(pfrom->csRecvGetData);
                pfrom->vRecvGetData2.insert(pfrom->vRecvGetData2.end(), invDeque.begin(), invDeque.end());
            }
        }
    }

    unsigned int msgsProcessed = 0;
    // Don't bother if send buffer is too full to respond anyway
    CNode *pfrom_original = pfrom;
    std::deque<std::pair<CNodeRef, CNetMessage> > vPriorityRecvQ_delay;
    while ((!pfrom->fDisconnect) && (pfrom->nSendSize < nMaxSendBufferSize) && (shutdown_threads.load() == false))
    {
        CNodeRef noderef;
        bool fIsPriority = false;
        CNetMessage msg;
        bool fUseLowPriorityMsg = true;
        bool fUsePriorityMsg = true;
        // try to complete the handshake before handling messages that require us to be fSuccessfullyConnected
        if (!pfrom->fSuccessfullyConnected)
        {
            TRY_LOCK(pfrom->cs_vRecvMsg, lockRecv);
            if (!lockRecv)
                break;
            if (pfrom->vRecvMsg_handshake.empty())
                break;

            // get the message from the queue
            // simply getting the front message should be sufficient, the only time a extversion or verack is sent is
            // once the previous has been processed so tracking which stage of the handshake we are on is overkill
            std::swap(msg, pfrom->vRecvMsg_handshake.front());
            pfrom->vRecvMsg_handshake.pop_front();
            pfrom->currentRecvMsgSize -= msg.size();
            msgsProcessed++;
        }
        else
        {
            {
                if (!pfrom->vRecvMsg_handshake.empty())
                {
                    LOCK(pfrom->cs_vRecvMsg);
                    if (!pfrom->vRecvMsg_handshake.empty()) // check again after locking
                    {
                        std::string frontCommand = pfrom->vRecvMsg_handshake.front().hdr.GetCommand();
                        if (frontCommand == NetMsgType::VERSION || frontCommand == NetMsgType::VERACK ||
                            frontCommand == NetMsgType::EXTVERSION)
                        {
                            pfrom->vRecvMsg_handshake.clear();
                            pfrom->fDisconnect = true;
                            dosMan.Misbehaving(pfrom, 1);
                            return error("received handshake message '%s' after successful initialization,"
                                         "disconnecting peer=%s",
                                frontCommand, pfrom->GetLogName());
                        }
                    }
                }
            }
            // Get next message to process checking whether it is a priority messasge and if so then
            // process it right away. It doesn't matter that the peer where the message came from is
            // different than the one we are currently procesing as we will switch to the correct peer
            // automatically. Furthermore by using and holding the CNodeRef we automatically maintain
            // a node reference to the priority peer.
            if (fUsePriorityMsg && fPriorityRecvMsg.load())
            {
                TRY_LOCK(cs_priorityRecvQ, locked);
                if (locked && !vPriorityRecvQ.empty())
                {
                    // Get the message out of queue.
                    std::swap(noderef, vPriorityRecvQ.front().first);
                    std::swap(msg, vPriorityRecvQ.front().second);
                    vPriorityRecvQ.pop_front();

                    if (vPriorityRecvQ.empty())
                        fPriorityRecvMsg.store(false);

                    // check if we should process the message.
                    CNode *pnode = noderef.get();
                    if (pnode->fDisconnect)
                    {
                        // if the node is to be disconnected dont bother responding
                        continue;
                    }
                    if (pnode->nSendSize > nMaxSendBufferSize)
                    {
                        // if the nodes send is full, delay the processing of this message until a time when
                        // send is not full
                        vPriorityRecvQ_delay.emplace_back(std::move(noderef), std::move(msg));
                        continue;
                    }

                    fIsPriority = true;
                    fUseLowPriorityMsg = false;
                }
                else if (locked && vPriorityRecvQ.empty())
                {
                    fPriorityRecvMsg.store(false);
                    fUseLowPriorityMsg = true;
                }
            }

            if (fUseLowPriorityMsg)
            {
                if (pfrom->vRecvMsg.empty())
                    break;

                // get the message from the queue
                if (!pfrom->vRecvMsg.pop_front(msg))
                    break;
            }

            // Check if this is a priority message and if so then modify pfrom to be the peer which
            // this priority message came from.
            if (fIsPriority)
                pfrom = noderef.get();
            else
                pfrom->currentRecvMsgSize -= msg.size();

            msgsProcessed++;
        }

        // if (fDebug)
        //    LOGA("%s(message %u msgsz, %u bytes, complete:%s)\n", __func__,
        //            msg.hdr.nMessageSize, msg.vRecv.size(),
        //            msg.complete() ? "Y" : "N");

        // Scan for message start
        if (memcmp(msg.hdr.pchMessageStart, pfrom->GetMagic(chainparams), MESSAGE_START_SIZE) != 0)
        {
            // Setting the cleanSubVer string allows us to present this peer in the bantable
            // with an unknown type.
            if (memcmp(msg.hdr.pchMessageStart, chainparams.MessageStart(), MESSAGE_START_SIZE) == 0)
            {
                pfrom->cleanSubVer = "Unknown Network application";
            }

            LOG(NET, "PROCESSMESSAGE: INVALID MESSAGESTART %s peer=%s\n", SanitizeString(msg.hdr.GetCommand()),
                pfrom->GetLogName());
            if (!pfrom->fWhitelisted)
            {
                // ban for 4 hours
                dosMan.Ban(pfrom->addr, pfrom->cleanSubVer, BanReasonInvalidMessageStart, 4 * 60 * 60);
            }
            fOk = false;
            break;
        }

        // Read header
        CMessageHeader &hdr = msg.hdr;
        if (!hdr.IsValid(pfrom->GetMagic(chainparams)))
        {
            LOG(NET, "PROCESSMESSAGE: ERRORS IN HEADER %s peer=%s\n", SanitizeString(hdr.GetCommand()),
                pfrom->GetLogName());
            continue;
        }
        std::string strCommand = hdr.GetCommand();

        // Message size
        unsigned int nMessageSize = hdr.nMessageSize;

        CDataStream &vRecv = msg.vRecv;

#if 0 // Do not waste my CPU calculating a checksum provided by an untrusted node
      // TCP already has one that is sufficient for network errors.  The checksum does not increase security since
      // an attacker can always provide a bad message with a good checksum.
      // This code is removed by comment so it is clear that it is a deliberate omission.
        if (hdr.nChecksum != 0)
        {
            uint256 hash = Hash(vRecv.begin(), vRecv.begin() + nMessageSize);
            unsigned int nChecksum = ReadLE32((unsigned char *)&hash);
            if (nChecksum != hdr.nChecksum)
            {
                LOG(NET, "%s(%s, %u bytes): CHECKSUM ERROR nChecksum=%08x hdr.nChecksum=%08x\n", __func__,
                    SanitizeString(strCommand), nMessageSize, nChecksum, hdr.nChecksum);
                continue;
            }
        }
#endif

        // Process message
        bool fRet = false;
        try
        {
            fRet = ProcessMessage(pfrom, strCommand, hdr.msgCookie, vRecv, msg.nStopwatch);
            if (shutdown_threads.load() == true)
            {
                return false;
            }
        }
        catch (const std::ios_base::failure &e)
        {
            pfrom->PushMessageWithCookie(NetMsgType::REJECT, hdr.msgCookie | 0xffff, strCommand, REJECT_MALFORMED,
                std::string("error parsing message"));
            if (strstr(e.what(), "end of data"))
            {
                // Allow exceptions from under-length message on vRecv
                LOG(NET,
                    "%s(%s, %u bytes): Exception '%s' caught, normally caused by a message being shorter than "
                    "its stated length\n",
                    __func__, SanitizeString(strCommand), nMessageSize, e.what());
            }
            else if (strstr(e.what(), "size too large"))
            {
                // Allow exceptions from over-long size
                LOG(NET, "%s(%s, %u bytes): Exception '%s' caught\n", __func__, SanitizeString(strCommand),
                    nMessageSize, e.what());
            }
            else
            {
                PrintExceptionContinue(&e, "ProcessMessages()");
            }
        }
        catch (const boost::thread_interrupted &)
        {
            throw;
        }
        catch (const std::exception &e)
        {
            PrintExceptionContinue(&e, "ProcessMessages()");
        }
        catch (...)
        {
            PrintExceptionContinue(nullptr, "ProcessMessages()");
        }

        if (!fRet)
            LOG(NET, "%s(%s, %u bytes) FAILED peer %s\n", __func__, SanitizeString(strCommand), nMessageSize,
                pfrom->GetLogName());

        // Swap back to the original peer if we just processed a priority message
        if (fIsPriority)
            pfrom = pfrom_original;

        if (msgsProcessed > MAX_GETDATA_REQUESTS)
            break; // let someone else do something periodically
    }

    if (!vPriorityRecvQ_delay.empty())
    {
        // re-add the priority messages we delayed back to the queue so that we can try them again later
        LOCK(cs_priorityRecvQ);
        vPriorityRecvQ.insert(vPriorityRecvQ.end(), vPriorityRecvQ_delay.begin(), vPriorityRecvQ_delay.end());
        if (!vPriorityRecvQ.empty())
            fPriorityRecvMsg.store(true);
    }

    return fOk;
}

bool SendMessages(CNode *pto)
{
    const Consensus::Params &consensusParams = Params().GetConsensus();
    {
        // First set fDisconnect if appropriate.
        pto->DisconnectIfBanned();

        // Check for an internal disconnect request and if true then set fDisconnect. This would typically happen
        // during initial sync when a peer has a slow connection and we want to disconnect them.  We want to then
        // wait for any blocks that are still in flight before disconnecting, rather than re-requesting them again.
        if (pto->fDisconnectRequest)
        {
            NodeId nodeid = pto->GetId();
            int nInFlight = requester.GetNumBlocksInFlight(nodeid);
            LOG(IBD, "peer %s, checking disconnect request with %d in flight blocks\n", pto->GetLogName(), nInFlight);
            if (nInFlight == 0)
            {
                pto->fDisconnect = true;
                LOG(IBD, "peer %s, disconnect request was set, so disconnected\n", pto->GetLogName());
            }
        }

        // Now exit early if disconnecting or the version handshake is not complete.  We must not send PING or other
        // connection maintenance messages before the handshake is done.
        if (pto->fDisconnect || !pto->fSuccessfullyConnected)
            return true;

        //
        // Message: ping
        //
        bool pingSend = false;
        if (pto->fPingQueued)
        {
            // RPC ping request by user
            pingSend = true;
        }
        if (pto->nPingNonceSent == 0 && pto->nPingUsecStart + PING_INTERVAL * 1000000 < (int64_t)GetStopwatchMicros())
        {
            // Ping automatically sent as a latency probe & keepalive.
            pingSend = true;
        }
        if (pingSend)
        {
            uint64_t nonce = 0;
            while (nonce == 0)
            {
                GetRandBytes((unsigned char *)&nonce, sizeof(nonce));
            }
            pto->fPingQueued = false;
            pto->nPingUsecStart = GetStopwatchMicros();
            pto->nPingNonceSent = nonce;
            pto->PushMessage(NetMsgType::PING, nonce);
        }

        // Check to see if there are any thin type blocks in flight that have gone beyond the
        // timeout interval. If so then we need to disconnect them so that the thintype data is nullified.
        // We could null the associated data here but that would possibly cause a node to be banned later if
        // the thin type block finally did show up, so instead we just disconnect this slow node.
        thinrelay.CheckForDownloadTimeout(pto);

        // Check for block download timeout and disconnect node if necessary. Does not require cs_main.
        int64_t nNow = GetStopwatchMicros();
        requester.DisconnectOnDownloadTimeout(pto, consensusParams, nNow);

        // Address refresh broadcast
        if (!IsInitialBlockDownload() && pto->nNextLocalAddrSend < nNow)
        {
            AdvertiseLocal(pto);
            pto->nNextLocalAddrSend = PoissonNextSend(nNow, AVG_LOCAL_ADDRESS_BROADCAST_INTERVAL);
        }

        //
        // Message: addr
        //
        if (pto->nNextAddrSend < nNow)
        {
            LOCK(pto->cs_vSend);
            pto->nNextAddrSend = PoissonNextSend(nNow, AVG_ADDRESS_BROADCAST_INTERVAL);
            std::vector<CAddress> vAddr;
            vAddr.reserve(pto->vAddrToSend.size());

            const char *msg_type;
            if (pto->m_wants_addrv2)
            {
                msg_type = NetMsgType::ADDRV2;
            }
            else
            {
                msg_type = NetMsgType::ADDR;
            }

            for (const CAddress &addr : pto->vAddrToSend)
            {
                if (!pto->addrKnown.contains(addr.GetKey()))
                {
                    pto->addrKnown.insert(addr.GetKey());
                    vAddr.push_back(addr);
                    // receiver rejects addr messages larger than 1000
                    if (vAddr.size() >= 1000)
                    {
                        pto->PushMessage(msg_type, vAddr);
                        vAddr.clear();
                    }
                }
            }
            if (!vAddr.empty())
            {
                pto->PushMessage(msg_type, vAddr);
            }
            pto->vAddrToSend.clear();
        }

        CNodeState statem(CAddress(), "");
        const CNodeState *state = &statem;
        {
            CNodeStateAccessor stateAccess(nodestate, pto->GetId());
            if (state == nullptr)
            {
                return true;
            }
            statem = *stateAccess;
        }

        // If a sync has been started check whether we received the first batch of headers requested within the timeout
        // period. If not then disconnect and ban the node and a new node will automatically be selected to start the
        // headers download.
        if ((state->fSyncStarted) && (state->nSyncStartTime < GetTime() - INITIAL_HEADERS_TIMEOUT) &&
            (!state->fFirstHeadersReceived) && !pto->fWhitelisted)
        {
            // pto->fDisconnect = true;
            LOGA("Initial headers were either not received or not received before the timeout\n", pto->GetLogName());
        }

        // Start block sync
        if (pindexBestHeader == nullptr)
            pindexBestHeader = chainActive.Tip();
        // Download if this is a nice peer, or we have no nice peers and this one might do.
        bool fFetch = state->fPreferredDownload || (nPreferredDownload.load() == 0 && !pto->fOneShot);
        if (!state->fSyncStarted && !fImporting && !fReindex)
        {
            // Only allow the downloading of headers from a single pruned peer.
            static int nSyncStartedPruned = 0;
            if (pto->fClient && nSyncStartedPruned >= 1)
                fFetch = false;

            // Only actively request headers from a single peer, unless we're close to today.
            if ((nSyncStarted < MAX_HEADER_REQS_DURING_IBD && fFetch) ||
                chainActive.Tip()->GetBlockTime() > GetAdjustedTime() - SINGLE_PEER_REQUEST_MODE_AGE)
            {
                const CBlockIndex *pindexStart = chainActive.Tip();
                /* If possible, start at the block preceding the currently
                   best known header.  This ensures that we always get a
                   non-empty list of headers back as long as the peer
                   is up-to-date.  With a non-empty response, we can initialise
                   the peer's known best block.  This wouldn't be possible
                   if we requested starting at pindexBestHeader and
                   got back an empty response.  */
                if (pindexStart->pprev)
                    pindexStart = pindexStart->pprev;
                // BU Bug fix for Core:  Don't start downloading headers unless our chain is shorter
                if (pindexStart->height() < pto->nStartingHeight)
                {
                    CNodeStateAccessor modableState(nodestate, pto->GetId());
                    modableState->fSyncStarted = true;
                    modableState->nSyncStartTime = GetTime();
                    modableState->fRequestedInitialBlockAvailability = true;
                    modableState->nFirstHeadersExpectedHeight = pindexStart->height();
                    nSyncStarted++;

                    if (pto->fClient)
                        nSyncStartedPruned++;

                    LOG(NET, "initial getheaders (%d) to peer=%s (startheight:%d)\n", pindexStart->height(),
                        pto->GetLogName(), pto->nStartingHeight);
                    pto->PushMessage(NetMsgType::GETHEADERS, chainActive.GetLocator(pindexStart), uint256());
                }
            }
        }

        // During IBD and when a new NODE_NETWORK peer connects we have to ask for if it has our best header in order
        // to update our block availability. We only want/need to do this only once per peer (if the initial batch of
        // headers has still not been etirely donwnloaded yet then the block availability will be updated during that
        // process rather than here).
        if (IsInitialBlockDownload() && !state->fRequestedInitialBlockAvailability &&
            state->pindexBestKnownBlock == nullptr && !fReindex && !fImporting)
        {
            if (!pto->fClient)
            {
                CNodeStateAccessor(nodestate, pto->GetId())->fRequestedInitialBlockAvailability = true;

                // We only want one single header so we pass a null CBlockLocator.
                pto->PushMessage(NetMsgType::GETHEADERS, CBlockLocator(), pindexBestHeader.load()->GetBlockHash());
                LOG(NET | BLK, "Requesting header for initial blockavailability, peer=%s block=%s height=%d\n",
                    pto->GetLogName(), pindexBestHeader.load()->GetBlockHash().ToString(),
                    pindexBestHeader.load()->height());
            }
        }

        // Resend wallet transactions that haven't gotten in a block yet
        // Except during reindex, importing and IBD, when old wallet
        // transactions become unconfirmed and spams other nodes.
        if (!fReindex && !fImporting && !IsInitialBlockDownload())
        {
            GetMainSignals().Broadcast(nTimeBestReceived.load());
        }

        //
        // Try sending block announcements via headers
        //
        {
            // If we have less than MAX_BLOCKS_TO_ANNOUNCE in our
            // list of block hashes we're relaying, and our peer wants
            // headers announcements, then find the first header
            // not yet known to our peer but would connect, and send.
            // If no header would connect, or if we have too many
            // blocks, or if the peer doesn't want headers, just
            // add all to the inv queue.
            std::vector<uint256> vBlocksToAnnounce;
            {
                // Make a copy so that we do not need to keep
                // cs_inventory which cannot be taken before cs_main.
                LOCK(pto->cs_inventory);
                vBlocksToAnnounce.swap(pto->vBlockHashesToAnnounce);
            }

            std::vector<CBlockHeader> vHeaders;
            bool fRevertToInv = (!state->fPreferHeaders || vBlocksToAnnounce.size() > MAX_BLOCKS_TO_ANNOUNCE);
            CBlockIndex *pBestIndex = nullptr; // last header queued for delivery

            // Ensure pindexBestKnownBlock is up-to-date
            requester.ProcessBlockAvailability(pto->id);

            if (!fRevertToInv)
            {
                bool fFoundStartingHeader = false;
                // Try to find first header that our peer doesn't have, and
                // then send all headers past that one.  If we come across any
                // headers that aren't on chainActive, give up.
                for (const uint256 &hash : vBlocksToAnnounce)
                {
                    CBlockIndex *pindex = nullptr;
                    pindex = LookupBlockIndex(hash);

                    // Skip blocks that we don't know about.
                    if (!pindex)
                        continue;

                    if (pBestIndex != nullptr && pindex->pprev != pBestIndex)
                    {
                        // This means that the list of blocks to announce don't
                        // connect to each other.
                        // This shouldn't really be possible to hit during
                        // regular operation (because reorgs should take us to
                        // a chain that has some block not on the prior chain,
                        // which should be caught by the prior check), but one
                        // way this could happen is by using invalidateblock /
                        // reconsiderblock repeatedly on the tip, causing it to
                        // be added multiple times to vBlocksToAnnounce.
                        // Robustly deal with this rare situation by reverting
                        // to an inv.
                        fRevertToInv = true;
                        break;
                    }
                    pBestIndex = pindex;
                    if (fFoundStartingHeader)
                    {
                        // add this to the headers message
                        vHeaders.push_back(pindex->GetBlockHeader());
                    }
                    else if (PeerHasHeader(state, pindex))
                    {
                        continue; // keep looking for the first new block
                    }
                    else if (pindex->pprev == nullptr || PeerHasHeader(state, pindex->pprev))
                    {
                        // Peer doesn't have this header but they do have the prior one.
                        // Start sending headers.
                        fFoundStartingHeader = true;
                        vHeaders.push_back(pindex->GetBlockHeader());
                    }
                    else
                    {
                        // Peer doesn't have this header or the prior one -- nothing will
                        // connect, so bail out.
                        fRevertToInv = true;
                        break;
                    }
                }
            }
            if (fRevertToInv)
            {
                // If falling back to using an inv, just try to inv the tip.
                // The last entry in vBlocksToAnnounce was our tip at some point
                // in the past.
                if (!vBlocksToAnnounce.empty())
                {
                    for (const uint256 &hashToAnnounce : vBlocksToAnnounce)
                    {
                        CBlockIndex *pindex = nullptr;
                        pindex = LookupBlockIndex(hashToAnnounce);

                        // Skip blocks that we don't know about.
                        if (!pindex)
                            continue;

                        // If the peer announced this block to us, don't inv it back.
                        // (Since block announcements may not be via inv's, we can't solely rely on
                        // setInventoryKnown to track this.)
                        if (!PeerHasHeader(state, pindex))
                        {
                            if (pto->fPeerWantsINV2)
                                pto->PushInventory(CInv2(MSG_BLOCK, hashToAnnounce));
                            else
                                pto->PushInventory(CInv(MSG_BLOCK, hashToAnnounce));
                            LOG(NET, "%s: sending inv peer=%d hash=%s\n", __func__, pto->id, hashToAnnounce.ToString());
                        }
                    }
                }
            }
            else if (!vHeaders.empty())
            {
                if (vHeaders.size() > 1)
                {
                    LOG(NET, "%s: %u headers, range (%s, %s), to peer=%d\n", __func__, vHeaders.size(),
                        vHeaders.front().GetHash().ToString(), vHeaders.back().GetHash().ToString(), pto->id);
                }
                else
                {
                    LOG(NET, "%s: sending header %s to peer=%d\n", __func__, vHeaders.front().GetHash().ToString(),
                        pto->id);
                }
                pto->PushMessage(NetMsgType::HEADERS, vHeaders);
                CNodeStateAccessor(nodestate, pto->GetId())->pindexBestHeaderSent = pBestIndex;
            }
        }

        //
        // Message: inventory
        //
        // We must send all INV's before returning otherwise, under very heavy transaction rates, we could end up
        // falling behind in sending INV's and vInventoryToSend could possibly get quite large.
        bool haveInv2Send = false;
        {
            LOCK(pto->cs_inventory);
            haveInv2Send = !pto->vInventoryToSend.empty() || !pto->vInventoryToSend2.empty();
        }
        if (haveInv2Send && !pto->fPeerWantsINV2)
        {
            std::vector<CInv> vInvSend;
            FastRandomContext rnd;
            while (1)
            {
                // Send message INV up to the MAX_INV_TO_SEND. Once we reach the max then send the INV message
                // and if there is any remaining it will be sent on the next iteration until vInventoryToSend is empty.
                int nToErase = 0;
                {
                    // BU - here we only want to forward message inventory if our peer has actually been requesting
                    // useful data or giving us useful data.  We give them 2 minutes to be useful but then choke off
                    // their inventory.  This prevents fake peers from connecting and listening to our inventory
                    // while providing no value to the network.
                    // However we will still send them block inventory in the case they are a pruned node or wallet
                    // waiting for block announcements, therefore we have to check each inv in pto->vInventoryToSend.
                    bool fChokeTxInv =
                        (pto->nActivityBytes == 0 && (GetStopwatchMicros() - pto->nStopwatchConnected) > 120 * 1000000);

                    // Find INV's which should be sent, save them to vInvSend, and then erase from vInventoryToSend.
                    LOCK(pto->cs_inventory);
                    int invsz = std::min((int)pto->vInventoryToSend.size(), MAX_INV_TO_SEND);
                    vInvSend.reserve(invsz);
                    for (const CInv &inv : pto->vInventoryToSend)
                    {
                        nToErase++;
                        if (inv.type == MSG_TX)
                        {
                            if (fChokeTxInv)
                                continue;
                            // randomly don't inv but always send inventory to spv clients
                            if (((rnd.rand32() % 100) < randomlyDontInv.Value()) && !pto->fClient)
                                continue;
                            // skip if we already know about this one
                            if (pto->filterInventoryKnown.contains(inv.hash))
                                continue;
                        }
                        vInvSend.push_back(inv);
                        pto->filterInventoryKnown.insert(inv.hash);

                        if (vInvSend.size() >= MAX_INV_TO_SEND)
                            break;
                    }

                    if (nToErase > 0)
                    {
                        pto->vInventoryToSend.erase(
                            pto->vInventoryToSend.begin(), pto->vInventoryToSend.begin() + nToErase);
                    }
                    else // exit out of the while loop if nothing was done
                    {
                        break;
                    }
                }

                // To maintain proper locking order we have to push the message when we do not hold cs_inventory which
                // was held in the section above.
                if (nToErase > 0)
                {
                    if (!vInvSend.empty())
                    {
                        pto->PushMessage(NetMsgType::INV, vInvSend);
                        vInvSend.clear();
                    }
                }
            }
        }
        if (haveInv2Send && pto->fPeerWantsINV2)
        {
            std::vector<CInv2> vInvSend;
            FastRandomContext rnd;
            while (1)
            {
                // Send message INV up to the MAX_INV_TO_SEND. Once we reach the max then send the INV message
                // and if there is any remaining it will be sent on the next iteration until vInventoryToSend is empty.
                int nToErase = 0;
                {
                    // BU - here we only want to forward message inventory if our peer has actually been requesting
                    // useful data or giving us useful data.  We give them 2 minutes to be useful but then choke off
                    // their inventory.  This prevents fake peers from connecting and listening to our inventory
                    // while providing no value to the network.
                    // However we will still send them block inventory in the case they are a pruned node or wallet
                    // waiting for block announcements, therefore we have to check each inv in pto->vInventoryToSend.
                    bool fChokeTxInv =
                        (pto->nActivityBytes == 0 && (GetStopwatchMicros() - pto->nStopwatchConnected) > 120 * 1000000);

                    // Find INV's which should be sent, save them to vInvSend, and then erase from vInventoryToSend.
                    LOCK(pto->cs_inventory);
                    int invsz = std::min((int)pto->vInventoryToSend2.size(), MAX_INV_TO_SEND);
                    vInvSend.reserve(invsz);
                    for (const CInv2 &inv : pto->vInventoryToSend2)
                    {
                        nToErase++;
                        if (inv.type == MSG_TX)
                        {
                            if (fChokeTxInv)
                                continue;
                            // randomly don't inv but always send inventory to spv clients
                            if (((rnd.rand32() % 100) < randomlyDontInv.Value()) && !pto->fClient)
                                continue;
                            // skip if we already know about this one
                            if (pto->filterInventoryKnown.contains(inv.hash))
                                continue;
                        }
                        vInvSend.push_back(inv);
                        pto->filterInventoryKnown.insert(inv.hash);

                        if (vInvSend.size() >= MAX_INV_TO_SEND)
                            break;
                    }

                    if (nToErase > 0)
                    {
                        pto->vInventoryToSend2.erase(
                            pto->vInventoryToSend2.begin(), pto->vInventoryToSend2.begin() + nToErase);
                    }
                    else // exit out of the while loop if nothing was done
                    {
                        break;
                    }
                }

                // To maintain proper locking order we have to push the message when we do not hold cs_inventory which
                // was held in the section above.
                if (nToErase > 0)
                {
                    if (!vInvSend.empty())
                    {
                        pto->PushMessage(NetMsgType::INV, vInvSend);
                        vInvSend.clear();
                    }
                }
            }
        }

        // Send any CAPD communication that might be pending
        if (pto->capd)
            pto->capd->FlushMessages();

        // If the chain is not entirely sync'd then look for new blocks to download.
        //
        // Also check an edge condition, where we've invalidated a chain and set the pindexBestHeader to the
        // new most work chain, as a result we may end up just connecting whatever blocks are in setblockindexcandidates
        // resulting in pindexBestHeader equalling the chainActive.Tip() causing us to stop checking for more blocks to
        // download (our chain will now not sync until the next block announcement is received). Therefore, if the
        // best invalid chain work is still greater than our chaintip then we have to keep looking for more blocks
        // to download.
        //
        // Use temporaries for the chain tip and best invalid because they are both atomics and either could
        // be nullified between the two calls.
        CBlockIndex *pTip = chainActive.Tip();
        CBlockIndex *pBestInvalid = pindexBestInvalid.load();
        if (!IsChainSyncd() || (pBestInvalid && pTip && pBestInvalid->chainWork() > pTip->chainWork()))
        {
            TRY_LOCK(cs_requestNextBlocks, locked);
            if (locked)
            {
                // Request the next blocks. Mostly this will get executed during IBD but sometimes even
                // when the chain is syncd a block will get request via this method.
                requester.RequestNextBlocksToDownload(pto);
            }
        }
    }
    return true;
}


bool HandleHeaderPathMessage(CDataStream &vRecv, CNode *pfrom, uint32_t msgCookie)
{
    LOG(NET, "Handling block path request from peer %s\n", pfrom->GetLogName());
    uint32_t fromHeight;
    uint256 fromHash;
    uint32_t toHeight;
    uint256 toHash;

    vRecv >> fromHeight >> fromHash >> toHeight >> toHash;

    CBlockIndex *from = nullptr;
    CBlockIndex *to = nullptr;

    std::vector<CBlockHeader> headers;

    {
        READLOCK(chainActive.cs_chainLock);

        // Find where we are going from and to
        if (fromHeight == std::numeric_limits<uint32_t>::max())
            from = chainActive.Tip();
        // returns nullptr if too big
        else if (fromHeight != 0 || fromHash == uint256())
            from = chainActive._idx(fromHeight);
        else
            from = chainActive._lookup(fromHash);

        if (toHeight == std::numeric_limits<uint32_t>::max())
            to = chainActive.Tip();
        else if (toHeight != 0 || fromHash == uint256())
            to = chainActive._idx(toHeight); // returns nullptr if too big
        else
            to = chainActive._lookup(toHash);

        // If we do not find these blocks, then give up
        if (from == nullptr || to == nullptr)
        {
            pfrom->PushMessageWithCookie(NetMsgType::REJECT, msgCookie | 0xFFFF, std::string(NetMsgType::GETHEADERPATH),
                REJECT_FORK, std::string("start or end header not on main chain"));
            return false;
        }
        // We only accept paths backwards.  Caller can just swap their from and to fields
        if (from->height() < to->height())
        {
            pfrom->PushMessageWithCookie(NetMsgType::REJECT, msgCookie | 0xFFFF, std::string(NetMsgType::GETHEADERPATH),
                REJECT_MALFORMED,
                std::string("start precedes end, but header paths must move from backwards through the chain"));
            return false;
        }

        // Just in case we were given a hash, get the height
        toHeight = to->height();
        fromHeight = from->height();

        uint32_t curHeight = fromHeight;

        std::vector<uint32_t> path;
        path.push_back(curHeight);
        while (curHeight != toHeight)
        {
            uint32_t nextHeight = GetConsensusAncestorHeight(curHeight);
            if (nextHeight < toHeight) // Ancestor jumps back too far, have to use previous
            { // Note this will switch the log vs linear path.
                curHeight = curHeight - 1;
                path.push_back(curHeight);
            }
            else if (nextHeight > toHeight)
            {
                curHeight = nextHeight;
                path.push_back(curHeight);
                // Heuristic, if we are in the linear path, and have a long way to go we need to switch back to the log
                // path even if not much progress is made for a few blocks. There's a better algorithm to generate the
                // optimal path but this will do. Investigate changing the ANCESTOR_HASH_IF_ODD to (height - 5040 &
                // ~0xff)+1 or similar to jump back a linear amount that also puts us 1 away from a log path with larger
                // jumps

                // 16 log hops is 65536, 16 linear hops is 80640 so 16 linear hops is where moving to the log path
                // starts to make sense
                if ((nextHeight & 1) && (nextHeight > toHeight + ANCESTOR_HASH_IF_ODD * 16))
                {
                    curHeight = curHeight - 1;
                    path.push_back(curHeight);
                }
            }
            else // We hit it!
            {
                curHeight = nextHeight;
                path.push_back(curHeight);
            }
        }

        // Ok we have a path, now produce the headers list to get there
        auto sz = path.size();
        headers.resize(path.size());

        for (size_t i = 0; i < sz; i++)
        {
            headers[i] = chainActive._idx(path[i])->header;
        }
    }

    pfrom->PushMessageWithCookie(NetMsgType::HEADERPATH, msgCookie | 0xFFFF, headers);
    return true;
}
