// Copyright (c) 2009-2010 Satoshi Nakamoto
// Copyright (c) 2009-2015 The Bitcoin Core developers
// Copyright (c) 2015-2022 The Bitcoin Unlimited developers
// Copyright (C) 2020 Tom Zander <tomz@freedommail.ch>
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include "protocol.h"

#include "compat.h"
#include "streams.h"
#include "util.h"
#include "utilstrencodings.h"

#ifndef WIN32
#include <arpa/inet.h>
#endif

namespace NetMsgType
{
const char *VERSION = "version";
const char *VERACK = "verack";
const char *ADDR = "addr";
const char *ADDRV2 = "addrv2";
const char *INV = "inv";
const char *GETDATA = "getdata";
const char *EXTGETDATA = "extgetdata";
const char *MERKLEBLOCK = "merkleblock";
const char *GETBLOCKS = "getblocks";
const char *GETHEADERS = "getheaders";
const char *TX = "tx";
const char *HEADERS = "headers";
const char *BLOCK = "block";
const char *GETADDR = "getaddr";
const char *MEMPOOL = "mempool";
const char *PING = "ping";
const char *PONG = "pong";
const char *NOTFOUND = "notfound";
const char *FILTERLOAD = "filterload";
const char *FILTERADD = "filteradd";
const char *FILTERCLEAR = "filterclear";
const char *FILTERSIZEXTHIN = "filtersizext";
const char *REJECT = "reject";
const char *SENDHEADERS = "sendheaders";
// BUIP010 Xtreme Thinblocks - begin section
const char *THINBLOCK = "thinblock";
const char *XTHINBLOCK = "xthinblock";
const char *XBLOCKTX = "xblocktx";
const char *GET_XBLOCKTX = "get_xblocktx";
const char *GET_XTHIN = "get_xthin";
const char *GET_THIN = "get_thin";
// BUIP010 Xtreme Thinblocks - end section
// BUIPXXX Graphene - begin section
const char *GRAPHENEBLOCK = "grblk";
const char *GRAPHENETX = "grblktx";
const char *GET_GRAPHENETX = "get_grblktx";
const char *GET_GRAPHENE = "get_grblk";
const char *GET_GRAPHENE_RECOVERY = "get_grrec";
const char *GRAPHENE_RECOVERY = "grrec";
// BUIPXXX Graphene - end section
// Mempool sync - begin section
const char *MEMPOOLSYNC = "memsync";
const char *MEMPOOLSYNCTX = "memsynctx";
const char *GET_MEMPOOLSYNC = "get_memsync";
const char *GET_MEMPOOLSYNCTX = "getmemsynctx";
// Mempool sync - end section
const char *XPEDITEDREQUEST = "req_xpedited";
const char *XPEDITEDBLK = "Xb";
const char *XPEDITEDTXN = "Xt";
const char *EXTVERSION = "extversion";
const char *XUPDATE = "xupdate";
const char *SENDCMPCT = "sendcmpct";
const char *CMPCTBLOCK = "cmpctblock";
const char *GETBLOCKTXN = "getblocktxn";
const char *BLOCKTXN = "blocktxn";

const char *GETHEADERPATH = "gethdrpath";
const char *HEADERPATH = "hdrpath";

const char *DSPROOF = "dsproof";

const char *REQTXVAL = "req-txval";
const char *RESTXVAL = "res-txval";

const char *CAPDPREFIX = "capd";
const char *CAPDGETINFO = "capdgetinfo";
const char *CAPDINFO = "capdinfo";
const char *CAPDINV = "capdinv";
const char *CAPDGETMSG = "capdgetmsg";
const char *CAPDMSG = "capdmsg";
const char *CAPDQUERY = "capdq";
const char *CAPDQUERYREPLY = "capdqreply";
const char *CAPDREMOVENOTIFY = "capdremove";

const char *TOKENINFO = "tokeninfo";
}; // namespace NetMsgType

/* clang-format off */
static std::map<const int, const char *> ppszTypeName = {
    // "ERROR",  Should never occur
    {MSG_TX, NetMsgType::TX},
    {MSG_BLOCK, NetMsgType::BLOCK},
    // "filtered block",  Should never occur
    {MSG_THINBLOCK, NetMsgType::THINBLOCK}, // thinblock or compact block
    {MSG_XTHINBLOCK, NetMsgType::XTHINBLOCK},
    {MSG_GRAPHENEBLOCK, NetMsgType::GRAPHENEBLOCK},
    {MSG_DOUBLESPENDPROOF, NetMsgType::DSPROOF},
    {MSG_TOKENINFO, NetMsgType::TOKENINFO}
};
/** All known message types. Keep this in the same order as the list of
 * messages above and in protocol.h.
 */
const static std::string allNetMessageTypes[] = {
    NetMsgType::VERSION,
    NetMsgType::VERACK,
    NetMsgType::ADDR,
    NetMsgType::ADDRV2,
    NetMsgType::INV,
    NetMsgType::GETDATA,
    NetMsgType::EXTGETDATA,
    NetMsgType::MERKLEBLOCK,
    NetMsgType::GETBLOCKS,
    NetMsgType::GETHEADERS,
    NetMsgType::TX,
    NetMsgType::HEADERS,
    NetMsgType::BLOCK,
    NetMsgType::GETADDR,
    NetMsgType::MEMPOOL,
    NetMsgType::PING,
    NetMsgType::PONG,
    NetMsgType::NOTFOUND,
    NetMsgType::FILTERLOAD,
    NetMsgType::FILTERADD,
    NetMsgType::FILTERCLEAR,
    NetMsgType::FILTERSIZEXTHIN,
    NetMsgType::REJECT,
    NetMsgType::SENDHEADERS,
    NetMsgType::THINBLOCK,
    NetMsgType::XTHINBLOCK,
    NetMsgType::XBLOCKTX,
    NetMsgType::GET_XBLOCKTX,
    NetMsgType::GET_XTHIN,
    NetMsgType::GET_THIN,
    NetMsgType::GRAPHENEBLOCK,
    NetMsgType::GRAPHENETX,
    NetMsgType::GET_GRAPHENETX,
    NetMsgType::GET_GRAPHENE,
    NetMsgType::GET_GRAPHENE_RECOVERY,
    NetMsgType::GRAPHENE_RECOVERY,
    NetMsgType::MEMPOOLSYNC,
    NetMsgType::MEMPOOLSYNCTX,
    NetMsgType::GET_MEMPOOLSYNC,
    NetMsgType::GET_MEMPOOLSYNCTX,
    NetMsgType::XPEDITEDREQUEST,
    NetMsgType::XPEDITEDBLK,
    NetMsgType::XPEDITEDTXN,
    NetMsgType::EXTVERSION,
    NetMsgType::XUPDATE,
    NetMsgType::SENDCMPCT,
    NetMsgType::CMPCTBLOCK,
    NetMsgType::GETBLOCKTXN,
    NetMsgType::BLOCKTXN,
    NetMsgType::GETHEADERPATH,
    NetMsgType::HEADERPATH,
    NetMsgType::DSPROOF,
    NetMsgType::REQTXVAL,
    NetMsgType::RESTXVAL,
    NetMsgType::CAPDINV,
    NetMsgType::CAPDGETMSG,
    NetMsgType::CAPDMSG,
    NetMsgType::CAPDGETINFO,
    NetMsgType::CAPDINFO,
    NetMsgType::CAPDQUERY,
    NetMsgType::CAPDQUERYREPLY,
    NetMsgType::TOKENINFO
};
/* clang-format on */

const static std::vector<std::string> allNetMessageTypesVec(allNetMessageTypes,
    allNetMessageTypes + ARRAYLEN(allNetMessageTypes));

CMessageHeader::CMessageHeader(const MessageStartChars &pchMessageStartIn, uint32_t _msgCookie)
{
    memcpy(pchMessageStart, pchMessageStartIn, MESSAGE_START_SIZE);
    memset(pchCommand, 0, sizeof(pchCommand));
    nMessageSize = -1;
    msgCookie = _msgCookie;
}

CMessageHeader::CMessageHeader(const MessageStartChars &pchMessageStartIn,
    const char *pszCommand,
    uint32_t _msgCookie,
    unsigned int nMessageSizeIn)
{
    memcpy(pchMessageStart, pchMessageStartIn, MESSAGE_START_SIZE);
    const size_t size = strnlen(pszCommand, COMMAND_SIZE);
    if (size != COMMAND_SIZE)
    {
        memset(pchCommand + size, '\0', COMMAND_SIZE - size);
    }
    memcpy(pchCommand, pszCommand, size);
    nMessageSize = nMessageSizeIn;
    msgCookie = _msgCookie;
}

std::string CMessageHeader::GetCommand() const
{
    return std::string(pchCommand, pchCommand + strnlen(pchCommand, COMMAND_SIZE));
}

bool CMessageHeader::IsValid(const MessageStartChars &pchMessageStartIn) const
{
    // Check start string
    if (memcmp(pchMessageStart, pchMessageStartIn, MESSAGE_START_SIZE) != 0)
        return false;

    // Check the command string for errors
    for (const char *p1 = pchCommand; p1 < pchCommand + COMMAND_SIZE; p1++)
    {
        if (*p1 == 0)
        {
            // Must be all zeros after the first zero
            for (; p1 < pchCommand + COMMAND_SIZE; p1++)
                if (*p1 != 0)
                    return false;
        }
        else if (*p1 < ' ' || *p1 > 0x7E)
            return false;
    }

    // Message size
    if (nMessageSize > MAX_SIZE)
    {
        LOGA("CMessageHeader::IsValid(): (%s, %u bytes) nMessageSize > MAX_SIZE\n", GetCommand(), nMessageSize);
        return false;
    }

    return true;
}


CAddress::CAddress() : CService() { Init(); }
CAddress::CAddress(CService ipIn, uint64_t nServicesIn) : CService(ipIn)
{
    Init();
    nServices = nServicesIn;
}
CAddress::CAddress(CService ipIn, uint64_t nServicesIn, uint32_t nTimeIn) : CService(ipIn)
{
    Init();
    nServices = nServicesIn;
    nTime = nTimeIn;
}

void CAddress::Init()
{
    nServices = NODE_NETWORK;
    nTime = 100000000;
}

CInv::CInv()
{
    type = 0;
    hash.SetNull();
}

CInv::CInv(int typeIn, const uint256 &hashIn)
{
    type = typeIn;
    hash = hashIn;
}

CInv::CInv(const std::string &strType, const uint256 &hashIn)
{
    bool fFound = false;
    for (auto &mi : ppszTypeName)
    {
        if (strType == mi.second)
        {
            type = mi.first;
            fFound = true;
            break;
        }
    }
    if (!fFound)
        throw std::out_of_range(strprintf("CInv::CInv(string, uint256): unknown type '%s'", strType));
    hash = hashIn;
}

bool operator<(const CInv &a, const CInv &b) { return (a.type < b.type || (a.type == b.type && a.hash < b.hash)); }
bool CInv::IsKnownType() const { return (type >= MSG_FIRST_TYPE && type <= MSG_LAST_TYPE); }
const char *CInv::GetCommand() const
{
    if (!IsKnownType())
        throw std::out_of_range(strprintf("CInv::GetCommand(): type=%d unknown type", type));
    return ppszTypeName[type];
}

std::string CInv::ToString() const { return strprintf("%s %s", GetCommand(), hash.ToString()); }

CInv2::CInv2()
{
    type = 0;
    hash.SetNull();
}

CInv2::CInv2(uint8_t typeIn, const uint256 &hashIn)
{
    type = typeIn;
    hash = hashIn;
}

CInv2::CInv2(const std::string &strType, const uint256 &hashIn)
{
    bool fFound = false;
    for (auto &mi : ppszTypeName)
    {
        if (strType == mi.second)
        {
            type = mi.first;
            fFound = true;
            break;
        }
    }
    if (!fFound)
        throw std::out_of_range(strprintf("CInv2::CInv2(string, uint256): unknown type '%s'", strType));
    hash = hashIn;
}

bool operator<(const CInv2 &a, const CInv2 &b) { return (a.type < b.type || (a.type == b.type && a.hash < b.hash)); }
bool CInv2::IsKnownType() const { return (type >= MSG_FIRST_TYPE && type <= MSG_LAST_TYPE); }
const char *CInv2::GetCommand() const
{
    if (!IsKnownType())
        throw std::out_of_range(strprintf("CInv::GetCommand(): type=%d unknown type", type));
    return ppszTypeName[type];
}
std::string CInv2::ToString() const { return strprintf("%s %s", GetCommand(), hash.ToString()); }

CExtInv::CExtInv()
{
    type = 0;
    hash.clear();
}

CExtInv::CExtInv(uint8_t typeIn, const std::vector<uint8_t> &hashIn)
{
    type = typeIn;
    hash = hashIn;
}

CExtInv::CExtInv(uint8_t typeIn, uint64_t &hashIn)
{
    type = typeIn;

    CDataStream ss(0, 0);
    ser_writedata64(ss, hashIn);
    for (auto c : ss)
        hash.push_back(c);
}

CExtInv::CExtInv(const std::string &strType, const std::vector<uint8_t> &hashIn)
{
    bool fFound = false;
    for (auto &mi : ppszTypeName)
    {
        if (strType == mi.second)
        {
            type = mi.first;
            fFound = true;
            break;
        }
    }
    if (!fFound)
        throw std::out_of_range(strprintf("CExtInv::CExtInv(string, std:vector<uint8_t>): unknown type '%s'", strType));
    hash = hashIn;
}

bool CExtInv::IsKnownType() const { return (type >= MSG_FIRST_EXT_TYPE && type <= MSG_LAST_EXT_TYPE); }
const char *CExtInv::GetCommand() const
{
    if (!IsKnownType())
        throw std::out_of_range(strprintf("CExtInv::GetCommand(): type=%d unknown type", type));
    return ppszTypeName[type];
}

std::string CExtInv::ToString() const
{
    // Limit the string size to 64 bytes maximum to prevent log file DOS attacks, but also to accommodate
    // larger (or smaller) than standard 32 byte inv's, as could be the case with subgroup token id's.
    size_t nMaxBytes = std::min((size_t)64, hash.size());
    return strprintf("%s %s", GetCommand(), GetHex(&hash.data()[0], nMaxBytes));
}

const std::vector<std::string> &getAllNetMessageTypes() { return allNetMessageTypesVec; }
