// Copyright (c) 2011-2015 The Bitcoin Core developers
// Copyright (c) 2015-2022 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#if defined(HAVE_CONFIG_H)
#include "nexa-config.h"
#endif

#include "optionsdialog.h"
#include "ui_optionsdialog.h"

#include "guiutil.h"
#include "nexaunits.h"
#include "optionsmodel.h"

#include "main.h" // for DEFAULT_SCRIPTCHECK_THREADS and MAX_SCRIPTCHECK_THREADS
#include "net.h" // for access to the network traffic shapers
#include "netbase.h"
#include "tweak.h"
#include "txdb.h" // for -cache.dbcache defaults

#ifdef ENABLE_WALLET
#include "wallet/wallet.h" // for CWallet::GetRequiredFee()
#endif

#include <boost/lexical_cast.hpp>
#include <thread>

#include <QDataWidgetMapper>
#include <QDir>
#include <QIntValidator>
#include <QLocale>
#include <QMessageBox>
#include <QTimer>

extern CTweak<bool> instantTxns;
extern CTweak<bool> autoTxns;
extern CTweak<bool> tokenWhitelist;

// Contains native language names that are missing from the QT database.  The key is the language code
// whereas the value is the language name as written in the native tounge.
std::map<QString, QString> mapNativeLanguageNames{{"pam", "Pilipino"}, {"la", "Latin"}, {"eo", "Esperanto"}};

OptionsDialog::OptionsDialog(QWidget *parent, bool enableWallet)
    : QDialog(parent), ui(new Ui::OptionsDialog), portValidator(1, 65536, this), // fix memory leaks
      proxyPortValidator(1, 65536, this), model(0), mapper(0)
{
    ui->setupUi(this);
    /* Main elements init */
    ui->threadsScriptVerif->setMinimum(-GetNumCores());
    ui->threadsScriptVerif->setMaximum(MAX_SCRIPTCHECK_THREADS);
    ui->reindexOnStartup->setEnabled(true);
    ui->resyncOnStartup->setEnabled(true);
    ui->rescanOnStartup->setEnabled(true);

/* Network elements init */
#ifndef USE_UPNP
    ui->mapPortUpnp->setEnabled(false);
#endif

    ui->proxyIp->setEnabled(false);
    ui->proxyPort->setEnabled(false);
    ui->proxyPort->setValidator(&portValidator); // new QIntValidator(1, 65535, this));

    ui->proxyIpTor->setEnabled(false);
    ui->proxyPortTor->setEnabled(false);
    ui->proxyPortTor->setValidator(&proxyPortValidator);

#ifdef ENABLE_WALLET
    if (instantTxns.Value())
    {
        ui->spendZeroConfChange->setEnabled(false);
        ui->instantTransactions->setChecked(true);
    }

    if (instantTxns.Value() || GetBoolArg("-spendzeroconfchange", false))
    {
        ui->autoConsolidation->setEnabled(true);
    }
    else
    {
        ui->autoConsolidation->setEnabled(false);
    }

    if (autoTxns.Value())
    {
        ui->autoConsolidation->setChecked(true);
    }
    else
    {
        ui->autoConsolidation->setChecked(false);
    }

    ui->tokenWhitelist->setEnabled(true);
    if (tokenWhitelist.Value())
    {
        ui->tokenWhitelist->setChecked(true);
    }
    else
    {
        ui->tokenWhitelist->setChecked(false);
    }

#endif


    connect(ui->connectSocks, SIGNAL(toggled(bool)), ui->proxyIp, SLOT(setEnabled(bool)));
    connect(ui->connectSocks, SIGNAL(toggled(bool)), ui->proxyPort, SLOT(setEnabled(bool)));
    connect(ui->connectSocks, SIGNAL(toggled(bool)), this, SLOT(updateProxyValidationState()));

    connect(ui->connectSocksTor, SIGNAL(toggled(bool)), ui->proxyIpTor, SLOT(setEnabled(bool)));
    connect(ui->connectSocksTor, SIGNAL(toggled(bool)), ui->proxyPortTor, SLOT(setEnabled(bool)));
    connect(ui->connectSocksTor, SIGNAL(toggled(bool)), this, SLOT(updateProxyValidationState()));

/* Window elements init */
#ifdef Q_OS_MAC
    /* remove Window tab on Mac */
    ui->tabWidget->removeTab(ui->tabWidget->indexOf(ui->tabWindow));
#endif

    /* remove Wallet tab in case of -disablewallet */
    if (!enableWallet)
    {
        ui->tabWidget->removeTab(ui->tabWidget->indexOf(ui->tabWallet));
    }

    /* Display elements init */
    QDir translations(":translations");

    ui->bitcoinAtStartup->setToolTip(ui->bitcoinAtStartup->toolTip().arg(tr(PACKAGE_NAME)));
    ui->bitcoinAtStartup->setText(ui->bitcoinAtStartup->text().arg(tr(PACKAGE_NAME)));

    ui->lang->setToolTip(ui->lang->toolTip().arg(tr(PACKAGE_NAME)));
    ui->lang->addItem(QString("(") + tr("default") + QString(")"), QVariant(""));
    Q_FOREACH (const QString &langStr, translations.entryList())
    {
        QLocale locale(langStr);
        QString languageName = locale.nativeLanguageName();
        if (languageName.isEmpty() && mapNativeLanguageNames.count(langStr))
        {
            languageName = mapNativeLanguageNames[langStr];
        }
        QString countryName = locale.nativeCountryName();

        /** check if the locale name consists of 2 parts (language_country) */
        if (langStr.contains("_"))
        {
            /** display language strings as "native language - native country (locale name)", e.g. "Deutsch -
             * Deutschland (de)" */
            ui->lang->addItem(languageName + QString(" - ") + countryName + QString(" (") + langStr + QString(")"),
                QVariant(langStr));
        }
        else
        {
            /** display language strings as "native language (locale name)", e.g. "Deutsch (de)" */
            ui->lang->addItem(languageName + QString(" (") + langStr + QString(")"), QVariant(langStr));
        }
    }
    ui->thirdPartyTxUrls->setPlaceholderText("https://example.com/tx/%s");

    ui->unit->setModel(new BitcoinUnits(this));

    /* Widget-to-option mapper */
    mapper = new QDataWidgetMapper(this);
    mapper->setSubmitPolicy(QDataWidgetMapper::ManualSubmit);
    mapper->setOrientation(Qt::Vertical);

    /* setup/change UI elements when proxy IPs are invalid/valid */
    ui->proxyIp->setCheckValidator(new ProxyAddressValidator(parent));
    ui->proxyIpTor->setCheckValidator(new ProxyAddressValidator(parent));
    connect(ui->proxyIp, SIGNAL(validationDidChange(QValidatedLineEdit *)), this, SLOT(updateProxyValidationState()));
    connect(
        ui->proxyIpTor, SIGNAL(validationDidChange(QValidatedLineEdit *)), this, SLOT(updateProxyValidationState()));
    connect(ui->proxyPort, SIGNAL(textChanged(const QString &)), this, SLOT(updateProxyValidationState()));
    connect(ui->proxyPortTor, SIGNAL(textChanged(const QString &)), this, SLOT(updateProxyValidationState()));
}

OptionsDialog::~OptionsDialog() { delete ui; }
void OptionsDialog::setModel(OptionsModel *_model)
{
    this->model = _model;

    if (_model)
    {
        /* check if client restart is needed and show persistent message */
        if (_model->isRestartRequired())
            showRestartWarning(true);

        QString strLabel = _model->getOverriddenByCommandLine();
        if (strLabel.isEmpty())
            strLabel = tr("none");
        ui->overriddenByCommandLineLabel->setText(strLabel);

        mapper->setModel(_model);
        setMapper();
        mapper->toFirst();

        updateDefaultProxyNets();
    }

    /* warn when one of the following settings changes by user action (placed here so init via mapper doesn't trigger
     * them) */

    /* Main */
    connect(ui->threadsScriptVerif, SIGNAL(valueChanged(int)), this, SLOT(showRestartWarning()));
    connect(ui->reindexOnStartup, SIGNAL(clicked(bool)), this, SLOT(showRestartWarning()));
    connect(ui->resyncOnStartup, SIGNAL(clicked(bool)), this, SLOT(showRestartWarning()));
    /* Wallet */
    connect(ui->spendZeroConfChange, SIGNAL(clicked(bool)), this, SLOT(setZeroConf()));
    connect(ui->instantTransactions, SIGNAL(clicked(bool)), this, SLOT(setInstant()));
    connect(ui->autoConsolidation, SIGNAL(clicked(bool)), this, SLOT(setAuto()));
    connect(ui->rescanOnStartup, SIGNAL(clicked(bool)), this, SLOT(showRestartWarning()));
    connect(ui->tokenWhitelist, SIGNAL(clicked(bool)), this, SLOT(setTokenWhitelist()));
    /* Network */
    connect(ui->allowIncoming, SIGNAL(clicked(bool)), this, SLOT(showRestartWarning()));
    connect(ui->connectSocks, SIGNAL(clicked(bool)), this, SLOT(showRestartWarning()));
    connect(ui->connectSocksTor, SIGNAL(clicked(bool)), this, SLOT(showRestartWarning()));
    /* Display */
    connect(ui->lang, SIGNAL(valueChanged()), this, SLOT(showRestartWarning()));
    connect(ui->thirdPartyTxUrls, SIGNAL(textChanged(const QString &)), this, SLOT(thirdPartyTxWarning()));
}

void OptionsDialog::setMapper()
{
    /* Main */
    mapper->addMapping(ui->bitcoinAtStartup, OptionsModel::StartAtStartup);
    mapper->addMapping(ui->threadsScriptVerif, OptionsModel::ThreadsScriptVerif);
    mapper->addMapping(ui->reindexOnStartup, OptionsModel::ReindexOnStartup);
    mapper->addMapping(ui->resyncOnStartup, OptionsModel::ResyncOnStartup);

    /* Wallet */
    mapper->addMapping(ui->spendZeroConfChange, OptionsModel::SpendZeroConfChange);
    mapper->addMapping(ui->coinControlFeatures, OptionsModel::CoinControlFeatures);
    mapper->addMapping(ui->instantTransactions, OptionsModel::InstantTransactions);
    mapper->addMapping(ui->autoConsolidation, OptionsModel::AutoConsolidation);
    mapper->addMapping(ui->rescanOnStartup, OptionsModel::RescanOnStartup);
    mapper->addMapping(ui->tokenWhitelist, OptionsModel::TokenWhitelist);

    /* Network */
    mapper->addMapping(ui->mapPortUpnp, OptionsModel::MapPortUPnP);
    mapper->addMapping(ui->allowIncoming, OptionsModel::Listen);

    mapper->addMapping(ui->connectSocks, OptionsModel::ProxyUse);
    mapper->addMapping(ui->proxyIp, OptionsModel::ProxyIP);
    mapper->addMapping(ui->proxyPort, OptionsModel::ProxyPort);

    mapper->addMapping(ui->connectSocksTor, OptionsModel::ProxyUseTor);
    mapper->addMapping(ui->proxyIpTor, OptionsModel::ProxyIPTor);
    mapper->addMapping(ui->proxyPortTor, OptionsModel::ProxyPortTor);

    /* Window */

#ifndef Q_OS_MAC
    mapper->addMapping(ui->minimizeToTray, OptionsModel::MinimizeToTray);
    mapper->addMapping(ui->minimizeOnClose, OptionsModel::MinimizeOnClose);
#endif

    /* Display */
    mapper->addMapping(ui->lang, OptionsModel::Language);
    mapper->addMapping(ui->unit, OptionsModel::DisplayUnit);
    mapper->addMapping(ui->thirdPartyTxUrls, OptionsModel::ThirdPartyTxUrls);
}

void OptionsDialog::setOkButtonState(bool fState) { ui->okButton->setEnabled(fState); }
void OptionsDialog::on_resetButton_clicked()
{
    if (model)
    {
        // confirmation dialog
        QMessageBox::StandardButton btnRetVal = QMessageBox::question(this, tr("Confirm options reset"),
            tr("Client restart required to activate changes.") + "<br><br>" +
                tr("Client will be shut down. Do you want to proceed?"),
            QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel);

        if (btnRetVal == QMessageBox::Cancel)
            return;

        /* reset all options and close GUI */
        model->Reset();
        QApplication::quit();
    }
}

void OptionsDialog::on_okButton_clicked()
{
    mapper->submit();
    accept();
    updateDefaultProxyNets();
}

void OptionsDialog::on_cancelButton_clicked() { reject(); }
void OptionsDialog::showRestartWarning(bool fPersistent)
{
    ui->statusLabel->setStyleSheet("QLabel { color: red; }");

    if (fPersistent)
    {
        ui->statusLabel->setText(tr("Client restart required to activate changes."));
    }
    else
    {
        ui->statusLabel->setText(tr("This change would require a client restart."));
        // clear non-persistent status label after 10 seconds
        // Todo: should perhaps be a class attribute, if we extend the use of statusLabel
        QTimer::singleShot(10000, this, SLOT(clearStatusLabel()));
    }
}

void OptionsDialog::setZeroConf(bool fPersistent)
{
    showRestartWarning(fPersistent);
    setAuto();
}
void OptionsDialog::setInstant(bool fPersistent)
{
#ifdef ENABLE_WALLET
    bool fInstant = (ui->instantTransactions->isEnabled() && ui->instantTransactions->isChecked());
    instantTxns.Set(fInstant);

    if (!fInstant)
        ui->spendZeroConfChange->setEnabled(true);
    else
        ui->spendZeroConfChange->setEnabled(false);

    setAuto();
#endif
}

void OptionsDialog::setAuto(bool fPersistent)
{
#ifdef ENABLE_WALLET
    bool fInstant = (ui->instantTransactions->isEnabled() && ui->instantTransactions->isChecked());
    bool fZeroConf = (ui->spendZeroConfChange->isEnabled() && ui->spendZeroConfChange->isChecked());
    bool fAuto = ui->autoConsolidation->isChecked();

    if (fInstant || fZeroConf)
    {
        ui->autoConsolidation->setEnabled(true);
        if (fAuto)
            autoTxns.Set(true);
        else
            autoTxns.Set(false);
    }
    else
    {
        ui->autoConsolidation->setEnabled(false);
        autoTxns.Set(false);
    }
#endif
}

void OptionsDialog::setTokenWhitelist()
{
#ifdef ENABLE_WALLET
    // should always be enabled but santiy check it
    if (ui->tokenWhitelist->isEnabled() && ui->tokenWhitelist->isChecked())
    {
        tokenWhitelist.Set(true);
    }
    else
    {
        tokenWhitelist.Set(false);
    }
#endif
}

void OptionsDialog::thirdPartyTxWarning(bool fPersistent)
{
    QString str = ui->thirdPartyTxUrls->displayText();

    const char *err = isInvalidThirdPartyTxUrlString(str);

    if (err != nullptr)
    {
        ui->statusLabel->setStyleSheet("QLabel { color: red; }");
        ui->statusLabel->setText(tr(err));
    }
    else // Its a valid URL
    {
        showRestartWarning(fPersistent);
    }
}

void OptionsDialog::clearStatusLabel() { ui->statusLabel->clear(); }
void OptionsDialog::updateProxyValidationState()
{
    QValidatedLineEdit *pUiProxyIp = ui->proxyIp;
    QValidatedLineEdit *otherProxyWidget = (pUiProxyIp == ui->proxyIpTor) ? ui->proxyIp : ui->proxyIpTor;
    if (pUiProxyIp->isValid() && (!ui->proxyPort->isEnabled() || ui->proxyPort->text().toInt() > 0) &&
        (!ui->proxyPortTor->isEnabled() || ui->proxyPortTor->text().toInt() > 0))
    {
        setOkButtonState(otherProxyWidget->isValid()); // only enable ok button if both proxys are valid
        ui->statusLabel->clear();
    }
    else
    {
        setOkButtonState(false);
        ui->statusLabel->setStyleSheet("QLabel { color: red; }");
        ui->statusLabel->setText(tr("The supplied proxy address is invalid."));
    }
}

void OptionsDialog::updateDefaultProxyNets()
{
    proxyType proxy;
    std::string strProxy;
    QString strDefaultProxyGUI;

    GetProxy(NET_IPV4, proxy);
    strProxy = proxy.proxy.ToStringIP() + ":" + proxy.proxy.ToStringPort();
    strDefaultProxyGUI = ui->proxyIp->text() + ":" + ui->proxyPort->text();
    (strProxy == strDefaultProxyGUI.toStdString()) ? ui->proxyReachIPv4->setChecked(true) :
                                                     ui->proxyReachIPv4->setChecked(false);

    GetProxy(NET_IPV6, proxy);
    strProxy = proxy.proxy.ToStringIP() + ":" + proxy.proxy.ToStringPort();
    strDefaultProxyGUI = ui->proxyIp->text() + ":" + ui->proxyPort->text();
    (strProxy == strDefaultProxyGUI.toStdString()) ? ui->proxyReachIPv6->setChecked(true) :
                                                     ui->proxyReachIPv6->setChecked(false);

    GetProxy(NET_TOR3, proxy);
    strProxy = proxy.proxy.ToStringIP() + ":" + proxy.proxy.ToStringPort();
    strDefaultProxyGUI = ui->proxyIp->text() + ":" + ui->proxyPort->text();
    (strProxy == strDefaultProxyGUI.toStdString()) ? ui->proxyReachTor->setChecked(true) :
                                                     ui->proxyReachTor->setChecked(false);
}

ProxyAddressValidator::ProxyAddressValidator(QObject *parent) : QValidator(parent) {}
QValidator::State ProxyAddressValidator::validate(QString &input, int &pos) const
{
    Q_UNUSED(pos);
    // Validate the proxy
    proxyType addrProxy = proxyType(CService(input.toStdString(), 9050), true);
    if (addrProxy.IsValid())
        return QValidator::Acceptable;

    return QValidator::Invalid;
}
