// Copyright (c) 2011-2015 The Bitcoin Core developers
// Copyright (c) 2015-2022 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include <boost/lexical_cast.hpp>

#if defined(HAVE_CONFIG_H)
#include "nexa-config.h"
#endif

#include "optionsmodel.h"

#include "guiutil.h"
#include "nexaunits.h"

#include "amount.h"
#include "init.h"
#include "main.h" // For DEFAULT_SCRIPTCHECK_THREADS
#include "net.h"
#include "tweak.h"
#include "txdb.h" // for -cache.dbcache defaults

#ifdef ENABLE_WALLET
#include "wallet/wallet.h"
#include "wallet/walletdb.h"
#endif

#include <QNetworkProxy>
#include <QSettings>
#include <QStringList>

extern CTweak<bool> instantTxns;
extern CTweak<bool> autoTxns;
extern CTweak<bool> tokenWhitelist;

OptionsModel::OptionsModel(QObject *parent, bool resetSettings) : QAbstractListModel(parent) { Init(resetSettings); }
void OptionsModel::addOverriddenOption(const std::string &option)
{
    strOverriddenByCommandLine += QString::fromStdString(option) + "=" + QString::fromStdString(mapArgs[option]) + " ";
}

// Writes all missing QSettings with their default values
void OptionsModel::Init(bool resetSettings)
{
    if (resetSettings)
        Reset();

    QSettings settings;

    // Ensure restart flag is unset on client startup
    setRestartRequired(false);

    // These are Qt-only settings:

    // Window
    if (!settings.contains("fMinimizeToTray"))
        settings.setValue("fMinimizeToTray", false);
    fMinimizeToTray = settings.value("fMinimizeToTray").toBool();

    if (!settings.contains("fMinimizeOnClose"))
        settings.setValue("fMinimizeOnClose", false);
    fMinimizeOnClose = settings.value("fMinimizeOnClose").toBool();

    // Display
    if (!settings.contains("nDisplayUnit"))
        settings.setValue("nDisplayUnit", BitcoinUnits::NEX);
    nDisplayUnit = settings.value("nDisplayUnit").toInt();

    if (!settings.contains("strThirdPartyTxUrls"))
        settings.setValue("strThirdPartyTxUrls", "https://explorer.nexa.org/tx/%s");
    strThirdPartyTxUrls = settings.value("strThirdPartyTxUrls", "").toString();

    if (!settings.contains("strThirdPartyTokenUrls"))
        settings.setValue("strThirdPartyTokenUrls", "https://explorer.nexa.org/token/%s");
    strThirdPartyTokenUrls = settings.value("strThirdPartyTokenUrls", "").toString();

    if (!settings.contains("fCoinControlFeatures"))
        settings.setValue("fCoinControlFeatures", false);
    fCoinControlFeatures = settings.value("fCoinControlFeatures", false).toBool();

    // These are shared with the core or have a command-line parameter
    // and we want command-line parameters to overwrite the GUI settings.
    //
    // If setting doesn't exist create it with defaults.
    //
    // If SoftSetArg() or SoftSetBoolArg() return false we were overridden
    // by command-line and show this in the UI.

    // Main
    if (!settings.contains("nThreadsScriptVerif"))
        settings.setValue("nThreadsScriptVerif", DEFAULT_SCRIPTCHECK_THREADS);
    if (!SoftSetArg("-par", settings.value("nThreadsScriptVerif").toString().toStdString()))
        addOverriddenOption("-par");

    // Start reindex and then turn off the reindex checkbox so that we do it only one time
    if (!settings.contains("fReindexOnStartup"))
        settings.setValue("fReindexOnStartup", false);
    if (!SoftSetArg("-reindex", settings.value("fReindexOnStartup").toString().toStdString()))
        addOverriddenOption("-reindex");
    settings.setValue("fReindexOnStartup", false);

    // Start resync and then turn off the resync checkbox so that we do it only one time
    if (!settings.contains("fResyncOnStartup"))
        settings.setValue("fResyncOnStartup", false);
    if (!SoftSetArg("-resync", settings.value("fResyncOnStartup").toString().toStdString()))
        addOverriddenOption("-resync");
    settings.setValue("fResyncOnStartup", false);

// Wallet
#ifdef ENABLE_WALLET
    // Start wallet rescan and then turn off the rescan checkbox so that we do it only one time
    if (!settings.contains("fRescanOnStartup"))
        settings.setValue("fRescanOnStartup", false);
    if (!SoftSetArg("-rescan", settings.value("fRescanOnStartup").toString().toStdString()))
        addOverriddenOption("-rescan");
    settings.setValue("fRescanOnStartup", false);

    if (!settings.contains("bSpendZeroConfChange"))
        settings.setValue("bSpendZeroConfChange", true);

    // Turn on instant transactions for QT users only on first launch
    if (!settings.contains("QT_instantTransactions"))
    {
        settings.setValue("QT_instantTransactions", true);
    }
    // after first launch set instant transactions to whatever QT setting was saved on shutdown
    instantTxns.Set(settings.value("QT_instantTransactions").toBool());

    // Set default value of auto consoldation to be the QT instant transactions value
    if (!settings.contains("fAutoConsolidation"))
    {
        settings.setValue("fAutoConsolidation", settings.value("QT_instantTransactions").toBool());
    }

    // Set the spendzeroconfchange options depending on whether instant transactions is on or not.
    if (instantTxns.Value())
    {
        SoftSetBoolArg("-spendzeroconfchange", true);

        // Because of instant transactions this option is automatically overridden.
        addOverriddenOption("-spendzeroconfchange");
    }
    else
    {
        if (!SoftSetBoolArg("-spendzeroconfchange", settings.value("bSpendZeroConfChange").toBool()))
            addOverriddenOption("-spendzeroconfchange");
    }

    // Set Auto Consolidation options depending on whether Instant Transactions or Spend Zero Conf are turned on
    if (instantTxns.Value() || settings.value("bSpendZeroConfChange").toBool())
    {
        if (settings.value("fAutoConsolidation").toBool())
            autoTxns.Set(true);
    }
    else
    {
        autoTxns.Set(false);
    }

    // Turn on token whitelist for QT users only on first launch
    if (!settings.contains("QT_tokenWhitelist"))
    {
        settings.setValue("QT_tokenWhitelist", true);
    }
    // after first launch set tokenWhitelist to whatever QT setting was saved on shutdown
    tokenWhitelist.Set(settings.value("QT_tokenWhitelist").toBool());
#endif

    // Network

    // Turn UPnP on the very first time a user launches the QT wallet
    if (!settings.contains("fUseUPnP") || !settings.contains("fUseUPnP_QT"))
    {
        // settings.setValue("fUseUPnP", DEFAULT_UPNP);
        settings.setValue("fUseUPnP", true);
        settings.setValue("fUseUPnP_QT", true);
    }
    if (!SoftSetBoolArg("-upnp", settings.value("fUseUPnP").toBool()))
        addOverriddenOption("-upnp");

    if (!settings.contains("fListen"))
        settings.setValue("fListen", DEFAULT_LISTEN);
    if (!SoftSetBoolArg("-listen", settings.value("fListen").toBool()))
        addOverriddenOption("-listen");

    if (!settings.contains("fUseProxy"))
        settings.setValue("fUseProxy", false);
    if (!settings.contains("addrProxy"))
        settings.setValue("addrProxy", "127.0.0.1:9050");
    // Only try to set -proxy, if user has enabled fUseProxy
    if (settings.value("fUseProxy").toBool() &&
        !SoftSetArg("-proxy", settings.value("addrProxy").toString().toStdString()))
        addOverriddenOption("-proxy");
    else if (!settings.value("fUseProxy").toBool() && !GetArg("-proxy", "").empty())
        addOverriddenOption("-proxy");

    if (!settings.contains("fUseSeparateProxyTor"))
        settings.setValue("fUseSeparateProxyTor", false);
    if (!settings.contains("addrSeparateProxyTor"))
        settings.setValue("addrSeparateProxyTor", "127.0.0.1:9050");
    // Only try to set -onion, if user has enabled fUseSeparateProxyTor
    if (settings.value("fUseSeparateProxyTor").toBool() &&
        !SoftSetArg("-onion", settings.value("addrSeparateProxyTor").toString().toStdString()))
        addOverriddenOption("-onion");
    else if (!settings.value("fUseSeparateProxyTor").toBool() && !GetArg("-onion", "").empty())
        addOverriddenOption("-onion");

    // Display
    if (!settings.contains("language"))
        settings.setValue("language", "");
    if (!SoftSetArg("-lang", settings.value("language").toString().toStdString()))
        addOverriddenOption("-lang");

    language = settings.value("language").toString();
}

void OptionsModel::Reset()
{
    QSettings settings;

    // Remove all entries from our QSettings object
    settings.clear();

    // default setting for OptionsModel::StartAtStartup - disabled
    if (GUIUtil::GetStartOnSystemStartup())
        GUIUtil::SetStartOnSystemStartup(false);
}

int OptionsModel::rowCount(const QModelIndex &parent) const { return OptionIDRowCount; }
// read QSettings values and return them
QVariant OptionsModel::data(const QModelIndex &index, int role) const
{
    if (role == Qt::EditRole)
    {
        QSettings settings;
        switch (index.row())
        {
        case StartAtStartup:
            return GUIUtil::GetStartOnSystemStartup();
        case ReindexOnStartup:
            return settings.value("fReindexOnStartup");
        case ResyncOnStartup:
            return settings.value("fResyncOnStartup");
        case MinimizeToTray:
            return fMinimizeToTray;
        case MapPortUPnP:
#ifdef USE_UPNP
            return settings.value("fUseUPnP");
#else
            return false;
#endif
        case MinimizeOnClose:
            return fMinimizeOnClose;

        // default proxy
        case ProxyUse:
            return settings.value("fUseProxy", false);
        case ProxyIP:
        {
            // contains IP at index 0 and port at index 1
            QStringList strlIpPort = settings.value("addrProxy").toString().split(":", QString::SkipEmptyParts);
            return strlIpPort.at(0);
        }
        case ProxyPort:
        {
            // contains IP at index 0 and port at index 1
            QStringList strlIpPort = settings.value("addrProxy").toString().split(":", QString::SkipEmptyParts);
            return strlIpPort.at(1);
        }

        // separate Tor proxy
        case ProxyUseTor:
            return settings.value("fUseSeparateProxyTor", false);
        case ProxyIPTor:
        {
            // contains IP at index 0 and port at index 1
            QStringList strlIpPort =
                settings.value("addrSeparateProxyTor").toString().split(":", QString::SkipEmptyParts);
            return strlIpPort.at(0);
        }
        case ProxyPortTor:
        {
            // contains IP at index 0 and port at index 1
            QStringList strlIpPort =
                settings.value("addrSeparateProxyTor").toString().split(":", QString::SkipEmptyParts);
            return strlIpPort.at(1);
        }

#ifdef ENABLE_WALLET
        case SpendZeroConfChange:
            return settings.value("bSpendZeroConfChange");
        case InstantTransactions:
            return settings.value("QT_instantTransactions");
        case AutoConsolidation:
            return settings.value("fAutoConsolidation");
        case RescanOnStartup:
            return settings.value("fRescanOnStartup");
        case TokenWhitelist:
            return settings.value("QT_tokenWhitelist");
#endif
        case DisplayUnit:
            return nDisplayUnit;
        case ThirdPartyTxUrls:
            return strThirdPartyTxUrls;
        case ThirdPartyTokenUrls:
            return strThirdPartyTokenUrls;
        case Language:
            return settings.value("language");
        case CoinControlFeatures:
            return fCoinControlFeatures;
        case DatabaseCache:
            return settings.value("nDatabaseCache");
        case ThreadsScriptVerif:
            return settings.value("nThreadsScriptVerif");
        case Listen:
            return settings.value("fListen");
        default:
            return QVariant();
        }
    }
    return QVariant();
}

const char *isInvalidThirdPartyTxUrlString(QString value)
{
    // Check that the URLs are valid, and https.  Requiring https ensures that certain schemes that auto-execute
    // cannot be used.  Although the user would need to explicitly configure such to happen, preventing
    // this configuration protects the average user, and there isn't much application for weird schemes.
    QStringList listUrls = value.split("|", QString::SkipEmptyParts);
    for (int i = 0; i < listUrls.size(); ++i)
    {
        // remove whitespace and replace our tx placeholder with some valid URL data for validity checking.
        QUrl url = QUrl(listUrls[i].replace("%s", "tx").trimmed(), QUrl::StrictMode);
        if (!url.isValid())
        {
            return ("URL is invalid");
        }
        if (url.scheme().toLower() != "https")
        {
            return ("URL must be https");
        }
    }
    return nullptr;
}

const char *isInvalidThirdPartyTokenUrlString(QString value)
{
    // Check that the URLs are valid, and https.  Requiring https ensures that certain schemes that auto-execute
    // cannot be used.  Although the user would need to explicitly configure such to happen, preventing
    // this configuration protects the average user, and there isn't much application for weird schemes.
    QStringList listUrls = value.split("|", QString::SkipEmptyParts);
    for (int i = 0; i < listUrls.size(); ++i)
    {
        // remove whitespace and replace our token placeholder with some valid URL data for validity checking.
        QUrl url = QUrl(listUrls[i].replace("%s", "token").trimmed(), QUrl::StrictMode);
        if (!url.isValid())
        {
            return ("URL is invalid");
        }
        if (url.scheme().toLower() != "https")
        {
            return ("URL must be https");
        }
    }
    return nullptr;
}

// write QSettings values
bool OptionsModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    bool successful = true; /* set to false on parse error */
    if (role == Qt::EditRole)
    {
        QSettings settings;
        switch (index.row())
        {
        case StartAtStartup:
            successful = GUIUtil::SetStartOnSystemStartup(value.toBool());
            break;
        case ReindexOnStartup:
            settings.setValue("fReindexOnStartup", value);
            setRestartRequired(true);
            break;
        case ResyncOnStartup:
            settings.setValue("fResyncOnStartup", value);
            setRestartRequired(true);
            break;
        case MinimizeToTray:
            fMinimizeToTray = value.toBool();
            settings.setValue("fMinimizeToTray", fMinimizeToTray);
            break;
        case MapPortUPnP: // core option - can be changed on-the-fly
            settings.setValue("fUseUPnP", value.toBool());
            MapPort(value.toBool());
            break;
        case MinimizeOnClose:
            fMinimizeOnClose = value.toBool();
            settings.setValue("fMinimizeOnClose", fMinimizeOnClose);
            break;

        // default proxy
        case ProxyUse:
            if (settings.value("fUseProxy") != value)
            {
                settings.setValue("fUseProxy", value.toBool());
                setRestartRequired(true);
            }
            break;
        case ProxyIP:
        {
            // contains current IP at index 0 and current port at index 1
            QStringList strlIpPort = settings.value("addrProxy").toString().split(":", QString::SkipEmptyParts);
            // if that key doesn't exist or has a changed IP
            if (!settings.contains("addrProxy") || strlIpPort.at(0) != value.toString())
            {
                // construct new value from new IP and current port
                QString strNewValue = value.toString() + ":" + strlIpPort.at(1);
                settings.setValue("addrProxy", strNewValue);
                setRestartRequired(true);
            }
        }
        break;
        case ProxyPort:
        {
            // contains current IP at index 0 and current port at index 1
            QStringList strlIpPort = settings.value("addrProxy").toString().split(":", QString::SkipEmptyParts);
            // if that key doesn't exist or has a changed port
            if (!settings.contains("addrProxy") || strlIpPort.at(1) != value.toString())
            {
                // construct new value from current IP and new port
                QString strNewValue = strlIpPort.at(0) + ":" + value.toString();
                settings.setValue("addrProxy", strNewValue);
                setRestartRequired(true);
            }
        }
        break;

        // separate Tor proxy
        case ProxyUseTor:
            if (settings.value("fUseSeparateProxyTor") != value)
            {
                settings.setValue("fUseSeparateProxyTor", value.toBool());
                setRestartRequired(true);
            }
            break;
        case ProxyIPTor:
        {
            // contains current IP at index 0 and current port at index 1
            QStringList strlIpPort =
                settings.value("addrSeparateProxyTor").toString().split(":", QString::SkipEmptyParts);
            // if that key doesn't exist or has a changed IP
            if (!settings.contains("addrSeparateProxyTor") || strlIpPort.at(0) != value.toString())
            {
                // construct new value from new IP and current port
                QString strNewValue = value.toString() + ":" + strlIpPort.at(1);
                settings.setValue("addrSeparateProxyTor", strNewValue);
                setRestartRequired(true);
            }
        }
        break;
        case ProxyPortTor:
        {
            // contains current IP at index 0 and current port at index 1
            QStringList strlIpPort =
                settings.value("addrSeparateProxyTor").toString().split(":", QString::SkipEmptyParts);
            // if that key doesn't exist or has a changed port
            if (!settings.contains("addrSeparateProxyTor") || strlIpPort.at(1) != value.toString())
            {
                // construct new value from current IP and new port
                QString strNewValue = strlIpPort.at(0) + ":" + value.toString();
                settings.setValue("addrSeparateProxyTor", strNewValue);
                setRestartRequired(true);
            }
        }
        break;

#ifdef ENABLE_WALLET
        case InstantTransactions:
            if (settings.value("QT_instantTransactions") != value)
            {
                settings.setValue("QT_instantTransactions", value);
                instantTxns.Set(value.toBool());
            }
            break;
        case SpendZeroConfChange:
            if (instantTxns.Value())
            {
                settings.setValue("bSpendZeroConfChange", true);
            }
            else
            {
                if (settings.value("bSpendZeroConfChange") != value)
                {
                    settings.setValue("bSpendZeroConfChange", value);
                    setRestartRequired(true);
                }
            }
            break;
        case AutoConsolidation:
            if (settings.value("fAutoConsolidation") != value)
            {
                settings.setValue("fAutoConsolidation", value);
                autoTxns.Set(value.toBool());
            }
            break;
        case RescanOnStartup:
            settings.setValue("fRescanOnStartup", value);
            setRestartRequired(true);
            break;
        case TokenWhitelist:
            settings.setValue("QT_tokenWhitelist", value.toBool());
            tokenWhitelist.Set(value.toBool());
            Q_EMIT tokenWhitelistButtonChanged(value.toBool());
            break;
#endif
        case DisplayUnit:
            setDisplayUnit(value);
            break;
        case ThirdPartyTxUrls:
            if (strThirdPartyTxUrls != value.toString())
            {
                const char *ret = isInvalidThirdPartyTxUrlString(value.toString());

                if (ret == nullptr)
                {
                    strThirdPartyTxUrls = value.toString();
                    settings.setValue("strThirdPartyTxUrls", strThirdPartyTxUrls);
                    setRestartRequired(true);
                }
            }
            break;
        case ThirdPartyTokenUrls:
            if (strThirdPartyTokenUrls != value.toString())
            {
                const char *ret = isInvalidThirdPartyTokenUrlString(value.toString());

                if (ret == nullptr)
                {
                    strThirdPartyTokenUrls = value.toString();
                    settings.setValue("strThirdPartyTokenUrls", strThirdPartyTokenUrls);
                    setRestartRequired(true);
                }
            }
            break;
        case Language:
            if (settings.value("language") != value)
            {
                settings.setValue("language", value);
                setRestartRequired(true);
            }
            break;
        case CoinControlFeatures:
            fCoinControlFeatures = value.toBool();
            settings.setValue("fCoinControlFeatures", fCoinControlFeatures);
            Q_EMIT coinControlFeaturesChanged(fCoinControlFeatures);
            break;
        case DatabaseCache:
            if (settings.value("nDatabaseCache") != value)
            {
                settings.setValue("nDatabaseCache", value);
                setRestartRequired(true);
            }
            break;
        case ThreadsScriptVerif:
            if (settings.value("nThreadsScriptVerif") != value)
            {
                settings.setValue("nThreadsScriptVerif", value);
                setRestartRequired(true);
            }
            break;
        case Listen:
            if (settings.value("fListen") != value)
            {
                settings.setValue("fListen", value);
                setRestartRequired(true);
            }
            break;
        default:
            break;
        }
    }

    Q_EMIT dataChanged(index, index);

    return successful;
}

/** Updates current unit in memory, settings and emits displayUnitChanged(newUnit) signal */
void OptionsModel::setDisplayUnit(const QVariant &value)
{
    if (!value.isNull())
    {
        QSettings settings;
        nDisplayUnit = value.toInt();
        settings.setValue("nDisplayUnit", nDisplayUnit);
        Q_EMIT displayUnitChanged(nDisplayUnit);
    }
}

bool OptionsModel::getProxySettings(QNetworkProxy &proxy) const
{
    // Directly query current base proxy, because
    // GUI settings can be overridden with -proxy.
    proxyType curProxy;
    if (GetProxy(NET_IPV4, curProxy))
    {
        proxy.setType(QNetworkProxy::Socks5Proxy);
        proxy.setHostName(QString::fromStdString(curProxy.proxy.ToStringIP()));
        proxy.setPort(curProxy.proxy.GetPort());

        return true;
    }
    else
        proxy.setType(QNetworkProxy::NoProxy);

    return false;
}

void OptionsModel::setRestartRequired(bool fRequired)
{
    QSettings settings;
    return settings.setValue("fRestartRequired", fRequired);
}

bool OptionsModel::isRestartRequired()
{
    QSettings settings;
    return settings.value("fRestartRequired", false).toBool();
}
