// Copyright (c) 2011-2015 The Bitcoin Core developers
// Copyright (c) 2015-2022 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#if defined(HAVE_CONFIG_H)
#include "nexa-config.h"
#endif

#include "rpcconsole.h"
#include "ui_debugwindow.h"

#include "bantablemodel.h"
#include "clientmodel.h"
#include "dosman.h"
#include "guiutil.h"
#include "platformstyle.h"

#include "chainparams.h"
#include "rpc/client.h"
#include "rpc/server.h"
#include "util.h"
#include "validation/validation.h"

#include <univalue.h>

#ifdef ENABLE_WALLET
#include <db_cxx.h>
#endif

#include <QKeyEvent>
#include <QMenu>
#include <QScrollBar>
#include <QSettings>
#include <QSignalMapper>
#include <QStringList>
#include <QThread>
#include <QTime>
#include <QTimer>

// TODO: add a scrollback limit, as there is currently none
// TODO: make it possible to filter out categories (esp debug messages when implemented)
// TODO: receive errors and debug messages through ClientModel

const int CONSOLE_HISTORY = 50;
const int INITIAL_TRAFFIC_GRAPH_MINS = 30;
const int INITIAL_TRANSACTION_GRAPH_MINS = 1440;
const QSize FONT_RANGE(4, 40);
const char fontSizeSettingsKey[] = "consoleFontSize";
const QString duration_format = "H:mm:ss";
const QString time_format = "MMM  d yyyy, HH:mm:ss";

const struct
{
    const char *url;
    const char *source;
} ICON_MAPPING[] = {{"cmd-request", ":/icons/tx_input"}, {"cmd-reply", ":/icons/tx_output"},
    {"cmd-error", ":/icons/tx_output"}, {"misc", ":/icons/tx_inout"}, {nullptr, nullptr}};

/* Object for executing console RPC commands in a separate thread.
 */
class RPCExecutor : public QObject
{
    Q_OBJECT

public Q_SLOTS:
    void request(const QString &command);

Q_SIGNALS:
    void reply(int category, const QString &command);
};

/** Class for handling RPC timers
 * (used for e.g. re-locking the wallet after a timeout)
 */
class QtRPCTimerBase : public QObject, public RPCTimerBase
{
    Q_OBJECT
public:
    QtRPCTimerBase(boost::function<void(void)> &_func, int64_t millis) : func(_func)
    {
        timer.setSingleShot(true);
        connect(&timer, SIGNAL(timeout()), this, SLOT(timeout()));
        timer.start(millis);
    }
    ~QtRPCTimerBase() {}
private Q_SLOTS:
    void timeout() { func(); }

private:
    QTimer timer;
    boost::function<void(void)> func;
};

class QtRPCTimerInterface : public RPCTimerInterface
{
public:
    ~QtRPCTimerInterface() {}
    const char *Name() { return "Qt"; }
    RPCTimerBase *NewTimer(boost::function<void(void)> &func, int64_t millis)
    {
        return new QtRPCTimerBase(func, millis);
    }
};


#include "rpcconsole.moc"

/**
 * Split shell command line into a list of arguments. Aims to emulate \c bash and friends.
 *
 * - Arguments are delimited with whitespace
 * - Extra whitespace at the beginning and end and between arguments will be ignored
 * - Text can be "double" or 'single' quoted
 * - The backslash \c \ is used as escape character
 *   - Outside quotes, any character can be escaped
 *   - Within double quotes, only escape \c " and backslashes before a \c " or another backslash
 *   - Within single quotes, no escaping is possible and no special interpretation takes place
 *
 * @param[out]   args        Parsed arguments will be appended to this list
 * @param[in]    strCommand  Command line to split
 */
bool parseCommandLine(std::vector<std::string> &args, const std::string &strCommand)
{
    enum CmdParseState
    {
        STATE_EATING_SPACES,
        STATE_ARGUMENT,
        STATE_SINGLEQUOTED,
        STATE_DOUBLEQUOTED,
        STATE_ESCAPE_OUTER,
        STATE_ESCAPE_DOUBLEQUOTED
    } state = STATE_EATING_SPACES;
    std::string curarg;
    Q_FOREACH (char ch, strCommand)
    {
        switch (state)
        {
        case STATE_ARGUMENT: // In or after argument
        case STATE_EATING_SPACES: // Handle runs of whitespace
            switch (ch)
            {
            case '"':
                state = STATE_DOUBLEQUOTED;
                break;
            case '\'':
                state = STATE_SINGLEQUOTED;
                break;
            case '\\':
                state = STATE_ESCAPE_OUTER;
                break;
            case ' ':
            case '\n':
            case '\t':
                if (state == STATE_ARGUMENT) // Space ends argument
                {
                    args.push_back(curarg);
                    curarg.clear();
                }
                state = STATE_EATING_SPACES;
                break;
            default:
                curarg += ch;
                state = STATE_ARGUMENT;
            }
            break;
        case STATE_SINGLEQUOTED: // Single-quoted string
            switch (ch)
            {
            case '\'':
                state = STATE_ARGUMENT;
                break;
            default:
                curarg += ch;
            }
            break;
        case STATE_DOUBLEQUOTED: // Double-quoted string
            switch (ch)
            {
            case '"':
                state = STATE_ARGUMENT;
                break;
            case '\\':
                state = STATE_ESCAPE_DOUBLEQUOTED;
                break;
            default:
                curarg += ch;
            }
            break;
        case STATE_ESCAPE_OUTER: // '\' outside quotes
            curarg += ch;
            state = STATE_ARGUMENT;
            break;
        case STATE_ESCAPE_DOUBLEQUOTED: // '\' in double-quoted text
            if (ch != '"' && ch != '\\')
                curarg += '\\'; // keep '\' for everything but the quote and '\' itself
            curarg += ch;
            state = STATE_DOUBLEQUOTED;
            break;
        }
    }
    switch (state) // final state
    {
    case STATE_EATING_SPACES:
        return true;
    case STATE_ARGUMENT:
        args.push_back(curarg);
        return true;
    default: // ERROR to end in one of the other states
        return false;
    }
}

void RPCExecutor::request(const QString &command)
{
    std::vector<std::string> args;
    if (!parseCommandLine(args, command.toStdString()))
    {
        Q_EMIT reply(RPCConsole::CMD_ERROR, QString("Parse error: unbalanced ' or \""));
        return;
    }
    if (args.empty())
        return; // Nothing to do
    try
    {
        std::string strPrint;
        // Convert argument list to JSON objects in method-dependent way,
        // and pass it along with the method name to the dispatcher.
        UniValue params(UniValue::VARR);
        for (unsigned int idx = 1; idx < args.size(); idx++)
        {
            const std::string &strVal = args[idx];
            params.push_back(strVal);
        }
        UniValue result = tableRPC.execute(args[0], params);

        // Format result reply
        if (result.isNull())
            strPrint = "";
        else if (result.isStr())
            strPrint = result.get_str();
        else
            strPrint = result.write(2);

        Q_EMIT reply(RPCConsole::CMD_REPLY, QString::fromStdString(strPrint));
    }
    catch (UniValue &objError)
    {
        try // Nice formatting for standard-format error
        {
            int code = find_value(objError, "code").get_int();
            std::string message = find_value(objError, "message").get_str();
            Q_EMIT reply(
                RPCConsole::CMD_ERROR, QString::fromStdString(message) + " (code " + QString::number(code) + ")");
        }
        catch (const std::runtime_error &) // raised when converting to invalid type, i.e. missing code or message
        { // Show raw JSON object
            Q_EMIT reply(RPCConsole::CMD_ERROR, QString::fromStdString(objError.write()));
        }
    }
    catch (const std::exception &e)
    {
        Q_EMIT reply(RPCConsole::CMD_ERROR, QString("Error: ") + QString::fromStdString(e.what()));
    }
}

RPCConsole::RPCConsole(const PlatformStyle *_platformStyle, QWidget *parent)
    : QWidget(parent), ui(new Ui::RPCConsole), clientModel(0), historyPtr(0), cachedNodeid(-1),
      platformStyle(_platformStyle), peersTableContextMenu(0), banTableContextMenu(0), consoleFontSize(0)
{
    ui->setupUi(this);
    GUIUtil::restoreWindowGeometry("nRPCConsoleWindow", this->size(), this);

    ui->openDebugLogfileButton->setToolTip(ui->openDebugLogfileButton->toolTip().arg(tr(PACKAGE_NAME)));

    if (platformStyle->getImagesOnButtons())
    {
        ui->openDebugLogfileButton->setIcon(platformStyle->SingleColorIcon(":/icons/export"));
    }
    ui->clearButton->setIcon(platformStyle->SingleColorIcon(":/icons/remove"));
    ui->fontBiggerButton->setIcon(platformStyle->SingleColorIcon(":/icons/fontbigger"));
    ui->fontSmallerButton->setIcon(platformStyle->SingleColorIcon(":/icons/fontsmaller"));

    // Install event filter for up and down arrow
    ui->lineEdit->installEventFilter(this);
    ui->messagesWidget->installEventFilter(this);

    connect(ui->clearButton, SIGNAL(clicked()), this, SLOT(clear()));
    connect(ui->fontBiggerButton, SIGNAL(clicked()), this, SLOT(fontBigger()));
    connect(ui->fontSmallerButton, SIGNAL(clicked()), this, SLOT(fontSmaller()));
    connect(ui->btnClearTrafficGraph, SIGNAL(clicked()), ui->trafficGraph, SLOT(clear()));

// set library version labels
#ifdef ENABLE_WALLET
    ui->berkeleyDBVersion->setText(DbEnv::version(0, 0, 0));
#else
    ui->label_berkeleyDBVersion->hide();
    ui->berkeleyDBVersion->hide();
#endif
    // Register RPC timer interface
    rpcTimerInterface = new QtRPCTimerInterface();
    // avoid accidentally overwriting an existing, non QTThread
    // based timer interface
    RPCSetTimerInterfaceIfUnset(rpcTimerInterface);

    startExecutor();
    setTrafficGraphRange(INITIAL_TRAFFIC_GRAPH_MINS);
    setTransactionGraphRange(INITIAL_TRANSACTION_GRAPH_MINS);

    ui->detailWidget->hide();
    ui->peerHeading->setText(tr("Select a peer to view detailed information."));

    QSettings settings;
    consoleFontSize = settings.value(fontSizeSettingsKey, QFontInfo(QFont()).pointSize()).toInt();
    clear();
}

RPCConsole::~RPCConsole()
{
    GUIUtil::saveWindowGeometry("nRPCConsoleWindow", this);
    Q_EMIT stopExecutor();
    RPCUnsetTimerInterface(rpcTimerInterface);
    delete rpcTimerInterface;
    delete ui;
}

bool RPCConsole::eventFilter(QObject *obj, QEvent *event)
{
    if (event->type() == QEvent::KeyPress) // Special key handling
    {
        QKeyEvent *keyevt = static_cast<QKeyEvent *>(event);
        int key = keyevt->key();
        Qt::KeyboardModifiers mod = keyevt->modifiers();
        switch (key)
        {
        case Qt::Key_Up:
            if (obj == ui->lineEdit)
            {
                browseHistory(-1);
                return true;
            }
            break;
        case Qt::Key_Down:
            if (obj == ui->lineEdit)
            {
                browseHistory(1);
                return true;
            }
            break;
        case Qt::Key_PageUp: /* pass paging keys to messages widget */
        case Qt::Key_PageDown:
            if (obj == ui->lineEdit)
            {
                QApplication::postEvent(ui->messagesWidget, new QKeyEvent(*keyevt));
                return true;
            }
            break;
        case Qt::Key_Return:
        case Qt::Key_Enter:
            // forward these events to lineEdit
            if (obj == autoCompleter->popup())
            {
                QApplication::postEvent(ui->lineEdit, new QKeyEvent(*keyevt));
                return true;
            }
            break;
        default:
            // Typing in messages widget brings focus to line edit, and redirects key there
            // Exclude most combinations and keys that emit no text, except paste shortcuts
            if (obj == ui->messagesWidget && ((!mod && !keyevt->text().isEmpty() && key != Qt::Key_Tab) ||
                                                 ((mod & Qt::ControlModifier) && key == Qt::Key_V) ||
                                                 ((mod & Qt::ShiftModifier) && key == Qt::Key_Insert)))
            {
                ui->lineEdit->setFocus();
                QApplication::postEvent(ui->lineEdit, new QKeyEvent(*keyevt));
                return true;
            }
        }
    }
    return QWidget::eventFilter(obj, event);
}

void RPCConsole::setClientModel(ClientModel *model)
{
    clientModel = model;
    ui->trafficGraph->setClientModel(model);
    ui->transactionGraph->setClientModel(model);
    ui->groupBoxInstantaneous->setTitle(
        "Instantaneous Rate (" + QString::number(TX_RATE_RESOLUTION_MILLIS / 1000) + "s)");
    ui->groupBoxSmoothed->setTitle("Smoothed Rate (" + QString::number(TX_RATE_SMOOTHING_SEC, 'f', 0) + "s)");
    if (model && clientModel->getPeerTableModel() && clientModel->getBanTableModel())
    {
        // Keep up to date with client
        setNumConnections(model->getNumConnections());
        connect(model, SIGNAL(numConnectionsChanged(int)), this, SLOT(setNumConnections(int)));

        setNumBlocks(
            model->getNumBlocks(), model->getLastBlockDate(), model->getVerificationProgress(nullptr, false), false);
        connect(model, SIGNAL(numBlocksChanged(int, QDateTime, double, bool)), this,
            SLOT(setNumBlocks(int, QDateTime, double, bool)));
        connect(model, SIGNAL(timeSinceLastBlockChanged(qint64)), this, SLOT(updateTimeSinceLastBlock(qint64)));

        updateTrafficStats(model->getTotalBytesRecv(), model->getTotalBytesSent());
        connect(model, SIGNAL(bytesChanged(quint64, quint64)), this, SLOT(updateTrafficStats(quint64, quint64)));

        connect(model, SIGNAL(mempoolSizeChanged(long, size_t)), this, SLOT(setMempoolSize(long, size_t)));
        connect(model, SIGNAL(orphanPoolSizeChanged(long)), this, SLOT(setOrphanPoolSize(long)));
        connect(model, SIGNAL(messagePoolSizeChanged(long)), this, SLOT(setMessagePoolSize(long)));
        connect(model, SIGNAL(transactionsPerSecondChanged(double, double, double)), this,
            SLOT(setTransactionsPerSecond(double, double, double)));
        connect(model, SIGNAL(thinBlockPropagationStatsChanged(const ThinBlockQuickStats &)), this,
            SLOT(setThinBlockPropagationStats(const ThinBlockQuickStats &)));
        connect(model, SIGNAL(compactBlockPropagationStatsChanged(const CompactBlockQuickStats &)), this,
            SLOT(setCompactBlockPropagationStats(const CompactBlockQuickStats &)));
        connect(model, SIGNAL(grapheneBlockPropagationStatsChanged(const GrapheneQuickStats &)), this,
            SLOT(setGrapheneBlockPropagationStats(const GrapheneQuickStats &)));

        // set up peer table
        ui->peerWidget->setModel(model->getPeerTableModel());
        ui->peerWidget->verticalHeader()->hide();
        ui->peerWidget->setEditTriggers(QAbstractItemView::NoEditTriggers);
        ui->peerWidget->setSelectionBehavior(QAbstractItemView::SelectRows);
        ui->peerWidget->setSelectionMode(QAbstractItemView::SingleSelection);
        ui->peerWidget->setContextMenuPolicy(Qt::CustomContextMenu);
        ui->peerWidget->setColumnWidth(PeerTableModel::Address, ADDRESS_COLUMN_WIDTH);
        ui->peerWidget->setColumnWidth(PeerTableModel::Subversion, SUBVERSION_COLUMN_WIDTH);
        ui->peerWidget->setColumnWidth(PeerTableModel::Ping, PING_COLUMN_WIDTH);
        ui->peerWidget->horizontalHeader()->setStretchLastSection(true);

        // create peer table context menu actions
        QAction *disconnectAction = new QAction(tr("&Disconnect Node"), this);
        QAction *banAction1h = new QAction(tr("Ban Node for") + " " + tr("1 &hour"), this);
        QAction *banAction24h = new QAction(tr("Ban Node for") + " " + tr("1 &day"), this);
        QAction *banAction7d = new QAction(tr("Ban Node for") + " " + tr("1 &week"), this);
        QAction *banAction365d = new QAction(tr("Ban Node for") + " " + tr("1 &year"), this);

        // create peer table context menu
        peersTableContextMenu = new QMenu(this);
        peersTableContextMenu->addAction(disconnectAction);
        peersTableContextMenu->addAction(banAction1h);
        peersTableContextMenu->addAction(banAction24h);
        peersTableContextMenu->addAction(banAction7d);
        peersTableContextMenu->addAction(banAction365d);

        // Add a signal mapping to allow dynamic context menu arguments.
        // We need to use int (instead of int64_t), because signal mapper only supports
        // int or objects, which is okay because max bantime (1 year) is < int_max.
        QSignalMapper *signalMapper = new QSignalMapper(this);
        signalMapper->setMapping(banAction1h, 60 * 60);
        signalMapper->setMapping(banAction24h, 60 * 60 * 24);
        signalMapper->setMapping(banAction7d, 60 * 60 * 24 * 7);
        signalMapper->setMapping(banAction365d, 60 * 60 * 24 * 365);
        connect(banAction1h, SIGNAL(triggered()), signalMapper, SLOT(map()));
        connect(banAction24h, SIGNAL(triggered()), signalMapper, SLOT(map()));
        connect(banAction7d, SIGNAL(triggered()), signalMapper, SLOT(map()));
        connect(banAction365d, SIGNAL(triggered()), signalMapper, SLOT(map()));
        connect(signalMapper, SIGNAL(mapped(int)), this, SLOT(banSelectedNode(int)));

        // peer table context menu signals
        connect(ui->peerWidget, SIGNAL(customContextMenuRequested(const QPoint &)), this,
            SLOT(showPeersTableContextMenu(const QPoint &)));
        connect(disconnectAction, SIGNAL(triggered()), this, SLOT(disconnectSelectedNode()));

        // peer table signal handling - update peer details when selecting new node
        connect(ui->peerWidget->selectionModel(),
            SIGNAL(selectionChanged(const QItemSelection &, const QItemSelection &)), this,
            SLOT(peerSelected(const QItemSelection &, const QItemSelection &)));
        // peer table signal handling - update peer details when new nodes are added to the model
        connect(model->getPeerTableModel(), SIGNAL(layoutChanged()), this, SLOT(peerLayoutChanged()));

        // set up ban table
        ui->banlistWidget->setModel(model->getBanTableModel());
        ui->banlistWidget->verticalHeader()->hide();
        ui->banlistWidget->setEditTriggers(QAbstractItemView::NoEditTriggers);
        ui->banlistWidget->setSelectionBehavior(QAbstractItemView::SelectRows);
        ui->banlistWidget->setSelectionMode(QAbstractItemView::SingleSelection);
        ui->banlistWidget->setContextMenuPolicy(Qt::CustomContextMenu);
        ui->banlistWidget->setColumnWidth(BanTableModel::Address, BANSUBNET_COLUMN_WIDTH);
        ui->banlistWidget->setColumnWidth(BanTableModel::Bantime, BANTIME_COLUMN_WIDTH);
        ui->banlistWidget->horizontalHeader()->setStretchLastSection(true);

        // create ban table context menu action
        QAction *unbanAction = new QAction(tr("&Unban Node"), this);

        // create ban table context menu
        banTableContextMenu = new QMenu(this);
        banTableContextMenu->addAction(unbanAction);

        // ban table context menu signals
        connect(ui->banlistWidget, SIGNAL(customContextMenuRequested(const QPoint &)), this,
            SLOT(showBanTableContextMenu(const QPoint &)));
        connect(unbanAction, SIGNAL(triggered()), this, SLOT(unbanSelectedNode()));

        // ban table signal handling - clear peer details when clicking a peer in the ban table
        connect(ui->banlistWidget, SIGNAL(clicked(const QModelIndex &)), this, SLOT(clearSelectedNode()));
        // ban table signal handling - ensure ban table is shown or hidden (if empty)
        connect(model->getBanTableModel(), SIGNAL(layoutChanged()), this, SLOT(showOrHideBanTableIfRequired()));
        showOrHideBanTableIfRequired();

        // Provide initial values
        ui->clientVersion->setText(model->formatFullVersion());
        ui->clientUserAgent->setText(model->formatSubVersion());
        ui->dataDir->setText(model->dataDir());
        ui->startupTime->setText(model->formatClientStartupTime());
        ui->networkName->setText(QString::fromStdString(Params().NetworkIDString()));

        // Setup autocomplete and attach it
        QStringList wordList;
        std::vector<std::string> commandList = tableRPC.listCommands();
        for (size_t i = 0; i < commandList.size(); ++i)
        {
            wordList << commandList[i].c_str();
        }

        autoCompleter = new QCompleter(wordList, this);
        ui->lineEdit->setCompleter(autoCompleter);
        autoCompleter->popup()->installEventFilter(this);
    }
}

static QString categoryClass(int category)
{
    switch (category)
    {
    case RPCConsole::CMD_REQUEST:
        return "cmd-request";
        break;
    case RPCConsole::CMD_REPLY:
        return "cmd-reply";
        break;
    case RPCConsole::CMD_ERROR:
        return "cmd-error";
        break;
    default:
        return "misc";
    }
}

void RPCConsole::fontBigger() { setFontSize(consoleFontSize + 1); }
void RPCConsole::fontSmaller() { setFontSize(consoleFontSize - 1); }
void RPCConsole::setFontSize(int newSize)
{
    QSettings settings;

    // don't allow a insane font size
    if (newSize < FONT_RANGE.width() || newSize > FONT_RANGE.height())
        return;

    // temp. store the console content
    QString str = ui->messagesWidget->toHtml();

    // replace font tags size in current content
    str.replace(QString("font-size:%1pt").arg(consoleFontSize), QString("font-size:%1pt").arg(newSize));

    // store the new font size
    consoleFontSize = newSize;
    settings.setValue(fontSizeSettingsKey, consoleFontSize);

    // clear console (reset icon sizes, default stylesheet) and re-add the content
    float oldPosFactor =
        1.0 / ui->messagesWidget->verticalScrollBar()->maximum() * ui->messagesWidget->verticalScrollBar()->value();
    clear(false);
    ui->messagesWidget->setHtml(str);
    ui->messagesWidget->verticalScrollBar()->setValue(
        oldPosFactor * ui->messagesWidget->verticalScrollBar()->maximum());
}

void RPCConsole::clear(bool clearHistory)
{
    ui->messagesWidget->clear();
    if (clearHistory)
    {
        history.clear();
        historyPtr = 0;
    }
    ui->lineEdit->clear();
    ui->lineEdit->setFocus();

    // Add smoothly scaled icon images.
    // (when using width/height on an img, Qt uses nearest instead of linear interpolation)
    for (int i = 0; ICON_MAPPING[i].url; ++i)
    {
        ui->messagesWidget->document()->addResource(QTextDocument::ImageResource, QUrl(ICON_MAPPING[i].url),
            platformStyle->SingleColorImage(ICON_MAPPING[i].source)
                .scaled(
                    QSize(consoleFontSize * 2, consoleFontSize * 2), Qt::IgnoreAspectRatio, Qt::SmoothTransformation));
    }

    // Set default style sheet
    QFontInfo fixedFontInfo(GUIUtil::fixedPitchFont());
    ui->messagesWidget->document()->setDefaultStyleSheet(
        QString("table { }"
                "td.time { color: #808080; font-size: %2; padding-top: 3px; } "
                "td.message { font-family: %1; font-size: %2; white-space:pre-wrap; } "
                "td.cmd-request { color: #006060; } "
                "td.cmd-error { color: red; } "
                "b { color: #006060; } ")
            .arg(fixedFontInfo.family(), QString("%1pt").arg(consoleFontSize)));

    message(CMD_REPLY,
        (tr("Welcome to the %1 RPC console.").arg(tr(PACKAGE_NAME)) + "<br>" +
            tr("Use up and down arrows to navigate history, and <b>Ctrl-L</b> to clear screen.") + "<br>" +
            tr("Type <b>help</b> for an overview of available commands.")),
        true);
}

void RPCConsole::keyPressEvent(QKeyEvent *event)
{
    if (windowType() != Qt::Widget && event->key() == Qt::Key_Escape)
    {
        close();
    }
}

void RPCConsole::message(int category, const QString &message, bool html)
{
    QTime time = QTime::currentTime();
    QString timeString = time.toString();
    QString out;
    out += "<table><tr><td class=\"time\" width=\"65\">" + timeString + "</td>";
    out += "<td class=\"icon\" width=\"32\"><img src=\"" + categoryClass(category) + "\"></td>";
    out += "<td class=\"message " + categoryClass(category) + "\" valign=\"middle\">";
    if (html)
        out += message;
    else
        out += GUIUtil::HtmlEscape(message, false);
    out += "</td></tr></table>";
    ui->messagesWidget->append(out);
}

void RPCConsole::setNumConnections(int count)
{
    if (!clientModel)
        return;

    QString connections = QString::number(count) + " (";
    connections += tr("In:") + " " + QString::number(clientModel->getNumConnections(CONNECTIONS_IN)) + " / ";
    connections += tr("Out:") + " " + QString::number(clientModel->getNumConnections(CONNECTIONS_OUT)) + ")";

    ui->numberOfConnections->setText(connections);
}

void RPCConsole::setNumBlocks(int count, const QDateTime &blockDate, double nVerificationProgress, bool fHeader)
{
    if (!fHeader)
    {
        ui->numberOfBlocks->setText(QString::number(count));
        ui->lastBlockTime->setText(blockDate.toString(time_format));

        // Get the blocksize at the current chaintip
        uint64_t nBlockSize = chainActive.Tip()->GetBlockHeader().size;

        ui->lastBlockSize->setText(QString::number(nBlockSize));
        if (nBlockSize == 0)
            ui->lastBlockSize->setText("N/A");
        else if (nBlockSize < 1000000)
            ui->lastBlockSize->setText(QString::number(nBlockSize / 1000.0, 'f', 2) + " KB");
        else
            ui->lastBlockSize->setText(QString::number(nBlockSize / 1000000.0, 'f', 2) + " MB");
    }
}

void RPCConsole::updateTimeSinceLastBlock(qint64 lastBlockTime)
{
    // Recompute seconds since last block locally as this is called asynchronously
    // otherwise the UI updates can become noticeably longer than 1 second
    qint64 secs = GetTime() - lastBlockTime;
    // Prevent negative duration (likely indicates local system clock sync issue)
    if (secs < 0)
        secs = 0;
    ui->lastBlockTime->setText(QDateTime::fromTime_t(lastBlockTime).toString(time_format) + " (" +
                               QDateTime::fromTime_t(secs).toUTC().toString(duration_format) + ")");
}

void RPCConsole::setMempoolSize(long numberOfTxs, size_t dynUsage)
{
    ui->mempoolNumberTxs->setText(QString::number(numberOfTxs));

    if (dynUsage < 1000000)
        ui->mempoolSize->setText(QString::number(dynUsage / 1000.0, 'f', 2) + " KB");
    else
        ui->mempoolSize->setText(QString::number(dynUsage / 1000000.0, 'f', 2) + " MB");
}

void RPCConsole::setOrphanPoolSize(long numberOfTxs) { ui->orphanPoolNumberTxs->setText(QString::number(numberOfTxs)); }
void RPCConsole::setMessagePoolSize(long numberOfMessages)
{
    ui->messagesInCapdPool->setText(QString::number(numberOfMessages));
}
QString FormatTps(double tps)
{
    // Format the output
    if (tps < 100)
        return QString::number(tps, 'f', 2);
    else
        return QString::number((uint64_t)tps);
}

void RPCConsole::setTransactionsPerSecond(double smoothedTps, double instantaneousTps, double peakTps)
{
    ui->transactionsPerSecond->setText(FormatTps(smoothedTps) + "  (peak: " + FormatTps(peakTps) + ")");

    QString displayWindowText = ui->transactionGraph->getDisplayWindowLabelText();
    ui->labelInstantaneousPeakDisplayed->setText(displayWindowText);
    ui->labelInstantaneousAvgDisplayed->setText(displayWindowText);
    ui->labelSmoothedPeakDisplayed->setText(displayWindowText);
    ui->labelSmoothedAvgDisplayed->setText(displayWindowText);

    ui->instantaneousPeakRuntime->setText(FormatTps(ui->transactionGraph->getInstantaneousTpsPeak_Runtime()));
    ui->instantaneousPeakSampled->setText(FormatTps(ui->transactionGraph->getInstantaneousTpsPeak_Sampled()));
    ui->instantaneousPeakDisplayed->setText(FormatTps(ui->transactionGraph->getInstantaneousTpsPeak_Displayed()));
    ui->instantaneousAvgRuntime->setText(FormatTps(ui->transactionGraph->getInstantaneousTpsAverage_Runtime()));
    ui->instantaneousAvgSampled->setText(FormatTps(ui->transactionGraph->getInstantaneousTpsAverage_Sampled()));
    ui->instantaneousAvgDisplayed->setText(FormatTps(ui->transactionGraph->getInstantaneousTpsAverage_Displayed()));

    ui->smoothedPeakRuntime->setText(FormatTps(ui->transactionGraph->getSmoothedTpsPeak_Runtime()));
    ui->smoothedPeakSampled->setText(FormatTps(ui->transactionGraph->getSmoothedTpsPeak_Sampled()));
    ui->smoothedPeakDisplayed->setText(FormatTps(ui->transactionGraph->getSmoothedTpsPeak_Displayed()));
    ui->smoothedAvgRuntime->setText(FormatTps(ui->transactionGraph->getSmoothedTpsAverage_Runtime()));
    ui->smoothedAvgSampled->setText(FormatTps(ui->transactionGraph->getSmoothedTpsAverage_Sampled()));
    ui->smoothedAvgDisplayed->setText(FormatTps(ui->transactionGraph->getSmoothedTpsAverage_Displayed()));
}

void RPCConsole::setThinBlockPropagationStats(const ThinBlockQuickStats &thin)
{
    if (!IsThinBlocksEnabled())
    {
        ui->blocksXThinTotals->setText(tr("Disabled"));
        ui->blocksXThin24hAverages->setText(tr("Disabled"));
        return;
    }

    // Total: n (Sent: i / Received: r) saved bw
    QString text = QString::number(thin.nTotalOutbound + thin.nTotalInbound) + " (Sent: ";
    text += QString::number(thin.nTotalOutbound) + " / Received: ";
    text += QString::number(thin.nTotalInbound) + ") saved ";
    text += QString::fromStdString(formatInfoUnit(thin.nTotalBandwidthSavings));

    ui->blocksXThinTotals->setText(text);

    // 24-hour Average: n (Sent: i / Received: r), Compression (i% / r%), ReRequests f (f%)
    text = QString::number(thin.nLast24hOutbound + thin.nLast24hInbound) + " (Sent: ";
    text += QString::number(thin.nLast24hOutbound) + " / Received: ";
    text += QString::number(thin.nLast24hInbound) + "), Compression (";
    text += QString::number(thin.fLast24hOutboundCompression, 'f', 1) + "% / ";
    text += QString::number(thin.fLast24hInboundCompression, 'f', 1) + "%), ReRequests ";
    text += QString::number(thin.nLast24hRerequestTx) + " (";
    text += QString::number(thin.fLast24hRerequestTxPercent, 'f', 1) + "%)";

    ui->blocksXThin24hAverages->setText(text);
}

void RPCConsole::setCompactBlockPropagationStats(const CompactBlockQuickStats &compact)
{
    if (!IsCompactBlocksEnabled())
    {
        ui->blocksCompactTotals->setText(tr("Disabled"));
        ui->blocksCompact24hAverages->setText(tr("Disabled"));
        return;
    }

    // Total: n (Sent: i / Received: r) saved bw
    QString text = QString::number(compact.nTotalOutbound + compact.nTotalInbound) + " (Sent: ";
    text += QString::number(compact.nTotalOutbound) + " / Received: ";
    text += QString::number(compact.nTotalInbound) + ") saved ";
    text += QString::fromStdString(formatInfoUnit(compact.nTotalBandwidthSavings));

    ui->blocksCompactTotals->setText(text);

    // 24-hour Average: n (Sent: i / Received: r), Compression (i% / r%), ReRequests f (f%)
    text = QString::number(compact.nLast24hOutbound + compact.nLast24hInbound) + " (Sent: ";
    text += QString::number(compact.nLast24hOutbound) + " / Received: ";
    text += QString::number(compact.nLast24hInbound) + "), Compression (";
    text += QString::number(compact.fLast24hOutboundCompression, 'f', 1) + "% / ";
    text += QString::number(compact.fLast24hInboundCompression, 'f', 1) + "%), ReRequests ";
    text += QString::number(compact.nLast24hRerequestTx) + " (";
    text += QString::number(compact.fLast24hRerequestTxPercent, 'f', 1) + "%)";

    ui->blocksCompact24hAverages->setText(text);
}

void RPCConsole::setGrapheneBlockPropagationStats(const GrapheneQuickStats &graphene)
{
    if (!IsGrapheneBlockEnabled())
    {
        ui->blocksGrapheneTotals->setText(tr("Disabled"));
        ui->blocksGraphene24hAverages->setText(tr("Disabled"));
        return;
    }

    // Total: n (Sent: i / Received: r) saved bw, Decode failures d
    QString text = QString::number(graphene.nTotalOutbound + graphene.nTotalInbound) + " (Sent: ";
    text += QString::number(graphene.nTotalOutbound) + " / Received: ";
    text += QString::number(graphene.nTotalInbound) + ") saved ";
    text += QString::fromStdString(formatInfoUnit(graphene.nTotalBandwidthSavings)) + ", Decode failures ";
    text += QString::number(graphene.nTotalDecodeFailures);

    ui->blocksGrapheneTotals->setText(text);

    // 24-hour Average: n (Sent: i / Received: r), Compression (i% / r%), ReRequests f (f%)
    text = QString::number(graphene.nLast24hOutbound + graphene.nLast24hInbound) + " (Sent: ";
    text += QString::number(graphene.nLast24hOutbound) + " / Received: ";
    text += QString::number(graphene.nLast24hInbound) + "), Compression (";
    text += QString::number(graphene.fLast24hOutboundCompression, 'f', 1) + "% / ";
    text += QString::number(graphene.fLast24hInboundCompression, 'f', 1) + "%), ReRequests ";
    text += QString::number(graphene.nLast24hRerequestTx) + " (";
    text += QString::number(graphene.fLast24hRerequestTxPercent, 'f', 1) + "%)";

    ui->blocksGraphene24hAverages->setText(text);
}

void RPCConsole::on_lineEdit_returnPressed()
{
    QString cmd = ui->lineEdit->text();
    ui->lineEdit->clear();

    if (!cmd.isEmpty())
    {
        message(CMD_REQUEST, cmd);
        Q_EMIT cmdRequest(cmd);
        // Remove command, if already in history
        history.removeOne(cmd);
        // Append command to history
        history.append(cmd);
        // Enforce maximum history size
        while (history.size() > CONSOLE_HISTORY)
            history.removeFirst();
        // Set pointer to end of history
        historyPtr = history.size();
        // Scroll console view to end
        scrollToEnd();
    }
}

void RPCConsole::browseHistory(int offset)
{
    historyPtr += offset;
    if (historyPtr < 0)
        historyPtr = 0;
    if (historyPtr > history.size())
        historyPtr = history.size();
    QString cmd;
    if (historyPtr < history.size())
        cmd = history.at(historyPtr);
    ui->lineEdit->setText(cmd);
}

void RPCConsole::startExecutor()
{
    QThread *thread = new QThread;
    RPCExecutor *executor = new RPCExecutor();
    executor->moveToThread(thread);

    // Replies from executor object must go to this object
    connect(executor, SIGNAL(reply(int, QString)), this, SLOT(message(int, QString)));
    // Requests from this object must go to executor
    connect(this, SIGNAL(cmdRequest(QString)), executor, SLOT(request(QString)));

    // On stopExecutor signal
    // - queue executor for deletion (in execution thread)
    // - quit the Qt event loop in the execution thread
    connect(this, SIGNAL(stopExecutor()), executor, SLOT(deleteLater()));
    connect(this, SIGNAL(stopExecutor()), thread, SLOT(quit()));
    // Queue the thread for deletion (in this thread) when it is finished
    connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater()));

    // Default implementation of QThread::run() simply spins up an event loop in the thread,
    // which is what we want.
    thread->start();
}

void RPCConsole::on_tabWidget_currentChanged(int index)
{
    if (ui->tabWidget->widget(index) == ui->tab_console)
        ui->lineEdit->setFocus();
    else if (ui->tabWidget->widget(index) != ui->tab_peers)
        clearSelectedNode();
}

void RPCConsole::on_openDebugLogfileButton_clicked() { GUIUtil::openDebugLogfile(); }
void RPCConsole::scrollToEnd()
{
    QScrollBar *scrollbar = ui->messagesWidget->verticalScrollBar();
    scrollbar->setValue(scrollbar->maximum());
}

void RPCConsole::on_sldGraphRange_valueChanged(int value)
{
    const int multiplier = 5; // each position on the slider represents 5 min
    int mins = value * multiplier;
    setTrafficGraphRange(mins);
}

QString RPCConsole::FormatBytes(quint64 bytes)
{
    if (bytes < 1024)
        return QString(tr("%1 B")).arg(bytes);
    if (bytes < 1024 * 1024)
        return QString(tr("%1 KB")).arg(bytes / 1024);
    if (bytes < 1024 * 1024 * 1024)
        return QString(tr("%1 MB")).arg(bytes / 1024 / 1024);

    return QString(tr("%1 GB")).arg(bytes / 1024 / 1024 / 1024);
}

void RPCConsole::setTrafficGraphRange(int mins)
{
    ui->trafficGraph->setGraphRangeMins(mins);
    ui->lblGraphRange->setText(GUIUtil::formatDurationStr(mins * 60));
}

void RPCConsole::updateTrafficStats(quint64 totalBytesIn, quint64 totalBytesOut)
{
    ui->lblBytesIn->setText(FormatBytes(totalBytesIn));
    ui->lblBytesOut->setText(FormatBytes(totalBytesOut));
}

void RPCConsole::on_sldTpsGraphRange_valueChanged(int value)
{
    // The first hour is in 5 minute steps
    static const int firstMultiplier = 5;
    // All subsequent hours are in 15 minute steps
    static const int secondMultiplier = 15;

    int mins;
    if (value <= 12)
        mins = value * firstMultiplier;
    else
        mins = 60 + (value - 12) * secondMultiplier;
    setTransactionGraphRange(mins);
}

void RPCConsole::setTransactionGraphRange(int mins)
{
    ui->transactionGraph->setTpsGraphRangeMins(mins);
    ui->lblTpsGraphRange->setText(GUIUtil::formatDurationStr(mins * 60));
}

void RPCConsole::peerSelected(const QItemSelection &selected, const QItemSelection &deselected)
{
    Q_UNUSED(deselected);

    if (!clientModel || !clientModel->getPeerTableModel() || selected.indexes().isEmpty())
        return;

    const CNodeCombinedStats *stats = clientModel->getPeerTableModel()->getNodeStats(selected.indexes().first().row());
    if (stats)
        updateNodeDetail(stats);
}

void RPCConsole::peerLayoutChanged()
{
    if (!clientModel || !clientModel->getPeerTableModel())
        return;

    const CNodeCombinedStats *stats = nullptr;
    bool fUnselect = false;
    bool fReselect = false;

    if (cachedNodeid == -1) // no node selected yet
        return;

    // find the currently selected row
    int selectedRow = -1;
    QModelIndexList selectedModelIndex = ui->peerWidget->selectionModel()->selectedIndexes();
    if (!selectedModelIndex.isEmpty())
    {
        selectedRow = selectedModelIndex.first().row();
    }

    // check if our detail node has a row in the table (it may not necessarily
    // be at selectedRow since its position can change after a layout change)
    int detailNodeRow = clientModel->getPeerTableModel()->getRowByNodeId(cachedNodeid);

    if (detailNodeRow < 0)
    {
        // detail node disappeared from table (node disconnected)
        fUnselect = true;
    }
    else
    {
        if (detailNodeRow != selectedRow)
        {
            // detail node moved position
            fUnselect = true;
            fReselect = true;
        }

        // get fresh stats on the detail node.
        stats = clientModel->getPeerTableModel()->getNodeStats(detailNodeRow);
    }

    if (fUnselect && selectedRow >= 0)
    {
        clearSelectedNode();
    }

    if (fReselect)
    {
        ui->peerWidget->selectRow(detailNodeRow);
    }

    if (stats)
        updateNodeDetail(stats);
}

void RPCConsole::updateNodeDetail(const CNodeCombinedStats *stats)
{
    // Update cached nodeid
    cachedNodeid = stats->nodeStats.nodeid;

    // List of services not defined by a nServices bit that we still want to display on the UI
    // as a peer advertised service (i.e. Compact Blocks)
    QStringList additionalServices;
    if (stats->nodeStats.fSupportsCompactBlocks)
        additionalServices.append("CMPCT");

    // update the detail ui with latest node information
    QString peerAddrDetails(QString::fromStdString(stats->nodeStats.addrName) + "<br />");
    peerAddrDetails += tr("(node id: %1)").arg(QString::number(stats->nodeStats.nodeid));
    if (!stats->nodeStats.addrLocal.empty())
        peerAddrDetails += "<br />" + tr("via %1").arg(QString::fromStdString(stats->nodeStats.addrLocal));
    ui->peerHeading->setText(peerAddrDetails);
    ui->peerServices->setText(GUIUtil::formatServicesStr(stats->nodeStats.nServices, additionalServices));
    ui->peerLastSend->setText(
        stats->nodeStats.nLastSend ? GUIUtil::formatDurationStr(GetTime() - stats->nodeStats.nLastSend) : tr("never"));
    ui->peerLastRecv->setText(
        stats->nodeStats.nLastRecv ? GUIUtil::formatDurationStr(GetTime() - stats->nodeStats.nLastRecv) : tr("never"));
    ui->peerBytesSent->setText(FormatBytes(stats->nodeStats.nSendBytes));
    ui->peerBytesRecv->setText(FormatBytes(stats->nodeStats.nRecvBytes));
    ui->peerConnTime->setText(GUIUtil::formatDurationStr(GetTime() - stats->nodeStats.nTimeConnected));
    ui->peerPingTime->setText(GUIUtil::formatPingTime(stats->nodeStats.dPingTime));
    ui->peerPingWait->setText(GUIUtil::formatPingTime(stats->nodeStats.dPingWait));
    ui->timeoffset->setText(GUIUtil::formatTimeOffset(stats->nodeStats.nTimeOffset));
    ui->peerVersion->setText(QString("%1").arg(QString::number(stats->nodeStats.nVersion)));
    ui->peerSubversion->setText(QString::fromStdString(stats->nodeStats.cleanSubVer));
    ui->peerDirection->setText(stats->nodeStats.fInbound ? tr("Inbound") : tr("Outbound"));
    ui->peerHeight->setText(QString("%1").arg(QString::number(stats->nodeStats.nStartingHeight)));
    ui->peerWhitelisted->setText(stats->nodeStats.fWhitelisted ? tr("Yes") : tr("No"));

    // This check fails for example if the lock was busy and
    // nodeStateStats couldn't be fetched.
    if (stats->fNodeStateStatsAvailable)
    {
        // Ban score is init to 0
        ui->peerBanScore->setText(QString("%1").arg(stats->nodeStateStats.nMisbehavior));

        // Sync height is init to -1
        if (stats->nodeStateStats.nSyncHeight > -1)
            ui->peerSyncHeight->setText(QString("%1").arg(stats->nodeStateStats.nSyncHeight));
        else
            ui->peerSyncHeight->setText(tr("Unknown"));

        // Common height is init to -1
        if (stats->nodeStateStats.nCommonHeight > -1)
            ui->peerCommonHeight->setText(QString("%1").arg(stats->nodeStateStats.nCommonHeight));
        else
            ui->peerCommonHeight->setText(tr("Unknown"));
    }

    ui->detailWidget->show();
}

void RPCConsole::resizeEvent(QResizeEvent *event) { QWidget::resizeEvent(event); }
void RPCConsole::showEvent(QShowEvent *event)
{
    // restore column state
    GUIUtil::restoreColumnConfiguration("nPeersTable", ui->peerWidget->horizontalHeader());
    GUIUtil::restoreColumnConfiguration("nBannedPeersTable", ui->banlistWidget->horizontalHeader());

    QWidget::showEvent(event);

    if (!clientModel || !clientModel->getPeerTableModel())
        return;

    // start PeerTableModel auto refresh
    clientModel->getPeerTableModel()->startAutoRefresh();
}

void RPCConsole::hideEvent(QHideEvent *event)
{
    QWidget::hideEvent(event);

    if (!clientModel || !clientModel->getPeerTableModel())
        return;

    // stop PeerTableModel auto refresh
    clientModel->getPeerTableModel()->stopAutoRefresh();

    // save column state
    GUIUtil::saveColumnConfiguration("nPeersTable", ui->peerWidget->horizontalHeader());
    GUIUtil::saveColumnConfiguration("nBannedPeersTable", ui->banlistWidget->horizontalHeader());
}

void RPCConsole::showPeersTableContextMenu(const QPoint &point)
{
    QModelIndex index = ui->peerWidget->indexAt(point);
    if (index.isValid())
        peersTableContextMenu->exec(QCursor::pos());
}

void RPCConsole::showBanTableContextMenu(const QPoint &point)
{
    QModelIndex index = ui->banlistWidget->indexAt(point);
    if (index.isValid())
        banTableContextMenu->exec(QCursor::pos());
}

void RPCConsole::disconnectSelectedNode()
{
    // Get currently selected peer address
    QString strNode = GUIUtil::getEntryData(ui->peerWidget, 0, PeerTableModel::Address);

    // Find the node, disconnect it and clear the selected node
    CNodeRef node = FindNodeRef(strNode.toStdString());
    if (node)
    {
        node->fDisconnect = true;
        clearSelectedNode();
    }
}

void RPCConsole::banSelectedNode(int bantime)
{
    if (!clientModel)
        return;

    // Get currently selected peer address
    QString strNode = GUIUtil::getEntryData(ui->peerWidget, 0, PeerTableModel::Address);

    // Find the node, ban it and clear the selected node
    CNodeRef bannedNode = FindNodeRef(strNode.toStdString());
    if (bannedNode)
    {
        std::string nStr = strNode.toStdString();
        std::string addr;
        int port = 0;
        SplitHostPort(nStr, port, addr);

        dosMan.Ban(CNetAddr(addr), bannedNode.get()->cleanSubVer, BanReasonManuallyAdded, bantime);
        bannedNode->fDisconnect = true;

        clearSelectedNode();
    }
}

void RPCConsole::unbanSelectedNode()
{
    if (!clientModel)
        return;

    // Get currently selected ban address
    QString strNode = GUIUtil::getEntryData(ui->banlistWidget, 0, BanTableModel::Address);
    CSubNet possibleSubnet(strNode.toStdString());

    if (possibleSubnet.IsValid())
    {
        dosMan.Unban(possibleSubnet);
    }
}

void RPCConsole::clearSelectedNode()
{
    ui->peerWidget->selectionModel()->clearSelection();
    cachedNodeid = -1;
    ui->detailWidget->hide();
    ui->peerHeading->setText(tr("Select a peer to view detailed information."));
}

void RPCConsole::showOrHideBanTableIfRequired()
{
    if (!clientModel)
        return;

    bool visible = clientModel->getBanTableModel()->shouldShow();
    ui->banlistWidget->setVisible(visible);
    ui->banHeading->setVisible(visible);
}

void RPCConsole::setTabFocus(enum TabTypes tabType) { ui->tabWidget->setCurrentIndex(tabType); }
