// Copyright (c) 2011-2015 The Bitcoin Core developers
// Copyright (c) 2015-2022 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include "walletmodel.h"

#include "addresstablemodel.h"
#include "guiconstants.h"
#include "guiutil.h"
#include "paymentserver.h"
#include "recentrequeststablemodel.h"
#include "tokentablemodel.h"
#include "transactiontablemodel.h"

#include "dstencode.h"
#include "keystore.h"
#include "main.h"
#include "sync.h"
#include "ui_interface.h"
#include "wallet/grouptokenwallet.h"
#include "wallet/wallet.h"
#include "wallet/walletdb.h" // for BackupWallet

#include <stdint.h>

#include <QDebug>
#include <QSet>
#include <QTimer>

extern CTweak<uint32_t> dataCarrierSize;
extern CTweak<bool> dataCarrier;
extern CTweak<bool> instantTxns;
extern CTweak<uint32_t> instantTxnsDelay;

// Used for instant transactions balance update
int64_t nStartCheck = GetTime();

WalletModel::WalletModel(const PlatformStyle *platformStyle,
    CWallet *_wallet,
    OptionsModel *_optionsModel,
    QObject *parent)
    : QObject(parent), wallet(_wallet), optionsModel(_optionsModel), addressTableModel(0), transactionTableModel(0),
      tokenTableModel(0), recentRequestsTableModel(0), cachedBalance(0), cachedUnconfirmedBalance(0),
      cachedImmatureBalance(0), cachedEncryptionStatus(Unencrypted), cachedNumBlocks(0)
{
    fHaveWatchOnly = wallet->HaveWatchOnly() || getWatchBalance() > 0;
    fForceCheckBalanceChanged = false;
    fInstantTransactions = instantTxns.Value();

    addressTableModel = new AddressTableModel(wallet, this);
    transactionTableModel = new TransactionTableModel(platformStyle, wallet, this);
    tokenTableModel = new TokenTableModel(platformStyle, wallet, this);
    recentRequestsTableModel = new RecentRequestsTableModel(wallet, this);

    // This timer will be fired repeatedly to update the balance
    pollTimer = new QTimer(this);
    connect(pollTimer, SIGNAL(timeout()), this, SLOT(pollBalanceChanged()));
    pollTimer->start(MODEL_UPDATE_DELAY1);

    subscribeToCoreSignals();
}

WalletModel::~WalletModel() { unsubscribeFromCoreSignals(); }
CAmount WalletModel::getBalance(const CCoinControl *coinControl) const { return wallet->GetBalance(); }
CAmount WalletModel::getUnconfirmedBalance() const { return wallet->GetUnconfirmedBalance(); }
CAmount WalletModel::getImmatureBalance() const { return wallet->GetImmatureBalance(); }
bool WalletModel::haveWatchOnly() const { return fHaveWatchOnly; }
CAmount WalletModel::getWatchBalance() const { return wallet->GetWatchOnlyBalance(); }
CAmount WalletModel::getWatchUnconfirmedBalance() const { return wallet->GetUnconfirmedWatchOnlyBalance(); }
CAmount WalletModel::getWatchImmatureBalance() const { return wallet->GetImmatureWatchOnlyBalance(); }
void WalletModel::updateStatus()
{
    EncryptionStatus newEncryptionStatus = getEncryptionStatus();

    if (cachedEncryptionStatus != newEncryptionStatus)
        Q_EMIT encryptionStatusChanged(newEncryptionStatus);
}

void WalletModel::pollBalanceChanged()
{
    // Get required locks upfront. This avoids the GUI from getting stuck on
    // periodical polls if the core is holding the locks for a longer time -
    // for example, during a wallet rescan.
    TRY_LOCK(wallet->cs_wallet, lockWallet);
    if (!lockWallet)
        return;

    // If instant transactions was modified then force a balance update. Although
    // we force a balance check again if instant txns is on, we need this switch here
    // for when instant txns is turned off.
    bool temp = instantTxns.Value();
    if (temp != fInstantTransactions)
    {
        fForceCheckBalanceChanged = true;
        fInstantTransactions = temp;
    }

    // If instant transactions is turned on then every few seconds force a balance check
    if (fInstantTransactions)
    {
        int64_t nNow = GetTime();
        if ((nNow - nStartCheck) > instantTxnsDelay.Value())
        {
            fForceCheckBalanceChanged = true;
            nStartCheck = nNow;
        }
    }

    // Balance and confirmations might have changed
    // NOTE: updating confirmation in the UI is a relatively expensive operation and should
    //       be limited as much as possible. Such as only done here when a new block arrives.
    if (chainActive.Height() != cachedNumBlocks)
    {
        cachedNumBlocks = chainActive.Height();
        checkBalanceChanged();
        fForceCheckBalanceChanged = false;

        if (transactionTableModel)
            transactionTableModel->updateConfirmations();
        if (tokenTableModel)
            tokenTableModel->updateConfirmations();
    }

    // Update balance when needed
    if (fForceCheckBalanceChanged)
    {
        checkBalanceChanged();
        fForceCheckBalanceChanged = false;
    }
}

void WalletModel::checkBalanceChanged()
{
    CAmount newBalance = getBalance();
    CAmount newUnconfirmedBalance = getUnconfirmedBalance();
    CAmount newImmatureBalance = getImmatureBalance();
    CAmount newWatchOnlyBalance = 0;
    CAmount newWatchUnconfBalance = 0;
    CAmount newWatchImmatureBalance = 0;
    if (haveWatchOnly())
    {
        newWatchOnlyBalance = getWatchBalance();
        newWatchUnconfBalance = getWatchUnconfirmedBalance();
        newWatchImmatureBalance = getWatchImmatureBalance();
    }

    if (cachedBalance != newBalance || cachedUnconfirmedBalance != newUnconfirmedBalance ||
        cachedImmatureBalance != newImmatureBalance || cachedWatchOnlyBalance != newWatchOnlyBalance ||
        cachedWatchUnconfBalance != newWatchUnconfBalance || cachedWatchImmatureBalance != newWatchImmatureBalance)
    {
        cachedBalance = newBalance;
        cachedUnconfirmedBalance = newUnconfirmedBalance;
        cachedImmatureBalance = newImmatureBalance;
        cachedWatchOnlyBalance = newWatchOnlyBalance;
        cachedWatchUnconfBalance = newWatchUnconfBalance;
        cachedWatchImmatureBalance = newWatchImmatureBalance;

        Q_EMIT balanceChanged(newBalance, newUnconfirmedBalance, newImmatureBalance, newWatchOnlyBalance,
            newWatchUnconfBalance, newWatchImmatureBalance);
    }
}

void WalletModel::updateTransaction()
{
    // Balance and number of transactions might have changed
    fForceCheckBalanceChanged = true;
}

void WalletModel::updateAddressBook(const QString &address,
    const QString &label,
    bool isMine,
    const QString &purpose,
    int status)
{
    if (addressTableModel)
        addressTableModel->updateEntry(address, label, isMine, purpose, status);
}

void WalletModel::updateWatchOnlyFlag(bool fHaveWatchonly)
{
    fHaveWatchOnly = fHaveWatchonly;
    Q_EMIT notifyWatchonlyChanged(fHaveWatchonly);
}

bool WalletModel::validateAddress(const QString &address) { return IsValidDestinationString(address.toStdString()); }
WalletModel::SendCoinsReturn WalletModel::prepareTransaction(WalletModelTransaction &transaction,
    const CCoinControl *coinControl)
{
    CAmount total = 0;
    bool fSubtractFeeFromAmount = false;
    QList<SendCoinsRecipient> recipients = transaction.getRecipients();
    std::vector<CRecipient> vecSend;

    if (recipients.empty())
    {
        return OK;
    }

    QSet<QString> setAddress; // Used to detect duplicates
    int nAddresses = 0;

    // Pre-check input data for validity
    Q_FOREACH (const SendCoinsRecipient &rcp, recipients)
    {
        if (rcp.fSubtractFeeFromAmount)
            fSubtractFeeFromAmount = true;

        if (rcp.paymentRequest.IsInitialized())
        { // PaymentRequest...
            CAmount subtotal = 0;
            const payments::PaymentDetails &details = rcp.paymentRequest.getDetails();
            for (int i = 0; i < details.outputs_size(); i++)
            {
                const payments::Output &out = details.outputs(i);
                if (out.amount() <= 0)
                    continue;
                subtotal += out.amount();
                const unsigned char *scriptStr = (const unsigned char *)out.script().data();
                CScript scriptPubKey(scriptStr, scriptStr + out.script().size());
                CAmount nAmount = out.amount();
                CRecipient recipient = {scriptPubKey, nAmount, rcp.fSubtractFeeFromAmount};
                vecSend.push_back(recipient);
            }
            if (subtotal <= 0)
            {
                return InvalidAmount;
            }
            total += subtotal;
        }
        else
        { // User-entered address / amount:
            if (!validateAddress(rcp.address))
            {
                return InvalidAddress;
            }
            if (rcp.amount <= 0 && rcp.labelPublic == "")
            {
                return InvalidAmount;
            }
            setAddress.insert(rcp.address);
            ++nAddresses;

            CScript scriptPubKey;
            if (rcp.labelPublic == "")
                scriptPubKey = GetScriptForDestination(DecodeDestination(rcp.address.toStdString()));
            else
            {
                scriptPubKey = GetScriptLabelPublic(rcp.labelPublic.toStdString());
                // remove duplicate address as rcp.adddress was copied in SendCoinsDialog::on_sendButton_clicked()
                --nAddresses;

                // Make sure scriptPubKey is within size constraints and that we are configured
                // to foward OP_RETURN transactions.
                if (!dataCarrier.Value() || scriptPubKey.size() > dataCarrierSize.Value())
                    return LabelPublicExceedsLimits;
            }

            CRecipient recipient = {scriptPubKey, rcp.amount, rcp.fSubtractFeeFromAmount};
            vecSend.push_back(recipient);

            total += rcp.amount;
        }
    }
    if (setAddress.size() != nAddresses)
    {
        return DuplicateAddress;
    }

    CAmount nBalance = getBalance(coinControl);

    if (total > nBalance)
    {
        return AmountExceedsBalance;
    }

    {
        LOCK(wallet->cs_wallet);

        transaction.newPossibleKeyChange(wallet);

        CAmount nFeeRequired = 0;
        int nChangePosRet = -1;
        std::string strFailReason;

        CWalletTx *newTx = transaction.getTransaction();
        CReserveKey *keyChange = transaction.getPossibleKeyChange();
        bool fCreated = wallet->CreateTransaction(
            vecSend, *newTx, *keyChange, nFeeRequired, nChangePosRet, strFailReason, coinControl);
        transaction.setTransactionFee(nFeeRequired);
        if (fSubtractFeeFromAmount && fCreated)
            transaction.reassignAmounts(nChangePosRet);

        if (!fCreated)
        {
            if (!fSubtractFeeFromAmount && (total + nFeeRequired) > nBalance)
            {
                return SendCoinsReturn(AmountWithFeeExceedsBalance);
            }
            Q_EMIT message(tr("Send Coins"), QString::fromStdString(strFailReason), CClientUIInterface::MSG_ERROR);
            return TransactionCreationFailed;
        }

        // reject absurdly high fee. (This can never happen because the
        // wallet caps the fee at maxTxFee. This merely serves as a
        // belt-and-suspenders check)
        if (nFeeRequired > maxTxFeeTweak.Value())
            return AbsurdFee;
    }

    return SendCoinsReturn(OK);
}

WalletModel::SendCoinsReturn WalletModel::sendCoins(WalletModelTransaction &transaction)
{
    QByteArray transaction_array; /* store serialized transaction */

    {
        LOCK(wallet->cs_wallet);
        CWalletTx *newTx = transaction.getTransaction();

        Q_FOREACH (const SendCoinsRecipient &rcp, transaction.getRecipients())
        {
            if (rcp.paymentRequest.IsInitialized())
            {
                // Make sure any payment requests involved are still valid.
                if (PaymentServer::verifyExpired(rcp.paymentRequest.getDetails()))
                {
                    return PaymentRequestExpired;
                }

                // Store PaymentRequests in wtx.vOrderForm in wallet.
                std::string key("PaymentRequest");
                std::string value;
                rcp.paymentRequest.SerializeToString(&value);
                newTx->vOrderForm.push_back(make_pair(key, value));
            }
            else if (!rcp.message.isEmpty())
            {
                // Message from normal nexa:URI
                // (nexa:123...?message=example)
                newTx->vOrderForm.push_back(make_pair("Message", rcp.message.toStdString()));
            }
            else if (!rcp.message.isEmpty()) // Message from normal nexa:URI (nexa:123...?message=example)
                newTx->vOrderForm.push_back(make_pair("Message", rcp.message.toStdString()));
        }

        CReserveKey *keyChange = transaction.getPossibleKeyChange();
        std::string error;
        if (!wallet->CommitTransaction(*newTx, *keyChange, error))
            return TransactionCommitFailed;

        CTransaction *t = (CTransaction *)newTx;
        CDataStream ssTx(SER_NETWORK, PROTOCOL_VERSION);
        ssTx << *t;
        transaction_array.append(&(ssTx[0]), ssTx.size());
    }

    // Add addresses / update labels that we've sent to the address book,
    // and emit coinsSent signal for each recipient
    Q_FOREACH (const SendCoinsRecipient &rcp, transaction.getRecipients())
    {
        // Don't touch the address book when we have a payment request
        if (!rcp.paymentRequest.IsInitialized())
        {
            std::string strAddress = rcp.address.toStdString();
            CTxDestination dest = DecodeDestination(strAddress);
            std::string strLabel = rcp.label.toStdString();
            {
                LOCK(wallet->cs_wallet);

                std::map<CTxDestination, CAddressBookData>::iterator mi = wallet->mapAddressBook.find(dest);

                // Check if we have a new address or an updated label
                if (mi == wallet->mapAddressBook.end())
                {
                    wallet->SetAddressBook(dest, strLabel, "send");
                }
                else if (mi->second.name != strLabel)
                {
                    wallet->SetAddressBook(dest, strLabel, ""); // "" means don't change purpose
                }
            }
        }
        Q_EMIT coinsSent(wallet, rcp, transaction_array);
    }
    // update balance immediately, otherwise there could be a short noticeable delay until pollBalanceChanged hits
    checkBalanceChanged();

    return SendCoinsReturn(OK);
}

OptionsModel *WalletModel::getOptionsModel() { return optionsModel; }
AddressTableModel *WalletModel::getAddressTableModel() { return addressTableModel; }
TokenTableModel *WalletModel::getTokenTableModel() { return tokenTableModel; }
TransactionTableModel *WalletModel::getTransactionTableModel() { return transactionTableModel; }
RecentRequestsTableModel *WalletModel::getRecentRequestsTableModel() { return recentRequestsTableModel; }
WalletModel::EncryptionStatus WalletModel::getEncryptionStatus() const
{
    if (!wallet->IsCrypted())
    {
        return Unencrypted;
    }
    else if (wallet->IsLocked())
    {
        return Locked;
    }
    else
    {
        return Unlocked;
    }
}

bool WalletModel::setWalletEncrypted(bool encrypted, const SecureString &passphrase)
{
    if (encrypted)
    {
        // Encrypt
        return wallet->EncryptWallet(passphrase);
    }
    else
    {
        // Decrypt -- TODO; not supported yet
        return false;
    }
}

bool WalletModel::setWalletLocked(bool locked, const SecureString &passPhrase)
{
    if (locked)
    {
        // Lock
        return wallet->Lock();
    }
    else
    {
        // Unlock
        return wallet->Unlock(passPhrase);
    }
}

bool WalletModel::changePassphrase(const SecureString &oldPass, const SecureString &newPass)
{
    bool retval;
    {
        LOCK(wallet->cs_wallet);
        wallet->Lock(); // Make sure wallet is locked before attempting pass change
        retval = wallet->ChangeWalletPassphrase(oldPass, newPass);
    }
    return retval;
}

bool WalletModel::backupWallet(const QString &filename) { return BackupWallet(*wallet, filename.toLocal8Bit().data()); }
bool WalletModel::restoreWallet(const QString &filename)
{
    return RestoreWallet(*wallet, filename.toLocal8Bit().data());
}
// Handlers for core signals
static void NotifyKeyStoreStatusChanged(WalletModel *walletmodel, CCryptoKeyStore *wallet)
{
    qDebug() << "NotifyKeyStoreStatusChanged";
    QMetaObject::invokeMethod(walletmodel, "updateStatus", Qt::QueuedConnection);
}

static void NotifyAddressBookChanged(WalletModel *walletmodel,
    CWallet *wallet,
    const CTxDestination &address,
    const std::string &label,
    bool isMine,
    const std::string &purpose,
    ChangeType status)
{
    QString strAddress = QString::fromStdString(EncodeDestination(address));
    QString strLabel = QString::fromStdString(label);
    QString strPurpose = QString::fromStdString(purpose);

    qDebug() << "NotifyAddressBookChanged: " + strAddress + " " + strLabel + " isMine=" + QString::number(isMine) +
                    " purpose=" + strPurpose + " status=" + QString::number(status);
    QMetaObject::invokeMethod(walletmodel, "updateAddressBook", Qt::QueuedConnection, Q_ARG(QString, strAddress),
        Q_ARG(QString, strLabel), Q_ARG(bool, isMine), Q_ARG(QString, strPurpose), Q_ARG(int, status));
}

static void NotifyTransactionChanged(WalletModel *walletmodel, CWallet *wallet, const uint256 &hash, ChangeType status)
{
    Q_UNUSED(wallet);
    Q_UNUSED(hash);
    Q_UNUSED(status);
    QMetaObject::invokeMethod(walletmodel, "updateTransaction", Qt::QueuedConnection);
}

static void ShowProgress(WalletModel *walletmodel, const std::string &title, int nProgress)
{
    // emits signal "showProgress"
    QMetaObject::invokeMethod(walletmodel, "showProgress", Qt::QueuedConnection,
        Q_ARG(QString, QString::fromStdString(title)), Q_ARG(int, nProgress));
}

static void NotifyWatchonlyChanged(WalletModel *walletmodel, bool fHaveWatchonly)
{
    QMetaObject::invokeMethod(walletmodel, "updateWatchOnlyFlag", Qt::QueuedConnection, Q_ARG(bool, fHaveWatchonly));
}

void WalletModel::subscribeToCoreSignals()
{
    // Connect signals to wallet
    wallet->NotifyStatusChanged.connect(boost::bind(&NotifyKeyStoreStatusChanged, this, boost::arg<1>()));
    wallet->NotifyAddressBookChanged.connect(boost::bind(NotifyAddressBookChanged, this, boost::arg<1>(),
        boost::arg<2>(), boost::arg<3>(), boost::arg<4>(), boost::arg<5>(), boost::arg<6>()));
    wallet->NotifyTransactionChanged.connect(
        boost::bind(NotifyTransactionChanged, this, boost::arg<1>(), boost::arg<2>(), boost::arg<3>()));
    wallet->ShowProgress.connect(boost::bind(ShowProgress, this, boost::arg<1>(), boost::arg<2>()));
    wallet->NotifyWatchonlyChanged.connect(boost::bind(NotifyWatchonlyChanged, this, boost::arg<1>()));
}

void WalletModel::unsubscribeFromCoreSignals()
{
    // Disconnect signals from wallet
    wallet->NotifyStatusChanged.disconnect(boost::bind(&NotifyKeyStoreStatusChanged, this, boost::arg<1>()));
    wallet->NotifyAddressBookChanged.disconnect(boost::bind(NotifyAddressBookChanged, this, boost::arg<1>(),
        boost::arg<2>(), boost::arg<3>(), boost::arg<4>(), boost::arg<5>(), boost::arg<6>()));
    wallet->NotifyTransactionChanged.disconnect(
        boost::bind(NotifyTransactionChanged, this, boost::arg<1>(), boost::arg<2>(), boost::arg<3>()));
    wallet->ShowProgress.disconnect(boost::bind(ShowProgress, this, boost::arg<1>(), boost::arg<1>()));
    wallet->NotifyWatchonlyChanged.disconnect(boost::bind(NotifyWatchonlyChanged, this, boost::arg<1>()));
}

// WalletModel::UnlockContext implementation
WalletModel::UnlockContext WalletModel::requestUnlock()
{
    bool was_locked = getEncryptionStatus() == Locked;
    if (was_locked)
    {
        // Request UI to unlock wallet
        Q_EMIT requireUnlock();
    }
    // If wallet is still locked, unlock was failed or cancelled, mark context as invalid
    bool valid = getEncryptionStatus() != Locked;

    return UnlockContext(this, valid, was_locked);
}

WalletModel::UnlockContext::UnlockContext(WalletModel *_wallet, bool _valid, bool _relock)
    : wallet(_wallet), valid(_valid), relock(_relock)
{
}

WalletModel::UnlockContext::~UnlockContext()
{
    if (valid && relock)
    {
        wallet->setWalletLocked(true);
    }
}

void WalletModel::UnlockContext::CopyFrom(const UnlockContext &rhs)
{
    // Transfer context; old object no longer relocks wallet
    *this = rhs;
    rhs.relock = false;
}

bool WalletModel::getPubKey(const CKeyID &address, CPubKey &vchPubKeyOut) const
{
    return wallet->GetPubKey(address, vchPubKeyOut);
}

bool WalletModel::IsSpendable(const CTxDestination &dest) const { return wallet->IsMine(dest) & ISMINE_SPENDABLE; }
// returns a list of COutputs from COutPoints
void WalletModel::getOutputs(const std::vector<COutPoint> &vOutpoints, std::vector<COutput> &vOutputs)
{
    LOCK(wallet->cs_wallet);
    for (const COutPoint &outpoint : vOutpoints)
    {
        if (!wallet->mapWallet.count(outpoint))
            continue;
        COutput &out = wallet->mapWallet[outpoint];
        int nDepth = out.GetDepthInMainChain();
        if (nDepth < 0)
            continue;
        if (!out.spendable())
            continue;
        vOutputs.push_back(out);
    }
}

bool WalletModel::isSpent(const COutPoint &outpoint) const
{
    LOCK(wallet->cs_wallet);
    return wallet->IsSpent(outpoint);
}

// AvailableCoins + LockedCoins grouped by wallet address (put change in one group with wallet address)
void WalletModel::listCoins(std::map<QString, std::vector<COutput> > &mapCoins) const
{
    std::vector<COutput> vCoins;
    wallet->AvailableCoins(vCoins);

    LOCK(wallet->cs_wallet); // ListLockedCoins, mapWallet
    std::vector<COutPoint> vLockedCoins;
    wallet->ListLockedCoins(vLockedCoins);

    // add locked coins
    for (const COutPoint &outpoint : vLockedCoins)
    {
        if (!wallet->mapWallet.count(outpoint))
            continue;
        COutput out = wallet->mapWallet[outpoint];
        assert(out.tx);
        if (out.i < 0)
            continue; // this is a transaction entry, not an outpoint (aka coin).
        int nDepth = out.tx->GetDepthInMainChain();
        if (nDepth < 0)
            continue;
        if ((out.i < (int)out.tx->vout.size()) && (wallet->IsMine(out.tx->vout[out.i]) == ISMINE_SPENDABLE))
            vCoins.push_back(out);
    }

    for (const COutput &out : vCoins)
    {
        COutput cout = out;

        while (wallet->IsChange(cout.tx->vout[cout.i]) && cout.tx->vin.size() > 0 && wallet->IsMine(cout.tx->vin[0]))
        {
            auto prior = wallet->mapWallet.find(cout.tx->vin[0].prevout);
            if (prior == wallet->mapWallet.end())
                break;
            cout = COutput(prior->second.tx, prior->second.i, isminetype::ISMINE_SPENDABLE);
        }

        CTxDestination address;
        if (!out.spendable() || !ExtractDestination(cout.tx->vout[cout.i].scriptPubKey, address))
            continue;
        mapCoins[QString::fromStdString(EncodeDestination(address))].push_back(out);
    }
}

bool WalletModel::isLockedCoin(const COutPoint &outpt) const
{
    LOCK(wallet->cs_wallet);
    return wallet->IsLockedCoin(outpt);
}

void WalletModel::lockCoin(const COutPoint &output)
{
    LOCK(wallet->cs_wallet);
    wallet->LockCoin(output);
}

void WalletModel::unlockCoin(const COutPoint &output)
{
    LOCK(wallet->cs_wallet);
    wallet->UnlockCoin(output);
}

void WalletModel::listLockedCoins(std::vector<COutPoint> &vOutpts)
{
    LOCK(wallet->cs_wallet);
    wallet->ListLockedCoins(vOutpts);
}

void WalletModel::loadReceiveRequests(std::vector<std::string> &vReceiveRequests)
{
    LOCK(wallet->cs_wallet);
    for (const auto &item : wallet->mapAddressBook)
        for (const auto &item2 : item.second.destdata)
            if (item2.first.size() > 2 && item2.first.substr(0, 2) == "rr") // receive request
                vReceiveRequests.push_back(item2.second);
}

bool WalletModel::saveReceiveRequest(const std::string &sAddress, const int64_t nId, const std::string &sRequest)
{
    CTxDestination dest = DecodeDestination(sAddress);

    std::stringstream ss;
    ss << nId;
    std::string key = "rr" + ss.str(); // "rr" prefix = "receive request" in destdata

    LOCK(wallet->cs_wallet);
    if (sRequest.empty())
        return wallet->EraseDestData(dest, key);
    else
        return wallet->AddDestData(dest, key, sRequest);
}

bool WalletModel::hdEnabled() const { return wallet->IsHDEnabled(); }

int WalletModel::AddTokenTracker(const CGroupTokenID &grpID, const std::string &strTokenTicker) const
{
    // cs_wallet locked inside method
    return wallet->AddTokenTracker(grpID, strTokenTicker);
}

int WalletModel::RemoveTokenTracker(const CGroupTokenID &grpID) const
{
    // cs_wallet locked inside method
    return wallet->RemoveTokenTracker(grpID);
}

void WalletModel::listTokenTrackers(std::map<CGroupTokenID, std::string> &mapTrackers) const
{
    LOCK(wallet->cs_wallet);
    mapTrackers = wallet->mapTokenTrackers;
}

void WalletModel::listTokens(std::map<CGroupTokenID, std::string> &mapTokens) const
{
    // Find all the coins that have a groupID
    // add the ID and the ticker to mapTokens
    {
        LOCK(wallet->cs_wallet);
        for (auto &iter : wallet->mapWallet)
        {
            const COutput &coin = iter.second;
            if (coin.isNull() || coin.txOnly())
            {
                continue;
            }
            CGroupTokenInfo tg(coin.GetScriptPubKey());
            if (tg.associatedGroup != NoGroup && mapTokens.count(tg.associatedGroup) == 0)
            {
                CGroupTokenID grpID = tg.associatedGroup;
                if (grpID.isSubgroup())
                {
                    grpID = grpID.parentGroup();
                }
                std::string strTokenTicker;
                if (GetGroupTicker(grpID, strTokenTicker) == false)
                {
                    strTokenTicker = "???";
                }
                mapTokens.emplace(tg.associatedGroup, strTokenTicker);
            }
        }
    }
}
