// Copyright (c) 2012-2015 The Bitcoin Core developers
// Copyright (c) 2015-2022 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include "key.h"

#include "base58.h"
#include "dstencode.h"
#include "script/script.h"
#include "test/test_nexa.h"
#include "uint256.h"
#include "util.h"
#include "utilstrencodings.h"

#include <string>
#include <vector>

#include <boost/test/unit_test.hpp>

#include <secp256k1.h>
#include <secp256k1_recovery.h>
#include <secp256k1_schnorr.h>

using namespace std;

static const std::string strSecret1 = "5HxWvvfubhXpYYpS3tJkw6fq9jE9j18THftkZjHHfmFiWtmAbrj";
static const std::string strSecret2 = "5KC4ejrDjv152FGwP386VD1i2NYc5KkfSMyv1nGy1VGDxGHqVY3";
static const std::string strSecret1C = "Kwr371tjA9u2rFSMZjTNun2PXXP3WPZu2afRHTcta6KxEUdm1vEw";
static const std::string strSecret2C = "L3Hq7a8FEQwJkW1M2GNKDW28546Vp5miewcCzSqUD9kCAXrJdS3g";
static const std::string addr1 = "1QFqqMUD55ZV3PJEJZtaKCsQmjLT6JkjvJ";
static const std::string addr2 = "1F5y5E5FMc5YzdJtB9hLaUe43GDxEKXENJ";
static const std::string addr1C = "1NoJrossxPBKfCHuJXT4HadJrXRE9Fxiqs";
static const std::string addr2C = "1CRj2HyM1CXWzHAXLQtiGLyggNT9WQqsDs";

static const std::string rawpub1 = "0b4c866585dd868a9d62348a9cd008d6a312937048fff31670e7e920cfc7a7447b5f0bba9e01e6fe473"
                                   "5c8383e6e7a3347a0fd72381b8f797a19f694054e5a69";
static const std::string rawpub2 = "183905ae25e815634ce7f5d9bedbaa2c39032ab98c75b5e88fe43f8dd8246f3c5473ccd4ab475e6a9e6"
                                   "620b52f5ce2fd15a2de32cbe905154b3a05844af70785";
static const std::string rawpub1c = "0b4c866585dd868a9d62348a9cd008d6a312937048fff31670e7e920cfc7a744";
static const std::string rawpub2c = "183905ae25e815634ce7f5d9bedbaa2c39032ab98c75b5e88fe43f8dd8246f3c";

static const std::string strAddressBad = "1HV9Lc3sNHZxwj4Zk6fB38tEmBryq2cBiF";

#ifdef KEY_TESTS_DUMPINFO
void dumpKeyInfo(uint256 privkey)
{
    CKey key;
    key.resize(32);
    memcpy(&secret[0], &privkey, 32);
    vector<unsigned char> sec;
    sec.resize(32);
    memcpy(&sec[0], &secret[0], 32);
    printf("  * secret (hex): %s\n", HexStr(sec).c_str());

    for (int nCompressed = 0; nCompressed < 2; nCompressed++)
    {
        bool fCompressed = nCompressed == 1;
        printf("  * %s:\n", fCompressed ? "compressed" : "uncompressed");
        CBitcoinSecret bsecret;
        bsecret.SetSecret(secret, fCompressed);
        printf("    * secret (base58): %s\n", bsecret.ToString().c_str());
        CKey key;
        key.SetSecret(secret, fCompressed);
        vector<unsigned char> vchPubKey = key.GetPubKey();
        printf("    * pubkey (hex): %s\n", HexStr(vchPubKey).c_str());
        printf("    * address (base58): %s\n", CBitcoinAddress(vchPubKey).ToString().c_str());
    }
}
#endif

// get r value produced by ECDSA signing algorithm
// (assumes ECDSA r is encoded in the canonical manner)
std::vector<uint8_t> get_r_ECDSA(std::vector<uint8_t> sigECDSA)
{
    std::vector<uint8_t> ret(32, 0);

    assert(sigECDSA[2] == 2);
    int rlen = sigECDSA[3];
    assert(rlen <= 33);
    assert(sigECDSA[4 + rlen] == 2);
    if (rlen == 33)
    {
        assert(sigECDSA[4] == 0);
        assert(sigECDSA.size() >= 37);
        std::copy(sigECDSA.begin() + 5, sigECDSA.begin() + 37, ret.begin());
    }
    else
    {
        assert(rlen <= 32);
        assert(int(sigECDSA.size()) >= 4 + rlen);
        std::copy(sigECDSA.begin() + 4, sigECDSA.begin() + 4 + rlen, ret.begin() + (32 - rlen));
    }
    return ret;
}

BOOST_FIXTURE_TEST_SUITE(key_tests, BasicTestingSetup)

BOOST_AUTO_TEST_CASE(internal_test)
{
    // test get_r_ECDSA (defined above) to make sure it's working properly
    BOOST_CHECK(get_r_ECDSA(ParseHex("3045022100c6ab5f8acfccc114da39dd5ad0b1ef4d39df6a721e8"
                                     "24c22e00b7bc7944a1f7802206ff23df3802e241ee234a8b66c40"
                                     "c82e56a6cc37f9b50463111c9f9229b8f3b3")) ==
                ParseHex("c6ab5f8acfccc114da39dd5ad0b1ef4d39df6a721e824c22e00b7bc7944a1f78"));
    BOOST_CHECK(get_r_ECDSA(ParseHex("3045022046ab5f8acfccc114da39dd5ad0b1ef4d39df6a721e8"
                                     "24c22e00b7bc7944a1f7802206ff23df3802e241ee234a8b66c40"
                                     "c82e56a6cc37f9b50463111c9f9229b8f3b3")) ==
                ParseHex("46ab5f8acfccc114da39dd5ad0b1ef4d39df6a721e824c22e00b7bc7944a1f78"));
    BOOST_CHECK(get_r_ECDSA(ParseHex("3045021f4b5f8acfccc114da39dd5ad0b1ef4d39df6a721e8"
                                     "24c22e00b7bc7944a1f7802206ff23df3802e241ee234a8b66c40"
                                     "c82e56a6cc37f9b50463111c9f9229b8f3b3")) ==
                ParseHex("004b5f8acfccc114da39dd5ad0b1ef4d39df6a721e824c22e00b7bc7944a1f78"));
    BOOST_CHECK(get_r_ECDSA(ParseHex("3045021e5f8acfccc114da39dd5ad0b1ef4d39df6a721e8"
                                     "24c22e00b7bc7944a1f7802206ff23df3802e241ee234a8b66c40"
                                     "c82e56a6cc37f9b50463111c9f9229b8f3b3")) ==
                ParseHex("00005f8acfccc114da39dd5ad0b1ef4d39df6a721e824c22e00b7bc7944a1f78"));
}

BOOST_AUTO_TEST_CASE(key_test1)
{
    SelectParams(CBaseChainParams::LEGACY_UNIT_TESTS);

    CBitcoinSecret bsecret1, bsecret2, bsecret1C, bsecret2C, baddress1;
    BOOST_CHECK(bsecret1.SetString(Params(), strSecret1));
    BOOST_CHECK(bsecret2.SetString(Params(), strSecret2));
    BOOST_CHECK(bsecret1C.SetString(Params(), strSecret1C));
    BOOST_CHECK(bsecret2C.SetString(Params(), strSecret2C));
    BOOST_CHECK(!baddress1.SetString(Params(), strAddressBad));

    CKey key1 = bsecret1.GetKey();
    BOOST_CHECK(key1.IsCompressed() == false);
    CKey key2 = bsecret2.GetKey();
    BOOST_CHECK(key2.IsCompressed() == false);
    CKey key1C = bsecret1C.GetKey();
    BOOST_CHECK(key1C.IsCompressed() == true);
    CKey key2C = bsecret2C.GetKey();
    BOOST_CHECK(key2C.IsCompressed() == true);

    CPubKey pubkey1 = key1.GetPubKey();
    CPubKey pubkey2 = key2.GetPubKey();
    CPubKey pubkey1C = key1C.GetPubKey();
    CPubKey pubkey2C = key2C.GetPubKey();

    std::vector<uint8_t> v;
    pubkey1.Raw(v);
    BOOST_CHECK(HexStr(v) == rawpub1);
    pubkey2.Raw(v);
    BOOST_CHECK(HexStr(v) == rawpub2);
    pubkey1C.Raw(v);
    BOOST_CHECK(HexStr(v) == rawpub1c);
    pubkey2C.Raw(v);
    BOOST_CHECK(HexStr(v) == rawpub2c);

    BOOST_CHECK(key1.VerifyPubKey(pubkey1));
    BOOST_CHECK(!key1.VerifyPubKey(pubkey1C));
    BOOST_CHECK(!key1.VerifyPubKey(pubkey2));
    BOOST_CHECK(!key1.VerifyPubKey(pubkey2C));

    BOOST_CHECK(!key1C.VerifyPubKey(pubkey1));
    BOOST_CHECK(key1C.VerifyPubKey(pubkey1C));
    BOOST_CHECK(!key1C.VerifyPubKey(pubkey2));
    BOOST_CHECK(!key1C.VerifyPubKey(pubkey2C));

    BOOST_CHECK(!key2.VerifyPubKey(pubkey1));
    BOOST_CHECK(!key2.VerifyPubKey(pubkey1C));
    BOOST_CHECK(key2.VerifyPubKey(pubkey2));
    BOOST_CHECK(!key2.VerifyPubKey(pubkey2C));

    BOOST_CHECK(!key2C.VerifyPubKey(pubkey1));
    BOOST_CHECK(!key2C.VerifyPubKey(pubkey1C));
    BOOST_CHECK(!key2C.VerifyPubKey(pubkey2));
    BOOST_CHECK(key2C.VerifyPubKey(pubkey2C));

    BOOST_CHECK(DecodeDestination(addr1) == CTxDestination(pubkey1.GetID()));
    BOOST_CHECK(DecodeDestination(addr2) == CTxDestination(pubkey2.GetID()));
    BOOST_CHECK(DecodeDestination(addr1C) == CTxDestination(pubkey1C.GetID()));
    BOOST_CHECK(DecodeDestination(addr2C) == CTxDestination(pubkey2C.GetID()));

    CPubKey pubkey1c = pubkey1;
    pubkey1c.Compress();
    BOOST_CHECK(pubkey1c.IsValid());
    BOOST_CHECK(pubkey1c.size() == 33);
    pubkey1c.Decompress();
    BOOST_CHECK(pubkey1c == pubkey1);
    BOOST_CHECK(pubkey1c.IsValid());
    BOOST_CHECK(pubkey1c.size() == 65);


    CPubKey pubkey2c = pubkey2;
    pubkey2c.Compress();
    BOOST_CHECK(pubkey2c.IsValid());
    BOOST_CHECK(pubkey2c.size() == 33);
    pubkey2c.Decompress();
    BOOST_CHECK(pubkey2c == pubkey2);
    BOOST_CHECK(pubkey2c.IsValid());
    BOOST_CHECK(pubkey2c.size() == 65);


    for (int n = 0; n < 16; n++)
    {
        string strMsg = strprintf("Very secret message %i: 11", n);
        uint256 hashMsg = Hash(strMsg.begin(), strMsg.end());

        // normal ECDSA signatures

        vector<unsigned char> sign1, sign2, sign1C, sign2C;

        BOOST_CHECK(key1.SignECDSA(hashMsg, sign1));
        BOOST_CHECK(key2.SignECDSA(hashMsg, sign2));
        BOOST_CHECK(key1C.SignECDSA(hashMsg, sign1C));
        BOOST_CHECK(key2C.SignECDSA(hashMsg, sign2C));

        BOOST_CHECK(pubkey1.VerifyECDSA(hashMsg, sign1));
        BOOST_CHECK(!pubkey1.VerifyECDSA(hashMsg, sign2));
        BOOST_CHECK(pubkey1.VerifyECDSA(hashMsg, sign1C));
        BOOST_CHECK(!pubkey1.VerifyECDSA(hashMsg, sign2C));

        BOOST_CHECK(!pubkey2.VerifyECDSA(hashMsg, sign1));
        BOOST_CHECK(pubkey2.VerifyECDSA(hashMsg, sign2));
        BOOST_CHECK(!pubkey2.VerifyECDSA(hashMsg, sign1C));
        BOOST_CHECK(pubkey2.VerifyECDSA(hashMsg, sign2C));

        BOOST_CHECK(pubkey1C.VerifyECDSA(hashMsg, sign1));
        BOOST_CHECK(!pubkey1C.VerifyECDSA(hashMsg, sign2));
        BOOST_CHECK(pubkey1C.VerifyECDSA(hashMsg, sign1C));
        BOOST_CHECK(!pubkey1C.VerifyECDSA(hashMsg, sign2C));

        BOOST_CHECK(!pubkey2C.VerifyECDSA(hashMsg, sign1));
        BOOST_CHECK(pubkey2C.VerifyECDSA(hashMsg, sign2));
        BOOST_CHECK(!pubkey2C.VerifyECDSA(hashMsg, sign1C));
        BOOST_CHECK(pubkey2C.VerifyECDSA(hashMsg, sign2C));

        // compact ECDSA signatures (with key recovery)

        vector<unsigned char> csign1, csign2, csign1C, csign2C;

        BOOST_CHECK(key1.SignCompact(hashMsg, csign1));
        BOOST_CHECK(key2.SignCompact(hashMsg, csign2));
        BOOST_CHECK(key1C.SignCompact(hashMsg, csign1C));
        BOOST_CHECK(key2C.SignCompact(hashMsg, csign2C));

        CPubKey rkey1, rkey2, rkey1C, rkey2C;

        BOOST_CHECK(rkey1.RecoverCompact(hashMsg, csign1));
        BOOST_CHECK(rkey2.RecoverCompact(hashMsg, csign2));
        BOOST_CHECK(rkey1C.RecoverCompact(hashMsg, csign1C));
        BOOST_CHECK(rkey2C.RecoverCompact(hashMsg, csign2C));

        BOOST_CHECK(rkey1 == pubkey1);
        BOOST_CHECK(rkey2 == pubkey2);
        BOOST_CHECK(rkey1C == pubkey1C);
        BOOST_CHECK(rkey2C == pubkey2C);

        // Schnorr signatures

        std::vector<uint8_t> ssign1, ssign2, ssign1C, ssign2C;

        BOOST_CHECK(key1.SignSchnorr(hashMsg, ssign1));
        BOOST_CHECK(key2.SignSchnorr(hashMsg, ssign2));
        BOOST_CHECK(key1C.SignSchnorr(hashMsg, ssign1C));
        BOOST_CHECK(key2C.SignSchnorr(hashMsg, ssign2C));

        std::vector<uint8_t> ssign1N, ssign2N, ssign1CN, ssign2CN;
        uint8_t nonce[32] = {1, 2, 3, 4, 5, 6, 7, 8};


        BOOST_CHECK(key1.SignSchnorrWithNonce(hashMsg, nonce, ssign1N));
        BOOST_CHECK(key2.SignSchnorrWithNonce(hashMsg, nonce, ssign2N));
        BOOST_CHECK(key1C.SignSchnorrWithNonce(hashMsg, nonce, ssign1CN));
        BOOST_CHECK(key2C.SignSchnorrWithNonce(hashMsg, nonce, ssign2CN));

        // This section executes tests specific to the operation of Schnorr signatures, which are described
        // in BIP340: https://bips.xyz/340.  You will need to be familiar with Schnorr to follow the rest of
        // this comment.
        // In review, the signature is formulated as: pairs (R, s) that satisfy s (*) G = R + hash(R || m) (*) P
        // where (*) is group multiplication by a scalar (or group addition N times).
        // and where R is "public nonce", which is the private nonce provided in the function call (*) G.
        // Group multiplication by a scalar (*) is also used in Schnorr to convert a private key to a public key.
        // So what this test does is it gets the "pubkey" of the nonce (so the public nonce), and then
        // breaks the signature into its R, s components (which is just splitting it in half)
        // and compares these two numbers.
        // This proves that SignSchnorrWithNonce really did create a signature using the provided nonce.
        auto nonceAsKey = CKey();
        nonceAsKey.Set(nonce, nonce + 32, false);
        auto pubNonce = nonceAsKey.GetPubKey();
        BOOST_CHECK(pubNonce.Compress());
        std::vector<uint8_t> rawPubNonce;
        BOOST_CHECK(pubNonce.Raw(rawPubNonce));
        BOOST_CHECK(std::equal(ssign1N.begin(), ssign1N.begin() + 32, rawPubNonce.begin()));
        // If you use the same nonce then the R part of the signature will be the same
        BOOST_CHECK(std::equal(ssign1N.begin(), ssign1N.begin() + 32, ssign2N.begin()));
        BOOST_CHECK(std::equal(ssign1N.begin(), ssign1N.begin() + 32, ssign2N.begin()));
        BOOST_CHECK(std::equal(ssign1N.begin(), ssign1N.begin() + 32, ssign1CN.begin()));
        BOOST_CHECK(std::equal(ssign1N.begin(), ssign1N.begin() + 32, ssign2CN.begin()));

        BOOST_CHECK(pubkey1.VerifySchnorr(hashMsg, ssign1));
        BOOST_CHECK(!pubkey1.VerifySchnorr(hashMsg, ssign2));
        BOOST_CHECK(pubkey1.VerifySchnorr(hashMsg, ssign1C));
        BOOST_CHECK(!pubkey1.VerifySchnorr(hashMsg, ssign2C));

        BOOST_CHECK(pubkey1.VerifySchnorr(hashMsg, ssign1N));
        BOOST_CHECK(!pubkey1.VerifySchnorr(hashMsg, ssign2N));
        BOOST_CHECK(pubkey1.VerifySchnorr(hashMsg, ssign1CN));
        BOOST_CHECK(!pubkey1.VerifySchnorr(hashMsg, ssign2CN));

        BOOST_CHECK(!pubkey2.VerifySchnorr(hashMsg, ssign1));
        BOOST_CHECK(pubkey2.VerifySchnorr(hashMsg, ssign2));
        BOOST_CHECK(!pubkey2.VerifySchnorr(hashMsg, ssign1C));
        BOOST_CHECK(pubkey2.VerifySchnorr(hashMsg, ssign2C));

        BOOST_CHECK(pubkey1C.VerifySchnorr(hashMsg, ssign1));
        BOOST_CHECK(!pubkey1C.VerifySchnorr(hashMsg, ssign2));
        BOOST_CHECK(pubkey1C.VerifySchnorr(hashMsg, ssign1C));
        BOOST_CHECK(!pubkey1C.VerifySchnorr(hashMsg, ssign2C));

        BOOST_CHECK(!pubkey2C.VerifySchnorr(hashMsg, ssign1));
        BOOST_CHECK(pubkey2C.VerifySchnorr(hashMsg, ssign2));
        BOOST_CHECK(!pubkey2C.VerifySchnorr(hashMsg, ssign1C));
        BOOST_CHECK(pubkey2C.VerifySchnorr(hashMsg, ssign2C));

        // check deterministicity of ECDSA & Schnorr
        BOOST_CHECK(sign1 == sign1C);
        BOOST_CHECK(sign2 == sign2C);
        BOOST_CHECK(ssign1 == ssign1C);
        BOOST_CHECK(ssign2 == ssign2C);

        // Extract r value from ECDSA and Schnorr. Make sure they are
        // distinct (nonce reuse would be dangerous and can leak private key).
        std::vector<uint8_t> rE1 = get_r_ECDSA(sign1);
        BOOST_CHECK(ssign1.size() == 64);
        std::vector<uint8_t> rS1(ssign1.begin(), ssign1.begin() + 32);
        BOOST_CHECK(rE1.size() == 32);
        BOOST_CHECK(rS1.size() == 32);
        BOOST_CHECK(rE1 != rS1);

        std::vector<uint8_t> rE2 = get_r_ECDSA(sign2);
        BOOST_CHECK(ssign2.size() == 64);
        std::vector<uint8_t> rS2(ssign2.begin(), ssign2.begin() + 32);
        BOOST_CHECK(rE2.size() == 32);
        BOOST_CHECK(rS2.size() == 32);
        BOOST_CHECK(rE2 != rS2);
    }

    // test ECDSA DER signature verification
    std::string strAddr = "nqtsq5g5507g8lmp3dl458hzwzty6sq6zsttqq2n36eel304";
    CTxDestination destination = DecodeDestination(strAddr, Params(CBaseChainParams::NEXA));
    BOOST_CHECK(IsValidDestination(destination) == true);

    std::string strDERsig =
        "MEUCIQCv0+pkaVamk2Y4iQbHT77oxrovU44UgXWdlbDFpEyiZgIgDVgJk09ATrRfFU0UbPIrO0ETk3iJiaS7VLwXLoL871w=";
    bool fInvalid = false;
    std::vector<unsigned char> vDERsig = DecodeBase64(strDERsig.c_str(), &fInvalid);
    BOOST_CHECK(fInvalid == false);

    std::string strMessage = "Hello";
    CHashWriter ss(SER_GETHASH, 0);
    ss << strMessageMagic;
    ss << strMessage;

    CPubKey pubkey;
    BOOST_CHECK(pubkey.RecoverCompact(ss.GetHash(), vDERsig));
    ScriptTemplateDestination *st = nullptr;
    BOOST_CHECK((st = std::get_if<ScriptTemplateDestination>(&destination)) != nullptr);
    CGroupTokenInfo groupInfo;
    std::vector<unsigned char> templateHash;
    BOOST_CHECK(GetScriptTemplate(st->toScript(), &groupInfo, &templateHash) == ScriptTemplateError::OK);
    BOOST_CHECK(templateHash == P2PKT_ID);
    ScriptTemplateDestination signedBy(P2pktOutput(pubkey));
    BOOST_CHECK(*st == signedBy);


    // test deterministic signing expected values

    std::vector<unsigned char> detsig, detsigc;
    string strMsg = "Very deterministic message";
    uint256 hashMsg = Hash(strMsg.begin(), strMsg.end());
    // ECDSA
    BOOST_CHECK(key1.SignECDSA(hashMsg, detsig));
    BOOST_CHECK(key1C.SignECDSA(hashMsg, detsigc));
    BOOST_CHECK(detsig == detsigc);
    BOOST_CHECK(detsig == ParseHex("3045022100c6ab5f8acfccc114da39dd5ad0b1ef4d39df6a721e8"
                                   "24c22e00b7bc7944a1f7802206ff23df3802e241ee234a8b66c40"
                                   "c82e56a6cc37f9b50463111c9f9229b8f3b3"));
    BOOST_CHECK(key2.SignECDSA(hashMsg, detsig));
    BOOST_CHECK(key2C.SignECDSA(hashMsg, detsigc));
    BOOST_CHECK(detsig == detsigc);
    BOOST_CHECK(detsig == ParseHex("304502210094dc5a77b8d5db6b42b66c29d7033cd873fac7a1272"
                                   "4a90373726f60bb9f852a02204eb4c98b9a2f5c017f9417ba7c43"
                                   "279c20c84bb058dc05b3beeb9333016b15bb"));

    // Compact
    BOOST_CHECK(key1.SignCompact(hashMsg, detsig));
    BOOST_CHECK(key1C.SignCompact(hashMsg, detsigc));
    BOOST_CHECK(detsig == ParseHex("1b8c56f224d51415e6ce329144aa1e1c1563e297a005f450df015"
                                   "14f3d047681760277e79d57502df27b8feebb001a588aa3a8c2bc"
                                   "f5b2367273c15f840638cfc8"));
    BOOST_CHECK(detsigc == ParseHex("1f8c56f224d51415e6ce329144aa1e1c1563e297a005f450df015"
                                    "14f3d047681760277e79d57502df27b8feebb001a588aa3a8c2bc"
                                    "f5b2367273c15f840638cfc8"));
    BOOST_CHECK(key2.SignCompact(hashMsg, detsig));
    BOOST_CHECK(key2C.SignCompact(hashMsg, detsigc));
    BOOST_CHECK(detsig == ParseHex("1c9ffc56b38fbfc0e3eb2c42dff99d2375982449f35019c1b3d56"
                                   "ca62bef187c5103e483a0ad481eaacc224fef4ee2995027300d5f"
                                   "2457f7a20c43547aeddbae6e"));
    BOOST_CHECK(detsigc == ParseHex("209ffc56b38fbfc0e3eb2c42dff99d2375982449f35019c1b3d56"
                                    "ca62bef187c5103e483a0ad481eaacc224fef4ee2995027300d5f"
                                    "2457f7a20c43547aeddbae6e"));

    // Schnorr
    BOOST_CHECK(key1.SignSchnorr(hashMsg, detsig));
    BOOST_CHECK(key1C.SignSchnorr(hashMsg, detsigc));
    BOOST_CHECK(detsig == detsigc);
    BOOST_CHECK(detsig == ParseHex("2c56731ac2f7a7e7f11518fc7722a166b02438924ca9d8b4d1113"
                                   "47b81d0717571846de67ad3d913a8fdf9d8f3f73161a4c48ae81c"
                                   "b183b214765feb86e255ce"));
    BOOST_CHECK(key2.SignSchnorr(hashMsg, detsig));
    BOOST_CHECK(key2C.SignSchnorr(hashMsg, detsigc));
    BOOST_CHECK(detsig == detsigc);
    BOOST_CHECK(detsig == ParseHex("e7167ae0afbba6019b4c7fcfe6de79165d555e8295bd72da1b8aa"
                                   "1a5b54305880517cace1bcb0cb515e2eeaffd49f1e4dd49fd7282"
                                   "6b4b1573c84da49a38405d"));
}

BOOST_AUTO_TEST_SUITE_END()
