// Copyright (c) 2018 The Bitcoin developers
// Copyright (c) 2018-2022 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#ifndef NEXA_TX_ORPHANPOOL
#define NEXA_TX_ORPHANPOOL

#include "net.h"
#include "primitives/block.h"
#include "primitives/transaction.h"
#include "sync.h"
#include "uint256.h"

#include <map>
#include <set>
#include <stdint.h>

extern CTweak<uint32_t> maxOrphanPool;
extern CTweak<uint32_t> orphanPoolExpiry;

class CTxOrphanPool
{
public:
    // General orphan pool READ/WRITE lock
    CSharedCriticalSection cs_orphanpool;

private:
    //! Used in EraseOrphansByTime() to track when the last time was we checked the cache for anything to delete
    std::atomic<int64_t> nLastOrphanCheck;

public:
    //! Current in memory footprint of all txns in the orphan pool.
    uint64_t nBytesOrphanPool GUARDED_BY(cs_orphanpool);

    struct COrphanTx
    {
        CTransactionRef ptx;
        NodeId fromPeer;
        int64_t nEntryTime;
        uint64_t nOrphanTxSize;
    };

    std::map<uint256, COrphanTx> mapOrphanTransactions GUARDED_BY(cs_orphanpool);
    std::map<uint256, std::set<uint256> > mapOrphanTransactionsByPrev GUARDED_BY(cs_orphanpool);

    // Used for syncronizing post block processing of the orphan pool
    CCriticalSection cs_blockprocessing;
    std::deque<ConstCBlockRef> vPostBlockProcessing GUARDED_BY(cs_blockprocessing);

    CCriticalSection cs_processorphans;
    std::deque<std::vector<CTransactionRef>> vProcessOrphans GUARDED_BY(cs_processorphans);

    CTxOrphanPool();

    //! Do we already have this orphan in the orphan pool
    bool AlreadyHaveOrphan(const uint256 &txid);

    //! Add a transaction to the orphan pool
    bool AddOrphanTx(const CTransactionRef ptx, NodeId peer);

    //! Erase an ophan tx from the orphan pool
    //! @return true if an orphan matching the hash was found in the orphanpool and successfully erased.
    bool EraseOrphanTx(const uint256 &hash);

    //! Expire old orphans from the orphan pool
    void EraseOrphansByTime();

    //! Limit the orphan pool size by either number of transactions or the max orphan pool size allowed.
    unsigned int LimitOrphanTxSize(unsigned int nMaxOrphans, uint64_t nMaxBytes);

    //! Return all the transaction hashes for transactions currently in the orphan pool.
    void QueryIds(std::vector<uint256> &vHashes);

    //! Set the last orphan check time (used primarily in testing)
    // FIXME SetLastOrphanCheck seems to be used only in unit tests
    // DoS_tests.cpp and txvalidationcache_tests.cpp
    void _SetLastOrphanCheck(int64_t nTime)
    {
        AssertLockHeld(cs_orphanpool);
        nLastOrphanCheck = nTime;
    }
    //! Orphan pool current number of transactions
    uint64_t GetOrphanPoolSize()
    {
        READLOCK(cs_orphanpool);
        return mapOrphanTransactions.size();
    }

    //! Orphan pool bytes used
    uint64_t GetOrphanPoolBytes()
    {
        READLOCK(cs_orphanpool);
        return nBytesOrphanPool;
    }

    //! Remove all orphans from the pool that are in this group of transactions
    void RemoveForBlock(const std::vector<CTransactionRef> &vtx);

    //! Clear the orphan pool
    void clear()
    {
        WRITELOCK(cs_orphanpool);
        mapOrphanTransactions.clear();
        mapOrphanTransactionsByPrev.clear();
        nBytesOrphanPool = 0;
    }

private:
    //! Return all the orphan pool data structures so they can be saved to disk
    std::vector<CTxOrphanPool::COrphanTx> AllTxOrphanPoolInfo() const;

public:
    //! Load the orphan pool from disk
    bool LoadOrphanPool();

    //! Save the orphan pool to disk
    bool DumpOrphanPool();
};
extern CTxOrphanPool orphanpool;

uint64_t ProcessOrphans(const std::vector<CTransactionRef> &vWorkQueue);

#endif
