// Copyright (c) 2010 Satoshi Nakamoto
// Copyright (c) 2009-2015 The Bitcoin Core developers
// Copyright (c) 2015-2023 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include "amount.h"
#include "cashaddrenc.h"
#include "chain.h"
#include "coincontrol.h"
#include "core_io.h"
#include "dstencode.h"
#include "grouptokenwallet.h"
#include "init.h"
#include "main.h"
#include "net.h"
#include "rpc/server.h"
#include "script/sign.h"

#include "timedata.h"
#include "txadmission.h"
#include "util.h"
#include "utilmoneystr.h"
#include "validation/validation.h"
#include "wallet.h"
#include "walletdb.h"

#include <limits>
#include <stdint.h>
#include <univalue.h>

using namespace std;

int64_t nWalletUnlockTime;
static CCriticalSection cs_nWalletUnlockTime;
static CCriticalSection serializeCreateTx;

extern CTweak<bool> instantTxns;


std::string HelpRequiringPassphrase()
{
    return pwalletMain && pwalletMain->IsCrypted() ?
               "\nRequires wallet passphrase to be set with walletpassphrase call." :
               "";
}

bool EnsureWalletIsAvailable(bool avoidException)
{
    if (!pwalletMain)
    {
        if (!avoidException)
            throw JSONRPCError(RPC_METHOD_NOT_FOUND, "Method not found (disabled)");
        else
            return false;
    }
    return true;
}

void EnsureWalletIsUnlocked()
{
    LOCK(pwalletMain->cs_wallet);
    if (pwalletMain->IsLocked())
        throw JSONRPCError(RPC_WALLET_UNLOCK_NEEDED,
            "ACTION REQUIRED: You must first unlock the wallet by running the rpc: \"walletpassphrase "
            "<your passphrase> <timeout in seconds>\"");
}

void WalletTxToJSON(const CWalletTx &wtx, UniValue &entry)
{
    int confirms = wtx.GetDepthInMainChain();
    entry.pushKV("confirmations", confirms);
    if (wtx.IsCoinBase())
        entry.pushKV("generated", true);
    if (confirms > 0)
    {
        entry.pushKV("blockhash", wtx.hashBlock.GetHex());
        entry.pushKV("blockindex", wtx.nIndex);
        auto *tmp = LookupBlockIndex(wtx.hashBlock);
        if (tmp)
            entry.pushKV("blocktime", tmp->GetBlockTime());
    }
    else
    {
        entry.pushKV("trusted", wtx.IsTrusted());
    }
    entry.pushKV("txid", wtx.GetId().GetHex());
    entry.pushKV("txidem", wtx.GetIdem().GetHex());
    UniValue conflicts(UniValue::VARR);
    for (const uint256 &conflict : wtx.GetConflicts())
    {
        conflicts.push_back(conflict.GetHex());
    }
    entry.pushKV("walletconflicts", conflicts);
    entry.pushKV("time", wtx.GetTxTime());
    entry.pushKV("timereceived", (int64_t)wtx.nTimeReceived);

    for (const auto &item : wtx.mapValue)
    {
        entry.pushKV(item.first, item.second);
    }
}

string AccountFromValue(const UniValue &value)
{
    string strAccount = value.get_str();
    if (strAccount == "*")
        throw JSONRPCError(RPC_WALLET_INVALID_ACCOUNT_NAME, "Invalid account name");
    return strAccount;
}

// Sort fee from lower to higher
class CompareCoinValue
{
public:
    bool operator()(const COutput &a, const COutput &b) const { return a.GetValue() < b.GetValue(); }
};
UniValue consolidate(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 2 || params.size() > 4)
        throw runtime_error(
            "consolidate (\"num\" \"toleave\")\n"
            "\nConsolidates spendable utxos in the wallet. (-spendzeroconfchange or -wallet.instant\n"
            "must be enabled)\n"
            "\nArguments:\n"
            "1. \"num\"       (number, required) number of coins to select for consolidation)\n"
            "2. \"toleave\"   (number, required) minimum number of coins to leave)\n"
            "3. \"verbosity\" (boolean, optional, default: false) show also the outpoint of every consolidated utxos "
            "(default: false)\n"
            "4. \"address\"   (string, optional, default: freshly generated addr from your wallet) address destination "
            "where to send coins to\n"
            "\nResult:\n"
            "{\n"
            "  \"amount\"       :     (numeric) Amount on NEX consolidated into the destination address\n"
            "  \"destination\"  :     \"xxxxx\" Address used as consolidation destinations\n"
            "  \"txids\"        : [   (array of string) The transaction ids\n"
            "     \"transactionid\"   (string) The transaction id\n"
            "     ,...\n"
            "  ],\n"
            "  \"txidems\"      : [    (array of string) The transaction idems\n"
            "     \"transactionidem\"  (string) The transaction idem\n"
            "     ,...\n"
            "  ],\n"
            "\"outpoints\"  : [ (array od string) the outpoints of all consolidated utxos (showed only if verbosity is "
            "true)\n"
            "  \"outpoint\"     (string) outpoint\n"
            "     ,...\n"
            "  ],\n"
            "\nExamples:\n" +
            HelpExampleCli("consolidate", "500 50") + HelpExampleRpc("consolidate", "500, 50"));

    // Check that spend zero conf change or instant transactions are turned on
    if (!instantTxns.Value() && !GetBoolArg("-spendzeroconfchange", DEFAULT_SPEND_ZEROCONF_CHANGE))
        throw runtime_error("You must have either -spendzeroconfchange and/or -wallet.instant enabled");

    EnsureWalletIsUnlocked();

    // Begin consolidation process
    LOCK(serializeCreateTx);
    LOCK(pwalletMain->cs_wallet);

    if (!pwalletMain->IsLocked())
        pwalletMain->TopUpKeyPool();

    CReserveKey reservekey(pwalletMain);
    CScript scriptPubKey;
    CTxDestination dest;
    if (params.size() >= 4)
    {
        dest = DecodeDestination(params[3].get_str());
        if (!IsValidDestination(dest))
        {
            throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, "Invalid coin address");
        }
        // Parse address
        scriptPubKey = GetScriptForDestination(dest);
    }
    else
    {
        CPubKey vchPubKey;
        if (!reservekey.GetReservedKey(vchPubKey))
        {
            throw JSONRPCError(RPC_WALLET_KEYPOOL_RAN_OUT, "Error: Keypool ran out, please call keypoolrefill first");
        }
        scriptPubKey = P2pktOutput(vchPubKey);
        ExtractDestination(scriptPubKey, dest);
    }

    bool fVerbose = false;
    if (params.size() >= 3)
    {
        if (params[2].isStr())
            fVerbose = InterpretBool(params[2].get_str());
        else if (params[2].isNum())
            fVerbose = (params[2].get_int() != 0);
        else
            fVerbose = params[2].get_bool();
    }

    uint32_t numUtxos = UINT32_MAX;
    uint32_t numUtxosToLeave = 1;
    if (params[0].isStr())
    {
        numUtxos = std::stoi(params[0].get_str());
    }
    else if (params[0].isNum())
    {
        numUtxos = params[0].get_int();
    }
    if (numUtxos <= 0)
        throw JSONRPCError(RPC_WALLET_ERROR, "Error: number of coins to consolidate must be a positive number");
    if (numUtxos > 20000)
        throw JSONRPCError(
            RPC_WALLET_ERROR, "Error: number of coins to consolidate must be less than or equal to 20000");

    if (params[1].isStr())
    {
        numUtxosToLeave = std::stoi(params[1].get_str());
    }
    else if (params[1].isNum())
    {
        numUtxosToLeave = params[1].get_int();
    }
    if (numUtxosToLeave <= 0)
        throw JSONRPCError(RPC_WALLET_ERROR, "Error: number of coins to leave must be a positive number");
    if (numUtxosToLeave > 20000)
        throw JSONRPCError(RPC_WALLET_ERROR, "Error: number of coins to leave must be less than or equal to 20000");

    uint32_t nTotalCoinsChosen = 0;
    uint32_t numToFetch = numUtxos + numUtxosToLeave;
    CAmount nValueExt = 0;
    uint32_t nInputs = 0;
    uint32_t neededTxs = 0;
    UniValue txIds(UniValue::VARR);
    UniValue txIdems(UniValue::VARR);
    UniValue outPts(UniValue::VARR);
    while (nTotalCoinsChosen < numUtxos)
    {
        // Get a list of available coins and select up to the max vin allowed
        CCoinControl coinControl;
        coinControl.fAllowOtherInputs = false;
        coinControl.fAllowWatchOnly = false;
        vector<COutput> sel;
        CAmount nValue = 0;
        uint32_t count = 0;
        pwalletMain->AvailableCoins(sel, &coinControl, false, &numToFetch);
        uint32_t nTotalAvailable = sel.size();
        if (sel.size() <= 1 || sel.size() <= numUtxosToLeave || nTotalCoinsChosen >= numUtxos)
            break;

        bool fTxCreated = false;
        std::string strError;
        CWalletTx wtxNew;
        while (sel.size() >= 2)
        {
            bool fDone = false;
            // select the input to consolidate
            for (const auto &coin : sel)
            {
                coinControl.Select(coin.GetOutPoint());
                if (fVerbose)
                    outPts.push_back(coin.GetOutPoint().ToString());
                nValue += coin.GetValue();
                count++;
                nTotalCoinsChosen++;
                if (count >= MAX_TX_NUM_VIN || nTotalCoinsChosen >= numUtxos ||
                    (nTotalAvailable - count) < numUtxosToLeave)
                {
                    fDone = true;
                    break;
                }
            }

            // prepare to create a tx
            CAmount nFeeRequired = 0;
            vector<CRecipient> vecSend;
            int nChangePosRet = -1;
            bool fSubtractFeeFromAmount = true;
            CRecipient recipient = {scriptPubKey, nValue, fSubtractFeeFromAmount};
            vecSend.push_back(recipient);

            // Create transaction
            if (!pwalletMain->CreateTransaction(
                    vecSend, wtxNew, reservekey, nFeeRequired, nChangePosRet, strError, &coinControl))
            {
                // reset counters
                fDone = false;
                nValue = 0;
                count = 0;
                nTotalCoinsChosen = 0;
                coinControl.UnSelectAll();
                sel.erase(sel.begin());

                fTxCreated = false;
            }
            else
            {
                nValueExt += nValue;
                nInputs += count;
                fTxCreated = true;
            }

            if (fDone && fTxCreated)
                break;
        }
        if (!fTxCreated)
            throw JSONRPCError(
                RPC_WALLET_ERROR, "Error: failed to create transaction during consolidation: " + strError);

        // Commit the transaction
        std::string error;
        if (!pwalletMain->CommitTransaction(wtxNew, reservekey, error))
        {
            throw JSONRPCError(RPC_WALLET_ERROR, error);
        }
        else
        {
            txIds.push_back(wtxNew.GetId().ToString());
            txIdems.push_back(wtxNew.GetIdem().ToString());
            neededTxs++;
        }
    }

    UniValue result(UniValue::VOBJ);
    result.pushKV("total_amount", FormatMoney(nValueExt));
    result.pushKV("destination", EncodeCashAddr(dest, Params()));
    result.pushKV("txids", txIds);
    result.pushKV("txidems", txIdems);
    if (fVerbose)
        result.pushKV("outpoints", outPts);
    return result;
}

UniValue getnewaddress(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 2)
        throw runtime_error("getnewaddress (\"type\" \"account\" )\n"
                            "\nReturns a new address for receiving payments.\n"
                            "If 'account' is specified (DEPRECATED), it is added to the address book \n"
                            "so payments received with the address will be credited to 'account'.\n"
                            "\nArguments:\n"
                            "1. \"type\" (string, optional) 'p2pkt' for pay to public key template or 'p2pkh' (legacy)"
                            "2. \"account\"        (string, optional) DEPRECATED. The account name for the address to "
                            "be linked to. If not provided, the default account \"\" is used. It can also be set to "
                            "the empty string \"\" to represent the default account. The account does not need to "
                            "exist, it will be created if there is no account by the given name.\n"
                            "\nResult:\n"
                            "\"address\"    (string) The new address\n"
                            "\nExamples:\n" +
                            HelpExampleCli("getnewaddress", "") + HelpExampleRpc("getnewaddress", ""));

    LOCK(pwalletMain->cs_wallet);

    // Parse the account first so we don't generate a key if there's an error
    string strAccount;
    if (params.size() > 1)
        strAccount = AccountFromValue(params[1]);

    string strAddrType = "p2pkt";
    if (params.size() > 0)
        strAddrType = AccountFromValue(params[0]);

    if (!pwalletMain->IsLocked())
        pwalletMain->TopUpKeyPool();

    // Generate a new key that is added to wallet
    CPubKey newKey;
    if (!pwalletMain->GetKeyFromPool(newKey))
        throw JSONRPCError(RPC_WALLET_KEYPOOL_RAN_OUT, "Error: Keypool ran out, please call keypoolrefill first");

    if (strAddrType == "p2pkh")
    {
        CKeyID keyID = newKey.GetID();
        pwalletMain->SetAddressBook(keyID, strAccount, "receive");
        return EncodeDestination(keyID);
    }
    else
    {
        ScriptTemplateDestination dest(P2pktOutput(newKey));
        pwalletMain->SetAddressBook(dest, strAccount, "receive");
        return EncodeDestination(dest);
    }
}

CTxDestination GetAccountAddress(string strAccount, bool bForceNew = false)
{
    CWalletDB walletdb(pwalletMain->strWalletFile);

    CAccount account;
    walletdb.ReadAccount(strAccount, account);

    bool bKeyUsed = false;

    // Check if the current key has been used
    if (account.vchPubKey.IsValid())
    {
        CScript scriptPubKey = GetScriptForDestination(account.vchPubKey.GetID());
        for (MapWallet::iterator it = pwalletMain->mapWallet.begin();
             it != pwalletMain->mapWallet.end() && account.vchPubKey.IsValid(); ++it)
        {
            CWalletTxRef wtx = (*it).second.tx;
            // Get every transaction exactly once, skipping outpoint and idem entries
            if (it->first.hash != wtx->GetId())
                continue;

            for (const CTxOut &txout : wtx->vout)
            {
                if (txout.scriptPubKey == scriptPubKey)
                    bKeyUsed = true;
            }
        }
    }

    // Generate a new key
    if (!account.vchPubKey.IsValid() || bForceNew || bKeyUsed)
    {
        if (!pwalletMain->GetKeyFromPool(account.vchPubKey))
            throw JSONRPCError(RPC_WALLET_KEYPOOL_RAN_OUT, "Error: Keypool ran out, please call keypoolrefill first");

        pwalletMain->SetAddressBook(account.vchPubKey.GetID(), strAccount, "receive");
        walletdb.WriteAccount(strAccount, account);
    }

    return account.vchPubKey.GetID();
}

UniValue getaccountaddress(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() != 1)
        throw runtime_error(
            "getaccountaddress \"account\"\n"
            "\nDEPRECATED. Returns the current address for receiving payments to this account.\n"
            "\nArguments:\n"
            "1. \"account\"       (string, required) The account name for the address. It can also be set to the empty "
            "string \"\" to represent the default account. The account does not need to exist, it will be created and "
            "a new address created  if there is no account by the given name.\n"
            "\nResult:\n"
            "\"address\"   (string) The account address\n"
            "\nExamples:\n" +
            HelpExampleCli("getaccountaddress", "") + HelpExampleCli("getaccountaddress", "\"\"") +
            HelpExampleCli("getaccountaddress", "\"myaccount\"") +
            HelpExampleRpc("getaccountaddress", "\"myaccount\""));

    LOCK(pwalletMain->cs_wallet);

    // Parse the account first so we don't generate a key if there's an error
    string strAccount = AccountFromValue(params[0]);

    UniValue ret(UniValue::VSTR);

    ret = EncodeDestination(GetAccountAddress(strAccount));
    return ret;
}


UniValue getrawchangeaddress(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 1)
        throw runtime_error("getrawchangeaddress\n"
                            "\nReturns a new address, for receiving change.\n"
                            "This is for use with raw transactions, NOT normal use.\n"
                            "\nResult:\n"
                            "\"address\"    (string) The address\n"
                            "\nExamples:\n" +
                            HelpExampleCli("getrawchangeaddress", "") + HelpExampleRpc("getrawchangeaddress", ""));

    LOCK(pwalletMain->cs_wallet);

    if (!pwalletMain->IsLocked())
        pwalletMain->TopUpKeyPool();

    CReserveKey reservekey(pwalletMain);
    CPubKey vchPubKey;
    if (!reservekey.GetReservedKey(vchPubKey))
        throw JSONRPCError(RPC_WALLET_KEYPOOL_RAN_OUT, "Error: Keypool ran out, please call keypoolrefill first");

    reservekey.KeepKey();

    CKeyID keyID = vchPubKey.GetID();

    return EncodeDestination(keyID);
}

UniValue setaccount(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 1 || params.size() > 2)
        throw runtime_error(
            "setaccount \"address\" \"account\"\n"
            "\nDEPRECATED. Sets the account associated with the given address.\n"
            "\nArguments:\n"
            "1. \"address\"  (string, required) The address to be associated with an account.\n"
            "2. \"account\"         (string, required) The account to assign the address to.\n"
            "\nExamples:\n" +
            HelpExampleCli("setaccount", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" \"tabby\"") +
            HelpExampleRpc("setaccount", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\", \"tabby\""));

    LOCK(pwalletMain->cs_wallet);

    CTxDestination dest = DecodeDestination(params[0].get_str());
    if (!IsValidDestination(dest))
    {
        throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, "Invalid coin address");
    }

    string strAccount;
    if (params.size() > 1)
        strAccount = AccountFromValue(params[1]);

    // Only add the account if the address is yours.
    if (IsMine(*pwalletMain, dest, chainActive.Tip()))
    {
        // Detect when changing the account of an address that is the 'unused
        // current key' of another account:
        if (pwalletMain->mapAddressBook.count(dest))
        {
            std::string strOldAccount = pwalletMain->mapAddressBook[dest].name;
            if (dest == GetAccountAddress(strOldAccount))
                GetAccountAddress(strOldAccount, true);
        }
        pwalletMain->SetAddressBook(dest, strAccount, "receive");
    }
    else
        throw JSONRPCError(RPC_MISC_ERROR, "setaccount can only be used with own address");

    return NullUniValue;
}

UniValue getaccount(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() != 1)
        throw runtime_error("getaccount \"address\"\n"
                            "\nDEPRECATED. Returns the account associated with the given address.\n"
                            "\nArguments:\n"
                            "1. \"address\"  (string, required) The address for account lookup.\n"
                            "\nResult:\n"
                            "\"accountname\"        (string) the account address\n"
                            "\nExamples:\n" +
                            HelpExampleCli("getaccount", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\"") +
                            HelpExampleRpc("getaccount", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\""));

    LOCK(pwalletMain->cs_wallet);

    CTxDestination dest = DecodeDestination(params[0].get_str());
    if (!IsValidDestination(dest))
    {
        throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, "Invalid coin address");
    }

    std::string strAccount;
    std::map<CTxDestination, CAddressBookData>::iterator mi = pwalletMain->mapAddressBook.find(dest);
    if (mi != pwalletMain->mapAddressBook.end() && !(*mi).second.name.empty())
    {
        strAccount = (*mi).second.name;
    }
    return strAccount;
}

UniValue getaddressesbyaccount(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() != 1)
        throw runtime_error("getaddressesbyaccount \"account\"\n"
                            "\nDEPRECATED. Returns the list of addresses for the given account.\n"
                            "\nArguments:\n"
                            "1. \"account\"  (string, required) The account name.\n"
                            "\nResult:\n"
                            "[                     (json array of string)\n"
                            "  \"address\"  (string) an address associated with the given account\n"
                            "  ,...\n"
                            "]\n"
                            "\nExamples:\n" +
                            HelpExampleCli("getaddressesbyaccount", "\"tabby\"") +
                            HelpExampleRpc("getaddressesbyaccount", "\"tabby\""));

    LOCK(pwalletMain->cs_wallet);

    string strAccount = AccountFromValue(params[0]);

    // Find all addresses that have the given account
    UniValue ret(UniValue::VARR);
    for (const std::pair<const CTxDestination, CAddressBookData> &item : pwalletMain->mapAddressBook)
    {
        const CTxDestination &dest = item.first;
        const std::string &strName = item.second.name;
        if (strName == strAccount)
        {
            ret.push_back(EncodeDestination(dest));
        }
    }
    return ret;
}

static void SendMoney(const CTxDestination &address, CAmount nValue, bool fSubtractFeeFromAmount, CWalletTx &wtxNew)
{
    // Check amount
    if (nValue <= 0)
        throw JSONRPCError(RPC_INVALID_PARAMETER, "Invalid amount");

    // Parse address
    CScript scriptPubKey = GetScriptForDestination(address);

    // Create and send the transaction
    {
        LOCK(serializeCreateTx);

        CReserveKey reservekey(pwalletMain);
        CAmount nFeeRequired = 0;
        std::string strError;
        vector<CRecipient> vecSend;
        int nChangePosRet = -1;
        CRecipient recipient = {scriptPubKey, nValue, fSubtractFeeFromAmount};
        vecSend.push_back(recipient);
        if (!pwalletMain->CreateTransaction(vecSend, wtxNew, reservekey, nFeeRequired, nChangePosRet, strError))
        {
            if (!fSubtractFeeFromAmount && nValue + nFeeRequired > pwalletMain->GetBalance())
                strError = strprintf("Error: This transaction requires a transaction fee of at least %s because of its "
                                     "amount, complexity, or use of recently received funds!",
                    FormatMoney(nFeeRequired));
            throw JSONRPCError(RPC_WALLET_ERROR, strError);
        }
        std::string error;
        if (!pwalletMain->CommitTransaction(wtxNew, reservekey, error))
            throw JSONRPCError(RPC_WALLET_ERROR, error);
    }
}

UniValue sendtoaddress(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 2 || params.size() > 5)
        throw runtime_error(
            "sendtoaddress \"address\" amount ( \"comment\" \"comment-to\" subtractfeefromamount )\n"
            "\nSend an amount to a given address.\n" +
            HelpRequiringPassphrase() +
            "\nArguments:\n"
            "1. \"address\"  (string, required) The address to send to.\n"
            "2. \"amount\"      (numeric or string, required) The amount in " +
            CURRENCY_UNIT +
            " to send. eg 10.81\n"
            "3. \"comment\"     (string, optional) A comment used to store what the transaction is for. \n"
            "                             This is not part of the transaction, just kept in your wallet.\n"
            "4. \"comment-to\"  (string, optional) A comment to store the name of the person or organization \n"
            "                             to which you're sending the transaction. This is not part of the \n"
            "                             transaction, just kept in your wallet.\n"
            "5. subtractfeefromamount  (boolean, optional, default=false) The fee will be deducted from the amount "
            "being sent.\n"
            "                             The recipient will receive less coins than you enter in the amount "
            "field.\n"
            "\nResult:\n"
            "\"txidem\"  (string) The txidem.\n"
            "\nExamples:\n" +
            HelpExampleCli("sendtoaddress", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" 10.81") +
            HelpExampleCli("sendtoaddress",
                "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" 10.81 \"donation\" \"seans outpost\"") +
            HelpExampleCli(
                "sendtoaddress", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" 10.81 \"\" \"\" true") +
            HelpExampleRpc("sendtoaddress",
                "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\", 10.81, \"donation\", \"seans outpost\""));

    CTxDestination dest = DecodeDestination(params[0].get_str());
    if (!IsValidDestination(dest))
    {
        throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, "Invalid coin address");
    }

    // Amount
    CAmount nAmount = AmountFromValue(params[1]);
    if (nAmount <= 0)
        throw JSONRPCError(RPC_TYPE_ERROR, "Invalid amount for send");

    // Wallet comments
    CWalletTx wtx;
    if (params.size() > 2 && !params[2].isNull() && !params[2].get_str().empty())
        wtx.mapValue["comment"] = params[2].get_str();
    if (params.size() > 3 && !params[3].isNull() && !params[3].get_str().empty())
        wtx.mapValue["to"] = params[3].get_str();

    bool fSubtractFeeFromAmount = false;
    if (params.size() > 4)
        fSubtractFeeFromAmount = params[4].get_bool();

    EnsureWalletIsUnlocked();

    SendMoney(dest, nAmount, fSubtractFeeFromAmount, wtx);

    return wtx.GetIdem().GetHex();
}

UniValue listaddressgroupings(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp)
        throw runtime_error("listaddressgroupings\n"
                            "\nLists groups of addresses which have had their common ownership\n"
                            "made public by common use as inputs or as the resulting change\n"
                            "in past transactions\n"
                            "\nResult:\n"
                            "[\n"
                            "  [\n"
                            "    [\n"
                            "      \"address\",     (string) The address\n"
                            "      amount,                 (numeric) The amount in " +
                            CURRENCY_UNIT +
                            "\n"
                            "      \"account\"             (string, optional) The account (DEPRECATED)\n"
                            "    ]\n"
                            "    ,...\n"
                            "  ]\n"
                            "  ,...\n"
                            "]\n"
                            "\nExamples:\n" +
                            HelpExampleCli("listaddressgroupings", "") + HelpExampleRpc("listaddressgroupings", ""));

    LOCK(pwalletMain->cs_wallet);

    UniValue jsonGroupings(UniValue::VARR);
    std::map<CTxDestination, CAmount> balances = pwalletMain->GetAddressBalances();
    for (const std::set<CTxDestination> &grouping : pwalletMain->GetAddressGroupings())
    {
        UniValue jsonGrouping(UniValue::VARR);
        for (const CTxDestination &address : grouping)
        {
            UniValue addressInfo(UniValue::VARR);
            addressInfo.push_back(EncodeDestination(address));
            addressInfo.push_back(ValueFromAmount(balances[address]));

            if (pwalletMain->mapAddressBook.find(address) != pwalletMain->mapAddressBook.end())
            {
                addressInfo.push_back(pwalletMain->mapAddressBook.find(address)->second.name);
            }
            jsonGrouping.push_back(addressInfo);
        }
        jsonGroupings.push_back(jsonGrouping);
    }
    return jsonGroupings;
}

UniValue listactiveaddresses(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp)
        throw runtime_error("listactiveaddresses\n"
                            "\nLists addresses \n"
                            "made public by common use as inputs or as the resulting change\n"
                            "\nResult:\n"
                            "[\n"
                            "]\n"
                            "\nExamples:\n" +
                            HelpExampleCli("listaddressgroupings", "") + HelpExampleRpc("listaddressgroupings", ""));

    LOCK(pwalletMain->cs_wallet);

    UniValue ret(UniValue::VOBJ);
    std::map<CTxDestination, CAmount> balances = pwalletMain->GetAddressBalances();
    for (const auto &b : balances)
    {
        ret.pushKV(EncodeDestination(b.first), ValueFromAmount(b.second));
    }
    return ret;
}

bool SignMessage(const UniValue &params, UniValue &sig, std::string &error)
{
    LOCK(pwalletMain->cs_wallet);

    try
    {
        EnsureWalletIsUnlocked();
    }
    catch (...)
    {
        error = "Wallet could not be unlocked";
        return false;
    }

    std::string strAddress = params[0].get_str();
    std::string strMessage = params[1].get_str();
    CTxDestination dest = DecodeDestination(strAddress);
    if (!IsValidDestination(dest))
    {
        error = "Invalid coin address";
        return false;
    }

    CKey key;
    if (!pwalletMain->GetKey(dest, key))
    {
        error = "Private key not available, or address has no single key";
        return false;
    }

    vector<unsigned char> vchSig = signmessage(strMessage, key);
    if (vchSig.empty())
    {
        error = "Sign failed";
        return false;
    }

    sig = EncodeBase64(&vchSig[0], vchSig.size());
    return true;
}
UniValue signmessage(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() != 2)
        throw runtime_error(
            "signmessage \"address\" \"message\"\n"
            "\nSign a message with the private key of an address.  This is NOT compatible with CHECKDATASIG"
            "\n (use signdata instead)." +
            HelpRequiringPassphrase() +
            "\n"
            "\nArguments:\n"
            "1. \"address\"  (string, required) The coin address to use for the private key.\n"
            "2. \"message\"         (string, required) The message to create a signature of.\n"
            "\nResult:\n"
            "\"signature\"          (string) The signature of the message encoded in base 64\n"
            "\nExamples:\n"
            "\nUnlock the wallet for 30 seconds\n" +
            HelpExampleCli("walletpassphrase", "\"mypassphrase\" 30") + "\nCreate the signature\n" +
            HelpExampleCli("signmessage", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" \"my message\"") +
            "\nVerify the signature\n" +
            HelpExampleCli(
                "verifymessage", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" \"signature\" \"my message\"") +
            "\nAs json rpc\n" +
            HelpExampleRpc("signmessage", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\", \"my message\""));


    UniValue signature;
    std::string error = "";
    if (!SignMessage(params, signature, error))
    {
        throw JSONRPCError(RPC_TYPE_ERROR, error);
    }
    else
        return signature;
}

UniValue signdata(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 3 || params.size() > 4)
        throw runtime_error(
            "signdata \"address\" \"msgFormat\" \"message\"\n"
            "\nSign message for use with the CHECKDATASIG instruction."
            "\nAs per the CHECKDATASIG operation, this RPC normally signs the SHA256 of"
            "\nthe provided message unless the 'hash' message format is specified."
            "\nIf using the 'hash' message format, provide the hex encoded SHA256 hash"
            "\nof the message intended to be passed to CHECKDATASIG.\n" +
            HelpRequiringPassphrase() +
            "\n"
            "\nArguments:\n"
            "1. \"address\"  (string, required) The address to use for the private key.\n"
            "2. \"msgFormat\"       (string, required) Use \"string\", \"hex\", or \"hash\" to specify the message "
            "encoding.\n"
            "3. \"message\"         (string, required) The message to create a signature of.\n"
            "4. \"verbose\"         (string, optional) pass 'verbose' to return additional info.\n"
            "\nResult:\n"
            "\"signature\"          (string) The signature of the message encoded in hex\n"
            "\nif 'verbose', return a dictionary containing the signature, pubkey and pubkey hash in hex format.\n"
            "\nExamples:\n"
            "\nUnlock the wallet for 30 seconds\n" +
            HelpExampleCli("walletpassphrase", "\"mypassphrase\" 30") + "\nCreate the signature\n" +
            HelpExampleCli(
                "signdata", "\"nexa:qq5lslagrktm5qtxfw4ltpd5krehhrh595fc04hv0k\" \"string\" \"my message\"") +
            HelpExampleCli("signdata", "\"nexa:qq5lslagrktm5qtxfw4ltpd5krehhrh595fc04hv0k\" \"hex\" \"01020304\"") +
            "\nAs json rpc\n" +
            HelpExampleRpc(
                "signdata", "\"nexa:qq5lslagrktm5qtxfw4ltpd5krehhrh595fc04hv0k\", \"string\", \"my message\""));

    LOCK(pwalletMain->cs_wallet);

    EnsureWalletIsUnlocked();

    string strAddress = params[0].get_str();
    string datatype = params[1].get_str();
    string strMessage = params[2].get_str();
    bool verbose = false;
    if (params.size() > 3)
        verbose = (params[3].get_str() == "verbose");

    CTxDestination dest = DecodeDestination(strAddress);
    if (!IsValidDestination(dest))
        throw JSONRPCError(RPC_TYPE_ERROR, "Invalid coin address");

    CKey key;
    if (!pwalletMain->GetKey(dest, key))
        throw JSONRPCError(RPC_WALLET_ERROR, "Private key not available, or address has no single key");

    uint256 hash;
    if (datatype == "string")
    {
        CSHA256().Write((const unsigned char *)strMessage.c_str(), strMessage.size()).Finalize(hash.begin());
    }
    else if (datatype == "hex")
    {
        if (!IsHex(strMessage))
            throw JSONRPCError(RPC_TYPE_ERROR, "Message is not hex data");
        auto data = ParseHex(strMessage.c_str());
        CSHA256().Write(data.data(), data.size()).Finalize(hash.begin());
    }
    else if (datatype == "hash")
    {
        if (!IsHex(strMessage))
            throw JSONRPCError(RPC_TYPE_ERROR, "Message is not hex data");
        if (strMessage.size() != 2 * sizeof(uint256))
            throw JSONRPCError(RPC_TYPE_ERROR, "Message is not a hex hash");
        hash.SetHex(strMessage);
        // nexad reads hashes backwards.  By reversing here, we ensure that
        // signdata(addr, "string", "foo") == signdata(addr, "hash", normalSHA256("foo"))
        hash.reverse();
    }
    else
    {
        throw JSONRPCError(RPC_TYPE_ERROR, "Invalid message format");
    }
    std::vector<uint8_t> sig;
    key.SignSchnorr(hash, sig);
    if (sig.empty())
        throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, "Sign failed");
    if (verbose)
    {
        UniValue ret(UniValue::VOBJ);
        ret.pushKV("msghash", hash.ToString());
        ret.pushKV("signature", GetHex(sig.data(), sig.size()));
        CPubKey pub = key.GetPubKey();
        ret.pushKV("pubkeyhash", pub.GetID().GetHex());
        ret.pushKV("pubkey", GetHex(pub.begin(), pub.size()));
        return ret;
    }
    return UniValue(GetHex(sig.data(), sig.size()));
}


UniValue getreceivedbyaddress(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 1 || params.size() > 2)
        throw runtime_error(
            "getreceivedbyaddress \"address\" ( minconf )\n"
            "\nReturns the total amount received by the given address in transactions with at "
            "least minconf confirmations.\n"
            "\nArguments:\n"
            "1. \"address\"           (string, required) The address for transactions.\n"
            "2. minconf             (numeric, optional, default=1) Only include transactions confirmed "
            "at least this many times.\n"
            "\nResult:\n"
            "amount   (numeric) The total amount in " +
            CURRENCY_UNIT +
            " received at this address.\n"
            "\nExamples:\n"
            "\nThe amount from transactions with at least 1 confirmation\n" +
            HelpExampleCli("getreceivedbyaddress", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\"") +
            "\nThe amount including unconfirmed transactions, zero confirmations\n" +
            HelpExampleCli("getreceivedbyaddress", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" 0") +
            "\nThe amount with at least 6 confirmation, very safe\n" +
            HelpExampleCli("getreceivedbyaddress", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" 6") +
            "\nAs a json rpc call\n" +
            HelpExampleRpc("getreceivedbyaddress", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\", 6"));

    LOCK(pwalletMain->cs_wallet);

    // address
    CTxDestination dest = DecodeDestination(params[0].get_str());
    if (!IsValidDestination(dest))
    {
        throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, "Invalid coin address");
    }
    CScript scriptPubKey = GetScriptForDestination(dest);
    if (!IsMine(*pwalletMain, scriptPubKey, chainActive.Tip()))
        return ValueFromAmount(0);

    // Minimum confirmations
    int nMinDepth = 1;
    if (params.size() > 1)
        nMinDepth = params[1].get_int();

    // Tally
    CAmount nAmount = 0;
    for (MapWallet::iterator it = pwalletMain->mapWallet.begin(); it != pwalletMain->mapWallet.end(); ++it)
    {
        CWalletTxRef wtx = it->second.tx;
        // Get every transaction exactly once, skipping outpoint and idem entries
        if (it->first.hash != wtx->GetId())
            continue;

        const CTransaction *tx = wtx.get();
        if (wtx->IsCoinBase() || !CheckFinalTx(tx))
            continue;

        auto txDepth = wtx->GetDepthInMainChain();
        for (const CTxOut &txout : wtx->vout)
            if (txout.scriptPubKey == scriptPubKey)
                if (txDepth >= nMinDepth)
                    nAmount += txout.nValue;
    }

    return ValueFromAmount(nAmount);
}


UniValue getreceivedbyaccount(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 1 || params.size() > 2)
        throw runtime_error(
            "getreceivedbyaccount \"account\" ( minconf )\n"
            "\nDEPRECATED. Returns the total amount received by addresses with <account> in transactions with at least "
            "[minconf] confirmations.\n"
            "\nArguments:\n"
            "1. \"account\"        (string, required) The selected account, may be the default account using \"\".\n"
            "2. minconf          (numeric, optional, default=1) Only include transactions confirmed at least this many "
            "times.\n"
            "\nResult:\n"
            "amount              (numeric) The total amount in " +
            CURRENCY_UNIT +
            " received for this account.\n"
            "\nExamples:\n"
            "\nAmount received by the default account with at least 1 confirmation\n" +
            HelpExampleCli("getreceivedbyaccount", "\"\"") +
            "\nAmount received at the tabby account including unconfirmed amounts with zero confirmations\n" +
            HelpExampleCli("getreceivedbyaccount", "\"tabby\" 0") +
            "\nThe amount with at least 6 confirmation, very safe\n" +
            HelpExampleCli("getreceivedbyaccount", "\"tabby\" 6") + "\nAs a json rpc call\n" +
            HelpExampleRpc("getreceivedbyaccount", "\"tabby\", 6"));

    LOCK(pwalletMain->cs_wallet);

    // Minimum confirmations
    int nMinDepth = 1;
    if (params.size() > 1)
        nMinDepth = params[1].get_int();

    // Get the set of pub keys assigned to account
    string strAccount = AccountFromValue(params[0]);
    set<CTxDestination> setAddress = pwalletMain->GetAccountAddresses(strAccount);

    // Tally
    CAmount nAmount = 0;
    for (MapWallet::iterator it = pwalletMain->mapWallet.begin(); it != pwalletMain->mapWallet.end(); ++it)
    {
        const CWalletTxRef wtx = it->second.tx;
        // Get every transaction exactly once, skipping outpoint and idem entries
        if (it->first.hash != wtx->GetId())
            continue;

        if (wtx->IsCoinBase() || !CheckFinalTx(wtx.get()))
            continue;

        int depth = wtx->GetDepthInMainChain();
        for (const CTxOut &txout : wtx->vout)
        {
            CTxDestination address;
            if (ExtractDestination(txout.scriptPubKey, address) && IsMine(*pwalletMain, address, chainActive.Tip()) &&
                setAddress.count(address))
                if (depth >= nMinDepth)
                    nAmount += txout.nValue;
        }
    }

    return ValueFromAmount(nAmount);
}


CAmount GetAccountBalance(CWalletDB &walletdb, const string &strAccount, int nMinDepth, const isminefilter &filter)
{
    CAmount nBalance = 0;

    // Tally wallet transactions
    for (MapWallet::iterator it = pwalletMain->mapWallet.begin(); it != pwalletMain->mapWallet.end(); ++it)
    {
        CWalletTxRef wtx = it->second.tx;
        // Get every transaction exactly once, skipping outpoint and idem entries
        if (it->first.hash != wtx->GetId())
            continue;
        int depth = wtx->GetDepthInMainChain();
        if (!CheckFinalTx(wtx.get()) || wtx->GetBlocksToMaturity() > 0 || depth < 0)
            continue;

        CAmount nReceived, nSent, nFee;
        wtx->GetAccountAmounts(strAccount, nReceived, nSent, nFee, filter);

        if (nReceived != 0 && depth >= nMinDepth)
            nBalance += nReceived;
        nBalance -= nSent + nFee;
    }

    // Tally internal accounting entries
    nBalance += walletdb.GetAccountCreditDebit(strAccount);

    return nBalance;
}

CAmount GetAccountBalance(const string &strAccount, int nMinDepth, const isminefilter &filter)
{
    CWalletDB walletdb(pwalletMain->strWalletFile);
    return GetAccountBalance(walletdb, strAccount, nMinDepth, filter);
}


UniValue getbalance(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 3)
        throw runtime_error("getbalance ( \"account\" minconf includeWatchonly )\n"
                            "\nIf account is not specified, returns the server's total available balance.\n"
                            "If account is specified (DEPRECATED), returns the balance in the account.\n"
                            "Note that the account \"\" is not the same as leaving the parameter out.\n"
                            "The server total may be different to the balance in the default \"\" account.\n"
                            "\nArguments:\n"
                            "1. \"account\"      (string, optional) DEPRECATED. The selected account, or \"*\" for "
                            "entire wallet. It may be the default account using \"\".\n"
                            "2. minconf          (numeric, optional, default=1) Only include transactions confirmed at "
                            "least this many times.\n"
                            "3. includeWatchonly (bool, optional, default=false) Also include balance in watchonly "
                            "addresses (see 'importaddress')\n"
                            "\nResult:\n"
                            "amount              (numeric) The total amount in " +
                            CURRENCY_UNIT +
                            " received for this account.\n"
                            "\nExamples:\n"
                            "\nThe total amount in the wallet\n" +
                            HelpExampleCli("getbalance", "") +
                            "\nThe total amount in the wallet at least 5 blocks confirmed\n" +
                            HelpExampleCli("getbalance", "\"*\" 6") + "\nAs a json rpc call\n" +
                            HelpExampleRpc("getbalance", "\"*\", 6"));

    if (params.size() == 0)
        return ValueFromAmount(pwalletMain->GetBalance());

    int nMinDepth = 1;
    if (params.size() > 1)
        nMinDepth = params[1].get_int();
    isminefilter filter = ISMINE_SPENDABLE;
    if (params.size() > 2)
        if (params[2].get_bool())
            filter = filter | ISMINE_WATCH_ONLY;

    if (params[0].get_str() == "*")
    {
        // Calculate total balance a different way from GetBalance()
        // (GetBalance() sums up all unspent TxOuts)
        // getbalance and "getbalance * 1 true" should return the same number
        CAmount nBalance = 0;
        for (MapWallet::iterator it = pwalletMain->mapWallet.begin(); it != pwalletMain->mapWallet.end(); ++it)
        {
            const CWalletTxRef wtx = it->second.tx;
            // Get every transaction exactly once, skipping outpoint and idem entries
            if (it->first.hash != wtx->GetId())
                continue;

            if (!CheckFinalTx(wtx.get()) || wtx->GetBlocksToMaturity() > 0 || wtx->GetDepthInMainChain() < 0)
                continue;

            CAmount allFee;
            string strSentAccount;
            list<COutputEntry> listReceived;
            list<COutputEntry> listSent;
            wtx->GetAmounts(listReceived, listSent, allFee, strSentAccount, filter);
            CAmount net = 0;
            if (wtx->GetDepthInMainChain() >= nMinDepth)
            {
                for (const COutputEntry &r : listReceived)
                {
                    LOGA("received %lu", r.amount);
                    nBalance += r.amount;
                    net += r.amount;
                }
            }
            for (const COutputEntry &s : listSent)
            {
                LOGA("sent %lu", s.amount);
                nBalance -= s.amount;
                net -= s.amount;
            }
            nBalance -= allFee;
        }
        return ValueFromAmount(nBalance);
    }

    string strAccount = AccountFromValue(params[0]);

    CAmount nBalance = GetAccountBalance(strAccount, nMinDepth, filter);

    return ValueFromAmount(nBalance);
}

UniValue getunconfirmedbalance(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 0)
        throw runtime_error("getunconfirmedbalance\n"
                            "Returns the server's total unconfirmed balance\n");

    // Nothing relies on cs_main, but by locking it here, we ensure that a chain reorg doesn't
    // cause AvailableCoins to give inconsistent results
    LOCK2(cs_main, pwalletMain->cs_wallet);

    return ValueFromAmount(pwalletMain->GetUnconfirmedBalance());
}


UniValue movecmd(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 3 || params.size() > 5)
        throw runtime_error(
            "move \"fromaccount\" \"toaccount\" amount ( minconf \"comment\" )\n"
            "\nDEPRECATED. Move a specified amount from one account in your wallet to another.\n"
            "\nArguments:\n"
            "1. \"fromaccount\"   (string, required) The name of the account to move funds from. May be the default "
            "account using \"\".\n"
            "2. \"toaccount\"     (string, required) The name of the account to move funds to. May be the default "
            "account using \"\".\n"
            "3. amount            (numeric) Quantity of " +
            CURRENCY_UNIT +
            " to move between accounts.\n"
            "4. minconf           (numeric, optional, default=1) Only use funds with at least this "
            "many confirmations.\n"
            "5. \"comment\"       (string, optional) An optional comment, stored in the wallet only.\n"
            "\nResult:\n"
            "true|false           (boolean) true if successful.\n"
            "\nExamples:\n"
            "\nMove 10.01 " +
            CURRENCY_UNIT + " from the default account to the account named tabby\n" +
            HelpExampleCli("move", "\"\" \"tabby\" 10.01") + "\nMove 0.01 " + CURRENCY_UNIT +
            " timotei to akiko with a comment and funds have 6 confirmations\n" +
            HelpExampleCli("move", "\"timotei\" \"akiko\" 10.01 6 \"happy birthday!\"") + "\nAs a json rpc call\n" +
            HelpExampleRpc("move", "\"timotei\", \"akiko\", 10.01, 6, \"happy birthday!\""));

    LOCK(pwalletMain->cs_wallet);

    string strFrom = AccountFromValue(params[0]);
    string strTo = AccountFromValue(params[1]);
    CAmount nAmount = AmountFromValue(params[2]);
    if (nAmount <= 0)
        throw JSONRPCError(RPC_TYPE_ERROR, "Invalid amount for send");
    if (params.size() > 3)
        // unused parameter, used to be nMinDepth, keep type-checking it though
        (void)params[3].get_int();
    string strComment;
    if (params.size() > 4)
        strComment = params[4].get_str();

    CWalletDB walletdb(pwalletMain->strWalletFile);
    if (!walletdb.TxnBegin())
        throw JSONRPCError(RPC_DATABASE_ERROR, "database error");

    int64_t nNow = GetAdjustedTime();

    // Debit
    CAccountingEntry debit;
    debit.nOrderPos = pwalletMain->IncOrderPosNext(&walletdb);
    debit.strAccount = strFrom;
    debit.nCreditDebit = -nAmount;
    debit.nTime = nNow;
    debit.strOtherAccount = strTo;
    debit.strComment = strComment;
    pwalletMain->AddAccountingEntry(debit, walletdb);

    // Credit
    CAccountingEntry credit;
    credit.nOrderPos = pwalletMain->IncOrderPosNext(&walletdb);
    credit.strAccount = strTo;
    credit.nCreditDebit = nAmount;
    credit.nTime = nNow;
    credit.strOtherAccount = strFrom;
    credit.strComment = strComment;
    pwalletMain->AddAccountingEntry(credit, walletdb);

    if (!walletdb.TxnCommit())
        throw JSONRPCError(RPC_DATABASE_ERROR, "database error");

    return true;
}


UniValue sendfrom(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 3 || params.size() > 6)
        throw runtime_error(
            "sendfrom \"fromaccount\" \"toaddress\" amount ( minconf \"comment\" \"comment-to\" )\n"
            "\nDEPRECATED (use sendtoaddress). Sent an amount from an account to an address." +
            HelpRequiringPassphrase() +
            "\n"
            "\nArguments:\n"
            "1. \"fromaccount\"       (string, required) The name of the account to send funds from. May be the "
            "default account using \"\".\n"
            "2. \"toaddress\"  (string, required) The address to send funds to.\n"
            "3. amount                (numeric or string, required) The amount in " +
            CURRENCY_UNIT +
            " (transaction fee is added on top).\n"
            "4. minconf               (numeric, optional, default=1) Only use funds with at least this many "
            "confirmations.\n"
            "5. \"comment\"           (string, optional) A comment used to store what the transaction is for. \n"
            "                                     This is not part of the transaction, just kept in your wallet.\n"
            "6. \"comment-to\"        (string, optional) An optional comment to store the name of the person or "
            "organization \n"
            "                                     to which you're sending the transaction. This is not part of the "
            "transaction, \n"
            "                                     it is just kept in your wallet.\n"
            "\nResult:\n"
            "\"transactionid\"        (string) The transaction id.\n"
            "\nExamples:\n"
            "\nSend 0.01 " +
            CURRENCY_UNIT + " from the default account to the address, must have at least 1 confirmation\n" +
            HelpExampleCli("sendfrom", "\"\" \"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" 10.01") +
            "\nSend 10.01 from the tabby account to the given address, funds must have at least 6 confirmations\n" +
            HelpExampleCli("sendfrom", "\"tabby\" \"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" 10.01 6 "
                                       "\"donation\" \"seans outpost\"") +
            "\nAs a json rpc call\n" +
            HelpExampleRpc("sendfrom", "\"tabby\", \"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\", 10.01, 6, "
                                       "\"donation\", \"seans outpost\""));

    std::string strAccount = AccountFromValue(params[0]);
    CTxDestination dest = DecodeDestination(params[1].get_str());
    if (!IsValidDestination(dest))
    {
        throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, "Invalid coin address");
    }
    CAmount nAmount = AmountFromValue(params[2]);
    if (nAmount <= 0)
        throw JSONRPCError(RPC_TYPE_ERROR, "Invalid amount for send");
    int nMinDepth = 1;
    if (params.size() > 3)
        nMinDepth = params[3].get_int();

    CWalletTx wtx;
    wtx.strFromAccount = strAccount;
    if (params.size() > 4 && !params[4].isNull() && !params[4].get_str().empty())
        wtx.mapValue["comment"] = params[4].get_str();
    if (params.size() > 5 && !params[5].isNull() && !params[5].get_str().empty())
        wtx.mapValue["to"] = params[5].get_str();

    EnsureWalletIsUnlocked();

    // Check funds, if an account is selected
    if (strAccount != "")
    {
        CAmount nBalance = GetAccountBalance(strAccount, nMinDepth, ISMINE_SPENDABLE);
        if (nAmount > nBalance)
            throw JSONRPCError(RPC_WALLET_INSUFFICIENT_FUNDS, "Account has insufficient funds");
    }

    SendMoney(dest, nAmount, false, wtx);

    return wtx.GetIdem().GetHex();
}


UniValue sendmany(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 2 || params.size() > 5)
        throw runtime_error(
            "sendmany \"fromaccount\" {\"address\":amount,...} ( minconf \"comment\" [\"address\",...] )\n"
            "\nSend multiple times. Amounts are double-precision floating point numbers." +
            HelpRequiringPassphrase() +
            "\n"
            "\nArguments:\n"
            "1. \"fromaccount\"         (string, required) DEPRECATED. The account to send the funds from. Should be "
            "\"\" for the default account\n"
            "2. \"amounts\"             (string, required) A json object with addresses and amounts\n"
            "    {\n"
            "      \"address\":amount   (numeric or string) The address is the key, the numeric amount (can be "
            "string) in " +
            CURRENCY_UNIT +
            " is the value\n"
            "      ,...\n"
            "    }\n"
            "3. minconf                 (numeric, optional, default=1) Only use the balance confirmed at least this "
            "many times.\n"
            "4. \"comment\"             (string, optional) A comment\n"
            "5. subtractfeefromamount   (string, optional) A json array with addresses.\n"
            "                           The fee will be equally deducted from the amount of each selected address.\n"
            "                           Those recipients will receive less coins than you enter in their "
            "corresponding amount field.\n"
            "                           If no addresses are specified here, the sender pays the fee.\n"
            "    [\n"
            "      \"address\"            (string) Subtract fee from this address\n"
            "      ,...\n"
            "    ]\n"
            "\nResult:\n"
            "\"transactionid\"          (string) The transaction id for the send. Only 1 transaction is created "
            "regardless of \n"
            "                                    the number of addresses.\n"
            "\nExamples:\n"
            "\nSend two amounts to two different addresses:\n" +
            HelpExampleCli("sendmany", "\"\" "
                                       "\"{\\\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\\\":10.01,"
                                       "\\\"nexa:qqqdf524crwrn659hattd40jn5ykz32xey7z6n5z28\\\":10.02}\"") +
            "\nSend two amounts to two different addresses setting the confirmation and comment:\n" +
            HelpExampleCli("sendmany",
                "\"\" "
                "\"{\\\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\\\":10.01,"
                "\\\"nexa:qqqdf524crwrn659hattd40jn5ykz32xey7z6n5z28\\\":10.02}\" 6 \"testing\"") +
            "\nSend two amounts to two different addresses, subtract fee from amount:\n" +
            HelpExampleCli("sendmany", "\"\" "
                                       "\"{\\\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\\\":10.01,\\\"nexa:"
                                       "qqqdf524crwrn659hattd40jn5ykz32xey7z6n5z28\\\":10.02}\" "
                                       "1 \"\" "
                                       "\"[\\\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\\\",\\\"nexa:"
                                       "qqqdf524crwrn659hattd40jn5ykz32xey7z6n5z28\\\"]\"") +
            "\nAs a json rpc call\n" +
            HelpExampleRpc("sendmany",
                "\"\", "
                "\"{\\\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\\\":10.01,"
                "\\\"nexa:qqqdf524crwrn659hattd40jn5ykz32xey7z6n5z28\\\":10.02}\", 6, \"testing\""));

    string strAccount = AccountFromValue(params[0]);
    UniValue sendTo = params[1].get_obj();
    int nMinDepth = 1;
    if (params.size() > 2)
        nMinDepth = params[2].get_int();

    CWalletTx wtx;
    wtx.strFromAccount = strAccount;
    if (params.size() > 3 && !params[3].isNull() && !params[3].get_str().empty())
        wtx.mapValue["comment"] = params[3].get_str();

    UniValue subtractFeeFromAmount(UniValue::VARR);
    if (params.size() > 4)
        subtractFeeFromAmount = params[4].get_array();

    std::set<CTxDestination> destinations;
    std::vector<CRecipient> vecSend;

    CAmount totalAmount = 0;
    std::vector<std::string> keys = sendTo.getKeys();
    for (const std::string &name_ : keys)
    {
        CTxDestination dest = DecodeDestination(name_);
        if (!IsValidDestination(dest))
        {
            throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, std::string("Invalid coin address: ") + name_);
        }

        if (destinations.count(dest))
        {
            throw JSONRPCError(RPC_INVALID_PARAMETER, std::string("Invalid parameter, duplicated address: ") + name_);
        }
        destinations.insert(dest);

        CScript scriptPubKey = GetScriptForDestination(dest);
        CAmount nAmount = AmountFromValue(sendTo[name_]);
        if (nAmount <= 0)
            throw JSONRPCError(RPC_TYPE_ERROR, "Invalid amount for send");
        totalAmount += nAmount;

        bool fSubtractFeeFromAmount = false;
        for (unsigned int idx = 0; idx < subtractFeeFromAmount.size(); idx++)
        {
            const UniValue &addr = subtractFeeFromAmount[idx];
            if (addr.get_str() == name_)
                fSubtractFeeFromAmount = true;
        }

        CRecipient recipient = {scriptPubKey, nAmount, fSubtractFeeFromAmount};
        vecSend.push_back(recipient);
    }

    EnsureWalletIsUnlocked();

    // Check funds
    {
        LOCK(serializeCreateTx);

        // If an account is provided we need to make sure it doesn't exceed our account balance.
        // Otherwise, skip this expensive step because coin selection will fail if the amount exceeds the balance.
        if (strAccount != "")
        {
            CAmount nBalance = GetAccountBalance(strAccount, nMinDepth, ISMINE_SPENDABLE);
            if (totalAmount > nBalance)
                throw JSONRPCError(RPC_WALLET_INSUFFICIENT_FUNDS, "Account has insufficient funds");
        }

        // Send
        CReserveKey keyChange(pwalletMain);
        CAmount nFeeRequired = 0;
        int nChangePosRet = -1;
        string strFailReason;
        bool fCreated =
            pwalletMain->CreateTransaction(vecSend, wtx, keyChange, nFeeRequired, nChangePosRet, strFailReason);
        if (!fCreated)
            throw JSONRPCError(RPC_WALLET_INSUFFICIENT_FUNDS, strFailReason);
        std::string error;
        if (!pwalletMain->CommitTransaction(wtx, keyChange, error))
            throw JSONRPCError(RPC_WALLET_ERROR, error);
    }
    return wtx.GetIdem().GetHex();
}

// Defined in rpc/misc.cpp
extern CScript _createmultisig_redeemScript(const UniValue &params);

UniValue addmultisigaddress(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 2 || params.size() > 3)
    {
        string msg =
            "addmultisigaddress nrequired [\"key\",...] ( \"account\" )\n"
            "\nAdd a nrequired-to-sign multisignature address to the wallet.\n"
            "Each key is an address or hex-encoded public key.\n"
            "If 'account' is specified (DEPRECATED), assign address to that account.\n"

            "\nArguments:\n"
            "1. nrequired      (numeric, required) The number of required signatures out of the n keys or "
            "addresses.\n"
            "2. \"keysobject\"   (string, required) A json array of addresses or hex-encoded public keys\n"
            "     [\n"
            "       \"address\"  (string) address or hex-encoded public key\n"
            "       ...,\n"
            "     ]\n"
            "3. \"account\"      (string, optional) DEPRECATED. An account to assign the addresses to.\n"

            "\nResult:\n"
            "\"address\"  (string) A address associated with the keys.\n"

            "\nExamples:\n"
            "\nAdd a multisig address from 2 addresses\n" +
            HelpExampleCli("addmultisigaddress", "2 "
                                                 "\"[\\\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\\\",\\\"nexa:"
                                                 "qqqdf524crwrn659hattd40jn5ykz32xey7z6n5z28\\\"]\"") +
            "\nAs json rpc call\n" +
            HelpExampleRpc("addmultisigaddress", "2, "
                                                 "\"[\\\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\\\",\\\"nexa:"
                                                 "qqqdf524crwrn659hattd40jn5ykz32xey7z6n5z28\\\"]\"");
        throw runtime_error(msg);
    }

    LOCK(pwalletMain->cs_wallet);

    string strAccount;
    if (params.size() > 2)
        strAccount = AccountFromValue(params[2]);

    // Construct using pay-to-script-hash:
    CScript inner = _createmultisig_redeemScript(params);
    CScriptID innerID(inner);
    pwalletMain->AddCScript(inner);

    pwalletMain->SetAddressBook(innerID, strAccount, "send");
    return EncodeDestination(innerID);
}


struct tallyitem
{
    CAmount nAmount;
    int nConf;
    vector<uint256> txidems;
    bool fIsWatchonly;
    tallyitem()
    {
        nAmount = 0;
        nConf = std::numeric_limits<int>::max();
        fIsWatchonly = false;
    }
};

UniValue ListReceived(const UniValue &params, bool fByAccounts)
{
    // Minimum confirmations
    int nMinDepth = 1;
    if (params.size() > 0)
        nMinDepth = params[0].get_int();

    // Whether to include empty accounts
    bool fIncludeEmpty = false;
    if (params.size() > 1)
        fIncludeEmpty = params[1].get_bool();

    isminefilter filter = ISMINE_SPENDABLE;
    if (params.size() > 2)
        if (params[2].get_bool())
            filter = filter | ISMINE_WATCH_ONLY;

    // Tally
    std::map<CTxDestination, tallyitem> mapTally;
    for (MapWallet::iterator it = pwalletMain->mapWallet.begin(); it != pwalletMain->mapWallet.end(); ++it)
    {
        const CWalletTxRef wtx = it->second.tx;
        // Get every transaction exactly once, skipping outpoint and idem entries
        if (it->first.hash != wtx->GetId())
            continue;

        CValidationState state;
        if (wtx->IsCoinBase() || !CheckFinalTx(wtx.get()))
            continue;

        int nDepth = wtx->GetDepthInMainChain();
        if (nDepth < nMinDepth)
            continue;

        for (const CTxOut &txout : wtx->vout)
        {
            CTxDestination address;
            if (!ExtractDestination(txout.scriptPubKey, address))
                continue;

            isminefilter mine = IsMine(*pwalletMain, address, chainActive.Tip());
            if (!(mine & filter))
                continue;

            tallyitem &item = mapTally[address];
            item.nAmount += txout.nValue;
            item.nConf = min(item.nConf, nDepth);
            item.txidems.push_back(wtx->GetIdem());
            if (mine & ISMINE_WATCH_ONLY)
                item.fIsWatchonly = true;
        }
    }

    // Reply
    UniValue ret(UniValue::VARR);
    std::map<std::string, tallyitem> mapAccountTally;
    for (const std::pair<const CTxDestination, CAddressBookData> &item : pwalletMain->mapAddressBook)
    {
        const CTxDestination &dest = item.first;
        const std::string &strAccount = item.second.name;
        std::map<CTxDestination, tallyitem>::iterator it = mapTally.find(dest);
        if (it == mapTally.end() && !fIncludeEmpty)
            continue;

        CAmount nAmount = 0;
        int nConf = std::numeric_limits<int>::max();
        bool fIsWatchonly = false;
        if (it != mapTally.end())
        {
            nAmount = (*it).second.nAmount;
            nConf = (*it).second.nConf;
            fIsWatchonly = (*it).second.fIsWatchonly;
        }

        if (fByAccounts)
        {
            tallyitem &item2 = mapAccountTally[strAccount];
            item2.nAmount += nAmount;
            item2.nConf = min(item2.nConf, nConf);
            item2.fIsWatchonly = fIsWatchonly;
        }
        else
        {
            UniValue obj(UniValue::VOBJ);
            if (fIsWatchonly)
            {
                obj.pushKV("involvesWatchonly", true);
            }
            obj.pushKV("address", EncodeDestination(dest));
            obj.pushKV("account", strAccount);
            obj.pushKV("satoshi", UniValue(nAmount));
            obj.pushKV("amount", ValueFromAmount(nAmount));
            obj.pushKV("confirmations", (nConf == std::numeric_limits<int>::max() ? 0 : nConf));
            if (!fByAccounts)
            {
                obj.pushKV("label", strAccount);
            }
            UniValue transactions(UniValue::VARR);
            if (it != mapTally.end())
            {
                for (const uint256 &item3 : (*it).second.txidems)
                {
                    transactions.push_back(item3.GetHex());
                }
            }
            obj.pushKV("txidems", transactions);
            ret.push_back(obj);
        }
    }

    if (fByAccounts)
    {
        for (map<string, tallyitem>::iterator it = mapAccountTally.begin(); it != mapAccountTally.end(); ++it)
        {
            CAmount nAmount = (*it).second.nAmount;
            int nConf = (*it).second.nConf;
            UniValue obj(UniValue::VOBJ);
            if ((*it).second.fIsWatchonly)
                obj.pushKV("involvesWatchonly", true);
            obj.pushKV("account", (*it).first);
            obj.pushKV("satoshi", UniValue(nAmount));
            obj.pushKV("amount", ValueFromAmount(nAmount));
            obj.pushKV("confirmations", (nConf == std::numeric_limits<int>::max() ? 0 : nConf));
            ret.push_back(obj);
        }
    }

    return ret;
}

UniValue listreceivedbyaddress(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 3)
        throw runtime_error(
            "listreceivedbyaddress ( minconf includeempty includeWatchonly)\n"
            "\nList balances by receiving address.\n"
            "\nArguments:\n"
            "1. minconf       (numeric, optional, default=1) The minimum number of confirmations before payments are "
            "included.\n"
            "2. includeempty  (bool, optional, default=false) Whether to include addresses that haven't received any "
            "payments.\n"
            "3. includeWatchonly (bool, optional, default=false) Whether to include watchonly addresses (see "
            "'importaddress').\n"

            "\nResult:\n"
            "[\n"
            "  {\n"
            "    \"involvesWatchonly\" : true,        (bool) Only returned if imported addresses were involved in "
            "transaction\n"
            "    \"address\" : \"receivingaddress\",  (string) The receiving address\n"
            "    \"account\" : \"accountname\",       (string) DEPRECATED. The account of the receiving address. The "
            "default account is \"\".\n"
            "    \"amount\" : x.xxx,                  (numeric) The total amount in " +
            CURRENCY_UNIT +
            " received by the address\n"
            "    \"confirmations\" : n,               (numeric) The number of confirmations of the most recent "
            "transaction included\n"
            "    \"label\" : \"label\"                (string) A comment for the address/transaction, if any\n"
            "  }\n"
            "  ,...\n"
            "]\n"

            "\nExamples:\n" +
            HelpExampleCli("listreceivedbyaddress", "") + HelpExampleCli("listreceivedbyaddress", "6 true") +
            HelpExampleRpc("listreceivedbyaddress", "6, true, true"));

    LOCK(pwalletMain->cs_wallet);

    return ListReceived(params, false);
}

UniValue listreceivedbyaccount(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 3)
        throw runtime_error(
            "listreceivedbyaccount ( minconf includeempty includeWatchonly)\n"
            "\nDEPRECATED. List balances by account.\n"
            "\nArguments:\n"
            "1. minconf      (numeric, optional, default=1) The minimum number of confirmations before payments are "
            "included.\n"
            "2. includeempty (bool, optional, default=false) Whether to include accounts that haven't received any "
            "payments.\n"
            "3. includeWatchonly (bool, optional, default=false) Whether to include watchonly addresses (see "
            "'importaddress').\n"

            "\nResult:\n"
            "[\n"
            "  {\n"
            "    \"involvesWatchonly\" : true,   (bool) Only returned if imported addresses were involved in "
            "transaction\n"
            "    \"account\" : \"accountname\",  (string) The account name of the receiving account\n"
            "    \"amount\" : x.xxx,             (numeric) The total amount received by addresses with this account\n"
            "    \"confirmations\" : n,          (numeric) The number of confirmations of the most recent transaction "
            "included\n"
            "    \"label\" : \"label\"           (string) A comment for the address/transaction, if any\n"
            "  }\n"
            "  ,...\n"
            "]\n"

            "\nExamples:\n" +
            HelpExampleCli("listreceivedbyaccount", "") + HelpExampleCli("listreceivedbyaccount", "6 true") +
            HelpExampleRpc("listreceivedbyaccount", "6, true, true"));

    LOCK(pwalletMain->cs_wallet);

    return ListReceived(params, true);
}

static void MaybePushAddress(UniValue &entry, const CTxDestination &dest)
{
    if (IsValidDestination(dest))
    {
        entry.pushKV("address", EncodeDestination(dest));
    }
}

void ListTransactions(const CWalletTxRef wtx,
    const string &strAccount,
    int nMinDepth,
    bool fLong,
    UniValue &ret,
    const isminefilter &filter)
{
    CAmount nFee;
    string strSentAccount;
    list<CGroupedOutputEntry> listReceived;
    list<CGroupedOutputEntry> listSent;
    wtx->GetAmounts(listReceived, listSent, nFee, strSentAccount, filter);

    bool fAllAccounts = (strAccount == string("*"));
    bool involvesWatchonly = wtx->IsFromMe(ISMINE_WATCH_ONLY);

    // Sent
    if ((!listSent.empty() || nFee != 0) && (fAllAccounts || strAccount == strSentAccount))
    {
        for (const CGroupedOutputEntry &s : listSent)
        {
            UniValue entry(UniValue::VOBJ);
            if (involvesWatchonly || (::IsMine(*pwalletMain, s.destination, chainActive.Tip()) & ISMINE_WATCH_ONLY))
                entry.pushKV("involvesWatchonly", true);
            entry.pushKV("account", strSentAccount);
            MaybePushAddress(entry, s.destination);
            entry.pushKV("category", "send");
            if (s.grp != NoGroup)
            {
                entry.pushKV("group", EncodeGroupToken(s.grp));
                entry.pushKV("groupAmount", -s.grpAmount);
            }
            entry.pushKV("satoshi", UniValue(-s.amount));
            entry.pushKV("amount", ValueFromAmount(-s.amount));
            if (pwalletMain->mapAddressBook.count(s.destination))
                entry.pushKV("label", pwalletMain->mapAddressBook[s.destination].name);
            entry.pushKV("vout", s.vout);
            entry.pushKV("fee", ValueFromAmount(-nFee));
            if (fLong)
                WalletTxToJSON(*wtx, entry);
            entry.pushKV("abandoned", wtx->isAbandoned());
            ret.push_back(entry);
        }
    }

    // Received
    if (listReceived.size() > 0 && wtx->GetDepthInMainChain() >= nMinDepth)
    {
        for (const CGroupedOutputEntry &r : listReceived)
        {
            string account;
            if (pwalletMain->mapAddressBook.count(r.destination))
                account = pwalletMain->mapAddressBook[r.destination].name;
            if (fAllAccounts || (account == strAccount))
            {
                UniValue entry(UniValue::VOBJ);
                if (involvesWatchonly || (::IsMine(*pwalletMain, r.destination, chainActive.Tip()) & ISMINE_WATCH_ONLY))
                    entry.pushKV("involvesWatchonly", true);
                entry.pushKV("account", account);
                MaybePushAddress(entry, r.destination);
                if (wtx->IsCoinBase())
                {
                    if (wtx->GetDepthInMainChain() < 1)
                        entry.pushKV("category", "orphan");
                    else if (wtx->GetBlocksToMaturity() > 0)
                        entry.pushKV("category", "immature");
                    else
                        entry.pushKV("category", "generate");
                }
                else
                {
                    entry.pushKV("category", "receive");
                }
                if (r.grp != NoGroup)
                {
                    entry.pushKV("group", EncodeGroupToken(r.grp));
                    entry.pushKV("groupAmount", r.grpAmount);
                }
                entry.pushKV("satoshi", UniValue(r.amount));
                entry.pushKV("amount", ValueFromAmount(r.amount));
                if (pwalletMain->mapAddressBook.count(r.destination))
                    entry.pushKV("label", account);
                entry.pushKV("vout", r.vout);
                if (fLong)
                    WalletTxToJSON(*wtx, entry);
                ret.push_back(entry);
            }
        }
    }
}

void AcentryToJSON(const CAccountingEntry &acentry, const string &strAccount, UniValue &ret)
{
    bool fAllAccounts = (strAccount == string("*"));

    if (fAllAccounts || acentry.strAccount == strAccount)
    {
        UniValue entry(UniValue::VOBJ);
        entry.pushKV("account", acentry.strAccount);
        entry.pushKV("category", "move");
        entry.pushKV("time", acentry.nTime);
        entry.pushKV("satoshi", UniValue(acentry.nCreditDebit));
        entry.pushKV("amount", ValueFromAmount(acentry.nCreditDebit));
        entry.pushKV("otheraccount", acentry.strOtherAccount);
        entry.pushKV("comment", acentry.strComment);
        ret.push_back(entry);
    }
}

UniValue listtransactions(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 4)
        throw runtime_error(
            "listtransactions ( \"account\" count from includeWatchonly)\n"
            "\nReturns up to 'count' most recent transactions skipping the most recent 'from' transactions for account "
            "'account'.\n"
            "\nArguments:\n"
            "1. \"account\"    (string, optional) DEPRECATED. The account name. Should be \"*\".\n"
            "2. count          (numeric, optional, default=10) The number of transactions to return\n"
            "3. from           (numeric, optional, default=0) The number of transactions to skip\n"
            "4. includeWatchonly (bool, optional, default=false) Include transactions to watchonly addresses (see "
            "'importaddress')\n"
            "\nResult:\n"
            "[\n"
            "  {\n"
            "    \"account\":\"accountname\",       (string) DEPRECATED. The account name associated with the "
            "transaction. \n"
            "                                                It will be \"\" for the default account.\n"
            "    \"address\":\"address\",    (string) The address of the transaction. Not present for \n"
            "                                                move transactions (category = move).\n"
            "    \"category\":\"send|receive|move\", (string) The transaction category. 'move' is a local (off "
            "blockchain)\n"
            "                                                transaction between accounts, and not associated with an "
            "address,\n"
            "                                                transaction id or block. 'send' and 'receive' "
            "transactions are \n"
            "                                                associated with an address, transaction id and block "
            "details\n"
            "    \"amount\": x.xxx,          (numeric) The amount in " +
            CURRENCY_UNIT +
            ". This is negative for the 'send' category, and for the\n"
            "                                         'move' category for moves outbound. It is "
            "positive for the 'receive' category,\n"
            "                                         and for the 'move' category for inbound funds.\n"
            "    \"vout\": n,                (numeric) the vout value\n"
            "    \"fee\": x.xxx,             (numeric) The amount of the fee in " +
            CURRENCY_UNIT +
            ". This is negative and only available for the \n"
            "                                         'send' category of transactions.\n"
            "    \"confirmations\": n,       (numeric) The number of confirmations for the transaction. Available for "
            "'send' and \n"
            "                                         'receive' category of transactions. Negative confirmations "
            "indicate the\n"
            "                                         transaction conflicts with the block chain\n"
            "    \"trusted\": xxx            (bool) Whether we consider the outputs of this unconfirmed transaction "
            "safe to spend.\n"
            "    \"blockhash\": \"hashvalue\", (string) The block hash containing the transaction. Available for "
            "'send' and 'receive'\n"
            "                                          category of transactions.\n"
            "    \"blockindex\": n,          (numeric) The index of the transaction in the block that includes it. "
            "Available for 'send' and 'receive'\n"
            "                                          category of transactions.\n"
            "    \"blocktime\": xxx,         (numeric) The block time in seconds since epoch (1 Jan 1970 GMT).\n"
            "    \"txid\": \"transactionid\", (string) The transaction id. Available for 'send' and 'receive' category "
            "of transactions.\n"
            "    \"txidem\": \"transaction idem\", (string) The transaction idem. Available for 'send' and 'receive' "
            "category "
            "of transactions.\n"
            "    \"time\": xxx,              (numeric) The transaction time in seconds since epoch (midnight Jan 1 "
            "1970 GMT).\n"
            "    \"timereceived\": xxx,      (numeric) The time received in seconds since epoch (midnight Jan 1 1970 "
            "GMT). Available \n"
            "                                          for 'send' and 'receive' category of transactions.\n"
            "    \"comment\": \"...\",       (string) If a comment is associated with the transaction.\n"
            "    \"label\": \"label\"        (string) A comment for the address/transaction, if any\n"
            "    \"otheraccount\": \"accountname\",  (string) For the 'move' category of transactions, the account the "
            "funds came \n"
            "                                          from (for receiving funds, positive amounts), or went to (for "
            "sending funds,\n"
            "                                          negative amounts).\n"
            "    \"abandoned\": xxx          (bool) 'true' if the transaction has been abandoned (inputs are "
            "respendable). Only available for the \n"
            "                                         'send' category of transactions.\n"
            "  }\n"
            "]\n"

            "\nExamples:\n"
            "\nList the most recent 10 transactions in the systems\n" +
            HelpExampleCli("listtransactions", "") + "\nList transactions 100 to 120\n" +
            HelpExampleCli("listtransactions", "\"*\" 20 100") + "\nAs a json rpc call\n" +
            HelpExampleRpc("listtransactions", "\"*\", 20, 100"));

    LOCK(pwalletMain->cs_wallet);

    string strAccount = "*";
    if (params.size() > 0)
        strAccount = params[0].get_str();
    int nCount = 10;
    if (params.size() > 1)
        nCount = params[1].get_int();
    int nFrom = 0;
    if (params.size() > 2)
        nFrom = params[2].get_int();
    isminefilter filter = ISMINE_SPENDABLE;
    if (params.size() > 3)
        if (params[3].get_bool())
            filter = filter | ISMINE_WATCH_ONLY;

    if (nCount < 0)
        throw JSONRPCError(RPC_INVALID_PARAMETER, "Negative count");
    if (nFrom < 0)
        throw JSONRPCError(RPC_INVALID_PARAMETER, "Negative from");

    UniValue ret(UniValue::VARR);

    const CWallet::TxItems &txOrdered = pwalletMain->wtxOrdered;

    // iterate backwards until we have nCount items to return:
    for (CWallet::TxItems::const_reverse_iterator it = txOrdered.rbegin(); it != txOrdered.rend(); ++it)
    {
        CWalletTxRef wtx = (*it).second.first;
        if (wtx != nullptr)
            ListTransactions(wtx, strAccount, 0, true, ret, filter);
        CAccountingEntry *const pacentry = (*it).second.second;
        if (pacentry != nullptr)
            AcentryToJSON(*pacentry, strAccount, ret);

        if ((int)ret.size() >= (nCount + nFrom))
            break;
    }
    // ret is newest to oldest

    if (nFrom > (int)ret.size())
        nFrom = ret.size();
    if ((nFrom + nCount) > (int)ret.size())
        nCount = ret.size() - nFrom;

    vector<UniValue> arrTmp = ret.getValues();

    vector<UniValue>::iterator first = arrTmp.begin();
    std::advance(first, nFrom);
    vector<UniValue>::iterator last = arrTmp.begin();
    std::advance(last, nFrom + nCount);

    if (last != arrTmp.end())
        arrTmp.erase(last, arrTmp.end());
    if (first != arrTmp.begin())
        arrTmp.erase(arrTmp.begin(), first);

    std::reverse(arrTmp.begin(), arrTmp.end()); // Return oldest to newest

    ret.clear();
    ret.setArray();
    ret.push_backV(arrTmp);

    return ret;
}

UniValue listtransactionsfrom(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 4)
        throw runtime_error(
            "listtransactionsfrom ( \"account\" count from includeWatchonly)\n"
            "\nReturns up to 'count' most recent transactions skipping the first (oldest) 'from' transactions for "
            "account 'account'.\n"
            "\nArguments:\n"
            "1. \"account\"    (string, optional) DEPRECATED. The account name. Should be \"*\".\n"
            "2. count          (numeric, optional, default=10) The number of transactions to return\n"
            "3. from           (numeric, optional, default=0) The number of transactions to skip\n"
            "4. includeWatchonly (bool, optional, default=false) Include transactions to watchonly addresses (see "
            "'importaddress')\n"
            "\nResult:\n"
            "[\n"
            "  {\n"
            "    \"account\":\"accountname\",       (string) DEPRECATED. The account name associated with the "
            "transaction. \n"
            "                                                It will be \"\" for the default account.\n"
            "    \"address\":\"address\",    (string) The address of the transaction. Not present for \n"
            "                                                move transactions (category = move).\n"
            "    \"category\":\"send|receive|move\", (string) The transaction category. 'move' is a local (off "
            "blockchain)\n"
            "                                                transaction between accounts, and not associated with an "
            "address,\n"
            "                                                transaction id or block. 'send' and 'receive' "
            "transactions are \n"
            "                                                associated with an address, transaction id and block "
            "details\n"
            "    \"amount\": x.xxx,          (numeric) The amount in " +
            CURRENCY_UNIT +
            ". This is negative for the 'send' category, and for the\n"
            "                                         'move' category for moves outbound. It is "
            "positive for the 'receive' category,\n"
            "                                         and for the 'move' category for inbound funds.\n"
            "    \"vout\": n,                (numeric) the vout value\n"
            "    \"fee\": x.xxx,             (numeric) The amount of the fee in " +
            CURRENCY_UNIT +
            ". This is negative and only available for the \n"
            "                                         'send' category of transactions.\n"
            "    \"confirmations\": n,       (numeric) The number of confirmations for the transaction. Available for "
            "'send' and \n"
            "                                         'receive' category of transactions. Negative confirmations "
            "indicate the\n"
            "                                         transaction conflicts with the block chain\n"
            "    \"trusted\": xxx            (bool) Whether we consider the outputs of this unconfirmed transaction "
            "safe to spend.\n"
            "    \"blockhash\": \"hashvalue\", (string) The block hash containing the transaction. Available for "
            "'send' and 'receive'\n"
            "                                          category of transactions.\n"
            "    \"blockindex\": n,          (numeric) The index of the transaction in the block that includes it. "
            "Available for 'send' and 'receive'\n"
            "                                          category of transactions.\n"
            "    \"blocktime\": xxx,         (numeric) The block time in seconds since epoch (1 Jan 1970 GMT).\n"
            "    \"txid\": \"transactionid\", (string) The transaction id. Available for 'send' and 'receive' category "
            "of transactions.\n"
            "    \"txidem\": \"transaction idem\", (string) The transaction idem. Available for 'send' and 'receive' "
            "category "
            "of transactions.\n"
            "    \"time\": xxx,              (numeric) The transaction time in seconds since epoch (midnight Jan 1 "
            "1970 GMT).\n"
            "    \"timereceived\": xxx,      (numeric) The time received in seconds since epoch (midnight Jan 1 1970 "
            "GMT). Available \n"
            "                                          for 'send' and 'receive' category of transactions.\n"
            "    \"comment\": \"...\",       (string) If a comment is associated with the transaction.\n"
            "    \"label\": \"label\"        (string) A comment for the address/transaction, if any\n"
            "    \"otheraccount\": \"accountname\",  (string) For the 'move' category of transactions, the account the "
            "funds came \n"
            "                                          from (for receiving funds, positive amounts), or went to (for "
            "sending funds,\n"
            "                                          negative amounts).\n"
            "    \"abandoned\": xxx          (bool) 'true' if the transaction has been abandoned (inputs are "
            "respendable). Only available for the \n"
            "                                         'send' category of transactions.\n"
            "  }\n"
            "]\n"

            "\nExamples:\n"
            "\nList the most recent 10 transactions in the systems\n" +
            HelpExampleCli("listtransactionsfrom", "") + "\nList transactions 100 to 120\n" +
            HelpExampleCli("listtransactionsfrom", "\"*\" 20 100") + "\nAs a json rpc call\n" +
            HelpExampleRpc("listtransactionsfrom", "\"*\", 20, 100"));

    LOCK(pwalletMain->cs_wallet);

    string strAccount = "*";
    if (params.size() > 0)
        strAccount = params[0].get_str();
    int nCount = 10;
    if (params.size() > 1)
        nCount = params[1].get_int();
    int nFrom = 0;
    if (params.size() > 2)
        nFrom = params[2].get_int();
    isminefilter filter = ISMINE_SPENDABLE;
    if (params.size() > 3)
        if (params[3].get_bool())
            filter = filter | ISMINE_WATCH_ONLY;

    if (nCount < 0)
        throw JSONRPCError(RPC_INVALID_PARAMETER, "Negative count");
    if (nFrom < 0)
        throw JSONRPCError(RPC_INVALID_PARAMETER, "Negative from");

    UniValue ret(UniValue::VARR);

    const CWallet::TxItems &txOrdered = pwalletMain->wtxOrdered;
    if (txOrdered.size() < (unsigned int)nFrom)
        return ret;
    CWallet::TxItems::const_iterator it = txOrdered.begin();
    std::advance(it, nFrom);

    for (int cnt = 0; (it != txOrdered.end()) && (cnt < nCount); ++it, ++cnt)
    {
        CWalletTxRef wtx = (*it).second.first;
        if (wtx != 0)
            ListTransactions(wtx, strAccount, 0, true, ret, filter);
        CAccountingEntry *const pacentry = (*it).second.second;
        if (pacentry != 0)
            AcentryToJSON(*pacentry, strAccount, ret);
    }

    return ret;
}


UniValue listaccounts(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 2)
        throw runtime_error(
            "listaccounts ( minconf includeWatchonly)\n"
            "\nDEPRECATED. Returns Object that has account names as keys, account balances as values.\n"
            "\nArguments:\n"
            "1. minconf          (numeric, optional, default=1) Only include transactions with at least this many "
            "confirmations\n"
            "2. includeWatchonly (bool, optional, default=false) Include balances in watchonly addresses (see "
            "'importaddress')\n"
            "\nResult:\n"
            "{                      (json object where keys are account names, and values are numeric balances\n"
            "  \"account\": x.xxx,  (numeric) The property name is the account name, and the value is the total "
            "balance for the account.\n"
            "  ...\n"
            "}\n"
            "\nExamples:\n"
            "\nList account balances where there at least 1 confirmation\n" +
            HelpExampleCli("listaccounts", "") + "\nList account balances including zero confirmation transactions\n" +
            HelpExampleCli("listaccounts", "0") + "\nList account balances for 6 or more confirmations\n" +
            HelpExampleCli("listaccounts", "6") + "\nAs json rpc call\n" + HelpExampleRpc("listaccounts", "6"));

    map<string, CAmount> mapAccountBalances;

    {
        // Locking cs_main ensures that the chain doesn't progress during our summation of balances.  This means that
        // the balances will be consistent, although they may not point to the tip.  This API should report the tip
        LOCK2(cs_main, pwalletMain->cs_wallet);

        int nMinDepth = 1;
        if (params.size() > 0)
            nMinDepth = params[0].get_int();
        isminefilter includeWatchonly = ISMINE_SPENDABLE;
        if (params.size() > 1)
            if (params[1].get_bool())
                includeWatchonly = includeWatchonly | ISMINE_WATCH_ONLY;

        for (const PAIRTYPE(const CTxDestination, CAddressBookData) & entry : pwalletMain->mapAddressBook)
        {
            if (IsMine(*pwalletMain, entry.first, chainActive.Tip()) & includeWatchonly) // This address belongs to me
                mapAccountBalances[entry.second.name] = 0;
        }

        for (MapWallet::iterator it = pwalletMain->mapWallet.begin(); it != pwalletMain->mapWallet.end(); ++it)
        {
            const CWalletTxRef wtx = it->second.tx;
            // Get every transaction exactly once, skipping outpoint and idem entries
            if (it->first.hash != wtx->GetId())
                continue;

            CAmount nFee;
            string strSentAccount;
            list<COutputEntry> listReceived;
            list<COutputEntry> listSent;
            int nDepth = wtx->GetDepthInMainChain();
            if (wtx->GetBlocksToMaturity() > 0 || nDepth < 0)
                continue;
            wtx->GetAmounts(listReceived, listSent, nFee, strSentAccount, includeWatchonly);
            mapAccountBalances[strSentAccount] -= nFee;
            for (const COutputEntry &s : listSent)
            {
                mapAccountBalances[strSentAccount] -= s.amount;
            }
            if (nDepth >= nMinDepth)
            {
                for (const COutputEntry &r : listReceived)
                {
                    if (pwalletMain->mapAddressBook.count(r.destination))
                        mapAccountBalances[pwalletMain->mapAddressBook[r.destination].name] += r.amount;
                    else
                        mapAccountBalances[""] += r.amount;
                }
            }
        }

        const list<CAccountingEntry> &acentries = pwalletMain->laccentries;
        for (const CAccountingEntry &entry : acentries)
        {
            mapAccountBalances[entry.strAccount] += entry.nCreditDebit;
        }
    }

    UniValue ret(UniValue::VOBJ);
    for (const PAIRTYPE(const string, CAmount) & accountBalance : mapAccountBalances)
    {
        ret.pushKV(accountBalance.first, ValueFromAmount(accountBalance.second));
    }
    return ret;
}

UniValue listsinceblock(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp)
        throw runtime_error(
            "listsinceblock ( \"blockhash\" target-confirmations includeWatchonly)\n"
            "\nGet all transactions in blocks since block [blockhash], or all transactions if omitted\n"
            "\nArguments:\n"
            "1. \"blockhash\"   (string, optional) The block hash to list transactions since\n"
            "2. target-confirmations:    (numeric, optional) The confirmations required, must be 1 or more\n"
            "3. includeWatchonly:        (bool, optional, default=false) Include transactions to watchonly addresses "
            "(see 'importaddress')"
            "\nResult:\n"
            "{\n"
            "  \"transactions\": [\n"
            "    \"account\":\"accountname\",       (string) DEPRECATED. The account name associated with the "
            "transaction. Will be \"\" for the default account.\n"
            "    \"address\":\"address\",    (string) The address of the transaction. Not present for "
            "move transactions (category = move).\n"
            "    \"category\":\"send|receive\",     (string) The transaction category. 'send' has negative amounts, "
            "'receive' has positive amounts.\n"
            "    \"amount\": x.xxx,          (numeric) The amount in " +
            CURRENCY_UNIT +
            ". This is negative for the 'send' category, and for the 'move' category for moves \n"
            "                                          outbound. It is positive for the 'receive' "
            "category, and for the 'move' category for inbound funds.\n"
            "    \"vout\" : n,               (numeric) the vout value\n"
            "    \"fee\": x.xxx,             (numeric) The amount of the fee in " +
            CURRENCY_UNIT +
            ". This is negative and only available for the 'send' category of transactions.\n"
            "    \"confirmations\": n,       (numeric) The number of confirmations for the transaction. Available for "
            "'send' and 'receive' category of transactions.\n"
            "    \"blockhash\": \"hashvalue\",     (string) The block hash containing the transaction. Available for "
            "'send' and 'receive' category of transactions.\n"
            "    \"blockindex\": n,          (numeric) The index of the transaction in the block that includes it. "
            "Available for 'send' and 'receive' category of transactions.\n"
            "    \"blocktime\": xxx,         (numeric) The block time in seconds since epoch (1 Jan 1970 GMT).\n"
            "    \"txid\": \"transactionid\",  (string) The transaction id. Available for 'send' and 'receive' "
            "category of transactions.\n"
            "    \"txidem\": \"transaction idem\",  (string) The transaction idem. Available for 'send' and 'receive' "
            "category of transactions.\n"
            "    \"time\": xxx,              (numeric) The transaction time in seconds since epoch (Jan 1 1970 GMT).\n"
            "    \"timereceived\": xxx,      (numeric) The time received in seconds since epoch (Jan 1 1970 GMT). "
            "Available for 'send' and 'receive' category of transactions.\n"
            "    \"abandoned\": xxx,         (bool) 'true' if the transaction has been abandoned (inputs are "
            "respendable). Only available for the 'send' category of transactions.\n"
            "    \"comment\": \"...\",       (string) If a comment is associated with the transaction.\n"
            "    \"label\" : \"label\"       (string) A comment for the address/transaction, if any\n"
            "    \"to\": \"...\",            (string) If a comment to is associated with the transaction.\n"
            "  ],\n"
            "  \"lastblock\": \"lastblockhash\"     (string) The hash of the last block\n"
            "}\n"
            "\nExamples:\n" +
            HelpExampleCli("listsinceblock", "") +
            HelpExampleCli("listsinceblock", "\"000000000000000bacf66f7497b7dc45ef753ee9a7d38571037cdb1a57f663ad\" 6") +
            HelpExampleRpc(
                "listsinceblock", "\"000000000000000bacf66f7497b7dc45ef753ee9a7d38571037cdb1a57f663ad\", 6"));

    LOCK(pwalletMain->cs_wallet);

    CBlockIndex *pindex = nullptr;
    int target_confirms = 1;
    isminefilter filter = ISMINE_SPENDABLE;

    if (params.size() > 0)
    {
        uint256 blockId;

        blockId.SetHex(params[0].get_str());
        pindex = LookupBlockIndex(blockId);
    }

    if (params.size() > 1)
    {
        target_confirms = params[1].get_int();

        if (target_confirms < 1)
            throw JSONRPCError(RPC_INVALID_PARAMETER, "Invalid parameter");
    }

    if (params.size() > 2)
        if (params[2].get_bool())
            filter = filter | ISMINE_WATCH_ONLY;

    int depth = pindex ? (1 + chainActive.Height() - pindex->height()) : -1;

    UniValue transactions(UniValue::VARR);

    for (MapWallet::iterator it = pwalletMain->mapWallet.begin(); it != pwalletMain->mapWallet.end(); it++)
    {
        CWalletTxRef tx = it->second.tx;
        // Get every transaction exactly once, skipping outpoint and idem entries
        if (it->first.hash != tx->GetId())
            continue;

        if (depth == -1 || tx->GetDepthInMainChain() < depth)
            ListTransactions(tx, "*", 0, true, transactions, filter);
    }

    CBlockIndex *pblockLast = chainActive[chainActive.Height() + 1 - target_confirms];
    uint256 lastblock = pblockLast ? pblockLast->GetBlockHash() : uint256();

    UniValue ret(UniValue::VOBJ);
    ret.pushKV("transactions", transactions);
    ret.pushKV("lastblock", lastblock.GetHex());

    return ret;
}

UniValue gettransaction(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 1 || params.size() > 2)
        throw runtime_error(
            "gettransaction \"txid or txidem\" ( includeWatchonly )\n"
            "\nGet detailed information about in-wallet transaction <txid/txidem>\n"
            "\nArguments:\n"
            "1. \"txid or txidem\"    (string, required) The transaction id or transaction idem\n"
            "2. \"includeWatchonly\"    (bool, optional, default=false) Whether to include watchonly addresses in "
            "balance calculation and details[]\n"
            "\nResult:\n"
            "{\n"
            "  \"amount\" : x.xxx,        (numeric) The transaction amount in " +
            CURRENCY_UNIT +
            "\n"
            "  \"confirmations\" : n,     (numeric) The number of confirmations\n"
            "  \"blockhash\" : \"hash\",    (string) The block hash\n"
            "  \"blockindex\" : xx,       (numeric) The index of the transaction in the block that includes it\n"
            "  \"blocktime\" : ttt,       (numeric) The time in seconds since epoch (1 Jan 1970 GMT)\n"
            "  \"txid\" : \"transaction id\",   (string) The transaction id.\n"
            "  \"txidem\" : \"transaction idem\",   (string) The transaction idem.\n"
            "  \"time\" : ttt,            (numeric) The transaction time in seconds since epoch (1 Jan 1970 GMT)\n"
            "  \"timereceived\" : ttt,    (numeric) The time received in seconds since epoch (1 Jan 1970 GMT)\n"
            "  \"bip125-replaceable\": \"yes|no|unknown\"  (string) Whether this transaction could be replaced due to "
            "BIP125 (replace-by-fee);\n"
            "                                                   may be unknown for unconfirmed transactions not in the "
            "txpool\n"
            "  \"details\" : [\n"
            "    {\n"
            "      \"account\" : \"accountname\",  (string) DEPRECATED. The account name involved in the transaction, "
            "can be \"\" for the default account.\n"
            "      \"address\" : \"address\",   (string) The address involved in the transaction\n"
            "      \"category\" : \"send|receive\",    (string) The category, either 'send' or 'receive'\n"
            "      \"group\": \"groupidentifier\",     (string) The token identifier (appears only if applicable)\n"
            "      \"groupAmount\": n,               (numeric) The token quantity (appears only if applicable)\n"
            "      \"amount\" : x.xxx,               (numeric) The amount in " +
            CURRENCY_UNIT +
            "\n"
            "      \"label\" : \"label\",              (string) A comment for the address/transaction, if any\n"
            "      \"vout\" : n,                     (numeric) the vout value\n"
            "    }\n"
            "    ,...\n"
            "  ],\n"
            "  \"hex\" : \"data\"         (string) Raw data for transaction\n"
            "}\n"

            "\nExamples:\n" +
            HelpExampleCli("gettransaction", "\"1075db55d416d3ca199f55b6084e2115b9345e16c5cf302fc80e9d5fbf5d48d\"") +
            HelpExampleCli(
                "gettransaction", "\"1075db55d416d3ca199f55b6084e2115b9345e16c5cf302fc80e9d5fbf5d48d\" true") +
            HelpExampleRpc("gettransaction", "\"1075db55d416d3ca199f55b6084e2115b9345e16c5cf302fc80e9d5fbf5d48d\""));

    LOCK(pwalletMain->cs_wallet);

    uint256 hash;
    hash.SetHex(params[0].get_str());

    isminefilter filter = ISMINE_SPENDABLE;
    if (params.size() > 1)
        if (params[1].get_bool())
            filter = filter | ISMINE_WATCH_ONLY;

    UniValue entry(UniValue::VOBJ);

    const CWalletTxRef wtx = pwalletMain->GetWalletTx(hash);
    if (!wtx)
        throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, "Invalid or non-wallet transaction id");

    CAmount nCredit = wtx->GetCredit(filter);
    CAmount nDebit = wtx->GetDebit(filter);
    CAmount nNet = nCredit - nDebit;
    CAmount nFee = (wtx->IsFromMe(filter) ? wtx->GetValueOut() - nDebit : 0);

    entry.pushKV("satoshi", UniValue(nNet - nFee));
    entry.pushKV("amount", ValueFromAmount(nNet - nFee));
    if (wtx->IsFromMe(filter))
        entry.pushKV("fee", ValueFromAmount(nFee));

    WalletTxToJSON(*wtx, entry);

    UniValue details(UniValue::VARR);
    ListTransactions(wtx, "*", 0, false, details, filter);
    entry.pushKV("details", details);

    string strHex = EncodeHexTx(static_cast<CTransaction>(*wtx));
    entry.pushKV("hex", strHex);

    return entry;
}

UniValue abandontransaction(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() != 1)
        throw runtime_error(
            "abandontransaction \"txid\" or \"txidem\"\n"
            "\nMark in-wallet transaction as abandoned\n"
            "This will mark this transaction and all its in-wallet descendants as abandoned which will allow\n"
            "for their inputs to be respent.  It can be used to replace \"stuck\" or evicted transactions.\n"
            "It only works on transactions which are not included in a block.  It removes transactions currently\n"
            "in the txpool.  It has no effect on transactions which are already conflicted or abandoned.\n"
            "\nArguments:\n"
            "1. \"txid or txidem\"    (string, required) The transaction id or idem\n"
            "\nResult:\n"
            "\nExamples:\n" +
            HelpExampleCli(
                "abandontransaction", "\"1075db55d416d3ca199f55b6084e2115b9345e16c5cf302fc80e9d5fbf5d48d\"") +
            HelpExampleRpc(
                "abandontransaction", "\"1075db55d416d3ca199f55b6084e2115b9345e16c5cf302fc80e9d5fbf5d48d\""));

    LOCK(pwalletMain->cs_wallet);

    uint256 hash;
    hash.SetHex(params[0].get_str());

    if (!pwalletMain->GetWalletTx(hash))
        throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, "Invalid or non-wallet transaction id");
    if (!pwalletMain->AbandonTransaction(hash))
        throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, "Transaction not eligible for abandonment");

    return NullUniValue;
}


UniValue backupwallet(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() != 1)
        throw runtime_error(
            "backupwallet \"destination\"\n"
            "\nSafely copies wallet.dat to destination, which can be a directory or a path with filename.\n"
            "\nArguments:\n"
            "1. \"destination\"   (string) The destination directory or file\n"
            "\nExamples:\n" +
            HelpExampleCli("backupwallet", "\"backup.dat\"") + HelpExampleRpc("backupwallet", "\"backup.dat\""));

    LOCK(pwalletMain->cs_wallet);

    string strDest = params[0].get_str();
    if (!BackupWallet(*pwalletMain, strDest))
        throw JSONRPCError(RPC_WALLET_ERROR, "Error: Wallet backup failed!");

    return NullUniValue;
}


UniValue keypoolrefill(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 1)
        throw runtime_error("keypoolrefill ( newsize )\n"
                            "\nFills the keypool." +
                            HelpRequiringPassphrase() +
                            "\n"
                            "\nArguments\n"
                            "1. newsize     (numeric, optional, default=100) The new keypool size\n"
                            "\nExamples:\n" +
                            HelpExampleCli("keypoolrefill", "") + HelpExampleRpc("keypoolrefill", ""));

    LOCK(pwalletMain->cs_wallet);

    // 0 is interpreted by TopUpKeyPool() as the default keypool size given by -keypool
    unsigned int kpSize = 0;
    if (params.size() > 0)
    {
        if (params[0].get_int() < 0)
            throw JSONRPCError(RPC_INVALID_PARAMETER, "Invalid parameter, expected valid size.");
        kpSize = (unsigned int)params[0].get_int();
    }

    EnsureWalletIsUnlocked();
    if (!pwalletMain->TopUpKeyPool(kpSize))
        throw runtime_error(strprintf("Keypool is already full: %d keys", pwalletMain->GetKeyPoolSize()));

    if (pwalletMain->GetKeyPoolSize() < kpSize)
        throw JSONRPCError(RPC_WALLET_ERROR, "Error refreshing keypool.");

    return NullUniValue;
}


static void LockWallet(CWallet *pWallet)
{
    LOCK(cs_nWalletUnlockTime);
    nWalletUnlockTime = 0;
    pWallet->Lock();
}

UniValue walletpassphrase(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (pwalletMain->IsCrypted() && (fHelp || params.size() != 2))
        throw runtime_error(
            "walletpassphrase \"passphrase\" timeout\n"
            "\nStores the wallet decryption key in memory for 'timeout' seconds.\n"
            "This is needed prior to performing transactions related to private keys such as sending coins\n"
            "\nArguments:\n"
            "1. \"passphrase\"     (string, required) The wallet passphrase\n"
            "2. timeout            (numeric, required) The time to keep the decryption key in seconds.\n"
            "\nNote:\n"
            "Issuing the walletpassphrase command while the wallet is already unlocked will set a new unlock\n"
            "time that overrides the old one.\n"
            "\nExamples:\n"
            "\nunlock the wallet for 60 seconds\n" +
            HelpExampleCli("walletpassphrase", "\"my pass phrase\" 60") +
            "\nLock the wallet again (before 60 seconds)\n" + HelpExampleCli("walletlock", "") +
            "\nAs json rpc call\n" + HelpExampleRpc("walletpassphrase", "\"my pass phrase\", 60"));

    LOCK(pwalletMain->cs_wallet);

    if (fHelp)
        return true;
    if (!pwalletMain->IsCrypted())
        throw JSONRPCError(
            RPC_WALLET_WRONG_ENC_STATE, "Error: running with an unencrypted wallet, but walletpassphrase was called.");

    // Note that the walletpassphrase is stored in params[0] which is not mlock()ed
    SecureString strWalletPass;
    strWalletPass.reserve(100);
    // TODO: get rid of this .c_str() by implementing SecureString::operator=(std::string)
    // Alternately, find a way to make params[0] mlock()'d to begin with.
    strWalletPass = params[0].get_str().c_str();

    if (strWalletPass.length() > 0)
    {
        if (!pwalletMain->Unlock(strWalletPass))
            throw JSONRPCError(RPC_WALLET_PASSPHRASE_INCORRECT, "Error: The wallet passphrase entered was incorrect.");
    }
    else
        throw runtime_error("walletpassphrase <passphrase> <timeout>\n"
                            "Stores the wallet decryption key in memory for <timeout> seconds.");

    pwalletMain->TopUpKeyPool();

    int64_t nSleepTime = params[1].get_int64();
    if (nSleepTime <= 0)
    {
        throw JSONRPCError(RPC_WALLET_PASSPHRASE_INCORRECT, "Error: The timeout period must be a positive number");
    }

    int64_t nLimit = 24 * 60 * 60; // one day
    if (nSleepTime > nLimit)
    {
        throw JSONRPCError(RPC_WALLET_PASSPHRASE_INCORRECT,
            strprintf("Error: The timeout period can not be greater than %ld seconds", nLimit));
    }

    LOCK(cs_nWalletUnlockTime);
    nWalletUnlockTime = GetTime() + nSleepTime;
    RPCRunLater("lockwallet", boost::bind(LockWallet, pwalletMain), nSleepTime);

    return NullUniValue;
}


UniValue walletpassphrasechange(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (pwalletMain->IsCrypted() && (fHelp || params.size() != 2))
        throw runtime_error("walletpassphrasechange \"oldpassphrase\" \"newpassphrase\"\n"
                            "\nChanges the wallet passphrase from 'oldpassphrase' to 'newpassphrase'.\n"
                            "\nArguments:\n"
                            "1. \"oldpassphrase\"      (string) The current passphrase\n"
                            "2. \"newpassphrase\"      (string) The new passphrase\n"
                            "\nExamples:\n" +
                            HelpExampleCli("walletpassphrasechange", "\"old one\" \"new one\"") +
                            HelpExampleRpc("walletpassphrasechange", "\"old one\", \"new one\""));

    LOCK(pwalletMain->cs_wallet);

    if (fHelp)
        return true;
    if (!pwalletMain->IsCrypted())
        throw JSONRPCError(RPC_WALLET_WRONG_ENC_STATE,
            "Error: running with an unencrypted wallet, but walletpassphrasechange was called.");

    // TODO: get rid of these .c_str() calls by implementing SecureString::operator=(std::string)
    // Alternately, find a way to make params[0] mlock()'d to begin with.
    SecureString strOldWalletPass;
    strOldWalletPass.reserve(100);
    strOldWalletPass = params[0].get_str().c_str();

    SecureString strNewWalletPass;
    strNewWalletPass.reserve(100);
    strNewWalletPass = params[1].get_str().c_str();

    if (strOldWalletPass.length() < 1 || strNewWalletPass.length() < 1)
        throw runtime_error("walletpassphrasechange <oldpassphrase> <newpassphrase>\n"
                            "Changes the wallet passphrase from <oldpassphrase> to <newpassphrase>.");

    if (!pwalletMain->ChangeWalletPassphrase(strOldWalletPass, strNewWalletPass))
        throw JSONRPCError(RPC_WALLET_PASSPHRASE_INCORRECT, "Error: The wallet passphrase entered was incorrect.");

    return NullUniValue;
}


UniValue walletlock(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (pwalletMain->IsCrypted() && (fHelp || params.size() != 0))
        throw runtime_error(
            "walletlock\n"
            "\nRemoves the wallet encryption key from memory, locking the wallet.\n"
            "After calling this method, you will need to call walletpassphrase again\n"
            "before being able to call any methods which require the wallet to be unlocked.\n"
            "\nExamples:\n"
            "\nSet the passphrase for 2 minutes to perform a transaction\n" +
            HelpExampleCli("walletpassphrase", "\"my pass phrase\" 120") +
            "\nPerform a send (requires passphrase set)\n" +
            HelpExampleCli("sendtoaddress", "\"nexa:qpnfq7r98lkm2h7jksy0h5mqpwvttxlc3q6zkwmcha\" 1.01") +
            "\nClear the passphrase since we are done before 2 minutes is up\n" + HelpExampleCli("walletlock", "") +
            "\nAs json rpc call\n" + HelpExampleRpc("walletlock", ""));

    LOCK(pwalletMain->cs_wallet);

    if (fHelp)
        return true;
    if (!pwalletMain->IsCrypted())
        throw JSONRPCError(
            RPC_WALLET_WRONG_ENC_STATE, "Error: running with an unencrypted wallet, but walletlock was called.");

    {
        LOCK(cs_nWalletUnlockTime);
        pwalletMain->Lock();
        nWalletUnlockTime = 0;
    }

    return NullUniValue;
}


UniValue encryptwallet(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (!pwalletMain->IsCrypted() && (fHelp || params.size() != 1))
        throw runtime_error(
            "encryptwallet \"passphrase\"\n"
            "\nEncrypts the wallet with 'passphrase'. This is for first time encryption.\n"
            "After this, any calls that interact with private keys such as sending or signing \n"
            "will require the passphrase to be set prior the making these calls.\n"
            "Use the walletpassphrase call for this, and then walletlock call.\n"
            "If the wallet is already encrypted, use the walletpassphrasechange call.\n"
            "Note that this will shutdown the server.\n"
            "\nArguments:\n"
            "1. \"passphrase\"    (string) The pass phrase to encrypt the wallet with. It must be at least 1 "
            "character, but should be long.\n"
            "\nExamples:\n"
            "\nEncrypt you wallet\n" +
            HelpExampleCli("encryptwallet", "\"my pass phrase\"") +
            "\nNow set the passphrase to use the wallet, such as for signing or sending coins\n" +
            HelpExampleCli("walletpassphrase", "\"my pass phrase\"") + "\nNow we can so something like sign\n" +
            HelpExampleCli("signmessage", "\"address\" \"test message\"") +
            "\nNow lock the wallet again by removing the passphrase\n" + HelpExampleCli("walletlock", "") +
            "\nAs a json rpc call\n" + HelpExampleRpc("encryptwallet", "\"my pass phrase\""));

    LOCK(pwalletMain->cs_wallet);

    if (fHelp)
        return true;
    if (pwalletMain->IsCrypted())
        throw JSONRPCError(
            RPC_WALLET_WRONG_ENC_STATE, "Error: running with an encrypted wallet, but encryptwallet was called.");

    // TODO: get rid of this .c_str() by implementing SecureString::operator=(std::string)
    // Alternately, find a way to make params[0] mlock()'d to begin with.
    SecureString strWalletPass;
    strWalletPass.reserve(100);
    strWalletPass = params[0].get_str().c_str();

    if (strWalletPass.length() < 1)
        throw runtime_error("encryptwallet <passphrase>\n"
                            "Encrypts the wallet with <passphrase>.");

    if (!pwalletMain->EncryptWallet(strWalletPass))
        throw JSONRPCError(RPC_WALLET_ENCRYPTION_FAILED, "Error: Failed to encrypt the wallet.");

    // BDB seems to have a bad habit of writing old data into
    // slack space in .dat files; that is bad if the old data is
    // unencrypted private keys. So:
    StartShutdown();
    return "wallet encrypted; server stopping, restart to run with encrypted wallet. The keypool has been "
           "flushed and a new HD seed was generated (if you are using HD). You need to make a new backup.";
}

UniValue lockunspent(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 1 || params.size() > 2)
        throw runtime_error(
            "lockunspent unlock [{\"txidem\":\"txidem\",\"vout\":n},...]\n"
            "lockunspent unlock [\"outpointhash\",...]\n"
            "\nUpdates list of temporarily unspendable outputs.\n"
            "Temporarily lock (unlock=false) or unlock (unlock=true) specified transaction outputs.\n"
            "A locked transaction output will not be chosen by automatic coin selection, when spending coins.\n"
            "Locks are stored in memory only. Nodes start with zero locked outputs, and the locked output list\n"
            "is always cleared (by virtue of process exit) when a node stops or fails.\n"
            "Also see the listunspent call\n"
            "\nArguments:\n"
            "1. unlock            (boolean, required) Whether to unlock (true) or lock (false) the specified "
            "transactions\n"
            "2. \"transactions\"  (string, required) A json array of objects. Each object the txidem (string) vout "
            "(numeric)\n"
            "     [           (json array of json objects)\n"
            "       {\n"
            "         \"txidem\":\"id\",    (string) The transaction id\n"
            "         \"vout\": n         (numeric) The output number\n"
            "       }\n"
            "       ,...\n"
            "     ]\n"

            "\nResult:\n"
            "true|false    (boolean) Whether the command was successful or not\n"

            "\nExamples:\n"
            "\nList the unspent transactions\n" +
            HelpExampleCli("listunspent", "") + "\nLock an unspent transaction\n" +
            HelpExampleCli("lockunspent", "false "
                                          "\"[{\\\"txidem\\\":"
                                          "\\\"a08e6907dbbd3d809776dbfc5d82e371b764ed838b5655e72f463568df1aadf0\\\","
                                          "\\\"vout\\\":1}]\"") +
            "\nList the locked transactions\n" + HelpExampleCli("listlockunspent", "") +
            "\nUnlock the transaction again\n" +
            HelpExampleCli("lockunspent", "true "
                                          "\"[{\\\"txidem\\\":"
                                          "\\\"a08e6907dbbd3d809776dbfc5d82e371b764ed838b5655e72f463568df1aadf0\\\","
                                          "\\\"vout\\\":1}]\"") +
            "\nAs a json rpc call\n" +
            HelpExampleRpc("lockunspent", "false, "
                                          "\"[{\\\"txidem\\\":"
                                          "\\\"a08e6907dbbd3d809776dbfc5d82e371b764ed838b565"
                                          "5e72f463568df1aadf0\\\",\\\"vout\\\":1}]\""));

    LOCK(pwalletMain->cs_wallet);

    if (params.size() == 1)
        RPCTypeCheck(params, {UniValue::VBOOL});
    else
        RPCTypeCheck(params, {UniValue::VBOOL, UniValue::VARR});

    bool fUnlock = params[0].get_bool();

    if (params.size() == 1)
    {
        if (fUnlock)
            pwalletMain->UnlockAllCoins();
        return true;
    }

    UniValue outputs = params[1].get_array();
    for (unsigned int idx = 0; idx < outputs.size(); idx++)
    {
        const UniValue &output = outputs[idx];
        COutPoint outpt;
        if (!output.isObject())
        {
            string outpointhash = output.get_str();
            if (!IsHex(outpointhash))
                throw JSONRPCError(RPC_INVALID_PARAMETER, "Invalid parameter, expected hex outpoint or object");
            outpt = COutPoint(uint256S(outpointhash));
        }
        else
        {
            const UniValue &o = output.get_obj();

            RPCTypeCheckObj(o, {{"txidem", UniValue::VSTR}, {"vout", UniValue::VNUM}});

            string txidem = find_value(o, "txidem").get_str();
            if (!IsHex(txidem))
                throw JSONRPCError(RPC_INVALID_PARAMETER, "Invalid parameter, expected hex txidem");

            int nOutput = find_value(o, "vout").get_int();
            if (nOutput < 0)
                throw JSONRPCError(RPC_INVALID_PARAMETER, "Invalid parameter, vout must be positive");

            outpt = COutPoint(uint256S(txidem), nOutput);
        }

        if (fUnlock)
            pwalletMain->UnlockCoin(outpt);
        else
            pwalletMain->LockCoin(outpt);
    }

    return true;
}

UniValue listlockunspent(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 0)
        throw runtime_error("listlockunspent\n"
                            "\nReturns list of temporarily unspendable outputs.\n"
                            "See the lockunspent call to lock and unlock transactions for spending.\n"
                            "\nResult:\n"
                            "[\n"
                            "  {\n"
                            "    \"txid\" : \"transactionid\",     (string) The transaction id locked\n"
                            "    \"vout\" : n                      (numeric) The vout value\n"
                            "  }\n"
                            "  ,...\n"
                            "]\n"
                            "\nExamples:\n"
                            "\nList the unspent transactions\n" +
                            HelpExampleCli("lockunspent", "") + "\nLock an unspent transaction\n" +
                            HelpExampleCli("lockunspent", "false "
                                                          "\"[{\\\"txid\\\":"
                                                          "\\\"a08e6907dbbd3d809776dbfc5d82e371b764ed838b5655e72f463568"
                                                          "df1aadf0\\\",\\\"vout\\\":1}]\"") +
                            "\nList the locked transactions\n" + HelpExampleCli("listlockunspent", "") +
                            "\nUnlock the transaction again\n" +
                            HelpExampleCli("lockunspent", "true "
                                                          "\"[{\\\"txid\\\":"
                                                          "\\\"a08e6907dbbd3d809776dbfc5d82e371b764ed838b5655e72f463568"
                                                          "df1aadf0\\\",\\\"vout\\\":1}]\"") +
                            "\nAs a json rpc call\n" + HelpExampleRpc("listlockunspent", ""));

    LOCK(pwalletMain->cs_wallet);

    vector<COutPoint> vOutpts;
    pwalletMain->ListLockedCoins(vOutpts);

    UniValue ret(UniValue::VARR);

    for (COutPoint &outpt : vOutpts)
    {
        UniValue o(UniValue::VOBJ);

        o.pushKV("outpoint", outpt.GetHex());
        ret.push_back(o);
    }

    return ret;
}

UniValue getwalletinfo(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() != 0)
        throw runtime_error(
            "getwalletinfo\n"
            "Returns an object containing various wallet state info.\n"
            "\nResult:\n"
            "{\n"
            "  \"walletversion\": xxxxx,     (numeric) the wallet version\n"
            "  \"syncblock\": string,         (hex hash) the wallet is synced to this block\n"
            "  \"syncheight\": xxxx,         (numeric) the wallet is synced to this height\n"
            "  \"balance\": xxxxxxx,         (numeric) the total confirmed balance of the wallet in " +
            CURRENCY_UNIT +
            "\n"
            "  \"unconfirmed_balance\": xxx, (numeric) the total unconfirmed balance of the wallet in " +
            CURRENCY_UNIT +
            "\n"
            "  \"immature_balance\": xxxxxx, (numeric) the total immature balance of the wallet in " +
            CURRENCY_UNIT +
            "\n"
            "  \"watchonly_balance\": xxxxxxx,         (numeric) the total watchonly confirmed balance of the wallet "
            "in " +
            CURRENCY_UNIT +
            "\n"
            "  \"unconfirmed_watchonly_balance\": xxx, (numeric) the total watchonly unconfirmed balance of the wallet "
            "in " +
            CURRENCY_UNIT +
            "\n"
            "  \"immature_watchonly_balance\": xxxxxx, (numeric) the total watchonly immature balance of the wallet "
            "in " +
            CURRENCY_UNIT +
            "\n"
            "  \"txcount\": xxxxxxx,         (numeric) the total number of transactions in the wallet\n"
            "  \"unspentcount\": xxxxxxx,    (numeric) the total number of unspent coins in wallet\n"
            "  \"keypoololdest\": xxxxxx,    (numeric) the timestamp (seconds since GMT epoch) of the "
            "oldest pre-generated key in the key pool\n"
            "  \"keypoolsize\": xxxx,        (numeric) how many new keys are pre-generated\n"
            "  \"unlocked_until\": ttt,      (numeric) the timestamp in seconds since epoch (midnight "
            "Jan 1 1970 GMT) that the wallet is unlocked for transfers, or 0 if the wallet is locked\n"
            "  \"paytxfee\": xxxx,           (numeric) the transaction fee configuration, set in " +
            "sat/KB\n"
            "  \"hdmasterkeyid\": \"<hash160>\", (hex string) the Hash160 of the hd master pubkey\n"
            "}\n"
            "\nExamples:\n" +
            HelpExampleCli("getwalletinfo", "") + HelpExampleRpc("getwalletinfo", ""));

    // I am locking main here to ensure that we accurately report what block the wallet is synced to
    LOCK2(cs_main, pwalletMain->cs_wallet);

    UniValue obj(UniValue::VOBJ);
    obj.pushKV("walletversion", pwalletMain->GetVersion());
    obj.pushKV("syncblock", chainActive.Tip()->header.GetHash().ToString());
    obj.pushKV("syncheight", chainActive.Tip()->height());
    obj.pushKV("balance", ValueFromAmount(pwalletMain->GetBalance()));
    obj.pushKV("unconfirmed_balance", ValueFromAmount(pwalletMain->GetUnconfirmedBalance()));
    obj.pushKV("immature_balance", ValueFromAmount(pwalletMain->GetImmatureBalance()));
    obj.pushKV("watchonly_balance", ValueFromAmount(pwalletMain->GetWatchOnlyBalance()));
    obj.pushKV("unconfirmed_watchonly_balance", ValueFromAmount(pwalletMain->GetUnconfirmedWatchOnlyBalance()));
    obj.pushKV("immature_watchonly_balance", ValueFromAmount(pwalletMain->GetImmatureWatchOnlyBalance()));
    obj.pushKV("txcount", (int)pwalletMain->mapWallet.size() / 3);

    // We need to count each coin because the unspent map may still contain a few
    // already spent coins which havn't been cleared yet.
    uint64_t nCountUnspent = 0;
    for (const auto &coin : pwalletMain->mapWalletUnspent)
    {
        if (!pwalletMain->IsSpent(coin.first))
        {
            // This also includes watch-only coins.
            nCountUnspent++;
        }
    }
    obj.pushKV("unspentcount", nCountUnspent);

    obj.pushKV("keypoololdest", pwalletMain->GetOldestKeyPoolTime());
    obj.pushKV("keypoolsize", (int)pwalletMain->GetKeyPoolSize());
    if (pwalletMain->IsCrypted())
        obj.pushKV("unlocked_until", nWalletUnlockTime);
    obj.pushKV("paytxfee", CFeeRate(payTxFeeTweak.Value()).GetFeePerK());
    CKeyID masterKeyID = pwalletMain->GetHDChain().masterKeyID;
    if (!masterKeyID.IsNull())
        obj.pushKV("hdmasterkeyid", masterKeyID.GetHex());
    return obj;
}

UniValue resendwallettransactions(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() != 0)
        throw runtime_error("resendwallettransactions\n"
                            "Immediately re-broadcast unconfirmed wallet transactions to all peers.\n"
                            "Intended only for testing; the wallet code periodically re-broadcasts\n"
                            "automatically.\n"
                            "Returns array of transaction ids that were re-broadcast.\n");

    LOCK(pwalletMain->cs_wallet);

    std::vector<uint256> txids = pwalletMain->ResendWalletTransactionsBefore(GetTime());
    UniValue result(UniValue::VARR);
    for (const uint256 &txid : txids)
    {
        result.push_back(txid.ToString());
    }
    return result;
}

UniValue listunspent(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() > 3)
        throw runtime_error(
            "listunspent ( minconf maxconf  [\"address\",...] )\n"
            "\nReturns array of unspent transaction outputs\n"
            "with between minconf and maxconf (inclusive) confirmations.\n"
            "Optionally filter to only include txouts paid to specified addresses.\n"
            "Results are an array of Objects, each of which has:\n"
            "{txid, vout, scriptPubKey, amount, confirmations}\n"
            "\nArguments:\n"
            "1. minconf          (numeric, optional, default=1) The minimum confirmations to filter\n"
            "2. maxconf          (numeric, optional, default=9999999) The maximum confirmations to filter\n"
            "3. \"addresses\"    (string) A json array of addresses to filter\n"
            "    [\n"
            "      \"address\"   (string) address\n"
            "      ,...\n"
            "    ]\n"
            "\nResult\n"
            "[                   (array of json object)\n"
            "  {\n"
            "    \"outpoint\" : \"hash\"     (string) prevout identifier \n"
            "    \"txid\" : \"txid\",        (string) the transaction id \n"
            "    \"txidem\" : \"txidem\",    (string) the transaction idem \n"
            "    \"vout\" : n,               (numeric) the vout value\n"
            "    \"address\" : \"address\",  (string) the address\n"
            "    \"account\" : \"account\",  (string) DEPRECATED. The associated account, or \"\" for the default "
            "account\n"
            "    \"scriptPubKey\" : \"key\", (string) the script key\n"
            "    \"amount\" : x.xxx,         (numeric) the transaction amount in " +
            CURRENCY_UNIT +
            "\n"
            "    \"confirmations\" : n       (numeric) The number of confirmations\n"
            "  }\n"
            "  ,...\n"
            "]\n"

            "\nExamples\n" +
            HelpExampleCli("listunspent", "") +
            HelpExampleCli("listunspent", "6 9999999 "
                                          "\"[\\\"nexa:qqqdf524crwrn659hattd40jn5ykz32xey7z6n5z28\\\","
                                          "\\\"1LtvqCaApEdUGFkpKMM4MstjcaL4dKg8SP\\\"]\"") +
            HelpExampleRpc("listunspent", "6, 9999999 "
                                          "\"[\\\"nexa:qqqdf524crwrn659hattd40jn5ykz32xey7z6n5z28\\\","
                                          "\\\"1LtvqCaApEdUGFkpKMM4MstjcaL4dKg8SP\\\"]\""));

    RPCTypeCheck(params, {UniValue::VNUM, UniValue::VNUM, UniValue::VARR});

    int nMinDepth = 1;
    if (params.size() > 0)
        nMinDepth = params[0].get_int();

    int nMaxDepth = 9999999;
    if (params.size() > 1)
        nMaxDepth = params[1].get_int();

    set<CTxDestination> destinations;
    if (params.size() > 2)
    {
        UniValue inputs = params[2].get_array();
        for (unsigned int idx = 0; idx < inputs.size(); idx++)
        {
            const UniValue &input = inputs[idx];
            CTxDestination address = DecodeDestination(input.get_str());
            if (!IsValidDestination(address))
                throw JSONRPCError(RPC_INVALID_ADDRESS_OR_KEY, string("Invalid coin address: ") + input.get_str());
            if (destinations.count(address))
                throw JSONRPCError(
                    RPC_INVALID_PARAMETER, string("Invalid parameter, duplicated address: ") + input.get_str());
            destinations.insert(address);
        }
    }

    UniValue results(UniValue::VARR);
    vector<COutput> vecOutputs;
    assert(pwalletMain != nullptr);
    // Nothing relies on cs_main, but by locking it here, we ensure that a chain reorg doesn't
    // cause AvailableCoins to give inconsistent results
    LOCK2(cs_main, pwalletMain->cs_wallet);
    pwalletMain->AvailableCoins(vecOutputs, nullptr, true);
    for (const COutput &out : vecOutputs)
    {
        int depth = out.GetDepthInMainChain();
        if (depth < nMinDepth || depth > nMaxDepth)
            continue;

        if (destinations.size())
        {
            CTxDestination address;
            if (!ExtractDestination(out.tx->vout[out.i].scriptPubKey, address))
                continue;

            if (!destinations.count(address))
                continue;
        }

        CAmount nValue = out.tx->vout[out.i].nValue;
        const CScript &pk = out.tx->vout[out.i].scriptPubKey;
        UniValue entry(UniValue::VOBJ);
        entry.pushKV("outpoint", out.GetOutPoint().hash.GetHex());
        entry.pushKV("txid", out.tx->GetId().GetHex());
        entry.pushKV("txidem", out.tx->GetIdem().GetHex());
        entry.pushKV("vout", out.i);
        CTxDestination address;
        if (ExtractDestination(out.tx->vout[out.i].scriptPubKey, address))
        {
            entry.pushKV("address", EncodeDestination(address));
            if (pwalletMain->mapAddressBook.count(address))
                entry.pushKV("account", pwalletMain->mapAddressBook[address].name);
        }
        entry.pushKV("scriptPubKey", HexStr(pk.begin(), pk.end()));
        if (pk.type == ScriptType::TEMPLATE)
            entry.pushKV("scriptType", "template");
        else if (pk.type == ScriptType::SATOSCRIPT)
            entry.pushKV("scriptType", "satoscript");
        else if (pk.type == ScriptType::PUSH_ONLY)
            entry.pushKV("scriptType", "pushonly");
        if (pk.IsPayToScriptHash())
        {
            CTxDestination address2;
            if (ExtractDestination(pk, address2))
            {
                const CScriptID &hash = std::get<CScriptID>(address2);
                CScript redeemScript;
                if (pwalletMain->GetCScript(hash, redeemScript))
                    entry.pushKV("redeemScript", HexStr(redeemScript.begin(), redeemScript.end()));
            }
        }
        entry.pushKV("satoshi", UniValue(nValue));
        entry.pushKV("amount", ValueFromAmount(nValue));
        entry.pushKV("confirmations", depth);
        entry.pushKV("spendable", out.spendable());
        results.push_back(entry);
    }

    return results;
}

UniValue fundrawtransaction(const UniValue &params, bool fHelp)
{
    if (!EnsureWalletIsAvailable(fHelp))
        return NullUniValue;

    if (fHelp || params.size() < 1 || params.size() > 2)
        throw runtime_error(
            "fundrawtransaction \"hexstring\" includeWatching\n"
            "\nAdd inputs to a transaction until it has enough in value to meet its out value.\n"
            "This will not modify existing inputs, and will add one change output to the outputs.\n"
            "Note that inputs which were signed may need to be resigned after completion since in/outputs have been "
            "added.\n"
            "The inputs added will not be signed, use signrawtransaction for that.\n"
            "Note that all existing inputs must have their previous output transaction be in the wallet.\n"
            "Note that all inputs selected must be of standard form and P2SH scripts must be"
            "in the wallet using importaddress or addmultisigaddress (to calculate fees).\n"
            "Only pay-to-pubkey, multisig, and P2SH versions thereof are currently supported for watch-only\n"
            "\nArguments:\n"
            "1. \"hexstring\"     (string, required) The hex string of the raw transaction\n"
            "2. includeWatching (boolean, optional, default false) Also select inputs which are watch only\n"
            "\nResult:\n"
            "{\n"
            "  \"hex\":       \"value\", (string)  The resulting raw transaction (hex-encoded string)\n"
            "  \"fee\":       n,         (numeric) Fee the resulting transaction pays\n"
            "  \"changepos\": n          (numeric) The position of the added change output, or -1\n"
            "}\n"
            "\"hex\"             \n"
            "\nExamples:\n"
            "\nCreate a transaction with no inputs\n" +
            HelpExampleCli("createrawtransaction", "\"[]\" \"{\\\"myaddress\\\":10.01}\"") +
            "\nAdd sufficient unsigned inputs to meet the output value\n" +
            HelpExampleCli("fundrawtransaction", "\"rawtransactionhex\"") + "\nSign the transaction\n" +
            HelpExampleCli("signrawtransaction", "\"fundedtransactionhex\"") + "\nSend the transaction\n" +
            HelpExampleCli("sendrawtransaction", "\"signedtransactionhex\""));

    RPCTypeCheck(params, {UniValue::VSTR, UniValue::VBOOL});

    // parse hex string from parameter
    CTransaction origTx;
    if (!DecodeHexTx(origTx, params[0].get_str()))
        throw JSONRPCError(RPC_DESERIALIZATION_ERROR, "TX decode failed");

    if (origTx.vout.size() == 0)
        throw JSONRPCError(RPC_INVALID_PARAMETER, "TX must have at least one output");

    bool includeWatching = false;
    if (params.size() > 1)
        includeWatching = params[1].get_bool();

    CMutableTransaction tx(origTx);
    CAmount nFee;
    string strFailReason;
    int nChangePos = -1;
    if (!pwalletMain->FundTransaction(tx, nFee, nChangePos, strFailReason, includeWatching))
        throw JSONRPCError(RPC_INTERNAL_ERROR, strFailReason);

    UniValue result(UniValue::VOBJ);
    result.pushKV("hex", EncodeHexTx(tx));
    result.pushKV("changepos", nChangePos);
    result.pushKV("fee", ValueFromAmount(nFee));
    result.pushKV("complete", true); // Match signrawtransaction RPC
    result.pushKV("txid", tx.GetId().GetHex());
    result.pushKV("txidem", tx.GetIdem().GetHex());

    return result;
}

extern UniValue dumpprivkey(const UniValue &params, bool fHelp); // in rpcdump.cpp
extern UniValue importprivkey(const UniValue &params, bool fHelp);
extern UniValue importprivatekeys(const UniValue &params, bool fHelp);
extern UniValue importaddress(const UniValue &params, bool fHelp);
extern UniValue importaddresses(const UniValue &params, bool fHelp);
extern UniValue importpubkey(const UniValue &params, bool fHelp);
extern UniValue dumpwallet(const UniValue &params, bool fHelp);
extern UniValue importwallet(const UniValue &params, bool fHelp);
extern UniValue importprunedfunds(const UniValue &params, bool fHelp);
extern UniValue removeprunedfunds(const UniValue &params, bool fHelp);

extern UniValue dumptokenset(const UniValue &params, bool fHelp);

/* clang-format off */
static const CRPCCommand commands[] = {
    //  category              name                        actor (function)           okSafeMode
    //  --------------------- ------------------------    -----------------------    ----------
    {"rawtransactions",       "fundrawtransaction",       &fundrawtransaction,       false},
    {"hidden",                "resendwallettransactions", &resendwallettransactions, true},
    {"wallet",                "abandontransaction",       &abandontransaction,       false},
    {"wallet",                "addmultisigaddress",       &addmultisigaddress,       true},
    {"wallet",                "backupwallet",             &backupwallet,             true},
    {"wallet",                "consolidate",              &consolidate,              true},
    {"wallet",                "dumpprivkey",              &dumpprivkey,              true},
    {"wallet",                "dumpwallet",               &dumpwallet,               true},
    {"wallet",                "encryptwallet",            &encryptwallet,            true},
    {"wallet",                "getaccountaddress",        &getaccountaddress,        true},
    {"wallet",                "getaccount",               &getaccount,               true},
    {"wallet",                "getaddressesbyaccount",    &getaddressesbyaccount,    true},
    {"wallet",                "getbalance",               &getbalance,               false},
    {"wallet",                "getnewaddress",            &getnewaddress,            true},
    {"wallet",                "getrawchangeaddress",      &getrawchangeaddress,      true},
    {"wallet",                "getreceivedbyaccount",     &getreceivedbyaccount,     false},
    {"wallet",                "getreceivedbyaddress",     &getreceivedbyaddress,     false},
    {"wallet",                "gettransaction",           &gettransaction,           false},
    {"wallet",                "getunconfirmedbalance",    &getunconfirmedbalance,    false},
    {"wallet",                "getwalletinfo",            &getwalletinfo,            false},
    {"wallet",                "importprivkey",            &importprivkey,            true},
    {"wallet",                "importprivatekeys",        &importprivatekeys,        true},
    {"wallet",                "importwallet",             &importwallet,             true},
    {"wallet",                "importaddress",            &importaddress,            true},
    {"wallet",                "importaddresses",          &importaddresses,          true},
    {"wallet",                "importprunedfunds",        &importprunedfunds,        true},
    {"wallet",                "importpubkey",             &importpubkey,             true},
    {"wallet",                "keypoolrefill",            &keypoolrefill,            true},
    {"wallet",                "listaccounts",             &listaccounts,             false},
    {"wallet",                "listaddressgroupings",     &listaddressgroupings,     false},
    {"wallet",                "listactiveaddresses",      &listactiveaddresses,      false},
    {"wallet",                "listlockunspent",          &listlockunspent,          false},
    {"wallet",                "listreceivedbyaccount",    &listreceivedbyaccount,    false},
    {"wallet",                "listreceivedbyaddress",    &listreceivedbyaddress,    false},
    {"wallet",                "listsinceblock",           &listsinceblock,           false},
    {"wallet",                "listtransactions",         &listtransactions,         false},
    {"wallet",                "listtransactionsfrom",     &listtransactionsfrom,     false},
    {"wallet",                "listunspent",              &listunspent,              false},
    {"wallet",                "lockunspent",              &lockunspent,              true},
    {"wallet",                "move",                     &movecmd,                  false},
    {"wallet",                "sendfrom",                 &sendfrom,                 false},
    {"wallet",                "sendmany",                 &sendmany,                 false},
    {"wallet",                "sendtoaddress",            &sendtoaddress,            false},
    {"wallet",                "setaccount",               &setaccount,               true},
    {"wallet",                "signmessage",              &signmessage,              true},
    {"wallet",                "signdata",                 &signdata,                 true},
    {"wallet",                "walletlock",               &walletlock,               true},
    {"wallet",                "walletpassphrasechange",   &walletpassphrasechange,   true},
    {"wallet",                "walletpassphrase",         &walletpassphrase,         true},
    {"wallet",                "removeprunedfunds",        &removeprunedfunds,        true},
    {"token",                 "dumptokenset",             &dumptokenset,             true},
};

void RegisterWalletRPCCommands(CRPCTable &table)
{
    for (auto cmd : commands)
        table.appendCommand(cmd);
}
