// Copyright (c) 2009-2010 Satoshi Nakamoto
// Copyright (c) 2009-2015 The Bitcoin Core developers
// Copyright (c) 2015-2022 The Bitcoin Unlimited developers
// Copyright (c) 2017 The Bitcoin developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include "wallet/walletdb.h"

#include "base58.h"
#include "consensus/tx_verify.h"
#include "consensus/validation.h"
#include "dstencode.h"
#include "fs.h"
#include "init.h"
#include "main.h" // For CheckTransaction
#include "protocol.h"
#include "serialize.h"
#include "sync.h"
#include "util.h"
#include "utiltime.h"
#include "utiltranslate.h"
#include "wallet/wallet.h"

#include <boost/scoped_ptr.hpp>
#include <boost/version.hpp>
#include <fstream>
#include <thread>

using namespace std;

static uint64_t nAccountingEntryNumber = 0;

//
// CWalletDB
//

bool CWalletDB::WriteTokenTracker(const CGroupTokenID &id, const std::string &strTokenTicker)
{
    nWalletDBUpdated++;
    return Write(std::make_pair(std::string("tokentracker"), id), strTokenTicker);
}

bool CWalletDB::EraseTokenTracker(const CGroupTokenID &id)
{
    nWalletDBUpdated++;
    return Erase(std::make_pair(std::string("tokentracker"), id));
}

bool CWalletDB::WriteName(const CTxDestination &address, const std::string &strName)
{
    if (!IsValidDestination(address))
        return false;

    nWalletDBUpdated++;
    return Write(std::make_pair(std::string("name"), EncodeLegacyAddr(address, Params())), strName);
}

bool CWalletDB::EraseName(const CTxDestination &address)
{
    // This should only be used for sending addresses, never for receiving
    // addresses, receiving addresses must always have an address book entry if
    // they're not change return.
    if (!IsValidDestination(address))
        return false;

    nWalletDBUpdated++;
    return Erase(std::make_pair(std::string("name"), EncodeLegacyAddr(address, Params())));
}

bool CWalletDB::WritePurpose(const CTxDestination &address, const std::string &strPurpose)
{
    if (!IsValidDestination(address))
        return false;

    nWalletDBUpdated++;
    return Write(std::make_pair(std::string("purpose"), EncodeLegacyAddr(address, Params())), strPurpose);
}

bool CWalletDB::ErasePurpose(const CTxDestination &address)
{
    if (!IsValidDestination(address))
        return false;

    nWalletDBUpdated++;
    return Erase(std::make_pair(std::string("purpose"), EncodeLegacyAddr(address, Params())));
}

bool CWalletDB::WriteTx(const CWalletTx &wtx)
{
    nWalletDBUpdated++;
    return Write(std::make_pair(std::string("txidem"), wtx.GetIdem()), wtx);
}

bool CWalletDB::EraseTx(uint256 idem)
{
    nWalletDBUpdated++;
    return Erase(std::make_pair(std::string("txidem"), idem));
}

bool CWalletDB::WriteKey(const CPubKey &vchPubKey, const CPrivKey &vchPrivKey, const CKeyMetadata &keyMeta)
{
    nWalletDBUpdated++;

    if (!Write(std::make_pair(std::string("keymeta"), vchPubKey), keyMeta, false))
        return false;

    // hash pubkey/privkey to accelerate wallet load
    std::vector<unsigned char> vchKey;
    vchKey.reserve(vchPubKey.size() + vchPrivKey.size());
    vchKey.insert(vchKey.end(), vchPubKey.begin(), vchPubKey.end());
    vchKey.insert(vchKey.end(), vchPrivKey.begin(), vchPrivKey.end());

    return Write(std::make_pair(std::string("key"), vchPubKey),
        std::make_pair(vchPrivKey, Hash(vchKey.begin(), vchKey.end())), false);
}

bool CWalletDB::WriteCryptedKey(const CPubKey &vchPubKey,
    const std::vector<unsigned char> &vchCryptedSecret,
    const CKeyMetadata &keyMeta)
{
    const bool fEraseUnencryptedKey = true;
    nWalletDBUpdated++;

    if (!Write(std::make_pair(std::string("keymeta"), vchPubKey), keyMeta))
        return false;

    if (!Write(std::make_pair(std::string("ckey"), vchPubKey), vchCryptedSecret, false))
        return false;
    if (fEraseUnencryptedKey)
    {
        Erase(std::make_pair(std::string("key"), vchPubKey));
        Erase(std::make_pair(std::string("wkey"), vchPubKey));
    }
    return true;
}

bool CWalletDB::WriteMasterKey(unsigned int nID, const CMasterKey &kMasterKey)
{
    nWalletDBUpdated++;
    return Write(std::make_pair(std::string("mkey"), nID), kMasterKey, true);
}

bool CWalletDB::WriteCScript(const uint160 &hash, const CScript &redeemScript)
{
    nWalletDBUpdated++;
    return Write(std::make_pair(std::string("cscript"), hash), *(const CScriptBase *)(&redeemScript), false);
}

bool CWalletDB::WriteWatchOnly(const CScript &dest)
{
    nWalletDBUpdated++;
    return Write(std::make_pair(std::string("watchs"), *(const CScriptBase *)(&dest)), '1');
}

bool CWalletDB::EraseWatchOnly(const CScript &dest)
{
    nWalletDBUpdated++;
    return Erase(std::make_pair(std::string("watchs"), *(const CScriptBase *)(&dest)));
}

bool CWalletDB::WriteBestBlock(const CBlockLocator &locator)
{
    nWalletDBUpdated++;
    // Write empty block locator so versions that require a merkle branch automatically rescan
    Write(std::string("bestblock"), CBlockLocator());
    return Write(std::string("bestblock_nomerkle"), locator);
}

bool CWalletDB::ReadBestBlock(CBlockLocator &locator)
{
    if (Read(std::string("bestblock"), locator) && !locator.vHave.empty())
        return true;
    return Read(std::string("bestblock_nomerkle"), locator);
}

bool CWalletDB::WriteOrderPosNext(int64_t nOrderPosNext)
{
    nWalletDBUpdated++;
    return Write(std::string("orderposnext"), nOrderPosNext);
}

bool CWalletDB::WriteDefaultKey(const CPubKey &vchPubKey)
{
    nWalletDBUpdated++;
    return Write(std::string("defaultkey"), vchPubKey);
}

bool CWalletDB::ReadPool(int64_t nPool, CKeyPool &keypool)
{
    return Read(std::make_pair(std::string("pool"), nPool), keypool);
}

bool CWalletDB::WritePool(int64_t nPool, const CKeyPool &keypool)
{
    nWalletDBUpdated++;
    return Write(std::make_pair(std::string("pool"), nPool), keypool);
}

bool CWalletDB::ErasePool(int64_t nPool)
{
    nWalletDBUpdated++;
    return Erase(std::make_pair(std::string("pool"), nPool));
}

bool CWalletDB::WriteMinVersion(int nVersion) { return Write(std::string("minversion"), nVersion); }
bool CWalletDB::ReadAccount(const string &strAccount, CAccount &account)
{
    account.SetNull();
    return Read(make_pair(string("acc"), strAccount), account);
}

bool CWalletDB::WriteAccount(const string &strAccount, const CAccount &account)
{
    return Write(make_pair(string("acc"), strAccount), account);
}

bool CWalletDB::WriteAccountingEntry(const uint64_t nAccEntryNum, const CAccountingEntry &acentry)
{
    return Write(std::make_pair(std::string("acentry"), std::make_pair(acentry.strAccount, nAccEntryNum)), acentry);
}

bool CWalletDB::WriteAccountingEntry_Backend(const CAccountingEntry &acentry)
{
    return WriteAccountingEntry(++nAccountingEntryNumber, acentry);
}

CAmount CWalletDB::GetAccountCreditDebit(const string &strAccount)
{
    list<CAccountingEntry> entries;
    ListAccountCreditDebit(strAccount, entries);

    CAmount nCreditDebit = 0;
    for (const CAccountingEntry &entry : entries)
    {
        nCreditDebit += entry.nCreditDebit;
    }

    return nCreditDebit;
}

void CWalletDB::ListAccountCreditDebit(const string &strAccount, list<CAccountingEntry> &entries)
{
    bool fAllAccounts = (strAccount == "*");

    Dbc *pcursor = GetCursor();
    if (!pcursor)
        throw runtime_error("CWalletDB::ListAccountCreditDebit(): cannot create DB cursor");
    unsigned int fFlags = DB_SET_RANGE;
    while (true)
    {
        // Read next record
        CDataStream ssKey(SER_DISK, CLIENT_VERSION);
        if (fFlags == DB_SET_RANGE)
            ssKey << std::make_pair(
                std::string("acentry"), std::make_pair((fAllAccounts ? string("") : strAccount), uint64_t(0)));
        CDataStream ssValue(SER_DISK, CLIENT_VERSION);
        int ret = ReadAtCursor(pcursor, ssKey, ssValue, fFlags);
        fFlags = DB_NEXT;
        if (ret == DB_NOTFOUND)
            break;
        else if (ret != 0)
        {
            pcursor->close();
            throw runtime_error("CWalletDB::ListAccountCreditDebit(): error scanning DB");
        }

        // Unserialize
        string strType;
        ssKey >> strType;
        if (strType != "acentry")
            break;
        CAccountingEntry acentry;
        ssKey >> acentry.strAccount;
        if (!fAllAccounts && acentry.strAccount != strAccount)
            break;

        ssValue >> acentry;
        ssKey >> acentry.nEntryNo;
        entries.push_back(acentry);
    }

    pcursor->close();
}


class CWalletScanState
{
public:
    unsigned int nKeys;
    unsigned int nCKeys;
    unsigned int nKeyMeta;
    bool fIsEncrypted;
    int nFileVersion;
    std::vector<CWalletTxRef> tx;

    CWalletScanState()
    {
        nKeys = nCKeys = nKeyMeta = 0;
        fIsEncrypted = false;
        nFileVersion = 0;
    }
};

bool ReadKeyValue(CWallet *pwallet,
    CDataStream &ssKey,
    CDataStream &ssValue,
    CWalletScanState &wss,
    string &strType,
    string &strErr)
{
    try
    {
        // Unserialize
        // Taking advantage of the fact that pair serialization
        // is just the two items serialized one after the other
        ssKey >> strType;
        if (strType == "name")
        {
            string strAddress;
            ssKey >> strAddress;
            ssValue >> pwallet->mapAddressBook[DecodeDestination(strAddress)].name;
        }
        else if (strType == "purpose")
        {
            std::string strAddress;
            ssKey >> strAddress;
            ssValue >> pwallet->mapAddressBook[DecodeDestination(strAddress)].purpose;
        }
        else if (strType == "txidem")
        {
            uint256 hash;
            ssKey >> hash;
            CWalletTxRef wtx = std::make_shared<CWalletTx>();
            ssValue >> *wtx;
            CValidationState state;
            if (!(CheckTransaction(std::static_pointer_cast<const CTransaction>(wtx), state) &&
                    (wtx->GetIdem() == hash) && state.IsValid()))
                return false;

            assert(wtx->nOrderPos != -1);
            wss.tx.push_back(wtx); // Defer adding all transactions until the end
        }
        else if (strType == "acentry")
        {
            string strAccount;
            ssKey >> strAccount;
            uint64_t nNumber;
            ssKey >> nNumber;
            if (nNumber > nAccountingEntryNumber)
                nAccountingEntryNumber = nNumber;

            CAccountingEntry acentry;
            ssValue >> acentry;
            assert(acentry.nOrderPos != -1);
        }
        else if (strType == "watchs")
        {
            CScript script;
            ssKey >> *(CScriptBase *)(&script);
            char fYes;
            ssValue >> fYes;
            if (fYes == '1')
                pwallet->LoadWatchOnly(script);

            // Watch-only addresses have no birthday information for now,
            // so set the wallet birthday to the beginning of time.
            pwallet->nTimeFirstKey = 1;
        }
        else if (strType == "key" || strType == "wkey")
        {
            CPubKey vchPubKey;
            ssKey >> vchPubKey;
            if (!vchPubKey.IsValid())
            {
                strErr = "Error reading wallet database: CPubKey corrupt";
                return false;
            }
            CKey key;
            CPrivKey pkey;
            uint256 hash;

            if (strType == "key")
            {
                wss.nKeys++;
                ssValue >> pkey;
            }
            else
            {
                CWalletKey wkey;
                ssValue >> wkey;
                pkey = wkey.vchPrivKey;
            }

            // Old wallets store keys as "key" [pubkey] => [privkey]
            // ... which was slow for wallets with lots of keys, because the public key is re-derived from the private
            // key
            // using EC operations as a checksum.
            // Newer wallets store keys as "key"[pubkey] => [privkey][hash(pubkey,privkey)], which is much faster while
            // remaining backwards-compatible.
            try
            {
                ssValue >> hash;
            }
            catch (...)
            {
            }

            bool fSkipCheck = false;

            if (!hash.IsNull())
            {
                // hash pubkey/privkey to accelerate wallet load
                std::vector<unsigned char> vchKey;
                vchKey.reserve(vchPubKey.size() + pkey.size());
                vchKey.insert(vchKey.end(), vchPubKey.begin(), vchPubKey.end());
                vchKey.insert(vchKey.end(), pkey.begin(), pkey.end());

                if (Hash(vchKey.begin(), vchKey.end()) != hash)
                {
                    strErr = "Error reading wallet database: CPubKey/CPrivKey corrupt";
                    return false;
                }

                fSkipCheck = true;
            }
            if (!key.Load(pkey, vchPubKey, fSkipCheck))
            {
                strErr = "Error reading wallet database: CPrivKey corrupt";
                return false;
            }

            if (!pwallet->LoadKey(key, vchPubKey))
            {
                strErr = "Error reading wallet database: LoadKey failed";
                return false;
            }
        }
        else if (strType == "mkey")
        {
            unsigned int nID;
            ssKey >> nID;
            CMasterKey kMasterKey;
            ssValue >> kMasterKey;
            if (pwallet->mapMasterKeys.count(nID) != 0)
            {
                strErr = strprintf("Error reading wallet database: duplicate CMasterKey id %u", nID);
                return false;
            }
            pwallet->mapMasterKeys[nID] = kMasterKey;
            if (pwallet->nMasterKeyMaxID < nID)
                pwallet->nMasterKeyMaxID = nID;
        }
        else if (strType == "ckey")
        {
            CPubKey vchPubKey;
            ssKey >> vchPubKey;
            if (!vchPubKey.IsValid())
            {
                strErr = "Error reading wallet database: CPubKey corrupt";
                return false;
            }
            vector<unsigned char> vchPrivKey;
            ssValue >> vchPrivKey;
            wss.nCKeys++;

            if (!pwallet->LoadCryptedKey(vchPubKey, vchPrivKey))
            {
                strErr = "Error reading wallet database: LoadCryptedKey failed";
                return false;
            }
            wss.fIsEncrypted = true;
        }
        else if (strType == "keymeta")
        {
            CPubKey vchPubKey;
            ssKey >> vchPubKey;
            CKeyMetadata keyMeta;
            ssValue >> keyMeta;
            wss.nKeyMeta++;

            pwallet->LoadKeyMetadata(vchPubKey, keyMeta);

            // find earliest key creation time, as wallet birthday
            if (!pwallet->nTimeFirstKey || (keyMeta.nCreateTime < pwallet->nTimeFirstKey))
                pwallet->nTimeFirstKey = keyMeta.nCreateTime;
        }
        else if (strType == "defaultkey")
        {
            ssValue >> pwallet->vchDefaultKey;
        }
        else if (strType == "pool")
        {
            int64_t nIndex;
            ssKey >> nIndex;
            CKeyPool keypool;
            ssValue >> keypool;
            pwallet->setKeyPool.insert(nIndex);

            // If no metadata exists yet, create a default with the pool key's
            // creation time. Note that this may be overwritten by actually
            // stored metadata for that key later, which is fine.
            CKeyID keyid = keypool.vchPubKey.GetID();
            if (pwallet->mapKeyMetadata.count(keyid) == 0)
                pwallet->mapKeyMetadata[keyid] = CKeyMetadata(keypool.nTime);
        }
        else if (strType == "version")
        {
            ssValue >> wss.nFileVersion;
            if (wss.nFileVersion == 10300)
                wss.nFileVersion = 300;
        }
        else if (strType == "cscript")
        {
            uint160 hash;
            ssKey >> hash;
            CScript script;
            ssValue >> *(CScriptBase *)(&script);
            if (!pwallet->LoadCScript(script))
            {
                strErr = "Error reading wallet database: LoadCScript failed";
                return false;
            }
        }
        else if (strType == "orderposnext")
        {
            ssValue >> pwallet->nOrderPosNext;
        }
        else if (strType == "destdata")
        {
            std::string strAddress, strKey, strValue;
            ssKey >> strAddress;
            ssKey >> strKey;
            ssValue >> strValue;
            if (!pwallet->LoadDestData(DecodeDestination(strAddress), strKey, strValue))
            {
                strErr = "Error reading wallet database: LoadDestData failed";
                return false;
            }
        }
        else if (strType == "hdchain")
        {
            CHDChain chain;
            ssValue >> chain;
            if (!pwallet->SetHDChain(chain, true))
            {
                strErr = "Error reading wallet database: SetHDChain failed";
                return false;
            }
        }
        else if (strType == "tokentracker")
        {
            CGroupTokenID tokenId;
            std::string strTokenTicker;
            ssKey >> tokenId;
            ssValue >> strTokenTicker;
            pwallet->mapTokenTrackers.emplace(tokenId, strTokenTicker);
        }
    }
    catch (...)
    {
        return false;
    }
    return true;
}

static bool IsKeyType(string strType)
{
    return (strType == "key" || strType == "wkey" || strType == "mkey" || strType == "ckey");
}

DBErrors CWalletDB::LoadWallet(CWallet *pwallet)
{
    pwallet->vchDefaultKey = CPubKey();
    CWalletScanState wss;
    bool fNoncriticalErrors = false;
    DBErrors result = DB_LOAD_OK;

    LOCK(pwallet->cs_wallet);
    try
    {
        int nMinVersion = 0;
        if (Read((string) "minversion", nMinVersion))
        {
            if (nMinVersion > CLIENT_VERSION)
                return DB_TOO_NEW;
            pwallet->LoadMinVersion(nMinVersion);
        }

        // Get cursor
        Dbc *pcursor = GetCursor();
        if (!pcursor)
        {
            LOGA("Error getting wallet database cursor\n");
            return DB_CORRUPT;
        }

        while (true)
        {
            // Read next record
            CDataStream ssKey(SER_DISK, CLIENT_VERSION);
            CDataStream ssValue(SER_DISK, CLIENT_VERSION);
            int ret = ReadAtCursor(pcursor, ssKey, ssValue);
            if (ret == DB_NOTFOUND)
                break;
            else if (ret != 0)
            {
                LOGA("Error reading next record from wallet database\n");
                return DB_CORRUPT;
            }

            // Try to be tolerant of single corrupt records:
            string strType, strErr;
            if (!ReadKeyValue(pwallet, ssKey, ssValue, wss, strType, strErr))
            {
                // losing keys is considered a catastrophic error, anything else
                // we assume the user can live with:
                if (IsKeyType(strType))
                    result = DB_CORRUPT;
                else
                {
                    // Leave other errors alone, if we try to fix them we might make things worse.
                    fNoncriticalErrors = true; // ... but do warn the user there is something wrong.
                    if (strType == "txidem")
                        // Rescan if there is a bad transaction record:
                        SoftSetBoolArg("-rescan", true);
                }
            }
            if (!strErr.empty())
                LOGA("%s\n", strErr);
        }
        pcursor->close();
    }
    catch (const boost::thread_interrupted &)
    {
        throw;
    }
    catch (...)
    {
        result = DB_CORRUPT;
    }

    // Now that all keys and P2SH addresses are loaded, we can accurately determine
    // every output in these transactions that this wallet controls.  These are added individually
    // into mapWallet
    int count = 0;
    for (const CWalletTxRef &wtx : wss.tx)
    {
        pwallet->AddToWallet(wtx, true, nullptr);

        // For very large wallets occassionaly clear the any spent items from the unspent map
        if (count > 50000)
        {
            pwallet->ClearAllSpent();
            count = 0;
        }
        count++;
    }

    if (fNoncriticalErrors && result == DB_LOAD_OK)
        result = DB_NONCRITICAL_ERROR;

    // Any wallet corruption at all: skip any rewriting or
    // upgrading, we don't want to make it worse.
    if (result != DB_LOAD_OK)
        return result;

    LOGA("nFileVersion = %d\n", wss.nFileVersion);

    LOGA("Keys: %u plaintext, %u encrypted, %u w/ metadata, %u total\n", wss.nKeys, wss.nCKeys, wss.nKeyMeta,
        wss.nKeys + wss.nCKeys);

    // nTimeFirstKey is only reliable if all keys have metadata
    if ((wss.nKeys + wss.nCKeys) != wss.nKeyMeta)
        pwallet->nTimeFirstKey = 1; // 0 would be considered 'no value'

    // Rewrite encrypted wallets of versions 0.4.0 and 0.5.0rc:
    if (wss.fIsEncrypted && (wss.nFileVersion == 40000 || wss.nFileVersion == 50000))
        return DB_NEED_REWRITE;

    if (wss.nFileVersion < CLIENT_VERSION) // Update
        WriteVersion(CLIENT_VERSION);

    pwallet->laccentries.clear();
    ListAccountCreditDebit("*", pwallet->laccentries);
    for (CAccountingEntry &entry : pwallet->laccentries)
    {
        pwallet->wtxOrdered.insert(make_pair(entry.nOrderPos, CWallet::TxPair(nullptr, &entry)));
    }

    return result;
}

DBErrors CWalletDB::FindWalletTx(CWallet *pwallet, vector<uint256> &vTxHash, vector<CWalletTxRef> &vWtx)
{
    pwallet->vchDefaultKey = CPubKey();
    bool fNoncriticalErrors = false;
    DBErrors result = DB_LOAD_OK;

    LOCK(pwallet->cs_wallet);
    try
    {
        int nMinVersion = 0;
        if (Read((string) "minversion", nMinVersion))
        {
            if (nMinVersion > CLIENT_VERSION)
                return DB_TOO_NEW;
            pwallet->LoadMinVersion(nMinVersion);
        }

        // Get cursor
        Dbc *pcursor = GetCursor();
        if (!pcursor)
        {
            LOGA("Error getting wallet database cursor\n");
            return DB_CORRUPT;
        }

        while (true)
        {
            // Read next record
            CDataStream ssKey(SER_DISK, CLIENT_VERSION);
            CDataStream ssValue(SER_DISK, CLIENT_VERSION);
            int ret = ReadAtCursor(pcursor, ssKey, ssValue);
            if (ret == DB_NOTFOUND)
                break;
            else if (ret != 0)
            {
                LOGA("Error reading next record from wallet database\n");
                return DB_CORRUPT;
            }

            string strType;
            ssKey >> strType;
            if (strType == "txidem")
            {
                uint256 hash;
                ssKey >> hash;

                CWalletTxRef wtx = MakeWalletTxRef();
                ssValue >> *wtx;

                vTxHash.push_back(hash);
                vWtx.push_back(wtx);
            }
        }
        pcursor->close();
    }
    catch (const boost::thread_interrupted &)
    {
        throw;
    }
    catch (...)
    {
        result = DB_CORRUPT;
    }

    if (fNoncriticalErrors && result == DB_LOAD_OK)
        result = DB_NONCRITICAL_ERROR;

    return result;
}


DBErrors CWalletDB::ZapSelectTx(CWallet *pwallet, vector<uint256> &vTxHashIn, vector<uint256> &vTxHashOut)
{
    // build list of wallet TXs and hashes
    vector<CWalletTxRef> vWtx;
    {
        vector<uint256> vTxHash;
        DBErrors err = FindWalletTx(pwallet, vTxHash, vWtx);
        if (err != DB_LOAD_OK)
        {
            return err;
        }
    }

    std::sort(vWtx.begin(), vWtx.end(), [](CWalletTxRef &a, CWalletTxRef &b) { return a->GetIdem() < b->GetIdem(); });
    std::sort(vTxHashIn.begin(), vTxHashIn.end());

    // erase each matching wallet TX
    bool delerror = false;
    vector<uint256>::iterator it = vTxHashIn.begin();
    for (auto &tx : vWtx)
    {
        uint256 hash = tx->GetIdem();
        while (it < vTxHashIn.end() && (*it) < hash)
        {
            it++;
        }
        if (it == vTxHashIn.end())
        {
            break;
        }
        else if ((*it) == hash)
        {
            pwallet->EraseFromRam(tx);
            if (!EraseTx(hash))
            {
                LOG(DBASE, "Transaction was found for deletion but returned database error: %s\n", hash.GetHex());
                delerror = true;
            }
            vTxHashOut.push_back(hash);
        }
    }

    if (delerror)
    {
        return DB_CORRUPT;
    }
    return DB_LOAD_OK;
}


DBErrors CWalletDB::ZapWalletTx(CWallet *pwallet, vector<CWalletTxRef> &vWtx)
{
    // build list of wallet TXs
    vector<uint256> vTxHash;
    DBErrors err = FindWalletTx(pwallet, vTxHash, vWtx);
    if (err != DB_LOAD_OK)
        return err;

    // erase each wallet TX
    for (uint256 &hash : vTxHash)
    {
        if (!EraseTx(hash))
            return DB_CORRUPT;
    }

    return DB_LOAD_OK;
}


void ThreadFlushWalletDB(const string &strFile)
{
    // Make this thread recognisable as the wallet flushing thread
    RenameThread("nexa-wallet");

    static bool fOneThread;
    if (fOneThread)
        return;
    fOneThread = true;
    if (!GetBoolArg("-flushwallet", DEFAULT_FLUSHWALLET))
        return;

    unsigned int nLastSeen = nWalletDBUpdated;
    unsigned int nLastFlushed = nWalletDBUpdated;
    int64_t nLastWalletUpdate = GetTime();
    while (true)
    {
        if (shutdown_threads.load() == true)
        {
            break;
        }

        MilliSleep(500);

        if (nLastSeen != nWalletDBUpdated)
        {
            nLastSeen = nWalletDBUpdated;
            nLastWalletUpdate = GetTime();
        }

        if (nLastFlushed != nWalletDBUpdated && GetTime() - nLastWalletUpdate >= 2)
        {
            TRY_LOCK(bitdb.cs_db, lockDb);
            if (lockDb)
            {
                // Don't do this if any databases are in use
                int nRefCount = 0;
                map<string, int>::iterator mi = bitdb.mapFileUseCount.begin();
                while (mi != bitdb.mapFileUseCount.end())
                {
                    nRefCount += (*mi).second;
                    mi++;
                }

                if (nRefCount == 0)
                {
                    map<string, int>::iterator mi2 = bitdb.mapFileUseCount.find(strFile);
                    if (mi2 != bitdb.mapFileUseCount.end())
                    {
                        LOG(DBASE, "Flushing wallet.dat\n");
                        nLastFlushed = nWalletDBUpdated;
                        int64_t nStart = GetTimeMillis();

                        // Flush wallet.dat so it's self contained
                        bitdb.CloseDb(strFile);
                        bitdb.CheckpointLSN(strFile);

                        bitdb.mapFileUseCount.erase(mi2++);
                        LOG(DBASE, "Flushed wallet.dat %dms\n", GetTimeMillis() - nStart);
                    }
                }
            }
        }
    }
}

bool BackupWallet(const CWallet &wallet, const string &strDest)
{
    if (!wallet.fFileBacked)
        return false;
    while (true)
    {
        {
            LOCK(bitdb.cs_db);
            if (!bitdb.mapFileUseCount.count(wallet.strWalletFile) || bitdb.mapFileUseCount[wallet.strWalletFile] == 0)
            {
                // Flush log data to the dat file
                bitdb.CloseDb(wallet.strWalletFile);
                bitdb.CheckpointLSN(wallet.strWalletFile);
                bitdb.mapFileUseCount.erase(wallet.strWalletFile);

                // Copy wallet.dat
                fs::path pathSrc = GetDataDir() / wallet.strWalletFile;
                fs::path pathDest(strDest);
                if (fs::is_directory(pathDest))
                    pathDest /= wallet.strWalletFile;

                // BU copy_file does not work with c++11, due to a link error in many versions of boost.  Return this
                // code to use when the default boost version in most distros fix this bug.
                //                try {
                // #if BOOST_VERSION >= 104000
                //                    boost::filesystem::copy_file(pathSrc, pathDest,
                //                    boost::filesystem::copy_option::overwrite_if_exists);
                // #else
                //                    boost::filesystem::copy_file(pathSrc, pathDest);
                // #endif
                //                    LOGA("copied wallet.dat to %s\n", pathDest.string());
                //                    return true;
                //                } catch (const boost::filesystem::filesystem_error& e) {
                //                    LOGA("error copying wallet.dat to %s - %s\n", pathDest.string(), e.what());
                //                    return false;
                //                }
                // Replacement copy code
                try
                {
                    ifstream source(pathSrc.string(), ios::binary);
                    ofstream dest(pathDest.string(), ios::binary);
                    if (!source)
                    {
                        LOGA("error opening source wallet file %s\n", pathSrc.string());
                        return false;
                    }
                    if (!dest)
                    {
                        LOGA("error opening destination wallet file %s\n", pathDest.string());
                        return false;
                    }
                    dest << source.rdbuf();
                    return true;
                }
                catch (std::ios_base::failure &e)
                {
                    LOGA("error copying wallet from %s to %s - %s\n", pathSrc.string(), pathDest.string(), e.what());
                    return false;
                }
                // end replacement copy code
            }
        }
        MilliSleep(100);
    }
    return false;
}

bool RestoreWallet(const CWallet &wallet, const string &strSrc)
{
    if (!wallet.fFileBacked)
        return false;
    int nBakCount = 1;
    fs::path pathBak = "";
    {
        LOCK(bitdb.cs_db);
        if (!bitdb.mapFileUseCount.count(wallet.strWalletFile) || bitdb.mapFileUseCount[wallet.strWalletFile] == 0)
        {
            // Flush log data to the dat file
            bitdb.CloseDb(wallet.strWalletFile);
            bitdb.CheckpointLSN(wallet.strWalletFile);
            bitdb.mapFileUseCount.erase(wallet.strWalletFile);

            // Rename wallet.dat to wallet.dat.bak(n).  If the *.bak file already exists then
            // increment the index value and keep looking until we find one that doesn't exist yet.
            do
            {
                string destBak = wallet.strWalletFile + ".bak" + std::to_string(nBakCount++);
                pathBak = GetDataDir() / destBak;
            } while (fs::exists(pathBak));

            fs::path pathSrc = GetDataDir() / wallet.strWalletFile;
            try
            {
                fs::rename(pathSrc, pathBak);
            }
            catch (std::ios_base::failure &e)
            {
                LOGA("error renaming wallet from %s to %s - %s\n", pathSrc.string(), pathBak.string(), e.what());
                return false;
            }
        }
        else
        {
            LOGA("error restoring wallet - nothing to do\n");
            return false;
        }
    }

    // Copy selected wallet to wallet.dat and then reload
    fs::path pathDest = GetDataDir() / wallet.strWalletFile;
    fs::path pathSrc(strSrc);
    if (fs::is_directory(pathSrc))
    {
        LOGA("error restoring wallet - can not restore a directory\n");
        return false;
    }

    if (pathSrc == pathDest)
    {
        uiInterface.ThreadSafeMessageBox(
            strprintf(_("You are trying to restore the same wallet which you are trying to replace.")), "",
            CClientUIInterface::MSG_ERROR | CClientUIInterface::BTN_OK);
        return false;
    }
    try
    {
        fs::copy_file(pathSrc, pathDest);
    }
    catch (std::ios_base::failure &e)
    {
        LOGA("error copying wallet from %s to %s - %s\n", pathSrc.string(), pathDest.string(), e.what());
        return false;
    }

    // Initiate a rescan of the wallet by setting the rescan flag and then
    // loading the wallet into memory.  Once done we can initiate a shutdown
    // which will save the updated wallet to disk.
    SetBoolArg("-rescan", true);
    try
    {
        bool ret = InitLoadWallet();
        if (!ret)
        {
            LOGA("error restoring wallet - InitLoadWallet() failed\n");
            return false;
        }
        else
        {
            bool fRet = uiInterface.ThreadSafeMessageBox(
                strprintf(_("\"Restore Wallet\" succeeded and a backup of the previous wallet was saved to: "
                            "%s.\n\n\nWhen you click \"OK\" Nexa will shutdown to complete the process."),
                    pathBak),
                "", CClientUIInterface::MSG_INFORMATION | CClientUIInterface::BTN_OK | CClientUIInterface::MODAL);
            if (fRet)
            {
                StartShutdown();
            }
        }
    }
    catch (std::runtime_error &e)
    {
        LOGA("Could not InitLoadWallet() %s - %s\n", pathSrc.string(), e.what());
        return false;
    }

    return true;
}
//
// Try to (very carefully!) recover wallet.dat if there is a problem.
//
bool CWalletDB::Recover(CDBEnv &dbenv, const std::string &filename, bool fOnlyKeys)
{
    // Recovery procedure:
    // move wallet.dat to wallet.timestamp.bak
    // Open the wallet normally and read out as many records as possible.
    // Call Salvage with fAggressive=true to
    // get as much data as possible.
    // Rewrite salvaged data to wallet.dat
    // Set -rescan so any missing transactions will be
    // found.
    int64_t now = GetTime();
    std::string newFilename = strprintf("wallet.%d.bak", now);

    int result = dbenv.dbenv->dbrename(nullptr, filename.c_str(), nullptr, newFilename.c_str(), DB_AUTO_COMMIT);
    if (result == 0)
        LOGA("Renamed %s to %s\n", filename, newFilename);
    else
    {
        // If error code value is greater than 0, it means a system error occured (even thou this would also be
        // triggered by an incompatible version of bdb (eg trying to open with bdb 5.3 a db created by an higher
        // version, eg 18 (on macos)). error code lower tha zero should specific berkeley db error. Have a look at this
        // page from oracle documentation web site for more details:
        //
        // https://docs.oracle.com/cd/E17076_02/html/programmer_reference/program_errorret.html
        //
        // Quoting the relevant part from the above docuemntation page:
        //
        // All values returned by Berkeley DB functions are less than 0 in order to avoid conflict with possible values
        // of errno. Specifically, Berkeley DB reserves all values from -30,800 to -30,999 to itself as possible error
        // values. There are a few Berkeley DB interfaces where it is possible for an application function to be called
        // by a Berkeley DB function and subsequently fail with an application-specific return. Such failure returns
        // will be passed back to the function that originally called a Berkeley DB interface. To avoid ambiguity about
        // the cause of the error, error values separate from the Berkeley DB error name space should be used.
        //
        LOGA("dbrename failed. Trying to rename %s into  %s, error code: %d, error string: %s\n Look in %s/db.log for "
             "more detailed error report.",
            filename, newFilename, result, dbenv.dbenv->strerror(result), GetDataDir().string());
        return false;
    }

    std::map<std::vector<unsigned char>, std::vector<unsigned char> > readData;
    {
        CWallet dummyWallet;
        CWalletScanState wss;
        {
            CDB cdb(newFilename);
            Dbc *cur = cdb.GetCursor();
            int count = 0;
            while (true)
            {
                string strType, strErr;
                CDataStream ssKey(SER_DISK, CLIENT_VERSION);
                CDataStream ssValue(SER_DISK, CLIENT_VERSION);
                int ret = cdb.ReadAtCursor(cur, ssKey, ssValue);
                if (ret == DB_NOTFOUND)
                    break;
                count++;
                LOCK(dummyWallet.cs_wallet);
                std::vector<unsigned char> keycopy(ssKey.begin(), ssKey.end());
                std::vector<unsigned char> valcopy(ssValue.begin(), ssValue.end());
                bool readok = ReadKeyValue(&dummyWallet, ssKey, ssValue, wss, strType, strErr);
                if (readok && IsKeyType(strType))
                {
                    readData[keycopy] = valcopy;
                }
            }
            LOGA("Recover, read %d records", count);
            cdb.Close();
        }
        bitdb.CloseDb(newFilename);
    }

    std::vector<CDBEnv::KeyValPair> salvagedData;
    bool fSuccess = dbenv.Salvage(newFilename, true, salvagedData);
    if (salvagedData.empty())
    {
        LOGA("Salvage(aggressive) found no records in %s.\n", newFilename);
        return false;
    }
    LOGA("Salvage(aggressive) found %u records\n", salvagedData.size());

    std::unique_ptr<Db> pdbCopy(new Db(dbenv.dbenv, 0));
    int ret = pdbCopy->open(nullptr, // Txn pointer
        filename.c_str(), // Filename
        "main", // Logical db name
        DB_BTREE, // Database type
        DB_CREATE, // Flags
        0);
    if (ret > 0)
    {
        LOGA("Cannot create database file %s\n", filename);
        return false;
    }
    CWallet dummyWallet;
    CWalletScanState wss;

    DbTxn *ptxn = dbenv.TxnBegin();
    for (CDBEnv::KeyValPair &row : salvagedData)
    {
        if (fOnlyKeys)
        {
            CDataStream ssKey(row.first, SER_DISK, CLIENT_VERSION);
            CDataStream ssValue(row.second, SER_DISK, CLIENT_VERSION);
            string strType, strErr;
            bool fReadOK;
            {
                // Required in LoadKeyMetadata():
                LOCK(dummyWallet.cs_wallet);
                fReadOK = ReadKeyValue(&dummyWallet, ssKey, ssValue, wss, strType, strErr);
            }
            if (!IsKeyType(strType))
                continue;
            if (!fReadOK)
            {
                LOGA("WARNING: CWalletDB::Recover skipping %s: %s\n", strType, strErr);
                continue;
            }
        }
        Dbt datKey(&row.first[0], row.first.size());
        Dbt datValue(&row.second[0], row.second.size());
        int ret2 = pdbCopy->put(ptxn, &datKey, &datValue, DB_NOOVERWRITE);
        if (ret2 > 0)
            fSuccess = false;
    }

    // Write any data we got from opening the database normally.  If we got any, this is more likely to be correct

    int nrec = 0;
    for (auto &i : readData)
    {
        Dbt datKey((void *)&i.first[0], i.first.size());
        Dbt datValue((void *)&i.second[0], i.second.size());
        int putresult = pdbCopy->put(ptxn, &datKey, &datValue, DB_OVERWRITE_DUP);
        assert(putresult == 0);
        nrec++;
    }

    ptxn->commit(0);
    pdbCopy->close(0);

    return fSuccess;
}

bool CWalletDB::Recover(CDBEnv &dbenv, const std::string &filename)
{
    return CWalletDB::Recover(dbenv, filename, false);
}

bool CWalletDB::WriteDestData(const CTxDestination &address, const std::string &key, const std::string &value)
{
    if (!IsValidDestination(address))
        return false;

    nWalletDBUpdated++;
    return Write(
        std::make_pair(std::string("destdata"), std::make_pair(EncodeLegacyAddr(address, Params()), key)), value);
}

bool CWalletDB::EraseDestData(const CTxDestination &address, const std::string &key)
{
    if (!IsValidDestination(address))
        return false;

    nWalletDBUpdated++;
    return Erase(std::make_pair(std::string("destdata"), std::make_pair(EncodeLegacyAddr(address, Params()), key)));
}


bool CWalletDB::WriteHDChain(const CHDChain &chain)
{
    nWalletDBUpdated++;
    return Write(std::string("hdchain"), chain);
}
