// Copyright (c) 2009-2010 Satoshi Nakamoto
// Copyright (c) 2009-2015 The Bitcoin Core developers
// Copyright (c) 2015-2022 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#ifndef NEXA_WALLET_WALLETDB_H
#define NEXA_WALLET_WALLETDB_H

#include "amount.h"
#include "key.h"
#include "primitives/transaction.h"
#include "script/standard.h" // for CTxDestination
#include "wallet/db.h"

#include <list>
#include <stdint.h>
#include <string>
#include <utility>
#include <vector>

static const bool DEFAULT_FLUSHWALLET = true;

class CAccount;
class CAccountingEntry;
struct CBlockLocator;
class CKeyPool;
class CMasterKey;
class CScript;
class CWallet;
class CWalletTx;
typedef std::shared_ptr<CWalletTx> CWalletTxRef;
class COutput;
class uint160;
class uint256;

/** Error statuses for the wallet database */
enum DBErrors
{
    DB_LOAD_OK,
    DB_CORRUPT,
    DB_NONCRITICAL_ERROR,
    DB_TOO_NEW,
    DB_LOAD_FAIL,
    DB_NEED_REWRITE
};

/* simple HD chain data model */
class CHDChain
{
public:
    uint32_t nExternalChainCounter;
    CKeyID masterKeyID; //!< master key hash160

    static const int CURRENT_VERSION = 1;
    int nVersion;

    CHDChain() { SetNull(); }
    ADD_SERIALIZE_METHODS;
    template <typename Stream, typename Operation>
    inline void SerializationOp(Stream &s, Operation ser_action)
    {
        READWRITE(this->nVersion);
        READWRITE(nExternalChainCounter);
        READWRITE(masterKeyID);
    }

    void SetNull()
    {
        nVersion = CHDChain::CURRENT_VERSION;
        nExternalChainCounter = 0;
        masterKeyID.SetNull();
    }
};

class CKeyMetadata
{
public:
    static const int VERSION_BASIC = 1;
    static const int VERSION_WITH_HDDATA = 10;
    static const int CURRENT_VERSION = VERSION_WITH_HDDATA;

    int nVersion = CKeyMetadata::CURRENT_VERSION;
    int64_t nCreateTime = 0; // 0 means unknown
    std::string hdKeypath; // optional HD/bip32 keypath
    CKeyID hdMasterKeyID; // id of the hd masterkey used to derive this key

    CKeyMetadata() { SetNull(); }
    CKeyMetadata(int64_t nCreateTime_)
    {
        SetNull();
        nCreateTime = nCreateTime_;
    }

    ADD_SERIALIZE_METHODS;

    template <typename Stream, typename Operation>
    inline void SerializationOp(Stream &s, Operation ser_action)
    {
        READWRITE(this->nVersion);
        READWRITE(nCreateTime);
        if (this->nVersion >= VERSION_WITH_HDDATA)
        {
            READWRITE(hdKeypath);
            READWRITE(hdMasterKeyID);
        }
    }

    void SetNull()
    {
        nVersion = CKeyMetadata::CURRENT_VERSION;
        nCreateTime = 0;
        hdKeypath.clear();
        hdMasterKeyID.SetNull();
    }
};

/** Access to the wallet database (wallet.dat) */
class CWalletDB : public CDB
{
public:
    CWalletDB(const std::string &strFilename, const char *pszMode = "r+", bool _fFlushOnClose = true)
        : CDB(strFilename, pszMode, _fFlushOnClose)
    {
    }

    // the wallet will not display tokens that are not explicitly tracked in the UI or in rpcs
    // if this wallet created the token or holds an authority for the token it is automatically added
    // to tracking, the goal of this list is to try to prevent some scams
    bool WriteTokenTracker(const CGroupTokenID &id, const std::string &strTokenTicker);
    bool EraseTokenTracker(const CGroupTokenID &id);

    bool WriteName(const CTxDestination &address, const std::string &strName);
    bool EraseName(const CTxDestination &address);

    bool WritePurpose(const CTxDestination &address, const std::string &purpose);
    bool ErasePurpose(const CTxDestination &address);

    bool WriteTx(const CWalletTx &wtx);
    bool EraseTx(uint256 idem);

    bool WriteKey(const CPubKey &vchPubKey, const CPrivKey &vchPrivKey, const CKeyMetadata &keyMeta);
    bool WriteCryptedKey(const CPubKey &vchPubKey,
        const std::vector<unsigned char> &vchCryptedSecret,
        const CKeyMetadata &keyMeta);
    bool WriteMasterKey(unsigned int nID, const CMasterKey &kMasterKey);

    bool WriteCScript(const uint160 &hash, const CScript &redeemScript);

    bool WriteWatchOnly(const CScript &script);
    bool EraseWatchOnly(const CScript &script);

    bool WriteBestBlock(const CBlockLocator &locator);
    bool ReadBestBlock(CBlockLocator &locator);

    bool WriteOrderPosNext(int64_t nOrderPosNext);

    bool WriteDefaultKey(const CPubKey &vchPubKey);

    bool ReadPool(int64_t nPool, CKeyPool &keypool);
    bool WritePool(int64_t nPool, const CKeyPool &keypool);
    bool ErasePool(int64_t nPool);

    bool WriteMinVersion(int nVersion);

    /// This writes directly to the database, and will not update the CWallet's cached accounting entries!
    /// Use wallet.AddAccountingEntry instead, to write *and* update its caches.
    bool WriteAccountingEntry(const uint64_t nAccEntryNum, const CAccountingEntry &acentry);
    bool WriteAccountingEntry_Backend(const CAccountingEntry &acentry);
    bool ReadAccount(const std::string &strAccount, CAccount &account);
    bool WriteAccount(const std::string &strAccount, const CAccount &account);

    /// Write destination data key,value tuple to database.
    bool WriteDestData(const CTxDestination &address, const std::string &key, const std::string &value);
    /// Erase destination data tuple from wallet database.
    bool EraseDestData(const CTxDestination &address, const std::string &key);

    CAmount GetAccountCreditDebit(const std::string &strAccount);
    void ListAccountCreditDebit(const std::string &strAccount, std::list<CAccountingEntry> &acentries);

    DBErrors LoadWallet(CWallet *pwallet);
    // Returns every transaction in the wallet
    DBErrors FindWalletTx(CWallet *pwallet, std::vector<uint256> &vTxHash, std::vector<CWalletTxRef> &vWtx);
    DBErrors ZapWalletTx(CWallet *pwallet, std::vector<CWalletTxRef> &vWtx);
    DBErrors ZapSelectTx(CWallet *pwallet, std::vector<uint256> &vHashIn, std::vector<uint256> &vHashOut);

    static bool Recover(CDBEnv &dbenv, const std::string &filename, bool fOnlyKeys);
    static bool Recover(CDBEnv &dbenv, const std::string &filename);

    //! write the hdchain model (external chain child index counter)
    bool WriteHDChain(const CHDChain &chain);

private:
    CWalletDB(const CWalletDB &);
    void operator=(const CWalletDB &);
};

bool BackupWallet(const CWallet &wallet, const std::string &strDest);
bool RestoreWallet(const CWallet &wallet, const std::string &strDest);
void ThreadFlushWalletDB(const std::string &strFile);

#endif // NEXA_WALLET_WALLETDB_H
