// Copyright (c) 2011-2015 The Bitcoin Core developers
// Copyright (c) 2015-2023 The Bitcoin Unlimited developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#define BOOST_TEST_MODULE Nexa Test Suite

#include "test_nexa.h"

#include "blockstorage/blockcache.h"
#include "chainparams.h"
#include "consensus/consensus.h"
#include "consensus/validation.h"
#include "crypto/sha256.h"
#include "fs.h"
#include "key.h"
#include "main.h"
#include "miner.h"
#include "pubkey.h"
#include "random.h"
#include "rpc/register.h"
#include "rpc/server.h"
#include "script/sigcache.h"
#include "script/sighashtype.h"
#include "test/testutil.h"
#include "txadmission.h"
#include "txdb.h"
#include "txmempool.h"
#include "ui_interface.h"
#include "validation/validation.h"

#include <memory>

#include <boost/program_options.hpp>
#include <boost/test/unit_test.hpp>
#include <thread>

uint256 insecure_rand_seed;
FastRandomContext insecure_rand_ctx;

FalseScriptImportedState fsis;

extern void noui_connect();

extern CTweak<uint32_t> limitFreeRelay;
extern uint64_t nMaxSendBufferSize;

BasicTestingSetup::BasicTestingSetup(const std::string &chainName)
{
    // Do not place the data created by these unit tests on top of any existing chain,
    // by overriding datadir to use a temporary if it isn't already overridden
    if (mapArgs.count("-datadir") == 0)
        mapArgs["-datadir"] = GetTempPath().string();
    SHA256AutoDetect();
    RandomInit();
    ECC_Start();
    SetupEnvironment();
    SetupNetworking();
    InitSignatureCache();
    fPrintToDebugLog = false; // don't want to write to debug.log file
    fCheckBlockIndex = true;
    SelectParams(chainName);
    noui_connect();

    // check args to see if user wants to accept non standard txs
    bool fAcceptNonStandard = GetBoolArg("-acceptnonstdtxn", !ModifiableParams().RequireStandard());
    // attempt to change if the chain requires standard based on what the user configured
    bool acceptedStandardChange = ModifiableParams().SetRequireStandard(fAcceptNonStandard);
    if (acceptedStandardChange == false)
    {
        printf("acceptnonstdtxn is not currently supported for %s chain", ModifiableParams().NetworkIDString().c_str());
        assert(false);
    }
}

BasicTestingSetup::~BasicTestingSetup() { ECC_Stop(); }
TestingSetup::TestingSetup(const std::string &chainName) : BasicTestingSetup(chainName)
{
    nMiningForkTime = Params().GetConsensus().nextForkActivationTime;

    insecure_rand_seed = GetRandHash();
    insecure_rand_ctx = FastRandomContext(insecure_rand_seed);

    const CChainParams &chainparams = Params();
    // Ideally we'd move all the RPC tests to the functional testing framework
    // instead of unit tests, but for now we need these here.
    RegisterAllCoreRPCCommands(tableRPC);
    ClearDatadirCache();
    pathTemp =
        GetTempPath() / strprintf("test_nexa_%lu_%i", (unsigned long)GetTime(), (int)(InsecureRandRange(1 << 30)));
    fs::create_directories(pathTemp);
    blockcache.Init();
    pblocktree = new CBlockTreeDB(1 << 20, "", true);
    pcoinsdbview = new CCoinsViewDB(1 << 23, true);
    pcoinsTip = new CCoinsViewCache(pcoinsdbview);
    txCommitQ = new CIndexedCommitQ();
    ptokenDesc = new CTokenDescriptionDB(1 << 20, false, false);
    ptokenMint = new CTokenMintageDB(1 << 20, false, false);

    bool worked = InitBlockIndex(chainparams);
    assert(worked);

    // -relay.limitFreeRelay is diabled by default but some tests rely on it so make sure to set it here.
    limitFreeRelay.Set(15);

    // Initial dbcache settings so that the automatic cache setting don't kick in and allow
    // us to accidentally use up our RAM on Travis, and also so that we are not prevented from flushing the
    // dbcache if the need arises in the unit tests (dbcache must be less than the DEFAULT_HIGH_PERF_MEM_CUTOFF
    // to allow all cache entries to be flushed).
    dbcacheTweak.Set(5);
    nCoinCacheMaxSize.store(5000000);

    // Set the max send buffer size.
    nMaxSendBufferSize = 10000 * 1000;

    // Make sure there are 3 script check threads running for each queue
    SoftSetArg("-par", std::to_string(3));
    PV.reset(new CParallelValidation());

    RegisterNodeSignals(GetNodeSignals());
}

TestingSetup::~TestingSetup()
{
    UnregisterNodeSignals(GetNodeSignals());
    threadGroup.interrupt_all();
    threadGroup.join_all();
    UnloadBlockIndex();
    delete pcoinsTip;
    delete pcoinsdbview;
    delete pblocktree;
    delete ptokenDesc;
    delete ptokenMint;
    fs::remove_all(pathTemp);
}

TestChain100Setup::TestChain100Setup() : TestingSetup(CBaseChainParams::REGTEST)
{
    // Generate a 100-block chain:
    coinbaseKey.MakeNewKey(true);
    CScript scriptPubKey = CScript() << ToByteVector(coinbaseKey.GetPubKey()) << OP_CHECKSIG;
    for (int i = 0; i < Params().GetConsensus().coinbaseMaturity; i++)
    {
        std::vector<CMutableTransaction> noTxns;
        CBlock b = CreateAndProcessBlock(noTxns, scriptPubKey);
        coinbaseTxns.push_back(*b.vtx[0]);
    }
}

//
// Create a new block with just given transactions, coinbase paying to
// scriptPubKey, and try to add it to the current chain.
//
CBlock TestChain100Setup::CreateAndProcessBlock(const std::vector<CMutableTransaction> &txns,
    const CScript &scriptPubKey)
{
    const CChainParams &chainparams = Params();
    std::unique_ptr<CBlockTemplate> pblocktemplate(new CBlockTemplate());

    CPubKey pubkey;
    pblocktemplate = BlockAssembler(chainparams).CreateNewBlock(P2pktOutput(pubkey));
    CBlockRef pblock = pblocktemplate->block;

    // Replace mempool-selected txns with just coinbase plus passed-in txns:
    pblock->vtx.resize(1);
    for (const CMutableTransaction &tx : txns)
    {
        pblock->vtx.push_back(MakeTransactionRef(tx));
    }

    // enfore LTOR ordering of transactions
    std::sort(pblock->vtx.begin() + 1, pblock->vtx.end(), NumericallyLessTxHashComparator());

    pblock->UpdateHeader(); // make sure the size field is properly calculated
    pblock->nonce.resize(3);
    bool worked = MineBlock(*pblock, 1 << 23, chainparams.GetConsensus());
    assert(worked);

    CValidationState state;
    worked = ProcessNewBlock(state, chainparams, nullptr, pblock, true, nullptr, false);
    // ProcessNewBlock will fail here in some negative tests so no: assert(worked);

    CBlock result = *pblock;
    return result;
}

TestChain100Setup::~TestChain100Setup() {}
CTxMemPoolEntry TestMemPoolEntryHelper::FromTx(const CMutableTransaction &tx, CTxMemPool *pool)
{
    CTransaction txn(tx);
    return FromTx(txn, pool);
}

CTxMemPoolEntry TestMemPoolEntryHelper::FromTx(const CTransaction &txn, CTxMemPool *pool)
{
    CAmount inChainValue = txn.GetValueOut();

    CTxMemPoolEntry ret(
        MakeTransactionRef(txn), nFee, nTime, dPriority, nHeight, inChainValue, spendsCoinbase, sigOpCount, lp);
    // ret.dbgName = dbgName;
    // dbgName = ""; // Reset to no name every time we pull a new entry from this object
    return ret;
}

void Shutdown(void *parg) { exit(0); }
void StartShutdown() { exit(0); }
bool ShutdownRequested() { return false; }
using namespace boost::program_options;

CService ipaddress(uint32_t i, uint32_t port)
{
    struct in_addr s;
    s.s_addr = i;
    return CService(CNetAddr(s), port);
}

struct StartupShutdown
{
    StartupShutdown()
    {
        options_description optDef("Options");
        optDef.add_options()("testhelp", "program options information")(
            "log_level", "set boost logging (all, test_suite, message, warning, error, ...)")(
            "log_nexa", value<std::string>()->required(), "nexa logging destination (console, none)");
        variables_map opts;
        store(parse_command_line(boost::unit_test::framework::master_test_suite().argc,
                  boost::unit_test::framework::master_test_suite().argv, optDef),
            opts);

        if (opts.count("testhelp"))
        {
            std::cout << optDef << std::endl;
            exit(0);
        }

        if (opts.count("log_nexa"))
        {
            std::string s = opts["log_nexa"].as<std::string>();
            if (s == "console")
            {
                /* To enable this, add
                   -- --log_nexa console
                   to the end of the test_nexa argument list. */
                Logging::LogToggleCategory(ALL, true);
                fPrintToConsole = true;
                fPrintToDebugLog = false;
            }
            else if (s == "none")
            {
                fPrintToConsole = false;
                fPrintToDebugLog = false;
            }
        }
    }
    ~StartupShutdown() { UnlimitedCleanup(); }
};

#if BOOST_VERSION >= 106500
BOOST_TEST_GLOBAL_FIXTURE(StartupShutdown);
#else
BOOST_GLOBAL_FIXTURE(StartupShutdown);
#endif

std::ostream &operator<<(std::ostream &os, const uint256 &num)
{
    os << num.ToString();
    return os;
}

CBlock TestBlock1()
{
    // Random real block ce8b7b9f336a5ebc1ec9d5a69b6ddc182e32ccd8db97b3a12713e120b8cf131c
    // With 290 txns
    CDataStream stream(
        ParseHex(
            "cf8fb188145bee521ee64dc91d8b88188c4b14c928fe4f0727c9af2f4db5c22018d8081d8fc1f3c5b20da5b893ce7b3d7ec7465525"
            "4200280c9f21c0e2862beb3b13ea7fd747afab672b895e8a184e7348142e2146b6f22b4e0f7365bb66ace39c344b7f000000000000"
            "0000000000000000000000000000000000000000000000000000696c186382eb72fdeaa5426c190000000000000000000000000000"
            "00000000000000000000000037f8000000000000812200000004ce3b0f25fd2201000002013bc19b3b0000000017005114441ccfea"
            "db90bc7b122e33acdb2643c95294fc920000000000000000000f6a0372f6000008000000000000000000000000000100c212be1d6f"
            "ff9e374eb8ca20832977f90352e22f25af551c90ced2205a37eb3c64222103fa29e66a139ba67ec42b57213c966e1ffc575c2ee247"
            "5420230e959b0627b0e0403293f55f80c9780eb29813bc85c4c6df03f801c82bcac44d09e9cfbbbd5bd522fe65bda1dd0a67b8effe"
            "56d0dcbd56d4652daeec439d15d3e7ed509e5a247391feffffff6a2e0505000000000201220200000000000017005114d8825b53f9"
            "46b54ab5a11880c302be441c086132016d2b05050000000017005114fd348ae63bc7ab51fc3e3aaed78f932f6562d08c28f6000000"
            "01002b4b71b23ad92b337c73ce49166502288656d8ae8f03985163ca85ec8315a8b96422210337c5229ccd52eb3d647f9ca6a4e488"
            "0d6d7b2308514a45c62242ccfc6cbac87b40b8448b792be2e446ccfaf3b2574e7ed494ec0f28db5d148b3156ad5e5e3811c6c8ba24"
            "85b34ef5d910827f1dfb513d1041ba9b8b04cc9fd57d407cdadd67376cfeffffffa0f8040500000000020122020000000000001700"
            "5114d8825b53f946b54ab5a11880c302be441c08613201a3f5040500000000170051143ae9a79726b6cb726d8bb96b270aa86b5515"
            "1c2d28f60000000100e9e092bebc37e85a728739cabb270a153a06c9efe9ec0881f65330ca7dde71c864222102378948a1e45096ff"
            "4148ec48bafdd663af860099abfefe5068a8c7d8b9cbfc4e4040e39d03966810f1f3a32abc281f4d2ad52736e6821cedbed00861b6"
            "fd9c68c82c2e8d4eef7d63d60082474ebc8f76683677fb774df0dc370dee34aa4ed92e92feffffff01970505000000000201220200"
            "000000000017005114d8825b53f946b54ab5a11880c302be441c08613201049405050000000017005114cd00a15780d05c36b351d9"
            "eadf0ce6944a22f84c28f60000000100bee21853c79a90ee7dc3d345c0b87f12f05d709ce11a9d2162296f4f1a93c59e6422210229"
            "76f961ce9c45002a4d2d933743b12fb87debd31a62aa5cd744a23a01d5b485403564500eabdb104ea1fe34c8ee5fc7dbf165476d63"
            "6baa500e3621633c1701d0be15fc1ca3af4643a0734daed1d9c78b1839b1a5781df75e3b6417606f9c6727feffffff940405050000"
            "00000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320197010505000000001700511404c5413f"
            "f8e6da28e3f3b63e64126354698dba4428f60000000100808cfffb2f7cf1e57224e77d3f527139b93e522ec9e02fbec165c5ea050d"
            "b25a642221022fcd385600360c58737a6fe9b6e4153dd05f2d686217847f8b3ca768adfcd70f40dce1c84ef0a465aa1dda4b1c7973"
            "d90b443e86cec6a5417c8df61a69d92fb5dd20c714c482615c7bf95e31edc696a9085d4ca92cb6c2c7c174c6b52318a523c0feffff"
            "ffc5d20505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201c8cf05050000000017"
            "005114c173cb83b472542504b50e0774a2e7d21b1c8aaf28f60000000100b5ffc0b7bb8adc8e03e4218478faa3ffb5a6b839d27c4e"
            "33b607d1d63075630364222102c7f39a485d7f10b49f66ef8fdc022a72e586d3e5afa6872bc5999a87260f78e240d242aa7156d204"
            "bd5054cecce3d34dc973d2359a9323b30947f071ec3f58cc8415c9fcba2dedae0afc7de85a4211e6e05aeecdedd06c52a96126f71e"
            "0aae64c0feffffff3d5b0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132014058"
            "0505000000001700511489d8b741fa8c89fec8c625fa1483a5d5c470408f28f60000000100f058e79beecebd50700a2277e9d425f4"
            "f2f74909f774c7b0c03995fc8a9bb4cf642221028d64584bebc279e4260ab1230207f4a4b4f45a89eda301c6c1bf65204a8005e440"
            "4d456b2745b94586c71c81673adf738d94f84576e704ccd855701be696f265daa9f6bdf2ed8d4b06db26bd54006a06a22e7fe9c342"
            "03ba6f233b88ed353c6a5ffefffffffd980705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be44"
            "1c08613201009607050000000017005114ba329114274d151ee263848a44ced9773777717428f60000000100ae6efafdc038511083"
            "da4e99f0310cbacaf8459f8a58e68858597251a52e6fad64222103a1b3f1583613f720359bd2c1adf563d63f46552262a7d92ef73e"
            "0f19767ca50640c04c2f628059ee0e1c309ac5eea646a2efa3b34bc3a5442e26aab4ab5ff9bc9ae531ed33bd31c42ccaab08aaec38"
            "2a04adf8e8f8ce97857f19ee88b5c8fc55aefeffffff81150705000000000201220200000000000017005114d8825b53f946b54ab5"
            "a11880c302be441c08613201841207050000000017005114b74b11cc18dd8b410175c67d5ca36e2b1e424c0828f600000001008dfc"
            "f78e51a63ef03c6d44ab263745ffb559bc1d34000cc175364b8eaae6491e642221035a9f203ae0166b7fbf20e7b73c90c3df4e1b26"
            "d340431c9933cafa012ca8ed084005b7d9641b44786724234834dee482147d21c9353f1b20b645b47e7fbaf2ca2737293ec13201c2"
            "dc62574814cd19f5903f6a44f1fc7e8e85b2f4ebb05236a034feffffff385f0605000000000201220200000000000017005114d882"
            "5b53f946b54ab5a11880c302be441c086132013b5c060500000000170051149f3c2c34ea60beead5340287ad01e0839321b34328f6"
            "0000000100db954515c2679754e4dfaf431c6f188fcd67b94b552c8407af3fc1454c796c65642221027761fa67f0f478fd8f6c712b"
            "b5d11fe9d14a041bc77bd4a33cba09c091542620408d6ca4a3f0bf63b12f35bb61b983460cf6e063f42ffbdc7e8ed00b0368a1fb37"
            "5c66f45f9296ed2a9f2f0c855b34bd12c7be78bc7b2e5da513cb7b45e0be03b5feffffff0393070500000000020122020000000000"
            "0017005114d8825b53f946b54ab5a11880c302be441c08613201069007050000000017005114e27e58a710127f1a4ac81c9f744e2e"
            "f32f1aa8f328f6000000010049fa45575943666a8a40e06fcd3d3ffe998a4254f0a61f1e6fac25dfc5bb5e5b64222102cce7edbc52"
            "43bc07cbe4f16d59ccd02313a9d51547a39ff03b34ae3d76d72179408ff1b6ea38c8f29c0c1ae2566bf96f430f0b5f9d39271a0f4c"
            "b9e3d2e4b1adad0916df340a54a9fff7f5bdcbf7c8a9402d2ede3bdd2b6c0b8ed546d4d2afcc62feffffff63330705000000000201"
            "220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132016630070500000000170051146ae5f1599430492c"
            "b088c7779a6030a71006a6ac28f6000000010065d449fc084c87ef9bb4e96369cb98e8f115e9a8fd285d6e409419650906e3a06422"
            "2102d641afda28c2fec35d7c7a47e096e18f3edf91bad3b515135c2201446c67f64b405ce7f2fe48dd921330aec1823eee9789d3f1"
            "880d60c310b1a79e1e319ed18ac51d14aa222906107e63aeed6afb6e248a8f422ece1e2c9d80cad767ec0e9c8f49feffffff8f0806"
            "05000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320192050605000000001700511459"
            "37fce110c43f89f3e48ee1075c4ab04c33e2f128f6000000010039f7ce499c8bb03304bdb7b4f53d6b753b1c45e4406cd31ff65900"
            "ec1fbd3a76642221036421da198063f10714eb95626155c0347463e414eccf7b7f0d3ace6525a22f6240c4c5f08503f845aa6b35b8"
            "37ca5cb98595d107aa52b8ce9626d6a0794f6df40c9a4da4027f9884477af35e381832d152c36cf2d63d47ea9a25de4f9152632aa6"
            "feffffff1e780705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201217507050000"
            "000017005114ce3adb97b2dfa6a346bb3400cd36a79ee499a9b228f600000001002578a34f72892ba88c38a9bfb73d6b67a00c15a6"
            "9999d9c835b2063d9880cff764222103073199b10ce5e24364af5b0fb9cb95948e89d2cad22a4f03bb0332606395349440cb9ebb19"
            "fd9a4c63b49c9a24f2e2afcdbac75fdaadc2bda8798808f453a0ec93e35c32b12fba53a876045a15759fb5fb47b46c232cd2fb30a6"
            "f376ebe1a7e2d3feffffff7e180705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132"
            "018115070500000000170051142fe298f4ee394022ec17702f84bc4b7d2952596228f6000000010096b0bf4a63afb76b668815358b"
            "cc8c786c449f08139f94fd4f2d13eba85d8533642221037b3790740ebfe31df74f17f4cb0d8dc229c50c0ae9ea914631b6e58077d4"
            "df7f40a7d35f2339e015341bb1b77156cc99861f43b754b55a22165551b7707d58c50f0653486eadc024c98d1d0b5e49c5f20716b8"
            "46f13f795075ac51b137d34bf7eefeffffff36600705000000000201220200000000000017005114d8825b53f946b54ab5a11880c3"
            "02be441c08613201395d0705000000001700511419116dee0d1c24ca39eab74bcceba1540fca546b28f6000000010080078beb57f7"
            "be8d1d2130034947318e269e1cff7a5784b3f5dc7fb59288b4f86422210297c4245da33e0776dc0c85d4ec28dfa84226ee52f3eb9c"
            "1477c92b4f22afb29f40b703aa7ea64340b9e191b97613cb05ed39b647da184ae656a64daf056505526d42f38092699a4138518ba7"
            "6e64ad58496149747624c4d515188d841cb47dabf1feffffff6f270705000000000201220200000000000017005114d8825b53f946"
            "b54ab5a11880c302be441c086132017224070500000000170051148693e30c0b8cb27731d469d962c18dfa73142a6628f600000001"
            "00c2fad4a2f46fc2783d847322b7a9fe2a55f67cc1ef050328131406af538f085264222103ffd2ae759ab13790e917aee25beda37e"
            "4362ea703df5bbcf948d1a95999190b340c508667fa0693ac60e57c16287699fb575cab707a6bd0973cb1c66f21fd79a40bcbebf75"
            "ff1bcf5596805d72ba6067bcdad608f872f05a5c3982a0c1e59cb14efeffffff9df807050000000002012202000000000000170051"
            "14d8825b53f946b54ab5a11880c302be441c08613201a0f5070500000000170051146559a9844cee2ef3201f2afa5e358d2f8e863a"
            "3528f600000001006ab9071a66f1c44a6107b2fa5b5025d214b56ca2b6222028b70092ee93f7689564222102183dfac757a852bb7a"
            "c91d1273d0506e08066919b8eaba6bcaabbe4b944fa62740316b740ca81db751be8e73970f1f40c6926d34834d4f9c6fae2bbd546a"
            "28d10e30ada693527e9af0e0ee67825e1db7c9108fde4a8e99da912ba7bd238914ed8efeffffff583d080500000000020122020000"
            "0000000017005114d8825b53f946b54ab5a11880c302be441c086132015b3a0805000000001700511498d33cca71b99e37c4841f53"
            "a285ce282f4a8ba228f60000000100229514c05435fa6bce783865c1d6c4f826f89e020ace635fa49d41b4ab69901a64222102de2f"
            "5f587ccc8c286814f4d3d15366ed96a2d739146c34bc3499ca4c0932ebcb409312fa0621765fa50c619f5fe79edb0762c51f1008fa"
            "ba68dc3d6b9fc45fc554a1ee6cd14bcbcaf28ad5fd3d944b5c5e038d1ff5d40c83a37f7a1b01aedbb9e8feffffff55430505000000"
            "000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132015840050500000000170051147dff024ec3"
            "c944cca502898de247b8952bc1dde328f60000000100c3a2cab62818eb07f0b60bfceec4bd5dd50b92926e5e9eea334880e137111e"
            "7d6422210379b801dce36a32cdaf2311e577433bd9f1625d050eabbecc85904d43152325314054b471386b118acd201b9c7b735832"
            "145810f5cc320596f9ba29ba10ef1d811201581e159056c323a6adcba7684a14c6c664b817df7f799eca2028b7d6546007feffffff"
            "3a5e0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132013d5b0505000000001700"
            "5114bd3899045914bd46efb5d28ef476419d49f45be128f60000000100566ac4a1fcc2357c8bc0235bfb9f2edbc5d75372fcb78d80"
            "3cd7ec0e81e9a6e064222102fbb2f23506a9105edf517ec4ddf8a2cdab10958f8a2e73106fbe52a49801ad50400e75900e7ca60ffe"
            "a9da12be6001f70d7372109f939258f2d36192ab0c83443a47ab252a9212c2ebfea11ed6eee42fcdff36b7893f3171cd1a988d984e"
            "847777feffffff9cfa0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132019ff706"
            "050000000017005114c1f434a7768d977fc4c920d236fe3e0c81022a0f28f600000001004f890109147b510a82553f09b9360923b7"
            "8d34292b0bfb32693dea930d68c78164222103c4f8e3b1472f3a5daa945f3b3c989aa27a472933ea3300f5d2feb3f3d6b1d5dc402c"
            "7d92d148098da6ebfdd2b553fe3724950ff7c03c9fbc30b98c4466741f438d65ffc2c5239977cbb41a5c7cc79b70c8a59d7c7941c3"
            "ee74c26ab58d1a6995d0fefffffff79e0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c"
            "08613201fa9b070500000000170051140ba82784d0076957b0411b9f93cfa03f9481d7e228f60000000100a5dc07b087d77d8778e6"
            "9222fad2e2a84c9a50e2b8b22d1017934113979d39f76422210273cee88c0abed58bf78527b7483906fd8c06f043dd98514d200da1"
            "3f3a645d96401d66c3a9cd66b107cbd6e433ee8d0ff759b8252180f2df0508341c46ac4d2ad2a61d8bab55c24f2e899bac1c8e7828"
            "5f1ed4560eb7c6291876545985e8656c98feffffff47500605000000000201220200000000000017005114d8825b53f946b54ab5a1"
            "1880c302be441c086132014a4d06050000000017005114bb2e990f8a935a66fa78eddc09ed6c033eb2990228f6000000010027cfa7"
            "d400ff62690e2bcfc12bf05ff4651cdebacfa0741b8011d8d990c048ac64222103f00f89f82dea617e8e6105b0cf293dbc484fb5fb"
            "d467722d66b647df35dd9954404c80ac0e910d48e1f01964dfde3f231a18cc066eb5042a22d03facc2bbcfa96ea65ee54898db51b1"
            "e01876cf631fa0eed9cace0d468a8eae392889aad16cb9d1feffffff9afe0405000000000201220200000000000017005114d8825b"
            "53f946b54ab5a11880c302be441c086132019dfb040500000000170051148eae81ef7a3a070e008df2213f67973a4ca1e7d628f600"
            "000001001725fd791e927b00fd4e279139dc03200bebe394c3c13a242a5b7d7ae266dcea6422210304beae9ea5f90d1c69dc39f50a"
            "46abdc2fa6ddb3ef964774a635bc4daa3adbac4054794a5e6948f12261279eb4e021ebb3cd7c25c741d4026cb14fec3a3e374596f9"
            "a286212fdfe86b4644a0fa734e03d6c2a7eb51b4d86fb78f54a5132dbcf1a0feffffffecab05050000000002012202000000000000"
            "17005114d8825b53f946b54ab5a11880c302be441c08613201efa805050000000017005114662129e6e1845a0510de5fae771ed6a5"
            "4e0c643828f60000000100fd2d6a0c683f9359acd1e3324cc1d111c0687d4eb4f400e7a218a9fd6e08a83264222103887e96463b6c"
            "1c1a8b29fe6cec1392e2fb6cd5e42adf25086889611df4b7e40b40a397b985da5b93b39337bc0ff40f57feb67f3e6bc2f91e926e36"
            "c35f24da45846db9aeb9bc3c42a3f2888498cc6eacd8f374ddf2908049f9bb550b5098583d11feffffff395d070500000000020122"
            "0200000000000017005114d8825b53f946b54ab5a11880c302be441c086132013c5a07050000000017005114d28c85067452d615c4"
            "b4aff75af31f8821211d5e28f60000000100aa68af1f6e4ec21d4b8f90a64e526a4b26d7ef864f540288eb23c15e455600fe642221"
            "02fd67a953a993c1642f00d40a8cfd8ae16469b8b843d94418dd003d723ddaf00740c507b78c7a21902ecde5cd8ac749cd050ab934"
            "e05c0abc6ebfa6a2fe7bc7cab9aa8bdb89d2b736fee8ab895797b8f4690a947fe1fb522e14a94fba61829be643feffffffc4d10705"
            "000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201c7ce0705000000001700511460bd"
            "d799cab60150f87f1dce1ee8a763e94a803b28f60000000100cd6f6c05c80151dacb4e85d7cd15dd2c1d0a4f813f120586b6b4b8bc"
            "bdd14c6864222103f4513b8ea0dbff82c8f2a6c8295c2aaaca7e43f86f44ab24bcf634619a8b989f40efdd1692df2cfbbe8771d5e1"
            "4a2064cec64efa023eb76b88229b45ddb5da9836a1cbcd71ca443449fe07304b2340686a3ceb0944cc4244441f2773ce1f6e0f9ffe"
            "ffffffb0e70505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201b3e40505000000"
            "001700511456071600e7c13c5564369712ccdccfeacbe5a49428f600000001005b833c2c1c0895b9064f206ff4230f4bc95d87daa9"
            "f2be55cdfbe9bafda9af7f6422210288d989c1a1e66f4c4d6f28eaa71373152be8fa1e7a83ee184bd37d838724e0de401c763ba576"
            "ece26a163cb4de4661cf3be974cc9d2d613ad239ce5e08bc2f0999ebbf1a8843ce721d808c2fd3664be5ee523971a6532ae77fa93d"
            "fb854ea90059feffffff9ef90505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201"
            "a1f60505000000001700511465b419c960c037a6046a4690d98e741028680d0528f600000001008bcf813d01b2fde530c48ba748a0"
            "cc32002daacd7797776cfe2f4d935482e8e0642221037418676ae5fc6b1731f7ca871723289d11be864d0a54c9966c0d42deda0b8d"
            "9840cdd8a7cdb9ecdc3e318d3ee368a3962db0ce82950924ebaec36c4b546c365b2f03c5e03a481f0903b6f296efa7188109166fd1"
            "ba8fa187945c788e3fab30fed1feffffffe3b40505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302"
            "be441c08613201e6b10505000000001700511438987efdc3c868a7bbc402d895c7c5382d04ce4928f60000000100eaef5d4a76551a"
            "a4f13495e6c13bb22c1d366db15d4ee822c5de9a8eaf3e0bcb6422210259b70cd7a1b82fa149e541daa482b711639b122bfceac99d"
            "6f5bd2fd0a937fff40b26130a8779039d65ca433504d44bceec4eb04a46fa2a737ee47e615a4e6fa4ddcd6db9446d586fad47b9d82"
            "72466fe33335afc17df19a8d48a84395d1ab064bfeffffff0f870705000000000201220200000000000017005114d8825b53f946b5"
            "4ab5a11880c302be441c0861320112840705000000001700511407b1562fd64af98703d8e9e1573a902a0119baf628f60000000100"
            "dacef93ae95c17791873586c1fbd5de1b76df3b5b39a88a869ea7da3383b227e64222102045162dc00eb6b9ed238126d196a45ce61"
            "4674b372b31e38306163f41c17c8e8405922305d8196929209898bfc0e57a377c40334646c8e47772b91da1f1c771a99c2a74e3acb"
            "33d5649230231cf6701aa9d0d8558fc8f6a713380c015335b6f7f6feffffff4c4c0505000000000201220200000000000017005114"
            "d8825b53f946b54ab5a11880c302be441c086132014f49050500000000170051146c83af20ef404c520892460952a9747d283f8428"
            "28f600000001000cfa6749d295861ee42af37a795c2fea60cd6c3532752a676d391bc1396c9df1642221039163907b3513f4635006"
            "cdab39ba76f768ac37189343108cbb8b0222a591529240684cc73ee3b6e835b077df1fb3c29b6afb0f3dede8981a294b01efdec225"
            "fe5bf70b11c4ff3602ffac18b8b0203757159fc7a4a79e4db999dd5716351ba34c93feffffff3e5906050000000002012202000000"
            "00000017005114d8825b53f946b54ab5a11880c302be441c08613201415606050000000017005114395d48a36e80c5ac69d65526a6"
            "7c3b9625d5dcd628f60000000100bcd3036052a3549bf2d31e01708652865e34657eb8622113927c188168c1b702642221033afa32"
            "0abefa190c9855a5539a10d3275ec704a98670b436c3eff6aef7ea26a9403b7c4615692ca50f65bff46745c57690e90b1a57defaee"
            "4b0ad3f0f55faa50617bca6359c76743ef91d293d4522650018abeebd94af6b3031cf1e819b7a28648feffffffdbbb060500000000"
            "0201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201deb806050000000017005114b7914c89e886"
            "ab655929d16f7b0360e83ea1860a28f60000000100fe70d55e7e9310e58693213b70ec0e71925ce11bc756b7a36ec73a90f1e77ae8"
            "64222102222b14cf026c2ce2560d9a2bf5771999a99526dd8c48d2432cb9e98398cb2269404504d47652f155791d3c80cf93f45380"
            "4dc84618945e71710545bca59bf3d64b37d4430042bd6ce2876a22ea9923bd53e565e1d3a427aa127061b1f81b1f4064feffffff48"
            "4e0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132014b4b070500000000170051"
            "1492d53a0862c631c64a01ef06bbac3750dcf589e428f6000000010065e8e98a37a27a26c15b819b6388d2a6a4b110b908c819139e"
            "07e74be912eb9164222103aef1c708314e5a300f0dd0a0e325e0cc2d38c40e2f622337da9f45a35cffccb34082985e189e929b61e0"
            "7e2684039fb6790ec861d55fd0465e0765253d61068884f6cbe5527accc52436753683d3b8e73c904a21eac81f3e3bff01dc3c7d1c"
            "974cfeffffff2e6a0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320131670505"
            "0000000017005114f36fb0135bdbc1512aed5d85b89516ce08d068cc28f600000001003f16353b7ffd81e9ddee41c32f90d1ce722c"
            "f67dc9ae7141386b94f07282f94d64222102a0bdc12f6d822971d6051015e2949df50d52cf9ad2ad6469ad8673ec9c3aecd2407f84"
            "5e834d8710a390cde7a45262990bf5c92bfce11c4173fad16e3fe5edf8361c82a9aea2ef852845044794fa7953619f14286a752276"
            "5e8f0c91f3e857d085feffffff74230605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08"
            "6132017720060500000000170051147abe08a5ed8253e82ba9e7080dc08e28e5f8428e28f60000000100190445ba1f2ccba17b90bf"
            "d0c3be98b4024327f3cdbfdd2a3cdca202a993d99564222102bfbe3399d49a69ccbc74682cfcc2f6f9ce25e4208980b9ede8724b3a"
            "ef177e244011ea663c74fe66498f86da7d81ef10f63ff9b4a8683942e61e032ef05df74fbaf16cfc6d552a8280dffb2d580285e5cb"
            "a1a9a6c3c1d78936229a6431b63d5fb3feffffff17800605000000000201220200000000000017005114d8825b53f946b54ab5a118"
            "80c302be441c086132011a7d06050000000017005114d380d0d4215ba85747e78371f6249f0bc7b6add828f60000000100f83e8af1"
            "05930297daa52bb28efb3d34c8475f10f3c4b0f46060fd5a76910c0e64222103d3cf56999852126fc91b87ea1fd3060a8864cf0f28"
            "ac6eabdaed9fc27d38989f40c52723b4d9e75453d263f1c4d84898630deaeed5f795cd172e0f2f877ddc8652ee1507c35789e94f46"
            "e7431cae97ec4082d164f2c197b5310b5e08c79f37f4c2feffffffe7af0605000000000201220200000000000017005114d8825b53"
            "f946b54ab5a11880c302be441c08613201eaac06050000000017005114db1ba3881b31109a5605b750a440dcf0ff012a9f28f60000"
            "0001004a5455ce284520f8d74e369e905f60390fa0324de8f3bf448988bdc769eb9047642221039f5c10f739e14bcc447771134ab4"
            "1e0e48108b08c460d6c29cb5398a11efcc43408a375a4391839597a9cf77b7ca53a51e03d2faeb4c7cca03f04371d3641833a86e68"
            "2407e7cb14fe74eb262a23446cb1ca82f14bea2b6291ea82c527764456dbfeffffff52430805000000000201220200000000000017"
            "005114d8825b53f946b54ab5a11880c302be441c0861320155400805000000001700511485d3f2ccd0cdb29d5c8b18b677d1d2d20c"
            "95171b28f60000000100e8b5e60ae400b9b0cf6ab2587d7e4f2ba9bb2c97551ccd4a5559ad066f9e58a064222102da0e35748bbebc"
            "c01cd7e999ca542513ae9a16a6471e73038531a2e1b9fd64314060a90181882549b3cf10cabe3e05bfba3552a42e300456947d35d8"
            "af110f01c69b4dde82e80c6e53e7e38809a56a1ba980f1262a68db8f7533aa302fe3b7d9a7feffffff752107050000000002012202"
            "00000000000017005114d8825b53f946b54ab5a11880c302be441c08613201781e07050000000017005114874ac132e1eb1f7ba40c"
            "7aee5228c6b9b7c54b0b28f60000000100fa334f48bfed3fe8b1a82538f613c911fd6587af975494d06dc5f45f5d675d3064222102"
            "2d3d8953c1f5ec8e431563c4389740a801ac34d40b61b0f6173c0175e6b7cd8b402b768b02f3942569e33cd18027ccdda1ec576d63"
            "724152a6110c698dedab02d5aa99726d779a2b674bc5ca58f3bcc7df4840d5545b03b54b57275b1663236494feffffffadea050500"
            "0000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201b0e705050000000017005114e65fc3"
            "aa573792acf551c652d109fd02056a4f4328f6000000010000b9f60afc6b4ba58736afe24aadf03689add69f839cd7d05336e39280"
            "8c302f64222102e13ff07d22c375f488c860b2b2db250648dde93fbcf81b7cd85846c72a1fc9d840d17447eed581a4ccfe16f0472e"
            "fb8c39c5434f7ec813cc7f853ea04f91b7dfae703b25c4687c7c95442ade4dd04922d5ae1e8b3ddc67de3c1871fe94dd7e534ffeff"
            "ffff2a6c0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132012d69070500000000"
            "17005114891642834548b4dace0684f9629e20699760a26e28f6000000010065a8025297463e2800cf44aa6b43ed1290790cef1e83"
            "26b33c17bdfdba5606bf6422210279c3ecc03ba656e12a42a50f9a8e3730cc2a1dcfaf0d3c93345fe332002e13074084911c169d3f"
            "36034bccb4f283bc9c746826db8125ecce257a1feab0f259a6cf45aabf834bb2229176557305cc0b17e18e79d9712bd8fa7dd9df91"
            "88bb6ca3f3feffffff25730505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320128"
            "70050500000000170051143d2e9c0043b57fa7b8d9057a796ab9ec81c0580228f60000000100cfbba33f71e56a990034b03056efbf"
            "965c5d97716c1bc4e5161fef1195255db264222103230d5c6254ce0f1278b636e7164c0f9d66f3cbddddbe93593c9df02e5cc45010"
            "40e3f4fcc5289dd48715589f4070ac1652ee818c80d8be699dfe8939671ab9ace1b423319d5a196bd52d35ece4fb8228d69a07cd84"
            "863397b63a00431556182742feffffff791c0805000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be"
            "441c086132017c1908050000000017005114b7b15bae5833b37ba4148e11ab3b973671caaa4628f600000001004e7ed6e3a76b58c9"
            "10d9705734c6c9759ca461236f1964ff10c6506ee7715d1b6422210375e25d8035762988eaf6596dc4e162fe8464f263629d8903e6"
            "c4c90590e5e1cf4034793d44a16ab29ac54d4c7262df2a8e42271ed6e290a5bb7331140216c2b6a9f11ca0ad1e79b43a7f9fe4e29a"
            "a35a819acc92a738d4d1690896a02caea80571feffffffb4e20605000000000201220200000000000017005114d8825b53f946b54a"
            "b5a11880c302be441c08613201b7df06050000000017005114af999c3a67a3db9fc0a9118819861ec943d2b51128f6000000010012"
            "48a7a396a5e989548b09499b03995af1f8284fe7422c9c105c28e9e1d1cdcc64222102058e5909a86d7d2d6e5275a6324956fea4cb"
            "e54218c4dc28c5acef61851e6aab408c20d98a973d8e67c2d6ea5bcdbeda66ea7185e8551a5cb0786c3e4c5024c8fb665557889219"
            "f80bd1d20fa80859cf5a6799933d9813461129cb7cf92f8f1a9ffeffffff880d0805000000000201220200000000000017005114d8"
            "825b53f946b54ab5a11880c302be441c086132018b0a080500000000170051142f600717c812941ef2d04cb63302eb6f59e12d1528"
            "f60000000100e02afbbe5b506ee346341ce65d680a1ed67db11f1975a639ec416a0cec094f2264222103938920caa9ed659acffe07"
            "40f9b0f81f9f3cc5d65cd667df88e95f9d6d3fa8aa406304b26c87a8f0d939599dd6573f074b2ad03b96bc7fb882e3cec3839453c3"
            "23ace7f44723ce67b5df2cf4982e5d1d944e0863bcc4417a2e8bbc7e9a52f1b1a9feffffff1b7b0705000000000201220200000000"
            "000017005114d8825b53f946b54ab5a11880c302be441c086132011e7807050000000017005114bebf1ca6b2dc5cd2bd2ae9f500ac"
            "523fd696e12428f60000000100042e5b70871864e979310e216973cd5fad4cb0132d11a2b5062eaaf826c9802d64222103dec3bc9c"
            "129ba5260ee8eaab29885f394df1646a584a8824ff30ca56dba282c340b6cb0131da4599413b6defaf6193e3f600cad0a54db6718b"
            "e742fc1c179a44f098a356cd693a5ad14531865ad6c58390a695afed206c717944a978bd2e01c6e2fefffffffe9905050000000002"
            "01220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201019705050000000017005114dfdb6ee69111ba"
            "399333a9924ee3f47cf455299c28f60000000100f8c4d61cc84fc1bf49807824c0d6f3b8c6eb98828db2f92ad48801619a6c57d164"
            "2221035250e486d3ed0cb85ada226322da3aab6e05c2aa9696964dfdaeb92c785a9f82401eb2f302ef1d70d1a1f7d61b26ecf54a68"
            "e3c5a04f8ce0088a3dd5a18daea0c3283372b9860e06f0dcf6b143cbca7ddb1d45b3c979cfd4b1155ce47f7f242e1dfeffffff2472"
            "0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201276f07050000000017005114"
            "512a8df8a55155edba4e20d27f5cc20f57517eee28f60000000100625e19f4ea67dfe079397a1e0ee23b1c28733c8914cde71644a3"
            "b6d7d250a5a264222103d97ff77a312e11cbf87036a1e84af264ac4738e588e9005a99f2109209ec4c8b40796b443152d30263da6e"
            "31f4ff186a9f427078e06ae9c3c4e25a9bb64e5cf12ecee521ab35bade07fae3a85ace41fb7497270171503ea4e01ec199c031362b"
            "76feffffff692d0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132016c2a070500"
            "00000017005114dbf1af92d110ca2d254002882d78d047fc6bff5128f600000001005d16db3cb895451c3124291d9cfc349df301f6"
            "9cc79155557da9c0b7347f083864222102759de7dc2259dc69ced7f1e313438218eb17d087bb7e36a729c0a08be5780a7d405def5b"
            "32f2f2fe3570f650724b7325b21be7e7b33e4c4ccd0ef058ca30ae96d9e54eac439aaa972c96541f9aecf067f131b49af95d45177e"
            "5a99bab6cf844aa9feffffff6c2a0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861"
            "32016f270705000000001700511444166823e0be284031c8874d4165ed37d60bcd5928f600000001004bff3ceff6404e9c01528d7d"
            "2b473c415d7f32735d4f5a8b5c85b30ff67b29c96422210387f83f8457874476b405a65f60cf81d1c4b6e8e412e09c0662ebdf2798"
            "9229c040c915b9327af73133cb5229889d51d52b2a0136e5515f6e9b0cde5dcac2ce946b9c998f1fe34ffba8f9733fdaead0b92bb4"
            "b6095ef60a1da9ba475e6907eb23fdfeffffffe1b50605000000000201220200000000000017005114d8825b53f946b54ab5a11880"
            "c302be441c08613201e4b20605000000001700511420b42e0a7e0ee087e3ce81c075293837e208d51328f6000000010088801647cd"
            "ceb3ae33b1cef59173148ba940ea761cc1fccc62000d3bc6cd0f1c642221038dea9f46716674fe689a8938694c9f4a896fe6d844e8"
            "5946ca56d1aec4209b7240663f54eb482e7cebbcab3e2e644268de1e0a00bace942a8298fffcde5967ac2fc1506b89966687d6e0c7"
            "1febd793641e21b1e646fb494dec24e315428829c8d8feffffff70280505000000000201220200000000000017005114d8825b53f9"
            "46b54ab5a11880c302be441c08613201732505050000000017005114871b2198221208d559dcc5236359f23b3af5c90c28f6000000"
            "0100d177cb0329ce8b039bb0f4c621ae42579ef03779f3fc055bbc4ab4856bfca14064222103f0c53e659f9fe5d769b0454ae02fdc"
            "0e1e29aad56286152a5a868203d579bf654005f101cfbba88dae16b329b343f6113864e90dc75a25581e2abe2280f2a76c69601048"
            "a84819e63273fb39f7f0d7d0a8dc9da97dd3f1603b02d01726f927a63afeffffff0d8b050500000000020122020000000000001700"
            "5114d8825b53f946b54ab5a11880c302be441c0861320110880505000000001700511461e1183931131b640ca93912d1e522b8b2a5"
            "131528f600000001000cdbb6d7771dd526b330b11f3b1e05aeffe338fc117ce8b3daa32e55ca09099564222103b6f4ed34708e6f3f"
            "2f13e08a49f4c31ce2c4a4442cf47fb0874338f35224fe0d4073a51cf70390dfb88e2422b07602a6195f2e5f86186d86c02ca810e1"
            "4be3e664d07d72e33c605b849cdf2268bbd01b236fad8732a2d8793bc9f1a7ba6056a20ffeffffff43520805000000000201220200"
            "000000000017005114d8825b53f946b54ab5a11880c302be441c08613201464f080500000000170051147e0fccad6dade8227f998b"
            "523e4ee3fb39340fb528f600000001001dbd5a6435511fbd8d64bf55ca62b8ed9a145c16db198b32ba7049bdcf03943d642221031f"
            "1d9f8d624187534b2e7df86ce370b40eed89ddca9bf30c22fae47803c4c90c4048dccaa1a9d12b7bc4d678dc20069500d2c964af69"
            "57eebb927b2ab1139b00722b73437cc045a1d89244eccbc688fae87ee22dd97320c99f3c58908c4b4047c9feffffffccca06050000"
            "00000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201cfc706050000000017005114cd858a65"
            "963baf3b22ac001795b1eaa251d9d66828f600000001003de7010959bb92738aea1ef6dc33e2b4a7dcb2b1c18627bbcc79a629d71c"
            "215464222103d640aaf5820d2ffe7a6bf0fda364b0cc3e83e8d9f984bd3c12698a0861fbebba40ee5db9897abc4ddbf3eea97d8b79"
            "5a2bbeb582149df1992c4a9876cef0b8a30ab8d1a71ee770ba2aaac4ce8d4b5fb127d0bbd1e2e608fc3b791671cd3656f487feffff"
            "fffb9c0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201fe9905050000000017"
            "00511408fb58affd409f5e607c401dafb2be964e25697c28f60000000100d7e5101e3066b2c8ee7274318d520721dbeebab0a5a0d6"
            "082c2f14491b7ab86364222102cfa0bbb13b5f785ffcee654252284520743e7c0e13c10ffeee83f6aac317662e409562c212a075ad"
            "0eaebbf640d170c798650f1b37ff45ad1cbad986bc891f59c4ee4ad605cfd8d02719b832ebf3a40983a9dbd5616541854cd21257af"
            "ade9ce69feffffffa8ee0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201abeb"
            "06050000000017005114605aadc6e16f666c89f6130dbee27bf31717147128f60000000100333257eb9df143449c92f495eeeac5c2"
            "220f01050a82bbdd5827d32d1c9b8a0b6422210232b675eeb50c46539d0a0211923e70f01e0af75a9e0d438e7ba10ffe6faf74c540"
            "625ceee5b6cdb1bc3f9eff5ee2388e3d3182176f237eb061084125400b3696aa217851b3bc207587b905c4aaeba69d580c580c6f4b"
            "746f0ef2b7505bf75e3621feffffffcacb0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be44"
            "1c08613201cdc807050000000017005114d0b229e375c7295d8a1a8f9282b37e51e9b5211728f60000000100a4050961872b43d564"
            "177e0e68a3da8a6d26b007cde9afeb050fda36ebed68f06422210259d89b3fe0ef02b59707783f709bdfcc679217e5183e32bb2107"
            "34e32bae3426409ff3aa6ada1b3e26ec17fd0ae7b3f6e1983b61eeaa8bfdaea1c040cf7a41879b03c7f8622561f6f9cb4698830124"
            "0b6c068cfa7e4479490b13034ebda1894cebfeffffffafe60705000000000201220200000000000017005114d8825b53f946b54ab5"
            "a11880c302be441c08613201b2e3070500000000170051143b8b3bc1b9d8a1355a6d2f12a583351847f2167228f600000001003fb2"
            "b5c1efa00ceab7ea75f4d27ab896a27acd8559537037bc02b428f460039f642221021e9b04633713eaad08f7a8bd2740d4eae97e3d"
            "c1c74be43b6078e21840a9853f404ae4d5a97bb2d29f4438078df557a58c56f06fccd6d5376d5d8a920646d087682ee34f5f28b90d"
            "8f255d315297ef53a04337241a2c679b897880ed2f24a98da2feffffff55400805000000000201220200000000000017005114d882"
            "5b53f946b54ab5a11880c302be441c08613201583d080500000000170051145b7e23d4bcb0568223cc2a3882fed8a1c703ab6228f6"
            "0000000100af893217307b13e8b3c65480f5c4a116b45f66819e3eee906e8dc101288a40296422210276606804fabe6da2cdebd750"
            "129e2179c3f35dc73437f6174794938b7e81c78a40fe91d8ee7eb7919751726904ba9fdbcfaaa9d20ed2b207d551c7234913f7176a"
            "1ec4793c7ca9aca275d9c7bb6f0ce5b7cd2c4e32b6b8fbaf500d3efcdaeeaafffeffffff7224070500000000020122020000000000"
            "0017005114d8825b53f946b54ab5a11880c302be441c08613201752107050000000017005114226618a577b3df081ab6f019394d79"
            "8e073dc9ef28f600000001001bae768d9219d16347ae6cae758d7e1bb82b39106889e42216a919f3446d034264222102af6f1a9d80"
            "e12541b0ed555a90f2b6aa191c14860461fe00909b9fd8c5dd4425406eb1a7ee3f3c2d2bf8ae8add7829aa6de7e15b4731f7b2b044"
            "e53982b1af433f6f93b3d2e2c469b803eac88e4bdeb9ccfae43b9d671ee4f54cf9378a068f64e9feffffffb2e30705000000000201"
            "220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201b5e007050000000017005114fddeb06c8afe0467"
            "bf036c38b9d57e81d9313a5b28f60000000100c1e8036186acc51c336f5f2eb25e44fe321f697cd79701e4313239f735105fb16422"
            "2103dabf5dc44dcb99d201fed5449826918ef22f1627a3006d74a6aa74d5f34eef1c40b82209fe5fafe1132dc8031c88b2fa4c3da7"
            "7409f1681026cd52c24163e0f82663aa0269ac26cfa3e6554cab1de333c9340a812efb19a0afa4bbb3a6e2fbda96feffffff881005"
            "05000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132018b0d05050000000017005114fd"
            "0931d412881132ed08ef05d8a0c3b0e9a033ea28f600000001008aa50d14ef18eba117f7c788db5ee0339168004c0de55fc9ac639c"
            "437ba297766422210218f4bbdbbd52fc564ebdac09feaa518f987660b4347c502d406f2e0d057bbb4c40aa69cf1643aa07c890d9e7"
            "10099d8167c9d5591a848a2fbc57831f4dce6c83c0a9bf6f17ed77ff8a075ba73e97cbf0101cfb6f489ef25acd9556f85147c30f33"
            "feffffff23740605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201267106050000"
            "0000170051146979c3300f4e0ff85c8a611d0fbf80f341a753ac28f600000001008f98a18da14e1e75be8621cca4e7b08cc7433837"
            "bc3280dd35836a389ae5ac92642221027ce988d41d1562dbdc888a6cd24cb3813803a06fc93eaf9346d6b8654cfc27ff40f89764fe"
            "c70845dad5da62c4897164ef40f1e333ea63dc7b09a25c70ad2119abb98cce070a5b9f78b7a12d01e093776855a9731bc181d03ace"
            "776cd8bb4b04cafeffffffff970605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132"
            "01029506050000000017005114dae7a3fc364a1f2c41ccc6cd8d76882b770ad9f528f60000000100fbda5b0314cce74950c6cd96c1"
            "45d015787946c84c95dfbb1e9d4d2f251cc1a6642221037938c9d143e5c972abe8d2f32b0f0fb92347767cc000dc7547c7fc103091"
            "f03540ce4ae222a5b76b7f993493c95d90bfc6592b1592ddcab69ee72d70429cbf6db96e8d60cd5c2d709d1f1b90d637df562d75e2"
            "29e655ca0862bacf845faca0a8b1feffffff464f0805000000000201220200000000000017005114d8825b53f946b54ab5a11880c3"
            "02be441c08613201494c08050000000017005114f529f68a6bb1b793319883dd226854c694aa993028f600000001007d4604be2c6a"
            "9d049e7ecd8b610d3ff22021c51b2d004208fa566669493b4dc064222102c4c1a99d8e6727415ccf3f7693095748f14e30d6a4c385"
            "b0412138aab04c856d40915183017a90094eda1c0e15f568f456f05d43f17d731c9b3a5e6302197c0a20eb48e2ad4172b6c3b03bc4"
            "fa7c7f92b12de98bf630e26daa7209ab91e3635d12feffffff098d0705000000000201220200000000000017005114d8825b53f946"
            "b54ab5a11880c302be441c086132010c8a070500000000170051149a62f08f1ee4c5a369aaa41c03f40cea86c71efe28f600000001"
            "0020380319a247a72057b2fe8c6588c1acdcd75086f1d5b80dda4c04c1997a32196422210351517db8401b8a82dc096ba27c274a21"
            "6490beacd4f3417eefe44341bda3e4ef4006dc00cedd8811e03d3641e1585c08977d787d3c9fd9224337b79eff6b98ff92fa35c39b"
            "8548040d4fd117e7055b30ea8dafeafa5d8e5369715dd2f2282106fbfeffffff435505050000000002012202000000000000170051"
            "14d8825b53f946b54ab5a11880c302be441c0861320146520505000000001700511427711269934af131987a963a4281fe1bf700a4"
            "6128f600000001003ca30ea8905ad52c776ed5081fdcc0ad7e151f1a9955f77bce5c1030bc2fea1a6422210202f2adde1abb4a45c2"
            "488816a58b0892a622b01d8f92f84e023d48c921846e8e404900b5bbd10a3f36947f930c97e95f76bb24b3bea6dffcc21679735325"
            "0b0b1548a8cc48521df7945cdbeb407e917defc776fdabb7a4ad91e4f465ed1b76e793feffffff5a3c070500000000020122020000"
            "0000000017005114d8825b53f946b54ab5a11880c302be441c086132015d39070500000000170051145dc6c3b560abac56f3ad66ae"
            "e3ad2c097556f84728f600000001003378774e39bbb1285f272953ad2cfabef7b3967dca260c23192ff69a940db6ed6422210325fb"
            "ac9f46a32ab4ad425ac8b47b094ec84c81903b202f257198fac0a84c6d1d4076bf5d7e84279a89162f3c2eca3e1aa34f7d172abaf0"
            "f9f0db0b5fa0029cd7697d4016601e5b2f82949183a5727cf7888302bb02467ae620898eb79d8dc96d26feffffffdeb80605000000"
            "000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201e1b506050000000017005114d289cf1f1d"
            "070b5e732b2578f5aa59ed175fde7f28f60000000100a52d0a3c7bced36c86804a36f2fb0746d0573cb15aa89947784e7aa3799c12"
            "426422210346c63986f2a1ab72bbf2958dd9c8c1d9849379a476e63f88bccb72f8f8a461314067ecff53af231b25e91eaec8427a2f"
            "e832e9b9237a30a6d442002e62debb130d627e9b5bcf999014ce7a6c8059b3665e85f7d54932b05a2884c139f96b73f92ffeffffff"
            "2b6d0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132012e6a0505000000001700"
            "51146a12ce5a64109faee6d4cc6e91d6bfc42490c15328f60000000100bf19cf44ca1a1ffc2cc6ed562e3b5d3aba177513b5a50666"
            "1d4bbe7e2404a6086422210225e09636fa8637919b3c8b71881ea24e9d83e62e3c2d9da85b8ce2a01733989b40896cd7cdbcefb431"
            "a34946f13d4c9ac717ce389f200fa50c538bf1ff57fc2534304b70f14a22f8221db95e2be28e4a07e593adbf0e971b3654a40730ba"
            "d09b81feffffffa7f00505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201aaed05"
            "0500000000170051149ba407b85d2dd137c85427d548e84a3b93c1545128f600000001009ef9e6a0df58a7a448d4758a7e402eb026"
            "f3b71388889e1462352cb77cbd44b3642221021ddc70584cd249a50a68d1fd50b9786154d4b2982c3554ee981047a1cdb79bd84039"
            "3a23e67cf48e318ef25a227b44514f6c96a1888fbf368b87ff8e8183cc63921259cd437028e0ec4ef6353264ce0db894dd9bb07499"
            "a9a1ce15c61688e0b16ffeffffffd1c60505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c"
            "08613201d4c305050000000017005114db692ece4b3c3782cf8e56e126a422b82bf3608328f6000000010069dc0a6b8eb6d970fab9"
            "c21baf8f522323338bb2801eb3411bdf8f12618d3b6c64222102d658967a5cf243d6801cd199a9c018dc62402d868e83b535507864"
            "168d679acf4009b2efa10968899dc51519eaca791411b0a7b00b5aef70425ff3a60c67f1b808f3322c768bdf115d910152adcff6d6"
            "14151ea487fe9adb4e676b438335e16b22feffffffd0c50705000000000201220200000000000017005114d8825b53f946b54ab5a1"
            "1880c302be441c08613201d3c207050000000017005114bb5ecd973200922d5174a7f81e792983657bd3be28f60000000100ff6767"
            "3c0e5477e54eef4b2b83d28008305c03ab02568059b4c749560a40c1cc64222102d223c248afc581f426e874da08ae52c5d5ca9358"
            "36e6c1ece81b425c915cd6224091d999e7622827da7576e65427c6dbe15f4d929d902519ad1628a5b631974dfbee96e3bbc5da2008"
            "64d7e33ba5074ba631a9df6050c1a37fc54101af4ba0fd7efeffffff32650605000000000201220200000000000017005114d8825b"
            "53f946b54ab5a11880c302be441c0861320135620605000000001700511415177726b8770c3d9ba6c9a9620c0cd420a5b25128f600"
            "00000100d702e7004826f421a227a36d600c8142d00c7575b76f01bf124017875154d60f642221024ded1370dd00e435b49d28084e"
            "8d228569c43fd36127be6c386ebdb8ca1afc14400ead08841cc24ebff35465cdbf1b90b659a6b7b2b796ee2737a5aa117f9c261f3c"
            "1a8ca3d59e104c00676dd47849e7a2cc51b4863170834cb05e823d2d2f51abfeffffffdcb907050000000002012202000000000000"
            "17005114d8825b53f946b54ab5a11880c302be441c08613201dfb607050000000017005114674231257c1ba31ec7264e1914246c88"
            "15e1c55928f6000000010004ffa91f4f1ecdb096ab133188c1c4ce0a393009c2f9a886a08c5443276b095f64222103355d92caf34e"
            "56ebc123e10b9519b8f0e2a2734058994a66bebcac7406beca2b40b104f7379adc77bd679cad5330ac421b5adc142e5b767de261b7"
            "87e0c472eb0d5a94bfe49bbbfb050b730cde09f93ded86c35f3ecf5e0375ebb2c272840f9d45feffffffb3e4050500000000020122"
            "0200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201b6e1050500000000170051142f8e9dfdcd66bcbdca"
            "a84a54966a5b857a19e43828f6000000010040c08dda11be7683aa7cc2cae52e42962fbd34695af4f982fb75df4c5c4e321a642221"
            "03c4a0972a764c56368a7b889867a5a5c1daed4069bc5e0a11d0160d71333b58b3405fe02c5dff1964f8df047c10b148c5129b5135"
            "81793a7b5c838b53d9e3095f6c9570e4343b50400807fa6f4f221b6393ef8e1ec7c2f17b06d1102a5f792ede1bfeffffff00960705"
            "000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132010393070500000000170051140e4f"
            "36872952049839e1561e6c0750c535a24bfa28f60000000100e97c11aebfe498676fa0fcfefd274c78e77267e8804a82d52cbbdd7a"
            "568483f564222103ffc2a87289abadc12205c955f5c05d4e438f7b159dad1d95bafa581dabbef529408011832c6c7ebfde28ed962d"
            "efef5645431209638996f67752b5b10b0f48fc2547ab256fcce8287ebddc03968b95d8544b7c7006d392daebc81ee5856cfc1bc8fe"
            "ffffff1a7d0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132011d7a0605000000"
            "00170051141b66861edff161f87127248da8acf50c63ca41a728f60000000100d9c3a02e48848bc8f9c4dad922bbc37d580c4a392b"
            "f2a9a9cf4890e61973e272642221020ccc8bfe4fb9ac9704d7c1a65c47de0b3cd61a4ac65f10db885b85e0fa853c5b407c662588b9"
            "4d5ef1df9b224963728a754810f8849ed4e0c1bce731414f90e5c658f52aab73559cb42bc797fdffc83008e0d4ef35e1393895f54c"
            "2cb68d35da43feffffff4e480705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201"
            "514507050000000017005114b882d5e55bf967c30c40722daf91da0e1dc28d9e28f6000000010055c5527cf5f285c98fbbda76094f"
            "dde6050ff3a9b3506becdbdcb037b7424f2c642221029f54fa37e0913e2a20c3c8dfa1055a3d2b7bc85ceba5a4ca780f83361f2ebb"
            "af40c9e3653145245714a59c750009537cd76166bce4b41785dbb184ab16e570b1e3bf497ef53034058da57cbc8a2c06bfa8745326"
            "b4c0d78c21f3f9e17127fe69d9feffffffb9de0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302"
            "be441c08613201bcdb050500000000170051145f0979b5466ccef8902c1fb62cb464316226f52028f60000000100e41a4edcb14f81"
            "82fe9d1ffb07c1907195b4f2f5ff956324027f253d5b89fa87642221032d15b4c2ef37a3d4f52e5e7ea8a3cd61a0165642fdaeaa71"
            "e1f5b98749c71fa0400765d0f5375717a28ab594c4675e007f2a1ab81becc6c98391dd0a0bf7c3e7e1580c4d78db2fe2fbc6063e4a"
            "53c9ecca503694e5ae0b9eed0563ee938d345f9bfeffffff54420705000000000201220200000000000017005114d8825b53f946b5"
            "4ab5a11880c302be441c08613201573f070500000000170051147b7e45a28aee56a8849aa174cf70b9624b8ee54f28f60000000100"
            "c2f470fda139df95dfad83f5ff72e896ccbbf6dde431183ecc2e3d32012116a664222103ed7836b9791e360bafbaffc16ec3543c9d"
            "5ddd32f65c2336125f335d817e4a6e40e5f415dca5d80424c608413885bfc8c243aad516243f1f6a7bb3048f2858d955e0f8cbc1fc"
            "0504d5a7b549a96ca748bba04cd0216453f7b8818c79349d3d5e25feffffff8d090705000000000201220200000000000017005114"
            "d8825b53f946b54ab5a11880c302be441c086132019006070500000000170051143c1a267e9a23c2cf1f8a294a389816d8e4eacf44"
            "28f60000000100a6c9f98ed5091146b1e98372a94026bcbd63550e781d88bab1d7031f03fa98326422210248a3d2b610c5634f19ce"
            "08f3fb9fc4e00985b9dfb863e8c0d45ff4287a178c9240b237c9243e9b38d21dfae5051353c9dd24a4ddd3b7350f0c50924f7d6ff3"
            "abddd624ef67a12e82065da288eec021295c4085490a7cfb354c226db4b72bfbe08afeffffff5b3d05050000000002012202000000"
            "00000017005114d8825b53f946b54ab5a11880c302be441c086132015e3a0505000000001700511482a4798e15f9a5cbb8e5cb91bf"
            "b9e5db57b0fbc328f60000000100cd70129ab30590ce663785186a4d4ed55e675dfe57ec05c5bfe9c584966511e764222102d70e83"
            "437c8865e073dbc8ec9d2f759d1fcb4e038e40abf9b64c4e86e72378664001f978acd1a123ae1db2b0a2d14a1746265628d9926923"
            "60db115e75d08fd807a92dea925bf30c0ebbf93266d6eba44ee27c91ba79a85ff444b5a571dee6252cfeffffff494f050500000000"
            "0201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132014c4c05050000000017005114acc0689d6ce3"
            "eca25c462a32145099cabef4528428f60000000100771afecaa092c08ff9457d68e9e65165e0eb5587924f1b68ce335b680e959518"
            "64222103e5070e9904bfcf9b79b1864db1956625401a605a9a40ad28372c71d33f3d550740645a5ad37e1bc04f066bd4a77edbbaa5"
            "8dd4f3ba6c8b924fafdf474fab668472e86148e745dac095ee2945924380a3af57a3fd480354fce7f273dd48fccfe544feffffffc6"
            "d00605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201c9cd060500000000170051"
            "148aaba8fbaf80005d7ee24a627acf75663e688eba28f600000001002dd0970442377519699694d16ed9dfc9d9731ed7ea80f82d8e"
            "a2d6485957815f64222102386810780cc1dfe347e0de85cd3dbff2101a5817e3609f1c3302e21cd3dc7f3040f9141cc3bc68f1a9d8"
            "dcddcfde33089515f163be5684d0a54a208766195960138257641d91dead33fce2f0a12811f73a63fcb72da1151ad8364377039d2f"
            "5e7efeffffffdabd0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201ddba0505"
            "0000000017005114854768d83917868dec100576be0207cebff072a728f600000001007e5eb4753109c100395ed4a3b855ff329569"
            "25976768d115c7570c6f8ea290406422210210f1afb23d1772699cf9b5cd4741d6ecf91121d0b53ca024b053b36991c116cc4005fb"
            "164dc3c78974f89b9cb6bbdf65038f4a9bf87a7dab61e709240554c631702f96570dba47ac2544491ccff8187ccff0b04164ca04ae"
            "19a5c999b8947dcfc7feffffff8b0d0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08"
            "6132018e0a05050000000017005114334773fcb390fac5f46716f4e33ea7aa558d115828f60000000100a47c27b8198bf3474204e5"
            "acc9e0a877b156dab2351f5c185b286dc0abf14c9764222103b25570ce83cfb17dd047d715a7db8357bb1c0bd9dbd03942205d254a"
            "fc57ccf74038bb43906e3507d7d43abf79c313521bb0d309321e6978b8dfdcec253470ee8425f497b51bd182232c28afbf3dbf2a35"
            "e370027c6a69b9b7d230cabe63a79e62feffffff5b3a0805000000000201220200000000000017005114d8825b53f946b54ab5a118"
            "80c302be441c086132015e3708050000000017005114dd26d9846dd448f4abb238f98965909f7cc3fed628f60000000100c6ce01af"
            "e94cfed27e0d5b9ecf688630aaa5f5e922e79cdef391bc8d62e483c56422210308a74c454a14d3e7aeac63f0aae880de112888bd9b"
            "c1d57f51634601f35f4c5640c780727edbc2b663eaca108b736cca9f859093792ff2e8b86ae86dfe5ce39f6427117d175a52720274"
            "f47a75efaa5b313712b1e274e89dd1302f7a29671417c3feffffff7c1c0505000000000201220200000000000017005114d8825b53"
            "f946b54ab5a11880c302be441c086132017f1905050000000017005114126f2b2b0ed3d32bd44b9087c7c9f856a1827f2228f60000"
            "0001003626a3b36edefc5f398223a1655be9c41afa6618a6c39ec3177bc590ecfb7fd464222102603b4910d1075169db06159dbfc4"
            "2737b70fbfa89cf9177c188ea1c7845d0fbc401f7cb352224773935b21c22023c7483dd475726f8ee8ed1d3b7b3bb53ea8aae34909"
            "9dd409a14ad478c214f075d6f073e105f60083fcd34b99aad555c1ccd512fefffffff99d0605000000000201220200000000000017"
            "005114d8825b53f946b54ab5a11880c302be441c08613201fc9a060500000000170051146e3c06d8118d589d3fcaab008ec3277840"
            "3b94f028f60000000100318d5f82d6f63954b78b993346b593375953819c273aa753571192d3734d989e64222102e1d35ae522812e"
            "a6ef9bb754d394788e81e2af00cf3b2cbcc0314642a048b2884002230f23a29af69bc1a8c74c2b1e02d0a0978b9dd7bf479531f552"
            "65c7c11005b546a05d143acd8bf3bada84f3f4ae1a7a6faec91e67bbfd624fc640db700f0ffeffffff5c3b06050000000002012202"
            "00000000000017005114d8825b53f946b54ab5a11880c302be441c086132015f3806050000000017005114aa56c99446bf79720b6c"
            "aa9256199c97c5ffbba428f60000000100e27f93235fd464deeeec518ec403767c0aa588c32f95d5965d0556db14fd3e9b64222102"
            "4581bb0a87a4ffa5426fab21df7473e4616728b00bb80fef20c92ff51d781c38402c5a99ccc8abc931b2cb172578bd0920e6768492"
            "889b73fe4f2de8089336a9856273500a2b2c445b23992ad972bf98447029c39b3dccabd1d43a346a37be5dcdfeffffff9dfb040500"
            "0000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201a0f8040500000000170051147a91df"
            "a958821b609ece5aeb48ec2b76292b0e0728f60000000100f16a50c237550dcffa5c46d93e0df3d221f6b2892b90ecc9e092fa2dc7"
            "9aa9b6642221027d048df34404e7f8e844667d6ea2eff7b4bd63f216452f64f3cf0e582b00a1dc40c72c0835335bbadda94fbd300f"
            "a3cb0c75c3d13606e17e9671a8bc1ebe6d8bc472da6ce4fbc8ce797339c3592c7cfce5f62e3087799854c19eff56dcc285d6d8feff"
            "ffffd9bc0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201dcb9070500000000"
            "17005114338eb6cd79995e87e631c67e736dffbc1a2427d328f600000001004ada14760440bdd4601fc11b32e5db9606eeea567449"
            "00196af1912885cd31d864222102b840f86b8eff5c27577bb81673cbed5bd8be06373f0dd2ca368fa5e3cce070f34024286bfeadea"
            "fd86f71806ca5500077c6b50edc4f1b3832bb92bd4ec5218b1ed1b3e9a457595ac461cf66935665f9be9799747f3dcc48ca51fe20f"
            "fc067e46f7feffffff44530605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320147"
            "50060500000000170051146d1f8a0e3ce899f7c92e48e4b3fc6af94c66436628f60000000100e3af88b095f12baa7ae63bec331e35"
            "8a409f3f2737f8c86774eed5935e3caf0b64222103ba8cda777b4e8a47ea85654a688ea1178da639b23ca6484ae836f1adad2e65b0"
            "40eb24279c10ca37876a1e86e2fff4249a60938b513339754c84dadb101fec055280310ba4c2d378c443bbab98e78bb687ef4859a9"
            "bafb6b807f7643c4caa04155feffffffd3c20705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be"
            "441c08613201d6bf0705000000001700511478ebfd9fad63fe50566cbc12f108c758189a91dd28f60000000100504e00f7073d7ffd"
            "125611b7105df2bf8bfd9985a8dabb99065e617b791ec1d864222102e069fed013444476a9da47bbddb3f953bd39063d1aa92d93a0"
            "e9df785be4ad3440ccbe78532d58bb13dcd8cc1994a8a04d57e7bb70dcf5f57c503e32a3e551b5b90bfddb787277162a30643a5f3b"
            "96bcec36b1d908d502d6031379574713de5233feffffffaee80605000000000201220200000000000017005114d8825b53f946b54a"
            "b5a11880c302be441c08613201b1e5060500000000170051148734ee1bfb76ecef2454c78d6da99f989b405aca28f6000000010072"
            "88a8349fa9359cca760bfd9905a3bd81e417b39a76dd1772d2ae2b7836570c64222103c460d3ca398c08aa22682070e853e3c20830"
            "c3456c064a5a2806413e964c2870409ec5b6a2200e97aa6b534a0f174a67159682b9bdaa5ea1b50502ca49a6d6a31d35b2571a7cdf"
            "8d577d0f11911004bc2ac388db7e8c77dedf874f8136a68d2ea0feffffff70250805000000000201220200000000000017005114d8"
            "825b53f946b54ab5a11880c302be441c086132017322080500000000170051147ed599ac96abb5532a9bdfd4edd526732f01bb1528"
            "f60000000100a175a04ff10ac718b168ae8d62d4e7e9e1b01e96935264140238d5a92ab8e2c16422210354db16af6b7052eb6f9408"
            "3be043acab69afe0720e44d4560563850b2c5e8b6040aedfb408c5edb9aacf3a41e381be878c4cdc940e08a65c1a2f1fdf945448a0"
            "7e903ba3e385c360c12da89f5a0c9b2ec42d869216af13fce526ec6e0665cabd45feffffff26710605000000000201220200000000"
            "000017005114d8825b53f946b54ab5a11880c302be441c08613201296e060500000000170051148fb42e756f85050aa625385f85c5"
            "859dd6995caf28f60000000100a50ab694d7348e4e111fd048092c4d44b83d169b33dd9400d623b6b41374d0a464222102d3688fef"
            "8dd008aa367cb6c50406bf1c631f92de96fcbd58ceae390433ac3fe840092f0da66b4f4401e7675c144b919ebf599e44f0cec21ea2"
            "ce221ed1ca0706178062155f328bdf20aff783ed658b79a0bbe88dda5f0b26105ef398d09f28f9a8feffffffb1e506050000000002"
            "01220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201b4e20605000000001700511427ac7846341720"
            "14fb8f752f90d72e2db9a362ce28f60000000100b90e712834f1ac779d6e1b36dfe7a4b09356e43de5a648f652d1b8a1db000e1264"
            "222103b8910f2531c8ab23bffab9a5149b62c62ffed4d9f638460e9406121a7c3dfde840cc3e32a7a2945107274d22114f1f7b5a31"
            "2390018a3a43da59a6fe86dafc6a0fe7bec4c9f18122b034aef3e1c488f2e5785ddaf7177fe54bb465e975f6741e64feffffff4055"
            "0805000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201435208050000000017005114"
            "9d9303225604fa0c75bd067fa6cec1d9e4507f4f28f60000000100b9d48541c78158e9b7f5a57a7f9e3de30ff0408a10f036750e0b"
            "25f9c460d97a64222103b776abf824a59786d3d61a54753d86d94e18473a01c5f7c210bef9cae156c62040d140b551224d467401d4"
            "19d6a11df8609e1ee87d0c90a5850497fb867eec63cda993efb21a3f11d14570806c5f6aee3ba1346a0984fd01afcae520eba9ca95"
            "acfeffffff66300705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201692d070500"
            "00000017005114010557a703b523e1120953210382f09ae0b9a5b928f60000000100cae005cbd46a3f5382a5cc5c2b365e7e9b2459"
            "4b3f3518b853c13973b91fa9da64222103b30946c044babe72919591c1dbca72dad0b8129cc2d4069a9a7a10b0d21c5a34409f5952"
            "eb2acf782ef467c114c550d4b5959850eed68738bbc8ff47b11b69395076261439d3603d7f743ea9402dc85a85f988a8e636baaa17"
            "1f74c5d81c0fc7cffeffffff60360705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861"
            "3201633307050000000017005114e5b80c6b4893e7ed486c4baef39d825e3a02e51128f600000001007bbb8d0d4f9012092008ec03"
            "bbc1ad72cfbe6d0039d99715f4083526894c7f2d64222103b74a030cd9cad1b4f8d6a50cb01f15adcf34c2ba9056647b5f496c4171"
            "ca551a40d895a2b213345ac4b79d5f6fceb8e285dd1c02d57dfa1b592eb1409052f653a4f798a6ff0d0547eb0fb24d120a22b7262c"
            "1eaeac2a951aaa2e4f2c5f7247b422feffffff97fe0705000000000201220200000000000017005114d8825b53f946b54ab5a11880"
            "c302be441c086132019afb07050000000017005114ee984dc3417c65bb52d866c77128d49ad3b549f828f60000000100d670433550"
            "c205b058f5476433397a546113a073369738eb0702148ba0c4e882642221033aeb0c2b7e17373c68c7b01fb451d0b961b806cf2b34"
            "c061ab1b837ecbd4ea3d4004ac6a37eeecc8e50974706500cacd87c122f9d502b516e4b8fd435b1790f658a3714464d95416ba2aa5"
            "be699bfa4e63a9d15ff1c9d2850e13b15c8a7157f60ffeffffff15810705000000000201220200000000000017005114d8825b53f9"
            "46b54ab5a11880c302be441c08613201187e07050000000017005114ce79375defe14e6f8af7c2748722a293a5a6856428f6000000"
            "010039c1545c5d85c983a3f4af8477914a1a795c1369c70ffdea470004059acfba9c6422210399bbd820d9073e9cff2878769a93cd"
            "d0286a5d70395ab3c87f904506de566e2b4030597848658605020c8a7ac8b09f1933a5ea5315f66c1b3aa12393c36f2b65bb089a7a"
            "7febaac626e2f2325234631b4617c09699d2e43bf597dbddbe02fa5755feffffff4058050500000000020122020000000000001700"
            "5114d8825b53f946b54ab5a11880c302be441c08613201435505050000000017005114d55f5959f4830a1791379a4e95c0b0bf2470"
            "2ad128f600000001001e241de4a0a3453ff49fb83cc3887a48fda7b96e3f11c61551da5b42e1678fb864222103130d0f88a370e739"
            "16ba99a4bea5d65b1a52ce3a9d3436d4fbefadea71fa373b4030d1757dc78dd243cfa247ee66465cea594dc29214b78066b2656341"
            "007549ebca1ddbe84a4f02c8de982badc939ae0b163c61f406206a01d4f32427f3ee1fb5feffffff3b5c0605000000000201220200"
            "000000000017005114d8825b53f946b54ab5a11880c302be441c086132013e5906050000000017005114a9edd20b70cb9e288746e3"
            "c5ddc5afaaa74b73bf28f60000000100ef6449d248093e31a9757da513789364d9a26211e0b8be9e2c04cad6fe8aa5476422210302"
            "7ad461315c1eeb4b7336ab03d21883c9223f2ad6eb0f68e889cce69bf4b89140126ef9d76db6672c95117fcf64c9e445c66ae9b6c1"
            "82179e97452f22d8b6fc15130e34ba4ab149ed7ec69ee98acc95ceff44448b9d756758f8b92e7b3d69542efeffffff8e0a05050000"
            "00000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201910705050000000017005114c4c4707a"
            "d37a39db83b355dfccfda213078031ab28f600000001001214514e993251ce83a0445f6405ac52cc1fdea3eb72d83f9dc9fedab3a2"
            "89dc64222103f55170457345af883724e118d02a7727847544b4bd6015a3ff363bcea01fa91240479b1d785e08b292b6ba7b20de7c"
            "3bd77e6261c6a392e1c7cb3696267e53ce8f3bde856c6ab4d811137f1240bf0537b6c6db8b50d46143cb636def1f4ca2ab24feffff"
            "ff761f0805000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201791c08050000000017"
            "005114d6627ded83f6a87c07d6828aa919ca601c93b68b28f600000001002ccbaa4a834b938f8d5e6cc34604adfc47d11a4e082099"
            "be62e9a7e2f630db43642221023a57e2524a3784e22b44bc5c6853b2dd6b5e26dd930aaf3b6068da1dd20bddc940d28bfaa7ba6b38"
            "5f01214e1f739e85baef7737b60041d7a942d4119ffe6bad2cdee11f489d5da5f81da9b79a1f3412118f336d77e9fc1ff0f091ad84"
            "ac7f3a05feffffff77200605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132017a1d"
            "06050000000017005114bb70c299eb92e115e02a46fa0c26c29823de64bd28f60000000100997c4e650e7268248c8d1bd955578016"
            "4714c85abb4db59c4f607cd0d08486f664222103bdbf14c95f90705d835d782a347c76df6ec99e9c334e07e504b11db4999321b340"
            "e808dfc5c5106834f16d90b9528625a43df098ccaad6c0ca2b8daa9905072ab7255ea9c16ba433b719870c139c5b90f16064ddac19"
            "83c8ecbd04a12cf3fd3b4efeffffff94010805000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be44"
            "1c0861320197fe07050000000017005114d30383e513c2af1c537e9d9cb82c8ff887055dc028f60000000100ee48bb2067dbec1cda"
            "614ecddaee676ff462a06b03becfb955e75badfca0f6fd64222103eb6b7a4e39460e5fa1006094d9d4d4bd6738cc4f94155fe25e56"
            "ee76651e9cc7409f68026d979ece97e2147767afb17bf9abe62cd7926cd5f5376fadb462312e623e9815e071a0a86bd895ed3889bd"
            "7a330932970052f672fab30c2de7605541befeffffff92050605000000000201220200000000000017005114d8825b53f946b54ab5"
            "a11880c302be441c0861320195020605000000001700511427c33e554c868512c87b747c1a528ff72e19aab128f60000000100538f"
            "1c82528e6168b06af7ae3d064bfa089aa808c06f3c725b261b4e7ea2e2db64222102b02a093f4c577a506f73ba8e95bab1a3c60679"
            "4b544974b9e7ac743ee8517451406738f4698661c915e931bbc0e1a7d0e9893f36cfa8bf3ae6af85942ca81df8710c2040e3345263"
            "f5cf200ce7eb372267eaed262ea9ae899f8a235167973e7682feffffffc0d60605000000000201220200000000000017005114d882"
            "5b53f946b54ab5a11880c302be441c08613201c3d306050000000017005114d852dabf086e077819af9014c505434942e9777028f6"
            "0000000100979a3dcc5b4ce8440ffcfc3cd02be40f3750232eb3785eaf8af7022cf4ffc70d642221022bbe391a24f4528a07669983"
            "e6923b3f0ee5847d0423bd24c9f0ac8fa439e6f640e207da084c6b1fcc9807a3bb019b7d6bd6a9382bf7f0476cd406269ebc03ddf3"
            "5fbfaaf7642e636c1cb5ca119663d5dde0f7735663d2984553cfa23dca196f66feffffff1483060500000000020122020000000000"
            "0017005114d8825b53f946b54ab5a11880c302be441c08613201178006050000000017005114ded02faf036e2be1d1254f3f8df1b2"
            "49d4ecdebf28f60000000100c2dd112d6b3c80d34b8655bb2fb8ffb0aead648e0731063dae9ad149ac08968764222102cfbcf04545"
            "4368fccce2ecb8c2a2d5d8f6936a53e9f6159b6de7241a07322de1406182683c5d0bb75a7e143a0655630bd5580a1a75576b81d877"
            "654b7c8fda940c0d5908b9e60e771ec0cecf4bc8847c9264918aff9dcafddc4d4dcc0090244caafeffffff4f460805000000000201"
            "220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201524308050000000017005114a307945e349f8f49"
            "4050c32e3940e8a3cc19e02a28f600000001006b6eebd8d12385300de4d876d4b0e6e658d7d02b86e1b5282051d7d1a5d8bf786422"
            "21032b7df674b87adb34b6c75348f5eac75df69bac0845c8044481dd27608c2a7a2640b5c9f1760425797dc3e9f0199c7f00bb4839"
            "95e9073236fda43a4acff729055d2e4994b006c6dbfd37b8b643cbba92fa062e8481bffbbffb904a2274114d215ffefffffff4a107"
            "05000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201f79e070500000000170051148e"
            "7b55e753f017480dcf753fb6beaaab4b0ccb7428f600000001006a9d3f7d1f7bec8f70bd20d77fd3d381092b7ad5095881b0065203"
            "1ef65a83a9642221027afe78ad7c3205fa8e2242883eda8778b2453afb95c838f2c2cc4a92571b72af403022f77155babfdb529d1e"
            "bd4b64cd0860470559781e7563e5fb5f2da6af6375794d5215c54b10de2b6433d810db0e699a0eb49d3d7827211915fec910f44072"
            "feffffffa3f20705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201a6ef07050000"
            "0000170051142ec1a331fa53ea612547c0a14aff7f2d26e82fc928f6000000010074890cf68d4d13172fcfd4ed08881c78df5d87e0"
            "dbb910e615a33add3fdc7b6864222103b0eb2870e1b75d59ddc683183e36d3052b895214a0f69ee6c3cd0dd5a7ee01d94050614cc1"
            "9263ed48fbbfac741a255c8c731e334ecb21b1f818013ecd01078c9c0340c0d0244b81e96a9e79aea9363a3f2a004155575fa2edee"
            "23dbaddbb137d9feffffffa9ec0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132"
            "01ace9070500000000170051142560b4eb4561c567700e1ceb8733ff440b3853cd28f60000000100b3306e565a595c771e95fbeb12"
            "0f366116217c8d207d222403fb957396dd30a864222103f7357acb79de4a0c9f62648ebda1b7e3dc13573191b869b169f91c16abef"
            "4abc40c8e07490258d1030911a5ae906a9adb6e0cc9c23aa4ba50cd5498a1c7dadc42af34e5e3aca73d56715b25803281fe203891d"
            "f61ae8ef82c32f7a0042e18fe12efeffffff0a8e0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c3"
            "02be441c086132010d8b05050000000017005114fce6bbebad3c738edd41663a6611757e9b9c435f28f6000000010092765a4b5b94"
            "ead0a213c4fa6d2648f864447504864c0ee6eb24f31d57c7077764222102f58e7ccda85b5ac6b9049ba2cab5cfc4b7d8587fa5a4b9"
            "dcde5aa903a63bce3c4013050745017c66f6422eca067ebce4003446b4d3fe6c525ecbb1ecf3ea2e2029c934cc582558d21be99199"
            "f6e5f6878413f212ffcbbb9245c8e817af142ce7bcfeffffff76220505000000000201220200000000000017005114d8825b53f946"
            "b54ab5a11880c302be441c08613201791f050500000000170051146c83da63dc3796cbb5e7910ffc1dc77c6d0e14ad28f600000001"
            "003a3c19970116fa34888e7515c3ece35db522dd07cde05628ac16483f02ae615a64222103579d858348536dc777c619f701731c99"
            "4feb8e85dde401ff71f25b305aae47bd40d4f73f464d229dbc852c636ec82c463f1ec7331113b0e6c050940d6a53e70c23e5b7be79"
            "a861052b6de66b20ed51013e6e280fdd89956f24199f9bdf107d0621feffffff346405050000000002012202000000000000170051"
            "14d8825b53f946b54ab5a11880c302be441c086132013761050500000000170051140c9a8ed71351b1893038cfa523f9681690b304"
            "8d28f600000001000685cda9109c6b72fd96b49225dbaa8a25a30aac9ae1f7d82bf6e0dc2c52740c64222103c3ae96f66938f034b2"
            "a9c789efcd83fa9e10bc82a5b0b8177151c74fa195c5ca402d54838795c67a6fb040e1ef1af23dfac6e283420562fc48e2e8c7b7dd"
            "1d33fae5b900d059b52504a64b6c1fa39d4ff8aa5a10e000eb18c3a9da2b790fe620ccfeffffffe2b3070500000000020122020000"
            "0000000017005114d8825b53f946b54ab5a11880c302be441c08613201e5b007050000000017005114e946927a4ed3c14432774fcc"
            "bb517d0bdf74561b28f600000001000ebc121716968f48e88152092d6c5ef12ec12ee64b91fed653b472fd7f3dec3e64222103a3be"
            "74634b48e9d47a890ca29633c80da984ace03315c615f880f1fa2c12022240577dd990be7dbe31e906cd74f03ef4523fd3c4805844"
            "35563a69d1d8f113c9ee6ca4808645017731afb875fd8dfd59fc4d106fe7986473e17c6d84ab5d84c496fefffffffc9a0605000000"
            "000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201ff97060500000000170051144a6309d301"
            "4f99267257661e9c34a81095d9356528f60000000100c13f9712ffd75608459e08cfbbedf7101b0884d1425dee723ec29d696022cf"
            "5d642221026b96d1377532aa49f7e03c80373e8dd9a4b0508cd76a32f3ae0e9869b792c6f6408695d3a93c851a00645c3da2176d59"
            "a3c0a9dac54a6f33e2884dbefb41e028312f2272d31a1efed715ea4f3c8eb45fbb6f2bd806a7cc70134555eb8c36497817feffffff"
            "7f190505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320182160505000000001700"
            "5114c9971eea1aaf217f42222b7289f8de43a386e46028f6000000010093ceec8b76e0ee3145b9e29e3a8e6d39c4dd9a5624fb4990"
            "13461fd04fb195c3642221033ec943e3ab312f8ced96e8aa7ceee3049c386334b88f38661938092fd9683d5740a47c0984bba46f92"
            "3e1e936b62f33022a4bd377bb276c976820bf29dcb692bd7e5b3c8b4a24ba848e1d059cd553f7ab4aacf6a233513d06690dacc226d"
            "32078dfeffffff82130805000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201851008"
            "0500000000170051143eb5bdfbda1119f191357cd75baf85042c36235a28f600000001003a3f749f3427a1e86799422128ba065099"
            "fcfb200936d2395797391d0716621e642221037f4d2477e6b1fe2853a125dcd7c10123dbb640ba50de9212063e3646bc58c562405d"
            "61f94907b0d351512d6573de074ad989ba929623a57e16e95d697caf49a64e54a166f4805de1aabe3f8e50465dbaa14e84b1be3859"
            "23378e4341210c3145fffeffffffb7df0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c"
            "08613201badc0605000000001700511461540e787f8514f157731ccbc158c0cb064d4b0228f60000000100cae9330cb88c21e14181"
            "e1438b55e92867f03afb874ff856051dc7b407b21def642221021e6d133d26372b94971413438869cbef4f10b76c71b14611151b81"
            "d7d3b7ef1f4022384171342e4af1c26fc39821f9f9408117b8030edbc32ecf7a399e92d689ea86b3875428c628bdf075c717056468"
            "556c52495f41f4596c90a4d401f571833afeffffff8b0a0805000000000201220200000000000017005114d8825b53f946b54ab5a1"
            "1880c302be441c086132018e0708050000000017005114b85effdcb344bf22440c8cc4c21a1de2f0f209af28f60000000100862557"
            "c7bf938d6e0a71b4cf5e63c9fc28be87e19c27417ccb97a9093963eed8642221024b03e98f945d5e65fe2b6c7d9919e07d83b4b8a7"
            "21a0b00c23b3061a08a78b244031ec53fc2682f333ad80852db2d5c439587e7bfcf6d7bcb45a4ea0bb8a8dd25645da091b9f0fea2c"
            "e231d2f3b80403b2862a487b6369f41ee8847d625e47d71cfeffffffa5f10605000000000201220200000000000017005114d8825b"
            "53f946b54ab5a11880c302be441c08613201a8ee060500000000170051142f09ceda43ac873c094e961801de9ad11fcd0f9e28f600"
            "0000010098306e9abd072a601646f6813da63456e49d818ff3c7eae632d55d4084653ed16422210255aca7bf9db5f405fe7b2cce36"
            "00a6db09b8d0724d06667326227acb9b323b924001d17cee50029c78b04b6bed806508c71290977aa8c46e40126748c1ad47b3ae7f"
            "000b560f4834f8f8878c2ee8d9b862b177a984a60f785f42dbb9b1479ec2a5feffffff712606050000000002012202000000000000"
            "17005114d8825b53f946b54ab5a11880c302be441c08613201742306050000000017005114344d249b768573ee12b4c0984a0e3230"
            "eca3a65228f6000000010031a6a73b9a3a1146bf15fbf795f046ac503ca9fe7a88985a85e352daa137d071642221033a9eb8b23ad7"
            "f80dd51ae2cee2ea82736bddbe651cd40880e06c4c4aa27782a44055d1991b51375c60bf1604ef90b11858292f2596c42134025ff0"
            "0cc08a3bf564259b79deeb8a9798619666ca79907e215234ff979a0cdec1fea99f5a14a1c266feffffff9104080500000000020122"
            "0200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201940108050000000017005114b955b8eb7f6abeabd8"
            "b482586ec5272b3be0599a28f600000001000ccbd82882c4c3425d78cde61b85dd27d1e43f86129aa748c0b09fa7311a1fd8642221"
            "02b2f0e3955d2120981a823e4400341211ff2937d910f346d36218c151c3189e59404beee4ee79ffd652f8751f0002c38749328ae5"
            "da6c2f0acb55b3d00f740f99f36aab40b8c0a5a0131c6939538567c7566aa26c8b6057564b0fe8285d4fcf9e18feffffff870f0705"
            "000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132018a0c070500000000170051143ba8"
            "87022c1f3909cc2c38293d260ef672a31baf28f60000000100795a71596185d613bc91274d131db2a4235b55aeba6946b2c860df40"
            "39eae62464222103c87cdc150d4370876b8e88165d872455c34551a6c2b3eb00a5f8fad265fdb4cd403262eb79439631f5d139f363"
            "0bde62a28185d01b3c963e0a67c4762f8ea049bb05fc1a4990dc345c0728e16e29d4fad072cb5e4727a80d3d3e069d4d9c4d3c18fe"
            "ffffff22760505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320125730505000000"
            "0017005114251b156ef6bc84abbd36739e970baf917cf42a7528f60000000100ccea3d58b487fbf702b46b1e541ade9397a58de1dc"
            "2cab26c271b3d7a633972e642221032460d9db712b0511ebc2582d583ab31bb4f109408e5a4b91d46cf306183e4c3b402ead881b06"
            "00c42cafc6bbcbb4af4e5071e173688601edfaa13b819437529cb493e676b70c5cd74d9235bb6f9fa19b78884bb8f25bc3bf66b696"
            "84645320899efeffffff4f490505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201"
            "524605050000000017005114538455e9824d023cc473997ce04453250a8f1cbd28f60000000100cb61eaa9c30767ed4229e71fb5a6"
            "3ce43a1276a6ca048eca32bfc171e45410b26422210279dd853175737b7dc6fe134648b02240e1299bfbef656df4b4f32cd543bfe0"
            "4b4033444b54690e2d3d6efdd1d8b307e957343f12c9134dfee2760da19b76440bfa75f77cdbebb472a528e0874455c1ceacfa28ff"
            "fd132f5ea4325b973dfa23000cfeffffff20770605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302"
            "be441c08613201237406050000000017005114be3f8e4e4cfce60512b1c571098650aaa8b6d52428f600000001008f8a660ae826fe"
            "256f0fd7557877f95d3a686546e7529fcd68542dfee3d150f0642221028a380029f334c5e955b0ec2f98490eda9b64b0d21d2d37b8"
            "c6d7925603e441ef40237e77549e051993759c483bef0af2289eda6d854e1434ab783cc88d35cf0fb292f8d4636d4874e2f2c3bc24"
            "f9ebe0b4639c70f5c21ae4856976bb6adcb3b4f5feffffff35620605000000000201220200000000000017005114d8825b53f946b5"
            "4ab5a11880c302be441c08613201385f060500000000170051147415f10d639aacf411958c0a188fe0599419408728f60000000100"
            "35eb04166b9148a903a0016b4dd419465609504237d417ba2b7f9f0901a2cb276422210394eb54e673cd3b051b2df245a67cd65596"
            "c737e3adc90150727d36f7c3937b344098b0eb44a2bfa198dcb583e3130bd4e09117818e574041dbe647441fc3dc55c809b3d15460"
            "2923900acc8173c886075249ac400e955b932cb68bb3aeeac5613afeffffff37610505000000000201220200000000000017005114"
            "d8825b53f946b54ab5a11880c302be441c086132013a5e0505000000001700511451224e5429323410ea418a6bf0a13d540b383f4d"
            "28f600000001003f90cfe5b6cdb9f30dcc98c9905f57ea29887bf2a30f4b731ccc345e9d0f9821642221032457f1caddca391d6e48"
            "75ac685b2327bf87aea8e8b7e8f831ed27207bc5707540cdb6b5399a537e799ac90d56479f1b7580c441e05cb2ffb862aa07e7b486"
            "bb29cf85325a302a874f0e2eb70b90ac252fa5842d17fd94618e726b0e8b7aa870d2feffffffa6ef07050000000002012202000000"
            "00000017005114d8825b53f946b54ab5a11880c302be441c08613201a9ec07050000000017005114629fab91804dbc52b29f1d93f3"
            "bf94803c1a471428f6000000010089bbc1e9d4bf80a9e9a8565c446e004ce6faabf2812826e41c3dc984f7b6951e64222103e15f17"
            "9dcc5a081865af5c2fc9fa92c72ca4470f9d51b846fa4db3cd713938934003c2cec22f00e6c67faa7d01a4c9c79f0d29f0f42a3eed"
            "6da3e308deed8d5a2bc5197a91996bce4dd1cff3a66788d20c429d48289932f990724133386e162a58feffffffbed7070500000000"
            "0201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201c1d407050000000017005114a3f34a40084e"
            "e86a456382ec349bf5e85ed91f5b28f600000001003420321b4f06d0c2a451272e9a423c412227f660cbb139406500b4131efbe0e5"
            "6422210359c0f9eab40bebcbec5d6af2b29699e4170335568d59c11b66ba1e78a3f78ea040cca1f9e171e94fb8c4d57dd787f7bb51"
            "85e4f35632aa9958ffe3ad2a3d92204147e45edff5b0455f17576a8897f41f82c5f6e9ae70a213fde1d4daaba7ea3c76feffffffee"
            "a70705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201f1a4070500000000170051"
            "145f3fe38c647616c1066a8b4cfd1f562de557057828f600000001001365dbf7f3fbaf983b266058c03df4220ec0ad4ee737a2ce85"
            "6539e44ebb1ce76422210237e5e259e9f98981deb951edbd6524c0bbf635f748a929375b0c33e54806cbaf40b64615e97b0b45a415"
            "823f79ff63f2f83b44ae23476097428e7ba017d53d878927f06764089ee2e52cb0a3b3b017da9b1d4f62bb901a86f6b83b052bb8de"
            "5fa1feffffff13850505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320116820505"
            "0000000017005114e66bc2b7451e38cc386353cca258dfa3d845394b28f60000000100305a5569478e96520a9917b2681cfd9dbfd3"
            "e3d0b808944ecb0700ba964fe9bf64222103d598e7f65c6ea2faf159ee9aea8fdcf86ec5d8f0752d3bb888d4916d8d86536c40cc68"
            "bc25ed1fae5aa8fd90f52c60824a6d3d0b123818eaa2ad8f6edef3cec03d30a4909e7e5e12682d6c7e84281c07f4f53736cb4f48f9"
            "4146473137d58aa8c7feffffff0e890605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08"
            "6132011186060500000000170051146dfcae427212f4f45af6e65d160e8e35f955068f28f60000000100968de7697d2133807d588a"
            "d44cededc1cfe303daa9b3a01878b525575424a9ab64222103d5da8cf4f613585bdce67a85633156323cd181fd49cca33c7a9e4981"
            "a8a146d8401b4d27bea15eb417da31f0ba1f532a20e97d6478bf7ea91e7ba68cc8b815d4cf79b280d04b320a31d7627f33957bd90e"
            "3f09003f192f707221db6ccadac83dd2feffffffc2d50505000000000201220200000000000017005114d8825b53f946b54ab5a118"
            "80c302be441c08613201c5d2050500000000170051142e6bdab72d9b237a00cafe8f69a38cfd8c6f90e528f6000000010072725832"
            "cf20c96b66d23461e7fa3ccf64345555081f107899f4097ba299107e64222102cea33713d66451400a584e987ba5d42993e6adef93"
            "e9003c7dfdf6776c954b00407785ad9c9ef2d89a766010d6b9214aeb1b0d412cc9fb3b24c63407e3eb424d58ca65e95aeea7a2cbfe"
            "65772214f9b4c9ba15e67150c6ec47928bf848e1c1ec40feffffff99fd0605000000000201220200000000000017005114d8825b53"
            "f946b54ab5a11880c302be441c086132019cfa06050000000017005114636dbff028f0955ed6c8a90d6ad411ab9df6421d28f60000"
            "000100866eed3e1e0ce6d924b51d122341338b7965a6432b797db510b8452f61cf43a464222102a0eb8adb5659cb775149ca059ad5"
            "448d5edf47f4fce97db8567f8c06008fa81540b767a5c16333bb16ec31d868c73938a73c7b7e1f6e4a6be6b58dec5214711b7209a7"
            "7a495802cf9fd91209d66792e5e2ebb5c68c51b4f919706119dd8e4ccecffefffffff89f0505000000000201220200000000000017"
            "005114d8825b53f946b54ab5a11880c302be441c08613201fb9c050500000000170051146e5eab52f480868a608f011d22bc87f093"
            "194e1528f60000000100566ab2ff6813f078aca19dbd4179c5b99cbd0af499df3f371f8f6910e5e6dde964222102833e1a931e091b"
            "cb557b8d54046778e1062090c81ba12cf12824c137db737993402e22d5b673f04600cd6a2dab0d5f4e3fd4ed954fbec7e7efb15e1a"
            "58140b89b9dc5080aebc0540f8171628361c01eea1fc5cbd3db8ce8901ea5a1f9d17ee3084feffffff316705050000000002012202"
            "00000000000017005114d8825b53f946b54ab5a11880c302be441c08613201346405050000000017005114c39240d208107fccd7f5"
            "1768f5809c1af6c0d97428f60000000100bff0c2fb3c506e11584f6bb8356cc0ee2d7f5937510d5e260a9471edc6bb083c64222102"
            "0b789fb5eb053eaea94f332a49e1b29120bae4a5b0bdeec22bc6dc32b8aa7614408f94de4352fb6d3fb25905ae50663f79a4b80ca3"
            "2a4a4e3d1b792a6628e8b146c2fd9391e13ab4686a6e0b5735c9742e14a6fd6e3e823058bbdaf4a2a5bd0d45feffffffa1f6050500"
            "0000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201a4f30505000000001700511402c0e4"
            "b59bcfdb702c29b9457d31ab1fb16384fe28f60000000100f825b4d28ed2addf301209d5820a4e114ad0bd26625a6aaa8f9ced2a00"
            "5cd13a642221029d2a3182a14e6947cce9af0cae79b24b8560a98347aa9f97064a2cd30285d9564005e8fe5a65f148b89a836ae41d"
            "bb373d764eb80a9a33970283d4c5c514d265680adf21f4961e81133c0f527e11b431b7c1064855fa327efbe37703acf6dd006dfeff"
            "ffff90060705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132019303070500000000"
            "17005114b2e0beab5a82ad2957f9908c744c258c457a6fec28f600000001008633197f7f72be6c37e8c77b7c7651415d00981d343c"
            "e8c7c71ec395929c80c964222102c7c3d13bcd20572bede3c16fb247138cb0755b874de3f06760e5063b9acfa1fe40e7e380018888"
            "8dfe58c24d770d3486d7676245a0234ab76225f9d854a22d3579c3c53f96134a98aa723da251f00d8d1301a32c1918935fc881e69b"
            "c248f2bbe6fefffffff1a40705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201f4"
            "a107050000000017005114e6abc88f9bc46f0b77e94ff01243c19af04c05d328f60000000100e936562918f96fda891687476ae4e2"
            "95ac93e930bbace6f8f4c089806c23b8f46422210206f2189b39810ae05b877f7a6063043d1a302f23d8c04a4e391dedfaf3da883c"
            "40f9a2e6e1fe59bd1faffefe1fc12af76ebf08e830002c6afcfd3ec5d615fd0a6f65ce00c0ceab34daee9e49defbffcf7af2556560"
            "ff6f2c6a7ea054bef970e6bafeffffffe5b00705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be"
            "441c08613201e8ad0705000000001700511447f9952caf039c6f0182d985ec6a1d852dd94e0028f600000001005d820205c2b3183d"
            "6d6d2ae0362a97582f1134fcc77ecd87b7fcee856497960564222102889bdfae4962aaa4a3b23a36609676fd3580dc08deeffdb089"
            "703ddd03fa803740f3c8d9e37fe1ceaeecece1f0477895cf1daaef4f8c8302cfd379dc12b62f5fb9f1487e50a6475f4653bd8b4e5f"
            "d4e9e41a675db1808a42f1c79ab009421d303ffeffffffd4c30505000000000201220200000000000017005114d8825b53f946b54a"
            "b5a11880c302be441c08613201d7c00505000000001700511483650028e3b2271150581eb6cec270cb19dd54db28f60000000100b6"
            "f31623d2af626830ae4ce0b3ea5b12bd62c4d37df52578f6e1d584f6d29e5364222103075ffcf3885df7062aaeb0fca47124b8346b"
            "d7e64ae958ded6084a824f2e769040f2f899fcaf06d9fc77e060f4cf818cda392e205a6332e6d47bf64495862c4e1aef878320201b"
            "e2554b968631f110ab955bd7b1d5e2e8a46c839ec1f3bfec2408feffffff5d390705000000000201220200000000000017005114d8"
            "825b53f946b54ab5a11880c302be441c086132016036070500000000170051147c51e4b024c3fb363f290128a27d5aeb61944af728"
            "f600000001008e217fa3f6399bd897849c3c454028de3fa9373fb41702f610d373adfb2e098864222102a9fe1b742d0207ab910189"
            "6ba752d1c99da57528347309eee381d814e06463d3404c7c2e34b73aca40ccaa3a94d82e15e4d8eaab6930119407dc8a4fe431985a"
            "cfd672ba338f26a70178c3110dc557b3e5b49ad4a88884e0eba0a07995f86e4974feffffff1c7c0505000000000201220200000000"
            "000017005114d8825b53f946b54ab5a11880c302be441c086132011f7905050000000017005114129596813e0a8ea9da4911786b49"
            "85e3309f67bf28f60000000100dd3a1f545118620d5f1f05830a368fe9aab260912778d73343846bf05f6be36964222103e454055a"
            "5096a379e340864c7826e9bcc2a6d94efb4c5b545f2490cac43966d6406097099a5b4efd9ef22a96f268fc78a07cd360f7ba9d3bec"
            "725fae64b6ca79fe8064aa2b346582da04554e3cedfa3b0325b417bd3e8711c1a57aa256dc0fde0cfeffffffd2c406050000000002"
            "01220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201d5c106050000000017005114646188574d44da"
            "34e9c70b39ed2a2c2e4275d55028f6000000010017faada6b8e2076a3d1ebb79225c9bc47e89c46814a90a0b9d2b809e992b077764"
            "22210237947e520193044473245e8e8804cbbe65472846f5305dfc8ec1e390d13aa71540c456d42f7732b03b1d806d5d4a98a6a20f"
            "de74132da58fcd4a524c8267d77e24e246796bb66c79bcbebd47887727ef55cf6fd4d8893ed3b534ece377ce28b47dfeffffff5246"
            "0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201554305050000000017005114"
            "16abef328a6c8c79d8f06a7f6130d6208f0d346728f60000000100e6576c67be62c0f70875141caf5ea5db49f32844c4416a03c714"
            "192537c0a9ec642221035223a4312c7b49b61b9148cb788515ffe15e5476c55a6b9e69f2f7c88e8ead7440ef24a7bec02251f9935f"
            "7bf658220fe19f230bd4ddb88f466fe76547b17e80231384902c9c34de147a5548ae7139e813b9713d077ee76dcecf24163b4b9422"
            "2cfeffffff83140605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132018611060500"
            "00000017005114430cb1b7cecb95c04ccda9eaece335fdca763f2128f60000000100aef9dafbe9b1279c1fe95cb427d4682c31c49b"
            "4c6ef6b7958a9f96ff96172669642221023d7b74f91d05f1dc6addf5c91d6690145e309e390660edc77b15ac0aeadb60a940830fce"
            "17e3c21eae5ae0e57098e725e1b565afdde615660d51a966464babc0650c08b3fabc02cbc762579062d9f29936f44ae13430aa7840"
            "ab4e24a70c2a30bdfeffffff7d1a0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861"
            "32018017060500000000170051145af6c615cb1ae600648bde82b1b1b1a80d74ceed28f600000001002736c749b843a405c900b44b"
            "420e473db852044931c0a0691ccfaaf3ef2fe6e6642221032dde0a21dcb6c8686031ab4471424798feab27f6f70cffd11e7c977a18"
            "b00deb40c65d431c20d716ea3d3f47def0999d7f870308e579f42ba2d36ce76eeb136e2be7581c3e63c31877cf8284e8b514c8b08c"
            "24cedf3252b6095e5ca0255216d483feffffff5e370805000000000201220200000000000017005114d8825b53f946b54ab5a11880"
            "c302be441c08613201613408050000000017005114c4131fd092b802738e38e5efc9599d41a803904728f6000000010094a1c2c892"
            "75af9499344b44748ee12e1ca7f8ce22dade4a65fe921968b9f23f64222102f7d040f40f7c1a6a70996a6e72a941853fa2baca46e5"
            "dcc5b67ab7887a792bcd4023f2554d9be48b66f142847b51efefb0e87ab1ed9e888bd410fbd057510340cf7edbb5cc455c7d74c9da"
            "40bb663e151961b6ed7485945305e024aa04b13307c8feffffff6d2b0505000000000201220200000000000017005114d8825b53f9"
            "46b54ab5a11880c302be441c0861320170280505000000001700511447e4b445c6adb74145f0b18279c624e76b13b2ab28f6000000"
            "010068b05f2ae97b084cd5f91822ff70678deaa54c37a7f8cbe497601323d60e66c36422210235b564a4af4be8cff146fb5125cfe9"
            "7b21347c716c3b04fe4273e7f9400c6c5a40ce98575f8b6ea76e5dddfa2b25d1969cd777d2daf600017013c7a7974d3209bb7a9267"
            "71f219db350ba9c8d58305d19ee6fe3b10a3dc3dcb285c05584a8256e1feffffffe4b2060500000000020122020000000000001700"
            "5114d8825b53f946b54ab5a11880c302be441c08613201e7af060500000000170051142120440fc650fb3f1a0c30b2d3384397b5cc"
            "bbc428f600000001003370be4c2358969f719d69934512c7079488bf838e5962eef3e5760745049ac6642221028d9c3a61c5a91360"
            "97e3b0066dafade750deab0d185ecbb8913315d0cbd2c59640f99db0cb5cd013fa9d99bb1d970a639e60d51e4b4df9185f67acb1fe"
            "55563cd259e32dcd4a56ffca855f6ccc04d4c39b2e59dccdf34d15750229ce12dd2ed29bfefffffff3a30605000000000201220200"
            "000000000017005114d8825b53f946b54ab5a11880c302be441c08613201f6a0060500000000170051142264fb2e1ba093fbe68661"
            "8bc3bf18573c404be128f600000001008d50abd09f5a2d9b5681e35692a1caf42c4bd3b3bcc55687bc8832bc834412ae64222103d1"
            "84ca3985bc009e4b5114cd2bc0ccc04dde2717af8f3cafa9b28a003fa7e0e640ec7c90c39250a8a4c2a0c0aee7811b797c225d06bc"
            "e6b26f033b8667349136a86aebcf15ca7adc9357c003a55a302e1e3974116b62635d16fc3c9d3c9d742785feffffff781e07050000"
            "00000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132017b1b070500000000170051144a2a850d"
            "fb3085c2cb1e132baa66b6c37d30290828f6000000010031d08476a89161710eecf3edd817ec49d3883b498829dea390f11ec711d0"
            "3d85642221029d4831903a37740500a2501d1c3cc6763c563ba2b099fef9774a42f8f80e9b93409a71b31e5bdfed03f89adf6bc852"
            "5f8d3aaeef4738312d32d2c2342849273db877ad35b583a55fa36936edbd6aadef9a951e5553b9848bb438ff52052e232f7afeffff"
            "ff3f570705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201425407050000000017"
            "005114cf5e69cef857ad55d019b453fc7dfdf317150b5928f600000001004e03c5378ab3cd3794922053fa8a5626a07fa9a532095a"
            "b65fd13b350cdebeba64222102f8aeecf0b05912ec8a5bd423c676d55df642a1d42666ba61fec86f87884e9ed2407fcd751d83bd3c"
            "0b707e79d17e4de17da86afa93a59b2fc3ff3d156821fa8d6d8b410bd9aaca040fb0076a3076a45b984f1e70d76a7527ae0f8c5c7e"
            "154e66d1feffffff93030705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132019600"
            "0705000000001700511446a29f791764fecaabf006f8c1b6af70226a75fa28f6000000010024b9d9aad1522c7b2abe8400cbedb6dd"
            "85bf8141d05efd47fcb2a640bbc8ec6264222103239325e207303b189f4201caa4eb7bf30395b63d86c9ec591705f000938f1f3440"
            "f7ab335cc031d89b17736c88f27d4f5cdd0bf7019f814ee40ae20f75902404ef74621d48b84100c26a991ff67fdc8b0bd92bfeaf24"
            "f70fd4e2ea3d837b4692d8feffffffc7ce0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be44"
            "1c08613201cacb070500000000170051145bc8bc71bd4dfda55660e0c002ecd37927fe880a28f600000001007a20b912bd71bcff32"
            "f4b50a42744d2b2383686af78ff8a93c17d99b7cca66f5642221038afb3afc16eadaa03ccd7a7f073c291648a13a6e6808268c5402"
            "d2d9da097cdc4000d0dbf4e83e371d7d9aeefe7f42019b797be3b9ffdb996d8755dd95208813f8de10fff2b71dd20c34e5668293a2"
            "4d8efe3fc5d72de303f824239753368a9b34fefffffff2a50505000000000201220200000000000017005114d8825b53f946b54ab5"
            "a11880c302be441c08613201f5a20505000000001700511430e952f823458f13c3a5413038c3a4c0c1ac9c4528f60000000100876e"
            "777cd7aa73fdbf01883eb07d03ee2b18b39833014871ceb962274d34260264222102b3ef1d296384caed5d171810fadca3aa2757a5"
            "1cf99213a7eecb85228c9722b840eea7214e8e518efee28ec9e3e9011b09f0b18195596b4de87f88d68cad2e45f50b95e2d5d27bd2"
            "145a241091a8607844f141f7b9e6a34da5dd4e62e5f3019465feffffffb8dd0705000000000201220200000000000017005114d882"
            "5b53f946b54ab5a11880c302be441c08613201bbda07050000000017005114b1b007caef508937cc6fef0269ac780cd42186ed28f6"
            "00000001009c80bacf77e4850f9481e086c5630893541f3af86068f8c12101084e9e72b04c642221032fe3d179a066c29ebebb3517"
            "a9b5d01816a1e6c28f53a9fdcc3a906aeee867a440e396164cf8b1d4ac9d87ecf95a6f90fae404ef1327af1f7585ae9e5baa0d4d40"
            "b7a8e3462502d332996676846b0a1b2230bfcbb02851cf6f3269a2a184af7d89feffffff8611060500000000020122020000000000"
            "0017005114d8825b53f946b54ab5a11880c302be441c08613201890e060500000000170051148a460e877b72f274a041d6cc343d16"
            "1bb860944428f600000001005c81142dbc96fa9bbdc17eb15b4284b0c7c6cb06e05f8caa75fdbeaf8e69b6f96422210243e5181129"
            "81278491addc4bf7e12957031d91abe467cf9adec4253e71b168df40e9ef5b7618fbb841c5e3ef813df402d4f6cada81396922f78e"
            "4011ce2fc1c6b2245efe04b0b3592772d55ad48d0919c4cb93d33206f0a987bc6010e2f894b4c2feffffffdfb60705000000000201"
            "220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201e2b30705000000001700511428014922aca4ac01"
            "e3192ad51eaeb7c17659028e28f60000000100c9560058428b8639489bf27c8662c435ba4bb4fb612730cbd1b7008443cef88c6422"
            "21032c1ee1274bee64b3bb14f6ad66b5ad53751d5f90812ca2a0adffc02cc7a78a7a407cd64575f68fd6d4ab45caf01cb973cdbaf4"
            "f96a72043c439949bdfd102942b351079244953737125e0ed9c1660ca61ed82a9f441ee4af26bbc094409cf5847bfeffffff9afb07"
            "05000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132019df8070500000000170051148c"
            "8230f55065be9dcb31eb6980f943d4e8b8738128f600000001001f0dfc3a1d76c356d84db14f030aa1d62233f7afb87be83146a5de"
            "1812a746c3642221025c21aab5ab1d407f4f62ccf77cedd85512785359fed2278bdb03dd141e4dafc540c3f4764f2c417d838d0bfe"
            "6f36c0f1dcfba87ae7bb7860d76747412c73bcbd13e7b7935a02aa2c0dd4c086f4c7ef317847621b688fd3ef07643c2aa3a0346d21"
            "feffffff65320605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201682f06050000"
            "0000170051146f9452070a183c3315e316cbc0f6975191922da428f60000000100fa5ff675aefe09cdaa6f5f7490a51f9fa9c0e0a8"
            "745b8510fcb5041b3276c5f7642221030ed313061e512d58ef111aba11b5f4a69f69f001a89a31fe993a883863b190e1406ca6d73c"
            "946ba6d8effab04ce0d8293f1db5c35b5eeaee689c598d9d85264b5d4541b5af5efb9d65012c76cd9646def4d1428eea371b006dbf"
            "e99996f5c816fafeffffff197f0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132"
            "011c7c050500000000170051144aa8dad16049cd04cd9a6dab816b0959ae7c296e28f600000001003fe6b52a2b2d443eb1d96e5fca"
            "f57e589dcfebf7e663dca0bfb0f96923ebe15f6422210325103dd6847abe2c97dd4f6979ae8354e58d96778845919dd6484a62a11e"
            "ac6e40b17827017a9371211d7f08e3dfb6b9dea700b369dfab0866aaf9a799e58066303db4d3332724dabd57cc76632183edd0524b"
            "69cc3d8e3fa3598e500e88d13133feffffff64310805000000000201220200000000000017005114d8825b53f946b54ab5a11880c3"
            "02be441c08613201672e080500000000170051145a3f72f97f19557c8457a88e87501bc424986bed28f600000001004d80f68b28fc"
            "b7b2bab6ee95cb5aea60b35ed7e0548b0c5bc8d07ac52b4e543964222103fe92579309451a1f24fedb1d9d86936e9dfb83279a823f"
            "6754be0ae10c4389aa40d2b3b7f7e4e5be7a6508976aefd91c0b07b1054661629c387db8d74abd3df22f84b8fde6053ea035c042eb"
            "d3218930b9b2d43f85d816f3a00286c457154206cffeffffff6b2c0605000000000201220200000000000017005114d8825b53f946"
            "b54ab5a11880c302be441c086132016e290605000000001700511433e5bbf3e789cfc3e3ae78c03ed8f0e558340f8628f600000001"
            "00314bb87bb193c36b4723a0ef70cd7b4a3e26c53c0c69494398f7774910382ecb642221029baa33d5fbc5ba752c68a7bb0677da9a"
            "f57db38e911e400ae6258bffe73d8348406ca3df1bb03a3a9dc2da44a720d2af89d25bc922bab5be5d71af1740e3d1ad8b83cc306e"
            "fcbe5458dcd963b6659e0e7aa69a3bdb4cfbca7d449ef063a9e33816feffffffbdd906050000000002012202000000000000170051"
            "14d8825b53f946b54ab5a11880c302be441c08613201c0d606050000000017005114dd199dc62892484244e5ac58e5ea7ae4628937"
            "cb28f600000001008aa07e7caf3a21181c94d2f35c8ab5df2903726afb98ad03a78d1efe14a157776422210229dfdabc2babfe84f8"
            "03e3e593dfbe1e2340381b38afedc4a3d5a31301036fec401da402e1f3949707b05e5cbb8dc8d796e36dc2e92d75b7f6e92e5e22b6"
            "cf7172e7865385b97fa4808ffe0d349c0a9ef4b88c1789c48adec3332c5f648230a430feffffffeda9060500000000020122020000"
            "0000000017005114d8825b53f946b54ab5a11880c302be441c08613201f0a6060500000000170051141778e161c93e761dc25b009f"
            "e265634cf89e1d6b28f600000001007601bc80cb33adb7c934cb27f624bf17c06a735cf779b058a84abddb0b568b8b64222102efc5"
            "c839830452a33fde4ef64c32feabb2b4b0885abef029772f76a6d69b500a406987191a5c51d9a895024ff41a11bb785c1937c1db31"
            "9f622f72c3578e9461b821578e5fce11af5aa61322a0853daa283f938d7618f4afeb4d5613184af5f48ffeffffffaaed0505000000"
            "000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201adea050500000000170051141cba261258"
            "cbc8a550e284e77285ca53d579c2b028f60000000100e49c2f96f7a2b560f2a909acebebcdb66669a87aee53ef703c92a518a09bec"
            "3b642221028bf44357358164e07873b66c74d473154194d1404da57df2ccd8ad5d7580290940bf593318134506508c1aa092f86319"
            "0afcb2dc823fdd140097ceeb8f98729b3329366a51f8bfa56be5870399873532ebdcbf59023bf38f77bbb76adf4d0f8565feffffff"
            "5e3a0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320161370505000000001700"
            "5114aa78dba7a8184448db7aac727cc4fd8f9f11500628f6000000010057d0393aaa8f7664aec687a4ea6209f83aa7791827caf82d"
            "8382afb2f3c58a5064222103c7c0bf861114acd39e76f4e41a1c0b5e348a1c35bcded485fded729c19108566405b46aa1c7212cb26"
            "0ce079b735da763f7847f82aeb9bd81d5d0a42a63ccb79e7f5e813d5a840931b03e47234e038d4d0ae623f218ff7c2218101e7af7a"
            "44df5afeffffff85100805000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201880d08"
            "05000000001700511450e2b1df5b9ca267e4fc83ac2bf8fa6cde38556228f600000001008fce83e4dd6cd6fd88e7565da4e6b2de68"
            "c3c8e7ce029448a5ea82bb6b7de3c3642221035655d54fa4b9d9fc3d68c0b082cecf03f56f61a75a9b2588b03f7b30379364cd40e9"
            "4839f1125bf7ded32c99044ff3142440d9e1446c4e3c34709518f8ae4da927d6382f50c4589fe0c8491c7a6ef5d43df1d31da43223"
            "f29c46b52e915bba44e7feffffff296e0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c"
            "086132012c6b06050000000017005114f91adc73ac5d0711cf4e52f2c1d16ccdd31c580428f60000000100ac08c20f51016d3d7da9"
            "014e5644c88fded06756286cc4a3279f9650aca9411364222102d9fb5b0fe8ea671a103a68bb42e23940eb07fff6406a92709b6a99"
            "be7fd32fa040ebac0e8e164c95b00eb11108ddda2377f69e891a39b827106cc49f5004d5cf705227eda90275a66c607aaec6dc46b8"
            "c9b981753566ab6adab6498870f7741bbbfeffffff0b8c0605000000000201220200000000000017005114d8825b53f946b54ab5a1"
            "1880c302be441c086132010e8906050000000017005114e42af7d267c255226c3365796debb4879409e56a28f60000000100bd7510"
            "acfb0335b8eaa7dcdfd9fa3d860dc277777818643ef534199c484c3fb66422210220eb6ba4c0e1c4434203b60b2464f3863c1ef7ea"
            "02d071ade1dc0267a98c054f40aab0f070ac77899b799619af2cff1fedf12cba7d9bb054cbfb07ed0944d2731cc8ddf3f7ea6ad69a"
            "d57e943d0e0cc7f2614d524f4fcbb53f4cea9053488f755bfeffffff62350605000000000201220200000000000017005114d8825b"
            "53f946b54ab5a11880c302be441c08613201653206050000000017005114546e0d7cd5e4c9f9cae3d278c56d0ea67c52834128f600"
            "000001002ff5fd2d59435dda7526a5b269aba145dd4b773e652cd64c084af54a64e9ed4a64222102a87b438e7b3d2188da61a8ee7c"
            "0a23ade37fd00a93f7c5c269d0920ca0ac47b7404cf02b3019a7414b0589bcabbad981ec99e33bbcc60755d1db67f1465963ed78b9"
            "df5afc9ba167dfe4ad0ed1fb4db9f9b55b212bfc8a720e98afd48d9ac9dd76feffffff276f07050000000002012202000000000000"
            "17005114d8825b53f946b54ab5a11880c302be441c086132012a6c07050000000017005114a756821a9e4ac6253d6634f00e4251bc"
            "07e01d1728f6000000010092ca2d350bccc0e624d9133be065489573453477eff87c925e47b32ca4c3ec5164222102d93786734085"
            "1775f123765fde9d28d7d43cc2552c359df7dcc8ed1cbe397a594020603b4456f720e732dc5bf751c14503ec58e219f48d793d5f12"
            "16403427e58063aa0a17e2dfb469fb156cc6e71def7d4684710a6cdeacb5e6ea4f524708359efeffffff672e080500000000020122"
            "0200000000000017005114d8825b53f946b54ab5a11880c302be441c086132016a2b0805000000001700511459fc8b994ed54d5da3"
            "5aedcc4fbdcc64dd53a35528f6000000010006b2472bbbfefec1267a2194d5ea809a77a9072606d492033778bb188c03ff52642221"
            "020959c437ef182fceea346b727d00eaedc7c41b769f304531acd1641fa8283396405b063799614b39e36776a4fa0fed1f054d0ead"
            "bbfcb28544e297960fcc14dfe0918675c0a916927dd1f75a06068cb7237cb5efcb56b58ea1be7e4ee30f8bd00bfeffffff73250505"
            "000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201762205050000000017005114e62e"
            "edd7555a2fd7de8389481644c3252728adfa28f600000001002b9e625433c2fa3b65e08d90bd03138b20f8437cd2d60ee2ffccc446"
            "0838939464222102bfa04e65372b0a86a7969f8e5efa3feee633ab48d56258c449235acc5d8ab5b840d7e385997c89d8488fd069bc"
            "8aa6d97192d58f66d89b9ebd23b3f25854c5316caf95cc1c1f6033d372961bcc0f131b70bef6506f5d0f887ffec337c02dfdfcc4fe"
            "ffffff4a4d0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132014d4a0605000000"
            "00170051148412e9babbebe04106bd88dbf9eaa4066aead90228f6000000010022e181a5e0b2b57a851f394640d1ddaa960e589845"
            "121e3661643eea572f79c16422210278c855d66e78cd99b63e1b46d19bc143b9f3bf4017f475b5aef94f41a04157d940861cd67c0f"
            "cc866d59899d3836979938b5a9347520bc8ebb1721d4b4bce1f81851f581ec7b7ba7fa5bd01e74b5c0b05c6aa191d0634d1a549702"
            "36f5f6b76ee9feffffff682f0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201"
            "6b2c06050000000017005114c6cea845b266fe0a2a8c8e7478e28eadb32b38e528f60000000100609023d9c9bea6993737f5880c90"
            "30749086a3175e5f57334a6baaf710c3fba664222102ed4ac8e77c2e13019988557dd3713cea825cce0e60984db9e6086e91e6f5ff"
            "5f406ea86caec511d13c1befb156028573c2ddf62dd15d79595049f65f0bfc50b8369769e3715e0597e3ecee80bad01783d49ffede"
            "8cc855a28ae9e68bd01d7a17cefeffffffebaa0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302"
            "be441c08613201eea707050000000017005114656b7748dc9529e5ee154a5ded7f7f0c0a33cb5728f60000000100fda4dbec6c3cc7"
            "bfe830c353ce07eeb2c694f6e92af604edeae57d57ec10eb7664222103f535fdc82b4688116e9f83e33aa85702fb02d9b33347596a"
            "2fc96ce90d6f0c19402c3b383a1deea2a22662457716ad9f8b25c44d06ff0bd7c50eb4c47944b0ffdea5f218ce059ae65e4c40dab4"
            "405557c222df225ad34898da2b788b3d9a8efa04feffffff45510705000000000201220200000000000017005114d8825b53f946b5"
            "4ab5a11880c302be441c08613201484e07050000000017005114d0a22c92ba08398d1d58eaf3cb8ec04ac3c3c55128f60000000100"
            "10b57cf2614bbe9ab120fba09dc6f387b85c3de56fdc46ea588dfa5feafdf67b64222103fa375563d759abadb036d704e227a6e715"
            "332f7f97b6b481143a79406e077291408a9fc2a0d8dc8875abd5f5fc2171e6afaec88ebd8f1be745501c9c616a1b7c1795ce18ad90"
            "82ebe6191a00dfcbba9c7b5ae1b774293cb696d91a7a2799af04a6feffffff07910505000000000201220200000000000017005114"
            "d8825b53f946b54ab5a11880c302be441c086132010a8e05050000000017005114a7ee89f04e6061e133616e83ec4746fe92947fec"
            "28f60000000100596ba2c94764a2fc8318a4b828b9a753cab0fd8fec8768c297b960054cb90126642221027768732ba4bb4e9b3073"
            "e208b561dc6052e1fc13b6d90bddf252031751e7dc8a40767b4de51b89040cb6e0feaefb5d40e6e0f6d284bcf73adee85f5de01d54"
            "0f3c26f0478152841635e3f23c133e4b1806a1c8b4abe5408565be31729bc3703388feffffffe6b105050000000002012202000000"
            "00000017005114d8825b53f946b54ab5a11880c302be441c08613201e9ae05050000000017005114a8c1d5931a6357978609d68224"
            "e4cb3c3a0f4fa928f6000000010016c67087be7521102c7fa4b8037d05c1ea0642f2a9d82235f8756ef640f05e2f64222103c29320"
            "d5725f18b19b45f6bd19c60dfce7f484796af590a86dcb27d7afc498424053de67ecbbe21e573a44751f9d59dd61a365b7ae75e462"
            "b314f0a8c43ef35b62561a152e01a3bfe5636c255164a4c81b0906ea873836c2a99f1464e947a1ab45feffffff4156060500000000"
            "0201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132014453060500000000170051141d9ce7764b31"
            "bb4ed08ff90f1401d2cf168f0dc428f60000000100b6df452ecfa61c444876ee9cba4c3d795df97f66111b53f37baa3cabf1a43ec7"
            "642221030b95647c7b81fad2610a77cbe03106df1bd3677d262dcc59d6beb634c3627bc840cfa828fad30a5dd36532d8e38bb60759"
            "aecb97d7b5bab12f7f823114a8243ae4157ef595d623533373c06cc5603bfff0cde15dad09871ee118a5c6e5ff9875e1feffffff11"
            "860605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132011483060500000000170051"
            "147ae5b0807bec2456da7caf14dc6633d19be4175728f600000001004bce6eb67f3c3b099052457cf0a7e626023bdc64786abf2869"
            "38af82733bb1df64222102e932ac072a43902c7a02861ba09dcae06b8f951509ebbf14724bb56240c5646d4079ad95e3d2146da8b5"
            "50cd66d345f6fe6d3f39db960d879bdc4f28ba03bfebc30015f751573085cace42fc9a88222b0b1ac439d032a6a26a167f2748df12"
            "8842feffffff4d4a0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320150470605"
            "0000000017005114e255028fae7daaf8d139084bc3eed5b0099d210c28f600000001000bb999ab77b0293e9a7116ad0ffacad4a89b"
            "7b7a5f5f32f94b145e16fe34eba3642221023cf8643423e8aa43322064418db40f130bcf4df7f6ecdb567c72c27f9728c77f404aec"
            "6000b06ae78fc09dd53f5eaf8a1373500311542678aae6de20c5f9ec36fe9bd3573f2846766a7db86c53d578928a10ff9281c966d8"
            "268df2572e472b2200feffffff82160505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08"
            "613201851305050000000017005114b0a91fac21767b8ef49643f508fe16761a51251128f6000000010048e2b075804e7a65acddc0"
            "a8b0c67f5dc1687acef4ed002423beb9d8e05968a264222102ea505bd876d05f9d828f50b9606973a063d94f8cbb01af93e7ed561c"
            "b9fd169b40bef5c784555fe03453ca3e05d789d339903c1af9cca48ac153373298e229f8b083006b572185144096371bc6c40639d3"
            "abd4e73fc11cf6dd4a3f7dfb7c408b78feffffffc3d30605000000000201220200000000000017005114d8825b53f946b54ab5a118"
            "80c302be441c08613201c6d006050000000017005114c8a4a1b435974426f2f6691428635e1d5ab88a3228f600000001002613b91a"
            "fcc2f40f0ed1e799c505a522b7d7145c2ae693af7f6c0125791a66306422210313eff3d8a4d22fa4ff77ffa61dabbbbe95e6c78505"
            "299ae2454223f9b209e93e40af4b6822e7dd2d8206ee914d2192e9e029c79721cbb7666588a529855d4625f51a412129f7b4ea7d00"
            "79830906f7515f02200a47ed06f84a3875f4b4e4e1ee8ffeffffff21750705000000000201220200000000000017005114d8825b53"
            "f946b54ab5a11880c302be441c08613201247207050000000017005114a78f6b9d1eb563f387a3c196309cc17f72e3f78328f60000"
            "000100e1fa10b6ba75c37b179e117acf16c47089fff478bffa95beacfd035e03ed7c2e64222103253da9d4fc694b9c0cc7b06fc87c"
            "aa893a12ccc7a50a434055a099ed59c74c314070b1e77a30a5437ef01d4cf134e3c3d8e235b37566e84f7be0ad9701fac37e019267"
            "7a1808468f543fdd9464a526a4cfbd6f6ed25608537e017159488ec52b02feffffff6a2b0805000000000201220200000000000017"
            "005114d8825b53f946b54ab5a11880c302be441c086132016d2808050000000017005114f7e28143e0fb847b65201560b220e40d46"
            "63285728f60000000100cb4f54425c7d2a8f45a68c4faf4a842b2b93714b4ea97ae58b5f3bb350bd6191642221020dfd2ae5082f02"
            "e332df969d4814b3fdc1af546af7ed3b8cdec9956cb5c93148402d0bdab8a22db3b4d660f1a57ec5c762f7193207a1281ed1fe1f7e"
            "07a5cafab2fa855408d59d0a1fce534c2f66f7bc15a2e4401501e8b1d4b733dcb9e2c7913ffeffffffe0b705050000000002012202"
            "00000000000017005114d8825b53f946b54ab5a11880c302be441c08613201e3b405050000000017005114a8f81a08c7275e5bc9dd"
            "d2706fcec4d17a09454f28f600000001005fd559fd1d46217c77e14a25807583cc1cea8aac742ef4022110a78b9e21a3d564222103"
            "39f5d2eec1a9be401d1f4ce6861d298f3612d1333a5f8edadb2bb32b5462a81140e0b45c5b4376ffdd79343e5a92276a1bdd1b971e"
            "e619c18a504f9fd49a38ef0378ef951e0f253ef21228d9a93f3e817439f940eed238c6adb54e5312e3b7ebd3feffffffd7c0050500"
            "0000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201dabd050500000000170051140189b3"
            "0bf9ab759ca591b5d5a267bfbcce3ed5ea28f60000000100e069f6d5bb22a2b2e7d26f6f67ce164f9077dd6e8a6ccacd9eff9df544"
            "291cd164222103e88c22880f61c450232d407472a8b9bc9b7167925c7bceb7b6821823c7d92a1140e2d7889c808b0b4fcd5b95f6d3"
            "b044f236241523eaeb94dd146523f1260dfcf14680cf83243178e1db6497277eb63ecea2399b7d16a7f1fdb0c35db9d1713037feff"
            "ffff84120705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201870f070500000000"
            "17005114d65ff2fb7f4900e73cc7790c15563cbd7187e6c128f60000000100de3c8cef575fe96cb947e6bb65c15d6465601de04cd0"
            "a7ad4c49fe3139397f89642221023a0002e277be0912fb544bbf3806fd8bb05335e90eec536d8d1b645691df929c404547a927d0aa"
            "ce3b879eac42e7a4f3407fc767ec556726054520a500d7205deeed330f3e336a5a2557aaf99451788c11920180b2501d4a46e2cc93"
            "ecceffa986feffffff95020605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320198"
            "ff05050000000017005114db542beec60c725e54c0ef98dbb917f94e302aea28f600000001007121f0e8b3cc4814b5a79078beba3b"
            "707c5087e1017e453e8e7aaa4a8d55e19d64222103aba4afa14cab9a9d8740833e3852ca91ae574dfcd75b5787bdb1a37ee1936485"
            "40723ec9b82c80149cb44eaf8f7480fbf60e888af66c7e131631784521b15dff28cb2a0d0b3e9207c48c3e6ac6bb527daa4da26887"
            "aa0948386e737302a4610564feffffff61370505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be"
            "441c0861320164340505000000001700511475000f2cceda47b2ff40abb5caec10c28e1e5ebe28f60000000100499c96b5d95ba66a"
            "c9a07d55b0ca60766a6244ac0860e95b67ac6a795f6b67ac642221028271e179cd4762f6e3e7fc083ce6a672b3d40a3c1551eb0ba0"
            "e8c134daff719a40456ec812dfcd27bda7f2ca4a72947cfbe8b15d5e88f4e94f293d28d7c1f34c70c3dbc4860b96e557f55c4574c9"
            "e863a35a3e697b07a93152179ba1b8f321b630feffffff67310505000000000201220200000000000017005114d8825b53f946b54a"
            "b5a11880c302be441c086132016a2e05050000000017005114508f311521fc4be421daa73d5ae3719244453a3a28f600000001000a"
            "ef1d09657d95c023a54ee0c61fdb85b22dfa5ce4b57fb20b537bda0e5f8618642221023a6db02f30b79a58f3cf5e4a95fbeca3cdfa"
            "81ed6b4543e16ba25e2fac7a60b140734616f39c3903ea039c0237a54e0d09776c7632a42eb338c3c4dccdf585d88847457e69dcc1"
            "e7e662d5fcc09a359805fea258992b1c5c5a36410a0cb6b2f0c4feffffff7c190805000000000201220200000000000017005114d8"
            "825b53f946b54ab5a11880c302be441c086132017f1608050000000017005114437b488938f78b38a08a3743bdc016e4540a8f8a28"
            "f600000001009dcfe9fadf2b26b98d86b8f58e8676d428429cad13dec742b57267a37e7d2a1f6422210268305318ed1cdca710fd46"
            "976db6e8cbe405746345d7171672126545e9594ce8407c6df709c0cfc775a192b70a0112b520864386166d40f86a2025c02c815282"
            "a877add660e6579c0c143d2356b2f00dd56ffa1235c902711c58139bb11c9fef80feffffff91070505000000000201220200000000"
            "000017005114d8825b53f946b54ab5a11880c302be441c0861320194040505000000001700511444b32bd362a7e176eb1bfd4ac782"
            "b6f67e72466428f60000000100aea5e9901684f50a30583dcbf4501367bd801a8169aa27c1a3b9457f667e0dfd64222102415015dd"
            "65abcd07a315fea85d3cf5c1af731ceb64666ffec35dd1640b6dcfe8403139eec55981fc306785a2af73f6e2886e238783a2f6068b"
            "b2c1e084521b15fd1d4d9103cdcd68b50e86c8b9e5559a59bc9ad569fc7a19d71e33bb42e50b342bfeffffff613408050000000002"
            "01220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320164310805000000001700511458cd03d661f4e3"
            "e081ad10b5ce1a92c21dcdb87428f60000000100777354399bef4cc38af46c2b69e704a8e100383901d4808a421c3b3cd52bcf9364"
            "2221028e62b7c2f494d7f2f44cbe515a75c91d858be7ea8d8ae0c24f3133b49352fb1340c7b0ef62c279b9b3b27fbba82bbda26f54"
            "76fc222cfa29756969cdb4ac182fa38011b595942d719a90582536778bc720b93122c2d1aa2b12f9cf6ea5a74c5134feffffffcec9"
            "0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201d1c605050000000017005114"
            "b426da4866874e46bcc30b86c41b4628c35d2d9a28f60000000100a223aaf0ff9489ec30d50aa71bcac4afd3e384a485f810cc7098"
            "38e8ce25ed4164222102a6fa55a079eb7ba6a3e342ce4e635e41333a34e9f3b06ce9958d3857664ef359406234bdc9eff6c21e1351"
            "da9a099491e16f38ff9c4d242593f759ca3e40ef57568006088205336814ac4ec123b724580d2394cbf34ed70e156e74ad25f1feee"
            "37feffffff593e0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132015c3b060500"
            "00000017005114c186144f93ae98b8f0ae27fdfb6dd30a62e2f7d428f600000001008587a2460dd20e98d18f4b15d83199c2193cdd"
            "4748e72e0dd9bd2cedfae5b19f642221020471336df80f2032cbfe94ce81a353f59b849877f6c7147c4a9f53969ed6695140c5103a"
            "e26e6e565fd394e755fb5fb0bab57aa2b06ba012818aa1f2b2834fd7e8e95dfaab4c792f5232e39f26325dd10fa7d9aebe2b5ef7df"
            "738ae9eb95e1b8b7feffffff8a0c0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861"
            "32018d0907050000000017005114ac0b129f54253a93be9086550e26a589cc9b3bcc28f600000001000bc59116466b38c4dc69467a"
            "51499d02a9b2d4051f03dfc914bb5cc985334f0664222102902f854382adf5053ae076d33f3358b8f14a52332623ea28516b761fb9"
            "8a81a8408ebac610b10ea6551afb2ababd4ba57a06bb5c63a6f862ee514112dca1ca32417432bff391be8a7fbc3b84fa0c00b3dd87"
            "79357a4f62907e8fe6a16856fcb321feffffff7a1d0605000000000201220200000000000017005114d8825b53f946b54ab5a11880"
            "c302be441c086132017d1a06050000000017005114a0b35a4cbf481b2a5e4c9b0afe2667c502ec03c828f60000000100c81bafed65"
            "f6fb2cb7985b8f0669ca2eb8e4bdb28239aaa1141fbf259a2617e464222103ba253aa9daf92473e215d5c0a2ca7992d5c57e89b333"
            "e57a001b45c4e0caca364012e365f5f262071d7cba9d184f90540b1e9fb4a719d546c6d3acb23c33e923c3f8a73270ae32f1b99861"
            "95f8c36e8835bf09fdff02692bd48d90fb9eca1f6dbffeffffff28700505000000000201220200000000000017005114d8825b53f9"
            "46b54ab5a11880c302be441c086132012b6d0505000000001700511459fdb5476ba434044cec8bcfe7d4fa72553bcf7728f6000000"
            "01008cb8f7ae8efa673e0fda152e9fa1c95dc66afbdeef50eac649a132adbb573af6642221030f87352ce4c6fec73358dc1f248c57"
            "a5b1bf9409cc28603aede9049cc7eb837640e6927ec1191194e4341870449ea05a65aa8f0714163bde65315bfdf7ec399f67cbd9a6"
            "365e15bc97f59ae17ba5dccc4b26f58432bcef0620fca9e77048adbe8ffeffffff0c8a070500000000020122020000000000001700"
            "5114d8825b53f946b54ab5a11880c302be441c086132010f870705000000001700511437f084708af66fa935c0f33788a121941ef4"
            "947d28f600000001002455c636addd67edb28e00207da2d40f799b2d806b0ba74ba978315d8ebe7e8f64222102740040ca9ea62673"
            "e37eca40c3fbe13dad760d7edb4917a0b8d6fdce269b5773402d8e76c57a7e551e4c8e756753a86518af54b0e2f7c7f83dd478be98"
            "142a49f0542dd95ff1c60a0af546c02649ee8c40aa16aff7b2d424a5df65ac655092094dfeffffffe8ad0705000000000201220200"
            "000000000017005114d8825b53f946b54ab5a11880c302be441c08613201ebaa0705000000001700511442ef4d46842603917270ca"
            "5c8dfc24ab13fc7e5628f600000001002b4c72c5b0a22e0a5ae010531eb0c85bc974e7a7b22625627eccbd5b17cf78c2642221021f"
            "f6e097dcb494f6de656df59a85f4057b0093b1f459b4aab72fa4742e5411b04031cb90b172aa4c4cf4e7d6a2148d4ef5fa872bcc1e"
            "5bcf20346c117adb6bed085edb0a508f91017c952cafa53cdb16202c0e80c2668ee2b93f994c24c511a820feffffff584005050000"
            "00000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132015b3d050500000000170051142d0634cc"
            "ebc0768f2d94b9cd7b4553d6506eaeaa28f600000001006b65f1abef780526e8f7a4b57fd0e40645682c3bc43b576e5b29b5218918"
            "05c86422210319717ef700cb3f931723c3d0575345dff42f9977386303d227ef57b6d744437b4093d55a55aacb1158d37eda666101"
            "89bcf8ab216612fd6c1bde6edd4e01079b75bdf40c200b0bdb2f887258f3ab25ec356cf772b670dd8379ebd75ca6b2a67736feffff"
            "ff10880505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201138505050000000017"
            "00511491b5f8e3ea1df09141d0a2555bbe4f4de7119d0628f60000000100d86d9127e918409e0219271ff22e09700bef106fcf66bb"
            "8d5115f00286817e9664222103a32efa9aa0da06f998829c8d4b65fc62727d49ebb83de832ca4d2afae43ef22c406165a819104cd4"
            "db34a2d0ba94eee7ad8614a28db58559fa1bf1b776d40571bf6d37cfb137406a3c45dd5f14497d97f9c1d21ac11f1e55cd9bf99945"
            "34dbd013feffffff04940505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132010791"
            "05050000000017005114fa25b04c20884c27a5321f3d1487c3e1c361192d28f600000001002de1dfe857404696c3ff9cfddbe3b4bf"
            "1b7d0511a90f9984b62e9ea3cd4164cb64222103063cc56e3451ba60245160fc1e4aa4806f669e1e08600f120c911f1886eb3fd740"
            "db7f051ebea17fdf9d265a7e82123b3ed68377e46491a5c15bc2587bfb71c3e20ed57248245f506080a3445142a565aeac4765955c"
            "cc9d7cec3d347373a12fd2feffffff3c5a0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be44"
            "1c086132013f570705000000001700511473a3e215798e444de0366b2c26b2e22ad28e072128f60000000100f3296d89c152b06953"
            "72b3bfe48f1856d1cea989563d8b2d2a175f8dafbb983564222102d32be50e9af2d577d0982458627bcb2251ebbf1233ef686a1d22"
            "9f67ab6880544001e010ab6e5aec6c9c9045c7956df46a850f64c3c140725f958b97efe6138f240da26b4de4dee074478c5edfd072"
            "718c736f01b092487e317f245167142d2d75feffffff46520505000000000201220200000000000017005114d8825b53f946b54ab5"
            "a11880c302be441c08613201494f0505000000001700511432ec72d9ad8238503a3ddd4f940073917636a87a28f60000000100c7b0"
            "8f4515cd414e947db4e745a5920e5acd51bb057f0d26960c1803d55aaee3642221022d5de3dd7ba2750ee710ccf0710c9a687f8cad"
            "233ba73dfca399fc9e00a07ebf40ecbf8308b78567e274ffe525c7a77bf382e5de099051fc42cabdd26da0eb38c64f4eabaa84f942"
            "09d91262b02cccdec6e522926cf8a5bd3a4c0d05ea90ba3f8efeffffff06900705000000000201220200000000000017005114d882"
            "5b53f946b54ab5a11880c302be441c08613201098d07050000000017005114e95170e06131fa6781e4da11c548f943674a30b428f6"
            "00000001003b9e769ade62bd5a9fd8ecb67561a94b2f48e1ef4ad0cb9f7c5ac52fb2653d2d64222103d1ade62cdbd0f2fb65bc3113"
            "0fc789cc13b730661231f8db215f0587eff219ee4043f14ba869499c23f21ee3e61f9c7ec2ab906fe40ba70251fbedd851818e6ae7"
            "4ccce80820a62e6315c61ee86724e76a53e122b9d8db4bcf93e8aad1fc9ad214feffffffddba050500000000020122020000000000"
            "0017005114d8825b53f946b54ab5a11880c302be441c08613201e0b70505000000001700511466d24b6e061a7c0b61cb65b7e6f044"
            "3f07d084c528f60000000100bcb1cc2c0d2c4352387428dcd65b962a971a251ee16f2b54638363082ab8217a6422210357bd305f1a"
            "2a571d98004a914cbc60fc197475eaac69a5c795f3ec63749684b740904349623aaa4bc318a73ed9476a78a8efdb1adfa679e2bcf8"
            "bb12858fb4c6b5908318a2e0f0746b42db1370282bf0d7d91e7343f5afcf1ed36dffad8cb702d2feffffff791f0505000000000201"
            "220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132017c1c05050000000017005114c31addc7416cba67"
            "4eb81b0dbdd556d819f00af128f600000001003b56d89b6bb9df547daacde50a6900f9206b950cb9f75ebdeb3d80446c413ea56422"
            "2102851ea9a976ce02413a873d6dabad9be9bd36fc9390e40fed65aa252896244032405493e1003eb8d79be9e4f31ef245d05729e9"
            "58160e6b6e6869318c1da86278bca59bc29c1d916f11a31fa7f4e5c43d7287a3757264c54084dcbef0b5ca549db9feffffffbfd805"
            "05000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201c2d505050000000017005114db"
            "97ff52a0578d6234bdb7d9e15189bf58c59b0528f60000000100074fb9f94b2d9c8c3a8a9132f0a4dc56731294a78af0bb49558744"
            "68be7b3b6b64222102c8fdc4b88c15b39d2027ab012b0a74262139f192113b41a0c08a609c6b3949c0405e11c66b73765a2a2e0f0f"
            "35eae68343c827f5fcd93f02f45c766d94173fe1e0ec3d37c53c518fa3a3c2a695abf24865bd302fb6272cc7edf68f55d1e22bf1a2"
            "feffffff05920605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201088f06050000"
            "000017005114426df9352a3c92c04b2ed2a5c79262410fe08af828f60000000100fcfc567e7d7e8d3a06f9af0a054fe4e62d1c7dfd"
            "1f7a9c3f51d26472e2835a0964222102fa88537e753cc2b78eaced008241eee970b2d367e5d952302b7ae8531eb3d61a409146ce74"
            "6fee02ef1bf19daa85bec2837d056eed374651ef6190048f2110cdc0ab9a2a336585646ef0d0b1e2de5830b778324c71b26b124194"
            "b30d03f46d0e9dfeffffffbadc0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132"
            "01bdd906050000000017005114fbf96db0dacd502b26b7d46cf309dcf7f7a1c22f28f6000000010046177808df40bd16646ddbadf4"
            "10237cbafe4063113c5bf334dbe123b41416a064222102e83839d2a97c11fe44a9d996537b040ee45b418e0065b9eeb310042d58e5"
            "c3d04078c6db0a16876577f932946fb90c83e629b6f783032964b519f738ec8dc8c3f05ba0efe11664c0d3d32e1c50d6755ae47695"
            "e7d2e66a4f59c31df0ce96b08c3afeffffffa4f30505000000000201220200000000000017005114d8825b53f946b54ab5a11880c3"
            "02be441c08613201a7f005050000000017005114f42ff8212e3f99ca459451ef611598eaf5f0de3728f6000000010021ce908791dd"
            "0e777a7e330a26c0fa104a223aa26a2727758fda32185e9d3f60642221034801dbc65bd55d781fbd429dd49ecc2e2bb90cd7c8b390"
            "a8fd1c3b7e71f697a040df26f75182e995d96054d2e6daf55a0abedac74a2af837c989462a2bb677de3b3023b58b12227230920a26"
            "3784a8dad77a87a3d74f5cfb968147eaeec4d29d58feffffff4c490805000000000201220200000000000017005114d8825b53f946"
            "b54ab5a11880c302be441c086132014f460805000000001700511401e0db2e64ea61001a7d7541b29e98d717341eff28f600000001"
            "006dcc493e919564b1e5707776db8396b97ded42180f433b5f851cbafd4ce2d39364222103f8864537749e337bba6c818dd9a93ea8"
            "a4d7b8ffce876b76333f694f5b025dc1408faa6db9e63d74cae933bc282c3ccfc752c0996697e5f1118a4de0c35e05cf8b26ea6f9c"
            "c5fd1b9132b347d58a7b42f30f2ae8ca663450db4526b5e980fed29dfeffffff801706050000000002012202000000000000170051"
            "14d8825b53f946b54ab5a11880c302be441c08613201831406050000000017005114394ec372d28068a0283caf9a80bc54728f9503"
            "5a28f60000000100698a1eb9d66a5776c848c7d1cb3f992b7635e679139d13b231d3e24359da370564222102c40acc7955cd308d3f"
            "77968f5036805f35e2055f67e84bcc842946c09839861440d4521c610b6bc7dc8b15f9d5b002d8267c8bcbd4b189ea79074f6d05a4"
            "782edc0b04e99e9a9e27751be7f2f2d0a627889976b6d9d1454c32fc54cfefe1783d26feffffff494c080500000000020122020000"
            "0000000017005114d8825b53f946b54ab5a11880c302be441c086132014c490805000000001700511408568a9c35d80bb7ec9e3062"
            "7883c8e7da9b622128f60000000100d7e1253dae12d41438fa0c7f4db2c68c0269da3fcd1891994a65e6a42d4d3fd264222102124d"
            "48307fe7d4efed378213dd20202b26771b4a8565f551bdaf44a6f707510240885cfa832f218dc98b5a3d4b8ff59d549079ca82b380"
            "694ec86afd4062c29313144a10cc921897c0f61039a1cdfd80388001e53bb43d5d39032360737eae5f2dfeffffff96000705000000"
            "000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320199fd06050000000017005114f8dd4ae17e"
            "af6164d095c48ea97412c1824d839128f6000000010079a118e404d1c726edb7cd2918fb26855b4cf8a968f0afc2fd0864e79fbecb"
            "aa64222103cf71f38772410a398d6628e74b8bd3975eeb0bf7e1ebd516192d53e772a7b49640a2f36d08473017f6ec72088893cfaa"
            "6c22fffa9baf60136f6c178926ecca310d1ff98a79da360addeb1fd6f28e09a6108a50a61740110bc124d28415ff5849e4feffffff"
            "97010505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132019afe0405000000001700"
            "5114a7ac19a2562eeaa04cea5ab372ef7267390904e828f60000000100e5d694559e0be0c4f71b47c59524f8975c43fd059b894075"
            "0b4acfeb80b65f5b6422210223d05f18707267607112926f91688e3295d32778afe2372423bf52cf32c02493405cbc873272788169"
            "06fc98a32f7439bc7cb90622acfcb84a5c7326122a856ae85323e67e15dd88827fc12125cec4fcfb21e42e6f285f90ca35a0bce060"
            "cebe45feffffff73220805000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201761f08"
            "0500000000170051141496a39a0e3f1206fc24a314defb92f45271092528f6000000010094af3108135a8710b679de7266b2617c35"
            "b8dcd77eb9b7450c05bb10df80858564222103cd060d0d18e75e55ded47bb6b41c659715bcf4fade69d64f7412a9e91925acb44002"
            "516ae4a2fc1d995e54b1b27bcaf41bb90d92d963b2f3d9eb29e88e7aa22ae263cbe662fd7086f42d79532be6e06920519c7925ebc3"
            "410eb024b2c9718a3172feffffff42540705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c"
            "086132014551070500000000170051144f72eead95a17dc6d6c3e1f266752a80dc940bd128f60000000100b5727017526f8bd21242"
            "f450206aef89ea265fd7ff7b7eef307ac532b7fa8d6a64222103b3d5771ffadb6e65cef4c60ba287e8f3b4474639256ca929acd5bd"
            "1b3744ccd540e751592c86f637d09c37a154fa4bf2d8f1fcbc054a48a66ef4ccd8a16e80ca0eed39fa3e3b4566108e644f60abb27a"
            "7a099eaf384cb228d3aa44c2860d2d655efeffffff30660705000000000201220200000000000017005114d8825b53f946b54ab5a1"
            "1880c302be441c08613201336307050000000017005114ee1f63524abadbf9e7fcad1dc69e9245c2e6d74b28f60000000100f953e5"
            "b61ce8f533b56b7b7ddc3164d22611352b432f6f24f2f864a0aaeac450642221032f41f6eb5d4a7d17ff6abbd79ef31469698f1049"
            "f44b5b2444a598c14e50e0de40df6c7797667b27073b0b76188e382dd462bc24261b15954bfc9097b03427142943ecadbabc47129e"
            "343f06f7430bbc2e31e2e3cd4e8914aea621cd829c40d177feffffff890e0605000000000201220200000000000017005114d8825b"
            "53f946b54ab5a11880c302be441c086132018c0b06050000000017005114b10001d7c3a94b3d25a2169b9c972c206f7b84b328f600"
            "00000100630252ea929c9f2257a5b6205b1437b3c0fbabf304a3508372176855fe15b38164222102e8152ed1b5c7d2d9ab532d4a83"
            "1653c5631331423134d1a3a55018920522b49d405845bbb8a95eb9cc95b734c5719297628d3c3463f24fce414090e870ba033debb9"
            "9c638594785e0f2233a363e4cf79cc9e50b53f01e8cb8ab6297295306eff36feffffff564106050000000002012202000000000000"
            "17005114d8825b53f946b54ab5a11880c302be441c08613201593e06050000000017005114ff766ef170891c92253f5d8340859c83"
            "0b5da91a28f60000000100f6c3c87f211030b62c8343449d3397cfe2969f95fb00cf716d8cfa00f4973cd664222102758f942b191c"
            "6cbda141b8ab4a5bcce9c9a5456b5bffe495baade40a8bc74245403827e54182e4ddca5bbbfb9cf479360c8641f648ba7d75ce2bbf"
            "cc4f7d02d44c3349e57501c65f806f9be75ca97ae81888858db0cf3099d7b14adf91853c1332feffffffa2f4060500000000020122"
            "0200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201a5f1060500000000170051140122d872f107ba0754"
            "2157ae912f97024f7f904528f60000000100713772277b375914f7c97d22cad89a62be741c811b860b9b235b7f95532f1812642221"
            "028fd6eea4cd24f4e60e88f81bf020194114cb19eebce26753e3cb634a5ad72967408d646c6a5ea97f2b6911cfecc20780597c89c0"
            "58362008bf756541c388fd90b072c370606d2ccdb011afa9009d9327275631fbc3e46f0e09b4382c01853f26effeffffff16820505"
            "000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201197f0505000000001700511428c5"
            "a46cc399507799bbc8f3e16e102b9ce0a39a28f600000001009ccaa11e89689ac18d719d327c88918a441fb31a166334ed0856430b"
            "bbf72fd864222102a50b6743be368e3c76a472422fb9955955c48716e25aca4ce8e19b7e223cafe140c8209cfc94c8b4eb1ece766e"
            "b7ecc198abd981b8b049769dc59dbee6313af7a6ae4220b9be5969ecb58ca5f0a4a8e8fd0dc6b1306d9b60813dcb318630ca2234fe"
            "fffffff5a20505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201f89f0505000000"
            "0017005114619fb690897e2bc8e3a22a482d712005ea22933628f60000000100c693186ad184696d5021466e87680f449ce1d955eb"
            "9a41342cf132e8e8dfd91964222103d4d22e0e46067bc9304019ad9db4a0c13f2e075941d2db878bf642faaf1f4af5404342672515"
            "8b38de090fdaa9d3b76c1fcb2cbfcc70fa2b535cd163c20436d82739cdf37ceb95a35764cab44d3078f483dbf4c95c4365afbba254"
            "e4983e8a52f5feffffff53440605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201"
            "5641060500000000170051149b0f143d37b9b8c6f800a0508ee288aa206535bc28f600000001000e80a841f9448369a760786ae92e"
            "e9a74639191954883dd245a75908877aef77642221038938ef935ecee7d77d7a2dfca5a6e1af779fe5ec11efbb9bcb07589b3a44b2"
            "67407f2e2f851e46bcaef3a9bf93c872db084738be55262af50cf4dd6fdd90ecb38d69a66b197c83ba565abbb98fde2ae426fedc8d"
            "32fccfdc76ccca4dc829002d58feffffffcbcc0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302"
            "be441c08613201cec905050000000017005114a48844be84c9a17a2804b784955f518ff0e840f528f60000000100410df5f29f01e7"
            "cbf0a38e6959c4bcfc9dbdac96c6e757971d3cd3d8b465427b64222102bd6ab87f57a3f3b44405f9b60bb8cb0fdf99bbffff6f99be"
            "bcefe36b9b7dda4240cda9d05790f4ddef3a69377c32dce4cdc99ed1227a54971a781793e3d761e757a1564157cdbcc88ee5da12a2"
            "c0eb0a9be786bb50e03044fdb48ddd06cb0eb569feffffff5f380605000000000201220200000000000017005114d8825b53f946b5"
            "4ab5a11880c302be441c0861320162350605000000001700511418a3361224abb8dbfb1dc041e7346e952c307eaf28f60000000100"
            "a142ae1f668c2ba7134a8be898c5e2bf8166d1bdd137aaade70816854c1c9fe16422210279b706371820a848ae92732e6575f620f6"
            "19c0125b06fd5ef1220e4b1ee645c140c174e185bd5810a3c6d0bead24d4aafb5cd12c68cbfffa8d92163bd9148d60ca7e25eb55c2"
            "c887f047e494250ec6a763638563c1a7b0afbd382f7edbb3582a45feffffffb5e00705000000000201220200000000000017005114"
            "d8825b53f946b54ab5a11880c302be441c08613201b8dd070500000000170051146be2c2933313d79e7d633c5fa39ff6ad9e0e1d0a"
            "28f6000000010059ea104b7cb92831553e333897e3c611ffdda4ae0215af2256679ec3d8185f8c642221020ad40c20ef149c888379"
            "cde5693528dbb0564929cc6905693450dea7379a29a9403f19a044a017ebbed362d12b5009e53102086c78664e02d11dcf37db96cf"
            "5d8ded5e554fa598783f8dd215f32bf4095b91432977d54e50f80acd597078261cd5feffffffabeb06050000000002012202000000"
            "00000017005114d8825b53f946b54ab5a11880c302be441c08613201aee8060500000000170051143b941ef3f785b746202691205c"
            "a86498c3ba115028f60000000100ba792d981d33ddc24391554caacf600dc2fc67c93752a890b7547f543112ab41642221026a8908"
            "bb017093f9af05f6ae4364d63ad595a1c7a906485b9b8b2a06516bd271408dd6d651a9a41d61959200fd9ed141a9e9a1dedd2a5502"
            "f695d7b0d791e1cc5b0dab71808912372ec831096d72de83e93cacab9fca3922226c108124cb8e4a88feffffffa0f5070500000000"
            "0201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201a3f2070500000000170051148e3453e47888"
            "46e5f6c06e523482bcaf8dd609ec28f600000001005323b9590225e0a625482216f03feb5c6c4b215071ad7d9726f0ea83ab726cf8"
            "6422210257ba5c628de7941b550c9ece5f13ec782aee8ff43271844d25cb0540916995f4407789acb75bf043c453cdd5fdde11b1e9"
            "af6d5199c131bf4f956b49118202ac4c3b56bf3f3de75bddee9c55322b294ca50845efab54ed645a08d5656b3bd327befeffffff6d"
            "280805000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132017025080500000000170051"
            "142557ce1e1bb725f48be6ec5b15ed85c7ee1771e828f6000000010049fe399007a4df2736dfc0f032dfe30b04389f9f4531993441"
            "47aaf23313660364222103021d18aa992728c253dcecb903ae2c4322f0a7cfe739e64e908eeca93c33c638406bbb22d66116055e76"
            "65e136a968fc1d49a9c03d94f2203c8b3cc0fc92fa0feaf9ee27e0b449490b618dae0e2a4ecf33b7ba6c98dbc4fa1c3b4ac12d3d90"
            "659dfeffffffc9cd0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201ccca0605"
            "0000000017005114fd8a86d661c5f52eef6f13ab6dd9fef04d2adf4328f60000000100fb343f4107757c4138b62100dadc4f0a238d"
            "b35f161a2b8f4515332aab7967796422210296f5c23c4eca17e965bc386a7eb767d1e664a8f84da1db2b5bbcca8e994b901640b703"
            "f1e6e544e26bb219b0c6e4c3e67c8b29e642c0e6593da4211368a388fee07062cbddc3133d899328bd1ec71b2a556fa980814986fd"
            "77afac9df8fcbb00e2feffffff9bfc0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08"
            "6132019ef9050500000000170051146e654970c61778349555dcd8c8a84c2b6e3d9eec28f600000001003e70e2cfaa36128a2f3af0"
            "6fd97fe248c8bc4300821a801a3c5df3ae904b082764222103b7da78b98cf8f68d9e659b4736b4085cb9bb3696b22b48e149ef80e4"
            "555e3ccd406788d0dc8e2b56da5847597183e9b66f8d3eda1d01506443b0f120b9f4a70ee3c495fc00adf2622af84021f3d94da327"
            "451320206d0d3b9b2965d610db0c7c79feffffff088f0605000000000201220200000000000017005114d8825b53f946b54ab5a118"
            "80c302be441c086132010b8c06050000000017005114cef8afe4350fdfc01798ada24901af853671149e28f60000000100961d0217"
            "d6bc286d8eca4a18b336d0880e6bd761687504a73cc834f355f09ea664222102ff7a2fd89c4fa5e208b5aeaa7bbf8acff19edd4794"
            "6c8b6b9174258a9309169b40aeb7ee5296e5a944aa03d7d43b2a656652dfd8de7a45b5043464106b8e586cdf91fa0f01b6dbd469c4"
            "72eeb977e6e4f830154c85a505e32ad1143d4c94ee3448feffffffd6bf0705000000000201220200000000000017005114d8825b53"
            "f946b54ab5a11880c302be441c08613201d9bc070500000000170051145484a88858a475a877f7b4ce1d67045c7391d81928f60000"
            "0001000107bac5c9a0777edcbe7ea681eeefdb59a6e6c2e9f732e4857c69b4e63052da64222103a77b763a544b745972bb60683bc9"
            "0220f04c42a6cfad54a924f707355f7bc92840fdd0a3fbe7508204bda41ec9b176304fd231bbf59ae98484fd36d72c4a68a52754a2"
            "7413df98bf5804e1cce35faf2260e29c1e3dda0e1f6635df2221b085e474fefffffff0a60605000000000201220200000000000017"
            "005114d8825b53f946b54ab5a11880c302be441c08613201f3a306050000000017005114a2fa6bc328ed78f372e2b8639a884e2551"
            "4a8ea628f6000000010001dacd4c772e6139a25e58427eba1ed9ba4278e1b3319cb56cbbdfc5d8fffe556422210286c22e049d1093"
            "006a56ea8a0976f48438c2e4dd53286c33528f95192a94ef3c40372eeabb3ffb2dff62619bf691e78c0a764dea71042f376f5dd261"
            "1bff9334ee257dd0d229bbdb622e2b291c4011926352b00a2062f4c52d8b993c8f36b1ef21feffffffd5c106050000000002012202"
            "00000000000017005114d8825b53f946b54ab5a11880c302be441c08613201d8be06050000000017005114ec98f7df969b82f22898"
            "e2e11985325749f3b71f28f60000000100d2716b2c9c5a06dc0810bff9ee6e03df7875590497a35e80eac53d44f2b1807164222103"
            "4fee60865a8b20737f1db62ae9853c3a87b731ba1ae1a045ed4269897a52eec940f53a6ecf9cc0a4f2be358d25a62de7c56330df55"
            "4c478d9222dd298f3799fcca760f1777a39d47be3e2d6f24ea250670d565c6287f0348a14121bebcb83b4f4bfeffffff1d7a060500"
            "0000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320120770605000000001700511423c325"
            "cd9007e3ccea2c7ab96613b6edd686156928f600000001007d7c50f65703f49fc45f2e2565a8e366ba9fb5760576f2136f79f98cc8"
            "e6047a64222103a0466e26dd9cef2b33cf3c41b4fa5cb300fd070100d1abe2fa7e624078574d6340c1e325c0eafea74ed0608086a4"
            "1bbe3d353a151155b11cf1defc1a0ec79fba55946315974da3ca712a60b602b700f9740879e62bbf24aa9a24ee1d77332719ebfeff"
            "ffff02950605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132010592060500000000"
            "17005114c7195f73893d4f3dccc23488db45b758766cf57928f600000001006d31cee067f4c3a06f663b53e71512f1ef75e89d17af"
            "2481bff56f9520882b4f642221034b221305b088a414aa67004be798d034dfb2ebccc425dca48908f63cedc26db74026b1ad13a9ad"
            "a0337acdf8a8eb3f9a6afc7ac510860679dd5f0d988eb9b0cad89cf786a585f7621f6834d87c1b181d2705ee88161a9536194fb5c0"
            "6e23f7e0d9feffffffbbda0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201be"
            "d7070500000000170051143ab93bd1c6020cd142453f9e6fadb63d7b8f382728f60000000100371719af0439eb8f5b3992e0aa55ed"
            "09bf76bd24dcb5f233935aac92bb23c00e64222102ffa0d67ee06b281c45caa2fdccf89377faf1aebdb97fb3ea5b2e0be762f0c9b5"
            "40fcfb2682c5f46d4bed5d29cf06f5ccf0a1427c1b271ba25ce62ef4ece047d2741cef8876a57399481a4bc01535f430568eb81614"
            "e9044a86eebf61d0c964d9f2feffffff8e070805000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be"
            "441c086132019104080500000000170051148ff168b3b53190470bebdf224b7f9f87c36487ca28f60000000100d0f87eef1bb6d4f2"
            "388c888e30e3c79201697d5929059a56bcc4ebf44d6329116422210266d0539a66f72fa46ca510ff518a2604bb2594f7a4a0f7d05d"
            "d4e79390c375ce402ad60fefa1850945822187d88d73b219a08c5fedef41b5eab8c62d85bb30c520986ec9e7d3b8ca4ac366522df1"
            "90aff26c4e820f540db3e90108109dd2c0715bfeffffff51450705000000000201220200000000000017005114d8825b53f946b54a"
            "b5a11880c302be441c08613201544207050000000017005114b17fdacfda807a3e4e0b84ff8765873e73abe7ad28f60000000100aa"
            "afee5161a2c3eb382f1202f80f1725f7a09a6ede80d1dcb22e73d666c3846264222102b7aaa0cd13429cf729d320fe9aa7c53383e5"
            "696f2c2bf01d3af3a1f81eed327e403421797d8cacc696b135c7db32c1ee420954bfef4af2cbf98f25d3f18067a1df7105ac71141f"
            "0c45c6b1c3a4dfa3bb8387f7e12263052a374020c0716f66b1e9feffffffc1d40705000000000201220200000000000017005114d8"
            "825b53f946b54ab5a11880c302be441c08613201c4d107050000000017005114f57f13899d12f0e5ac334f53b720c875247c2dd628"
            "f6000000010035888b5045cfed8d7dd72aa65d96e18a85029a296257053f0f975aac54761f5464222102ff09c07d21f7d14ab16a44"
            "cfe299eaaa73ddd65148d9e1038ea87c6ab68a7a604067160d44138ad5de689c39bcf2936eb8712bbeb99585836010dd5a5686f0f7"
            "21c3965fd2af29a1577f8b6ba93ade78b40225fe174cfa735463340366318bafe9feffffff98ff0505000000000201220200000000"
            "000017005114d8825b53f946b54ab5a11880c302be441c086132019bfc0505000000001700511459449af177c167deb564754aefbf"
            "8bbabdf0bc3028f60000000100ccb0af2773c734958b1f7efaa3bc62530319825667ef6b2e70aa16f4d83b5c65642221025570c28d"
            "2f45eb33f88d9ba0fbf1ba6b02a7457f7152504fa8702c0bba954bdd405f238f7c51f199d7794d413f9ab0823b250d420545004f36"
            "b292856dcf4959ebde696fdef17fe1f5472cc8ce110dc8b5cfd99bfc70acf282dd0c72d127559305feffffffcdc807050000000002"
            "01220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201d0c507050000000017005114168101baaa97ee"
            "4de5d2606bb5ca2e41b6eb2c0628f60000000100a6ad4bea5f38f3051c0e04a8ceb227fd0c2970fa119b8e0deb30fedc3f2c252f64"
            "222103dc1a2d891a51c87e79c0dfe91de3b72934730a48cfabe888af49be46e67d346840bde1f868afe9d8b88958af7122543ec358"
            "478fb4db29c3e1f67f78268c87a32a730946abe249fddab050c8d66bbbac276bdca294aeb054dcaf67d3568cfffc9efeffffff2d69"
            "0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201306607050000000017005114"
            "de6f76f6b48c986f4e33503119d725fca5fae18628f600000001008db3755ef19dc3e4d09f2f125251a5e9132c44eaaf7eaa0eada5"
            "8284c74ce16164222103413fb4e9c6a2db1190bdb382faf51f2cb6ef172d24f5d263691ba52c8930ccb5405d33adc1da5e85f42f71"
            "a718ebac6f969404d81bbcb3f5cc81f1578dbf918b926812ce88f2317601a04bd1d516e01dc28cbfd891ec55eec07ea1d29b0886a2"
            "f0feffffff85130505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132018810050500"
            "00000017005114a44dee9de98a57c3cf0580e4930f2bedff9f6f2c28f600000001004e75e35951bea93e0f42611c8b9518517624b6"
            "c861128ef989eccd40f5d5f17b64222102150cb1d4ecc9a42e450c53cbd9ca77b7f76f92376da3f6aa20243c9b79bb7ee4402b341d"
            "e7994c4190973c4189ad7f68f95e36ab0754b01436121977f3763cd650eb3a310facc7c9f3d5825bfd344d7dcaaaf4d0513387460f"
            "ef7c64691b9a3c80fefffffff6a00605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861"
            "3201f99d06050000000017005114263111a1af87701c4a9cea5bf1bfd7628fb790fc28f6000000010088dc6376dcd6e9142185ab66"
            "ff73cf55a77ff33c19a223c1836e2db7ebf0f3e464222102f5f05e5f5e0d97fdd087c17b7ceb6909e97f36b8c88677f9fa17209fc2"
            "dd8c684059c95a40dbbf7e87603e4053f2d090420649aed14ebe95feedac6bfcde10fba29b09d463b8be53f606ffa154aa220e33d7"
            "93ad4f1291db8cafad611c6164ab3afeffffffefa80505000000000201220200000000000017005114d8825b53f946b54ab5a11880"
            "c302be441c08613201f2a505050000000017005114c1d49fc7b05100634a6da7202dd8a154a3adf88c28f6000000010034707a1b86"
            "660378f07f47ae09a290bba3a873a4c5a2ef5130eb0eecf3f438fd642221032b13e4ee62ff42a1043260848d7763b2e2f2a589d345"
            "7e3132d3ea7eeacef33c409526cf4a093ba1230f37ff9c2145154b8f6c5e681f1346720861a6ac97b3a04db99a232033ed9c834c0c"
            "589d9a07308f245e951a92f5f492e3c145375433a439feffffffb6e10505000000000201220200000000000017005114d8825b53f9"
            "46b54ab5a11880c302be441c08613201b9de050500000000170051149bf1b77fe7e31e303dbb050225ebc8d1f426a42828f6000000"
            "01004c85c54bc37eb2429f4fbdba81b969e3dca09e5e932582e53e235d4eb63f57d264222102497cc60b39ad8793754369c1822787"
            "e24f44b758e683e1ed9ed974491fbcc37740b8c5979bdff8985745103ca7b91595d553b8fc98e300d9f6c65c7b43c0ecbff4986c0c"
            "d3c9ca38b60c57a1df0450d4d54009ad52d6e547b59cf7be01c824a966feffffff1f79050500000000020122020000000000001700"
            "5114d8825b53f946b54ab5a11880c302be441c0861320122760505000000001700511421b5d8b2b5c6828329fd5a81986efe12715a"
            "0dcd28f60000000100d1671711bf779b55dce2a4913fe71a5ebb4fe96a5edc8bdadb7e7a2e231c3fb6642221033f67bdf3e115026c"
            "54034d8d86c0da3e3c762a814a83966a777ec5bf8334fa0140f75c795aa5777c30d7ba3a870a1100cbf9d0ceda453632f74df4e8bb"
            "c9e8109b65cb2fb483cc97c5ef5ab215ed8222dd36d740fba71dc9a16ca03f68ad74013efeffffff8c0b0605000000000201220200"
            "000000000017005114d8825b53f946b54ab5a11880c302be441c086132018f080605000000001700511436a0c18ea6bb4c7e8e5aa8"
            "f09233178c464e6d4228f60000000100f5ce8f1e144b750159c9440b2f40416a6e8e89143cdd10a57d6918b3baca629564222103c1"
            "544b955b2b99cf95a8aac58389a550b193defd4ce5b3effd68d9e63f3d9d7f40024bf18b5c8e961cb221a57ea1d0e4faf83e064d16"
            "48e58e3123eb5c406228f6782a19269bc828f712133ceba05384656cb85f2b057f66eceedcfefbca07fa02fefffffface907050000"
            "00000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201afe607050000000017005114327021b1"
            "4c1e50e48e8edde892f4d4fce57007f328f60000000100362c8991897e45adc6675da385209516d48f0718f0fb3a786f8d7d628e73"
            "f5df64222102c90cb99e1387b0aad801a897b95ccbd3e01222bd66ea825bbe175abf288a2bed4056fcc3c2c359c390cbbca31ec818"
            "999a958789fc74fb7097d8702c88d9b538f5296af603d2e5cd809571356fbe47f013ac02ea0c0ad5ef7a7e4571104b95dc22feffff"
            "ffeaac0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201eda906050000000017"
            "00511427b001a82462f6a9b44b36838ca94730366ca14328f60000000100428ea6d102760088c9c5c6b90f1a694bfc5494f79abf20"
            "d1bbbf23f5caab0c676422210354b07fef21c8aa4257d9e948af78e4be6fd70b0bf9eab6d5157726bbf4236b48400207f8998a0e6a"
            "0b74f774cff582e8688646067034015e1a3ebc56a8ee5013e25f100c6c028c96398d0d8b49cd327d265dd000b3c4773619a9a71394"
            "8ca0a9ecfeffffffd8be0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201dbbb"
            "06050000000017005114118cc389cdbec16ed6553c19197d7682154614a828f600000001004524ed31249b56bc72033892ec025416"
            "0d1286cd3818057f59b3f66f64bfe49c6422210354bbac53a6dae5bd9f7c279c78a4d96832a954d76d0211c17e2db15e3988a5ad40"
            "f046e2c44927beabe833cee1cb9d526e9b375455aab96194bc4fc13369a2f35238ed3c8f643f46b44ff97a5c1cdcfed5799b7b428d"
            "d41b464b57dead036ac62dfeffffff6e290605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be44"
            "1c086132017126060500000000170051140815906bc6850b5c2429879f945823d642b71c1a28f60000000100704973075a95a4bddd"
            "ed871bee3f50a52ffa514269854265520cca2016fc9e1a642221025af1087d873e4e45ebaa2e0b9aee67f35dcdb58c98b3b46efa4e"
            "94ffd3c1b3a740590d984811a058693b7f969af3d2516441e2cc81082efaf011326901dbc72dd94fcda376e77ff5978f5384a0f197"
            "c4f36197c0c4271931ec9e5fdad48d2e9e1afeffffff7b1b0705000000000201220200000000000017005114d8825b53f946b54ab5"
            "a11880c302be441c086132017e18070500000000170051149685ebdf322e162acc60db1dd2af085f0db251cc28f60000000100cf97"
            "841f0dd9e38e7cc22f9d407885809356b6b0b43473ab00f00a130136289364222103417a55a01fc9a0dabab4e52ed9864c255e1c79"
            "c513c2a61a984d54a9b0414998400acf1864e6c61e0bf8f97978a6239817a2a845bdbefedbb040f7ad61f28aab685c5be4a26202f0"
            "c1ed37b48f6d00f305e34937dcaba719e63574b2dbd457b825feffffff33630705000000000201220200000000000017005114d882"
            "5b53f946b54ab5a11880c302be441c0861320136600705000000001700511418bb6e13ab224497c5258c21968afb97995b882628f6"
            "0000000100317fb839df9e2c4a56f9bd229bc5ffabd925634fbc11877f55c8710da3cbf925642221025c4c080d6374a2fe8b283c46"
            "7b8bf792077bf80c041cfab1427216239562dcb34094b3e59f04157057d29512dc07e5071ed1ccc99a3850f6bbf21f627b11ef0acb"
            "cc4a15dc26730a20c61581051bc5295eecdc6deee594015e8c94cfab3d24aa81feffffff5047060500000000020122020000000000"
            "0017005114d8825b53f946b54ab5a11880c302be441c086132015344060500000000170051140e6e1e0b59f5f876977aef135f86b6"
            "f81232a2cb28f600000001004226d69867a6f4c8091bfe65ff2d5a033ae4c4faf73715828394bb5309cee55364222103ac59d7660c"
            "764ddedc32a1227ea61aa1011f01ccda307cbffe509c68efabc21640e3d5e01ea01231a47e27fedaa070a13388d1ec2c45553af278"
            "d0de133d2177e5dede7f1ab64f37121515cbb23ab3052289a781c0679131f5f18053ecd83e0607feffffffbcdb0505000000000201"
            "220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201bfd805050000000017005114802aff278ac00941"
            "0645218f6ddc3db35ee0a06128f60000000100da3dc294e545c43ef331183ba72def856537830b4e282f71aa172b4a64d7c41e6422"
            "2103143567102f98ae2f10b7b6b61dd0c79a2a050cb9f4f7d9d71a908e010f6a3f8e4030c8e225249772de936d37284307a16ccd9c"
            "0c967c640ecb41f2b94d276021cb44d91d67c941fc7bc9cf3dac888698afb6459a184acb31f03e2b3396ce024de8feffffff128407"
            "05000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320115810705000000001700511404"
            "b42347b39339e71078b952966119d6e8ee26f128f600000001006f04a2d122e305a6c83af07fc913fa62bba64e96fba152eae1fb71"
            "740673900064222102fc26881d06b6f616ec406a1704725af75b763d415b83db9d7c9c0f8f8b901e7d40f8922745617fa1521ef54a"
            "4c87ff61ec0cbda08de5db468132a1695f0e9446b6d97b6ce93b0dde509892b649a712aa299e9f5438f469a832cb2ebce0c4cc7bf5"
            "feffffff573f0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132015a3c07050000"
            "0000170051149ed2b9b0baf9e808ce6145caf2dc873b6f2d71e128f6000000010035ac029c561642ddd5b672078d32c5faba137206"
            "520045b3a903a45f2fd19db8642221021c64b4957561cd58b520317b614e8b99fe1bbfcbff4def9afd03a635a32747fe40d6dddb5b"
            "f6a4dd82edf082c19224cd57d5bf1bf2f8219b51e234a17a6b3860f5470084db1dfd012e86660eb81567f1c302c49b6bc0a090c9c1"
            "4878ba9fe43c13feffffffe9ae0505000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132"
            "01ecab0505000000001700511468195cf1763074b358bc74cb1586510bc7e7ff5028f60000000100e59b10acf6a79c54385ed43d27"
            "87a1ea8a286649902d00b76d3b1445ffc45b7764222103845ef788c56564ce21c0e4de3537fd71d36a9a060d70c240a7680e8eabea"
            "0ced40d8244820807b6e73ed5c36cce83c45ade6c873a8f903c82fb9ed6402f8fcf1229e161358af3c76dd186f89df0013c9209b89"
            "f1300dd1709080487cec2f7c656cfeffffff2c6b0605000000000201220200000000000017005114d8825b53f946b54ab5a11880c3"
            "02be441c086132012f68060500000000170051148b598764a727d6e60d815382868ef56ae31260fa28f60000000100cda143652435"
            "e823de721962f153d83f39239286a1362df7f4395ef4d9b7ace5642221024c178a07666df02dd1c748281f67ad0e3170983bbd9b23"
            "68ec24310dd745015240772e6e4a9c065751590fb386b65e48e19d2e9d49fcfa19fc3a713d1f82443700dd58d6a6d09ad7529d759c"
            "ebea2adfcaee14f84484686017089b5497397479a8feffffff187e0705000000000201220200000000000017005114d8825b53f946"
            "b54ab5a11880c302be441c086132011b7b07050000000017005114d3400013824059bbb9cb7a808e50e1a8afd4f38928f600000001"
            "00a726ed1d0342edf78962d022f07def1729506ce2f46601db295f14e2aa25bcfb642221023c34cac33056b3709b522824e41c80a9"
            "1c8266e17049a3bb9e670caf0051553840843aa36b92359aa95ab5709538ec60ba94f8e2853b68243a29a3097b661b19208179619d"
            "89b6cf0eb9270ffa65fb945448e29547e3698502b24d2ed423671035feffffff9ff706050000000002012202000000000000170051"
            "14d8825b53f946b54ab5a11880c302be441c08613201a2f406050000000017005114f057f3421fb556ce83e8bd51748b8eb0d40814"
            "7128f60000000100cbe15ca6498a64b7a09c727b6cdbbb6f732602fbd1aa1c8e57ffb312911a98566422210256652d6e96216d0552"
            "c42d89b66a736904786c79c6850f69d28b5af3978b650b40511ac80f21e28220325027d823df91aebc55454a63eeb6b724c29dd08c"
            "ac2494f48673e931e5367c6466f29e17d7596569afd34233875fe175c102a47d5f081efeffffff6434050500000000020122020000"
            "0000000017005114d8825b53f946b54ab5a11880c302be441c086132016731050500000000170051145e553b5c37e2dc1d9ea84420"
            "94089752e220631d28f6000000010021feba9aa209be59af43256c3cf2b5a252e78329db4d9a92080f8be02669d829642221020a12"
            "87ca966f9a1a8d5966064e9b15fff9810c706bb80553cdd8e1881f66ec35405f970f39a42b66489286709855e37b55fd0560af25c4"
            "4f52e4d0957942d74f17c6a90e6f483954207ab473f3999d63019f5d79eef7945c9c0230299eab5bdc98feffffffcfc70605000000"
            "000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c08613201d2c406050000000017005114ebca2a585e"
            "0f383242a2af9a6c8dd259d6b3940928f6000000010005d9f6b6ff329083f579696f97704567eaf430b318720f974d681d48f6b5a7"
            "b664222103a984a48ff969716ef88a4f29f3d8d87ab50babcfe644402c96b8c736ae515fd9405e608972383023ae2185b4c5192f16"
            "09aba495ecf246751dcad1ba94d2df3598bb8b05539964dd080ce933d98767d26537228e15aba90a2586cf3a4cf6fbb7b5feffffff"
            "2f680605000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c0861320132650605000000001700"
            "5114ce02b1799491b8307bac44e15a38f0da1e1d5b3228f60000000100bbf623393d02517f65657367793c0ec63e63a836a06b0728"
            "45324e3f331c309364222103d77307a9bfed11b1674aee80cdad125ccad81a4ac6b41741b80a652be4c0f6f840f9dcd4b416d6562e"
            "9c3bfc4b1d284818d808ada2d6d55203042ff0aaaa984c81b308698e024d033975142db4020c471a6eb80e15a7d950fc68fba391c3"
            "40b83ffeffffff4b4b0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c086132014e4807"
            "050000000017005114f86e3dc9c68aad78e787867f3ea5c38f391ed19528f60000000100a1067d8634200661f1f4095dab22eb751b"
            "12f11e13689e73609302a062a7d48e6422210326b0cc33fac5d0bba94f077e7ad801b61c2dcde3503e1319a53320a08d76013d40f6"
            "b1a460232a0f6922ced691e29637536a01488d602049cc8d4281aeb3cf1a5e38fd32631a6b6a3b59b328cc0d3d0e77917369995681"
            "280ed34481502fa7fadcfefffffffa9b0705000000000201220200000000000017005114d8825b53f946b54ab5a11880c302be441c"
            "08613201fd980705000000001700511428f1db662dadbc64696cf104176bb64f279ca83f28f600000001000bd3dea4d40818174c7b"
            "a2acf0bb61f4c885bb0a991c54d03ec6fbebe013b4b964222103c4b7c210316948502894b81b2bf5a294a836931dc7d75e2f25f297"
            "92fb53c8e140a2d8fbd55caf2c2e37fd99d96565094f672dc59f49335979b6d43ef1dc4dfd0a9f1357f1ab14e84092d05966048026"
            "2012594320a2aba9668bf777d83f6b82acfeffffff7f160805000000000201220200000000000017005114d8825b53f946b54ab5a1"
            "1880c302be441c08613201821308050000000017005114883e43ddc6f658c0b31b14dd04951d2fc312395a28f600000001004a84e6"
            "a4704b5c27d1307b0e128f31ac70426e00914cba2dc404849dbcfc86866422210311482f7f759956b9acbb6fb1b5cc99c02a35e8df"
            "d513ff48e5dbba1c1e38a63e40993db18a400395142b397e98f038a7aaea504e3fce25f985437b41ed0ceffc1975706b71146e9abb"
            "1cec85bb5d9987fe7d4917c502d5b59dda7901f3f53d38dffeffffffc8cf0505000000000201220200000000000017005114d8825b"
            "53f946b54ab5a11880c302be441c08613201cbcc050500000000170051143e3828335507445fddde87aa1cad899aba46d97b28f600"
            "00"),
        SER_NETWORK, PROTOCOL_VERSION);
    CBlock block;
    stream >> block;
    return block;
};

CBlock TestBlock2()
{
    // Random real block 7efbd25a904baaa26ca7d93023381f4f77f040527957afac21cb21c87dbce1b5
    // With 4 txns
    CDataStream stream(
        ParseHex(
            "d10f314fdb101c4c3968d31a0f10bc2c61a92c7f5a15f0ccf2bbff10dcf3bc6050df021d6862befaaba8f1d98e7bff973f07e485e3"
            "5a3187d51c0d9afa85f08cae52a29d292be055c1b3f3bc43ff869e198c2bdc0c9bf79c73159968dfdd239f4bfd62e2000000000000"
            "00000000000000000000000000000000000000000000000000001d11506384eb4a42d952720a3a0000000000000000000000000000"
            "0000000000000000000000001c04000000000000040000001034001cae2e0100000000000000000000040000020194cf9a3b000000"
            "0017005114ea46099fdf3716e7586e4f637d2af76537132d910000000000000000000f6a034a7601092f506f6f6c3133372f000000"
            "00000100051e827c42270a3e5a415d137ab713c9131486f9f58be06ca1840976ba39154a642221022cf1e440c5b4ba86ae795ac319"
            "d033d4a4c76a479d7e353960f3616f79aa832c4009851ca0341fecd722274805daa677e5456434a1e2632f9e2667da951159c93c2d"
            "bb74c2af62edd34ba26ca2cefbf8911fed74de351ab54a8ba47693c18f0b06fefffffffc0300000000000001012202000000000000"
            "402028514a0929b5688b01daac98edf421e2f5abe96a78cc5dee3a3bb75cffd3000008c45f0000000000fc5114484e0f55f9d177da"
            "c80f34cdb6c6be5e4c33ae24000000000001005c754786dddf436d71fdf306f7075027ca4f379c48400ca6dbd27a05ed6ee9f66422"
            "210240c08f49933afb9274daf73f72e52d608bb1b6684869cb612d8c355d5b40fadc4093001b38101be058a537a08631876f3dc5fe"
            "0e0d5be8d47eb217ea155c759fc11dc8a4571066ef4a1da75230f1cebeaacec11370dc42700db753d06c77307c74feffffffa08601"
            "000000000003000000000000000000126a0438564c050461736466046173646600000122020000000000004020e1ca1d3422f400ea"
            "89c97d19e98d106da81ccb07b9ddb99bbd8cf1ac605400000825120000000000fc511425151b58158a6aa3c8d147a96f7beeb8c671"
            "b1c201a1820100000000001700511469f1ce2681c6a71cd2d3ca88ec32e59912f03b0700000000000100f750548f106fe1f9e16c06"
            "a58a5f12cbd10d8740eac205f7dfba465c1ba2c54d642221028f4959bf4f8752af0ca52cf1b213eb74bcdb7122410ddabecde20a72"
            "b4cede7e40f56512837e625c924a07487da04e44627c7d087d368d0234e9a2b35117e55efba776e45e8db0678194025d2520c03125"
            "a0b06409bb694e3ba3c32bdec089651efeffffffa18201000000000003000000000000000000126a0438564c050461736466046173"
            "646600000122020000000000004020d2357885ad9256bcea26db44b8a901e98f42fe6730bfba9562ec58b7b2b400000819f6020000"
            "0000fc51143b4abe2eb61626e30ea6a186fbd5dae1c14d709001a27e01000000000017005114ad926ed9fee896835674da40ad59fc"
            "1590cfa9e100000000"),
        SER_NETWORK, PROTOCOL_VERSION);
    CBlock block;
    stream >> block;
    return block;
}

CBlock TestBlock3()
{
    // Random real block a246f6fdd8bc24d763cbfc0d8bb7acd347e17f2c87929a295f9720a6bcc333e1
    // With 7 txns
    CDataStream stream(
        ParseHex(
            "f9c886072f67bb9345d29e4506f2ca4c5271fbaca811b5ab1e3e0bf419d971211cdb021dc30992a396915d9d0cd7df3bf558271ab1"
            "0c8bff464dc2ee2391cc8da6d03e393b661c660181bde06b5122122c2dd8fceda2e1dbe7425c0cdb1f11dbf1e9173c000000000000"
            "000000000000000000000000000000000000000000000000000055b4466384c35843fe93b60d330000000000000000000000000000"
            "000000000000000000000000d3080000000000000700000004e7732f02070000020167d49a3b000000001700511449da58946a34e6"
            "59a940088daeeaa5e23bd9b9f50000000000000000000f6a035862010008000000000000000000000000000200389a8f90a19e1084"
            "1981b28186c2d9e9409d9ea8eb7668a9a6616635177d02db6422210260af51019bb43ab375f9a8495ca7a64f6b85dbefbbfaf160ca"
            "e134cbc33c663240478252dc16bbf5254c139e265367c323a3f0664afb67612c9872159310dec7c24a7068195451267918f5dd2beb"
            "fddfa10c247f0bebae214c4be5a8c52bf48bd6feffffff220200000000000000f485d4c9897cde91be9332b7e46bfac5c8331887ad"
            "ece5e1dab3dcaf82680a0b642221034b0a8dcf0586b076173b2f0ffe4a650daf529c0c385c91f153aebfbe1f46dea840d0b53801c7"
            "70701098b0e2c7420f1bf81873f7333f56edfc38cdb28e8ea4c985b8ab87c72ee0d42fb4cc45e7ef8a07427aa63c18187fe0ae791b"
            "50d587608274feffffff8f80000000000000030122020000000000003a20837f103d409a778d0710998c22e803e7d62f18e5e8d5e2"
            "6a1c863c34aaad00000255a45114fe5cb9730dc41bbce9df98911383e1904610fb7a0122020000000000004020837f103d409a778d"
            "0710998c22e803e7d62f18e5e8d5e26a1c863c34aaad00000800000000000000fc5114c2fe632d4382ac3603014ae76780184ad59a"
            "029f01687c00000000000017005114d6a74ff4518e1f1b31926776c056a3b3d4aad214000000000001008a295287dc6900e5224d47"
            "475509d48c530ef172a4b97f4b51a197e6b6f746376422210237ce90ad73e20bb19320a8705924545f2fc114f70434e29b28b4d672"
            "089d4be440743a70a2426b3cc9313f823ccc5f6f8107bf15b9e311686c64f971b30ee02e2bc71299b08d47a2261d448b39da7673ef"
            "1654242095b74f5c9e2a0ee07c73c62ffefffffff41a0100000000000201b888000000000000170051144fef0963fbabae5d73a2c3"
            "406a979acb5af2d4fc01619100000000000017005114c07f46b83c62fa02a75f47c7d6adbe5a79e93e9256620100000100b40fd66e"
            "72d0ceed9279ced92faccb3501106a651e831bac7e6db70484dd3f9264222103160c63d3615f95450be2cabe2c99de96cd45934900"
            "d98bea130acab876d2d74c406a24e19e008d09949eea58524af12559e24064e4968c0d9d7dbd61b974f0a7c20050ef3abcd1af15be"
            "96e6d4e29fd14c3bb44f8b9b262b6def8d5c64746bc884feffffff7e8d00000000000003000000000000000000166a0438564c0506"
            "4a515247454e064a515247454e00000122020000000000004020ff9d42f693e6311b619d2ce78c4a58b0f6c585899732e8d0bb1f2d"
            "cf2efa000008c6610000000000fc51144fef0963fbabae5d73a2c3406a979acb5af2d4fc017b89000000000000170051149a371461"
            "9825a9018bb642ff7d503ea6c6d93e0f00000000000100df1a6e27fdf0ad52a92a36a2e801a396d0437823231148579a8f32bae6c3"
            "c68e64222103b94d5ed6db68e34a1d36fbb1e99f422dbfc01756b4c18507ceb30a121826b1ab40e4d06fd98f68aecff35f4b688e9d"
            "7e5537644c826adb64868cf5111b8d1dde68e5de7e9caf651dda204ea6fa8b4648435d1a9ea17f1420c92f9e26bde85f0d15feffff"
            "ff918400000000000003000000000000000000156a0438564c0505444f4e45520644c3b66e65720000012202000000000000402083"
            "7f103d409a778d0710998c22e803e7d62f18e5e8d5e26a1c863c34aaad000008734f0000000000fc5114b70c2d648c7dd46592be11"
            "73dbfd8bc235bd5323018f8000000000000017005114070d87e13a0eac905f4a0608a7c88f8a5863f31f000000000001007b76fbdd"
            "2b0f7e59d5dd20d4cce86325448f92e0fcf0a8b8bd266c72b9592073642221039368e2513c9f2feb401eebf03aa23d0b8afec59987"
            "d6d6b9c135a184b4d3f9dc40c9074ce56ac74f416f8e2cabe3d17277c983e9bb1846b5ebd2d0b521ad31c4bfe99f30f88452343726"
            "95145f015670ca0d1a4a65b93593021ba212908a3ae9c2feffffff6191000000000000020122020000000000004020c6f7cb3f7683"
            "a1aa1f9cd5782522272a26f276c8953cb3a2ed0551899ec0000008ec7a0100000000fc5114b2a24e01757446c43ad29218af120eff"
            "104b2ecf017e8d000000000000170051148f53153b30b498038a9635cb33d99d43353a307400000000000200e8cbfd093ccbdc5690"
            "0c2d9d6c0573981739289c006bb6ac0d6a87ea80b4bd786422210350810a98296c06bae1336fb0c962ea2d39ecd7956c2d88f07d05"
            "0122d37973d040f1c43f0345e355d28e2f3e554b1e9137abe6f8e45f4905c798d96da42699f549ae30c1f57ddd22e53ea11f012f64"
            "1d8f375d5a2df1f4aeb47d66d271b98b356cfeffffff220200000000000000792d68a2974a4e703caed06ee88c44b507aa25c0212e"
            "234961798dd93f4608026422210350810a98296c06bae1336fb0c962ea2d39ecd7956c2d88f07d050122d37973d040f1c43f0345e3"
            "55d28e2f3e554b1e9137abe6f8e45f4905c798d96da42699f549ae30c1f57ddd22e53ea11f012f641d8f375d5a2df1f4aeb47d66d2"
            "71b98b356cfeffffffb888000000000000030122020000000000003a20ff9d42f693e6311b619d2ce78c4a58b0f6c585899732e8d0"
            "bb1f2dcf2efa000002450051144fef0963fbabae5d73a2c3406a979acb5af2d4fc0122020000000000004020ff9d42f693e6311b61"
            "9d2ce78c4a58b0f6c585899732e8d0bb1f2dcf2efa00000800000000000000fc51142df759de93d0e6e808e68d5e0ede1b8f9ca51d"
            "240191840000000000001700511442d30ffcd78ea69665239f2577cbf0755e48ea4a00000000"),
        SER_NETWORK, PROTOCOL_VERSION);
    CBlock block;
    stream >> block;
    return block;
}

CBlock TestBlock4()
{
    // Random real block 6ffe14863fdc8c09660027074a8b013a261d190d91225592393f2d64f8c25f56
    // With 9 txns
    CDataStream stream(
        ParseHex(
            "18966c0d9ae285d724f06de1747df66642a1ab4d8674b37ef19ba8be5994f31a86e1021d8d3a9c7e2f6710414b821dd991746ac88a"
            "4379ed6679200265a2bd0d8b619ea3341370fc05ec45c40940860a5d986758f487ddf03ff776e551afe538b4fc12a9000000000000"
            "0000000000000000000000000000000000000000000000000000e3ef4f6384ea7b8192a402ef390000000000000000000000000000"
            "000000000000000000000000c90f0000000000000900000004e4647b02090000020028da9a3b000000001976a914ec743f997fcadb"
            "586c8b3aab17d369c7d82d714d88ac0000000000000000000d6a03fb750100060000000000000000000000020068e4142e319b4720"
            "eaff99cb73fa8b1d199e062143baaefeb1927971b7aeac9264222103113f96d857af038487ef2f21d7d0185c9945a8c98e2469081a"
            "06364402c6607a4019694e2f7abf2eca72a41499f3d211127b25374494fc06170d9245c8b0ca954e8b8e0b324c5fc3e4aacb4c2fd5"
            "945a81c82f5b57f516321da88da7275d165230feffffff220200000000000000381d4c41b1850858e8e431b76c9699379b3db336eb"
            "94ec7ad7e9bfca3237a011642221030a071aa32f9d1bb3b1915a90feb06b9853c144cd092e270621a3c62c41e21bd04029b78df18f"
            "b494d4d72b2d72c13c19191d0a6b213ad77f2d7be92f38d4039a6064597a6bf0d5bcd3d8b8099481c8ba0af443664c03e5c866a841"
            "626491ca99bcfefffffffe5b0e0000000000030122020000000000003a2079d8dfd50b0079834d43e9f4b1d5ab7f8118086441d27e"
            "8d649d59e7fbbb000002400051142a32d55a6517b054c221507016eb0576c1cd1bba012202000000000000402079d8dfd50b007983"
            "4d43e9f4b1d5ab7f8118086441d27e8d649d59e7fbbb00000800000000000000fc51147643432f4fe92826fc8b840258ed0e544c3e"
            "54bf01d7570e00000000001700511433f1de7e6079befeb833b088d5282f2dcb40c971000000000002005b996de6c6c420be54c9ba"
            "ce23088ab9cf1a208314f0e9538468e033761b08ed64222102e99339b421de29bc7c3920291693a484d5b34ed8353eb428d34996b7"
            "ca4735bf40d8bc2e7b7279b12b00204000c5db249f911e61d7801dda200f6067bf479b72224015d33582109c290bfa247d5412b7ca"
            "69cb507f7dea0929ad54e7d0bc2851bbfeffffff2202000000000000008c51ff73838a5a105a4d287aea89fdd0392c1e25170e3ca8"
            "00b967e53fee6e6e64222103301542af49fe52d217a6ba57c7ed66b18b68e1d9e32b90a342509948e5cdf69e40d076fa34594c7c61"
            "1e3eb440b02213926c391fe0762b784a6c81f2c0a3f54496c3740015a41711d9c84706871f33db12d59463acc0dbe8808731082bb0"
            "aac1d9feffffff3b470e0000000000030122020000000000003a20cb76f875d07b7c9a0cc7624641cec2a5b4720fcaae6757f6a1b5"
            "e56fe78b000002400051142a32d55a6517b054c221507016eb0576c1cd1bba0122020000000000004020cb76f875d07b7c9a0cc762"
            "4641cec2a5b4720fcaae6757f6a1b5e56fe78b00000800000000000000fc511475baa8dd1678bd83621c05c32254d57cb7b9c0d301"
            "14430e000000000017005114730c946c2c448325df2ad333718ef64786cec75a000000000002005702f4a3a1bd3fb887e26b6d3139"
            "d73c3b50bcef5da3b1be7894f293ffda8d1164222102160e40f66d63e0afa4d9c96a96cf88d43578326120870ed89bf19291fffaf9"
            "98401a747250eb3f7939aa0cbaf197081155c9c2ce727362ad37eaa1491541882514801a3bd67f339ae0c73cb45db1432687f8bc33"
            "ec6df4be6b89e184ea0af5a5f8feffffff22020000000000000096efb71d6748146e49f20043b5509e1f770a5d90e9165bea61da2c"
            "02616b47ef64222103f28fb87c95faf9b8dfaa64f5b7b6fb24ddad9056e6949851a9f84489bfacd3b340fac50bed973ca085218927"
            "cc58772c3563533586b0d5b8e2eebe5b96233672855e06eda70597d2b17dc4a39dc29b40a3cd7c9c29375108c9022cc994ae61b6fc"
            "feffffff894f0e0000000000030122020000000000003a203b444b2b52c93657d104b03cb9cf8a3941cddc926133004f62cef4a8c4"
            "e5000002400051142a32d55a6517b054c221507016eb0576c1cd1bba01220200000000000040203b444b2b52c93657d104b03cb9cf"
            "8a3941cddc926133004f62cef4a8c4e500000800000000000000fc5114a95d5f3849fd9b71ee696bb68e7540fb3156abaf01624b0e"
            "00000000001700511405c3e30f33ca0180fc261b4fb9a5ce715221618d0000000000020012f5d64be204860dd9f4b49a2646ed78aa"
            "84f47797567da64fba89d97d10793264222103df20d599f92e4de85c91c457ff0fce8c0a3c10429abdbd92923bda719f48c1be4022"
            "5d6c375a3a5de4bcfef8c565665379e38628c00b7cec9d8cce25260578ddf32ea3024d52c4fdc86ae1e64a74eaca7a875ed649e9e4"
            "09b06e2a846a67cf68c5feffffff220200000000000000fd42bafcb0abab8e11b1f66df22fe359521c0feb1f4fc9f9086a0ebc042f"
            "373464222102416b5e7362c3038898dfd480c54e74ce7035e40a4c2b2d7afcb3f5a60846bd9d402b305779097433079f1c1c60c9db"
            "6054ead7adf76b2a69cb93fcdbaac0ccf425bb17d1beb4529d7d5739fd359ab2dd45e08f05d06cb3ab9270dc68b43dbb2d8ffeffff"
            "ffd7570e0000000000030122020000000000003a2060610451710e87fd4f692dc519104a12fdfd529997a294423239cb113d760000"
            "02400051142a32d55a6517b054c221507016eb0576c1cd1bba012202000000000000402060610451710e87fd4f692dc519104a12fd"
            "fd529997a294423239cb113d7600000800000000000000fc5114a2ede8e10b070f44b5c1e7dd8ad17b3315bc69ab01b0530e000000"
            "000017005114e57d087403a9ad8e919ac9b4d7c6646465da4ce500000000000200a7cf574626473357569b3caa6a6cf3d1c66e1ee6"
            "dfb7b478d957c4dcf5769e2264222103d9ac273f90922f47a7b66ac0bbcb73365c6b5429c8e313c1549fe75aaaa30c8e40fc482d36"
            "cd879669278a532af772aba10fc34b348624c8dbcb760b28515b49c795d38aacdfce1827902ebb9a8e2c1f0314ebfc73d8c3688305"
            "7f9ed2b16de6cafeffffff22020000000000000054ebdfbd3c0a48f84f55924aadd023ec113b944e636ff8d7afd7eab12dd667a664"
            "222102ee0d879746fe6f8789a839745725d7c3223196d02c30f6f9b9eb9b10620337d74081b885213d01f8505b715372dec2c7fef8"
            "02a56cf08498f524fdd0d7750612cb4ab806bd5fa56166de8ed2af82564559393e92b9c08134f20e65456e0cc05cf6feffffff1443"
            "0e0000000000030122020000000000003a2077e2e62f245ade305292f01822e931bf92d7fd6844587c3807bbf1c899f80000024000"
            "51142a32d55a6517b054c221507016eb0576c1cd1bba012202000000000000402077e2e62f245ade305292f01822e931bf92d7fd68"
            "44587c3807bbf1c899f800000800000000000000fc511440ede15726f1526e329fa9edd7e8e2a1632db7e201ed3e0e000000000017"
            "0051147020dfaa8aba434d39fd0c6f4101f973c816f1c200000000000200214a11fcc0aba0849ee35e4216280516b21e5d97f7ce31"
            "6c80bb17773411fdd1642221038f3d83c580187cc7668d6a68c29e02ffff72256286bedd93ec3dac58072ae9d440ee0c08197bdb08"
            "0e3e51e6871a15b89d5e6ee586a1b0ad7cc3be1f3d554b77fdd40d61e6e44c3190eefea77f2f25beb0c5d5088fe97a73b064bd1f0b"
            "591cd6e2feffffff220200000000000000898f618f56d6dd523d685d4e806cfad6e5279aee928fbf972c8a23e3d7fb13b864222103"
            "2cc50734872d0bdb012b8569eee802127179a216f24710adafd0f9494150e65e4037248e3563c33a152f8159300bea1bdf13824b92"
            "5d484c007dfe8f5ccddf9a288ea85a757e3626c85dbe7e4fd6ce032b6eff0017c05080aa88cca47bf3ba3b2cfeffffff25600e0000"
            "000000030122020000000000003a202e01d761b767e1f4d121c8ac0a7299a27a3a89d77f8915058da0b12f60b7000002400051142a"
            "32d55a6517b054c221507016eb0576c1cd1bba01220200000000000040202e01d761b767e1f4d121c8ac0a7299a27a3a89d77f8915"
            "058da0b12f60b700000800000000000000fc5114dc66d73ea34160e4cac0213f8777f0f73757013301fe5b0e000000000017005114"
            "2b63516691a033ff604d1bb8d457f1aa344ef74c00000000000200677481f9e14e793bf91ac99eb4715e4f1339464fccdc77379b94"
            "3b3e9d11ed6564222102de5f351c1ff8cef2a106f4c086383b953512acb12caed3a2b2c689cc4ef05ec3404108ab34cb395b4f9b57"
            "15c098ff3b0e8d901d58cc3caf85acffd781bcfb8216cd727d9c8e5bfbc473bd937cbd32f043e0691586cfaed55cb06c8798d9a48f"
            "37feffffff2202000000000000004301ae27056b7159081499fd17df586573ba8fa4c5e0cbfd641ca46c55e88edb64222103ed4483"
            "34bf0d12b22d2f253673fa84743befd0a092cc0b984703775017edc83b406c9477c5390833b8c2dedb8f15e5270d578cead0c81012"
            "9c8bd08082c17689b8a9f6e04ce8d1898f2648e6c9b39a84b66d05b6cf044dee5f625aef7377361bbcfeffffff624b0e0000000000"
            "030122020000000000003a207f55500d7ecbe0cd991f2aac3963b4ca6fe945e3d95819da6d91c466ddfe000002400051142a32d55a"
            "6517b054c221507016eb0576c1cd1bba01220200000000000040207f55500d7ecbe0cd991f2aac3963b4ca6fe945e3d95819da6d91"
            "c466ddfe00000800000000000000fc5114102152024714b3b4aaaa28945cbda37f5f2945fe013b470e00000000001700511425e968"
            "88c2034f1377013a1340f2a58d1bd1c86f00000000000200d27dfbaa309502b65c4f503f03a332e6a7dfcaa24b2c4ec750ef178412"
            "f7fb53642221022d0d7932dcfc1678339979525c8e1536c2fbe81f2a1c40ed23bd285fdfee1b3a40c179f66d02d3d644a7380ee3f8"
            "68d34f03572fd318373f17ff512e0a7f695c025178b1325a095b30e68e7c44c7b4f06ad032b4c65508d632bcecbe1d213ed780feff"
            "ffff220200000000000000dc854d465f1d2477270470902961c220c1cb3b1323e1f9ff34a0380f7dae642a642221020675e7bc45b1"
            "32efa3ac51648a428a7ff3d6bc14edaa3c1039f14c3f1e7ea91940bf0283a61dec7b11e608f4fe1725f7e1d6c0c569eb9b64434b91"
            "2764f4eebe5ad06256f29e253f281fe0c43ffd1a7c164271b6234f7e4280aa278b01d948d0d1feffffffb0530e0000000000030122"
            "020000000000003a20efec13697aaf7f1b2736b603a2a3fae32674ad655b5df8ee46689c2b00d7000002400051142a32d55a6517b0"
            "54c221507016eb0576c1cd1bba0122020000000000004020efec13697aaf7f1b2736b603a2a3fae32674ad655b5df8ee46689c2b00"
            "d700000800000000000000fc51147bddbbc2b94d917d356cc0471aa5c74818a7162401894f0e000000000017005114991e4552c243"
            "8d77f6bdc043aa00cf05aaf3abad00000000"),
        SER_NETWORK, PROTOCOL_VERSION);
    CBlock block;
    stream >> block;
    return block;
}
